//#define DEBUG

#ifdef _WIN32
 #include <windows.h>
 #include <sys/utime.h>
#else
 #include <pthread.h>      // pthread_create
 #include <dlfcn.h>        // dlopen

 #define HANDLE            void *
 #define HMODULE           HANDLE
 #define LPVOID            void *

 #define SOCKET            int
 #define INVALID_SOCKET    -1
 #include <utime.h>
#endif

#include "AnsiString.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
//#include <iostream>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "debugger.h"

#ifdef _WIN32
 #include <io.h>
//#include <dir.h>
 #include <direct.h>
#else
//#include <unistd.h>
 #include <sys/select.h>
 #include <netinet/tcp.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
#endif

//#define STATIC_LINK

#ifdef STATIC_LINK
 #include "ConceptLibrary.h"
#endif

#ifdef _WIN32
 #define CMP_FUNC        stricmp
#else
 #define CMP_FUNC        strcmp
//#define LINUX_PATH "/opt/Concept/bin/"
 #define D_TEMP_DIR      "/tmp"
#endif

#define KEY_EXTENSION    ".appkey"
#define KEY_FILE         "generic.appkey"

#ifdef _WIN32
 #define LIBRARY         "concept-core.2.0.threadsafe.dll"
#else
 #define LIBRARY         "libconceptcore.2.0.threadsafe.so"
//#define LIBRARY      "concept-core.1.0alpha.so"
#endif

#define INI_FILE          "concept.ini"

#ifdef _WIN32
 #define TEMPLATE_PIPE    "ConceptCGIPipe"
#else
 #define TEMPLATE_PIPE    "ConceptCGIPipeXXXXXXXX.tmp"
#endif

extern "C" {
#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_request.h>
#include <http_core.h>
#include <http_log.h>
#include <apr_strings.h>
#include <apr_optional.h>
#include <apr_lib.h>
#include <util_script.h>
}
#include "semhh.h"

typedef int (*GET_VARIABLE_PROC)(int operation, void *VDESC, void *CONTEXT, int Depth, char *VariableName, char *buffer, int buf_size);
typedef int (*DEBUGGER_CALLBACK)(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED);
typedef int (*NOTIFY_PARENT)(int pipe_out, int apid, int msg_id, int len, char *data);

extern "C" {
typedef int (*API_SETNOTIFY)(NOTIFY_PARENT np);
typedef int (*API_INTERPRETER2)(char *filename, char *inc_dir, char *lib_dir, void *fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, void *userdata);
typedef int (*API_INTERPRETER)(char *filename, char *inc_dir, char *lib_dir, void *fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent);
}

typedef char * (*GETENV_EMULATOR)(const char *name, void *userdata);
typedef int (*HEADER_EMULATOR)(const char *key, const char *value, void *userdata);
typedef void (*PRINT_EMULATOR)(char *data, int len, void *userdata);

typedef char * (*FGETS_EMULATOR)(char *str, int size, void *userdata);
typedef int (*FGETC_EMULATOR)(void *userdata);


struct EnvContainer {
    GETENV_EMULATOR getenv;
    HEADER_EMULATOR header;
    PRINT_EMULATOR  print;

    FGETC_EMULATOR  fgetc;
    FGETS_EMULATOR  fgets;

    request_rec     *r;

    AnsiString      namer;
    AnsiString      namew;
    int             client_sock;
    int             c_piper;
    int             c_pipew;

    int             done_thread;
    char            *filename;

    char            *post_data;
    int             post_data_len;
    int             post_data_index;

    HHSEM           sem_lock;
};

#ifdef _WIN32
void switch_to_binary() {
    int hndl = fileno(stdout);

    setmode(hndl, O_BINARY);                 // handle to binary mode
    //ioctl( hndl, 1, (info & 0xff) | 0x20 );  // device to raw mode
    //stdout->flags |= _F_BIN;                 // stream to binary mode
}

#endif


static const char *lookup_header(const char *name, EnvContainer *ctx) {
    return(apr_table_get(ctx->r->headers_in, name));
}

//static APR_OPTIONAL_FN_TYPE(ssl_var_lookup) *concept_ssl_lookup = NULL;
//static APR_OPTIONAL_FN_TYPE(ssl_is_https) *concept_is_https = NULL;
AnsiString buffer2;
static char *lookup_variable(char *var, EnvContainer *ctx) {
    const char  *result;
    request_rec *r     = ctx->r;
    apr_size_t  varlen = strlen(var);

    /* fast exit */
    if (varlen < 4)
        return(apr_pstrdup(r->pool, ""));

    result = NULL;

    /* fast tests for variable length variables (sic) first */
    if (var[3] == ':') {
        if (var[4] && !strncasecmp(var, "ENV", 3)) {
            var   += 4;
            result = apr_table_get(r->notes, var);

            if (!result)
                result = apr_table_get(r->subprocess_env, var);
            if (!result)
                result = getenv(var);
        }

        /*else if (var[4] && !strncasecmp(var, "SSL", 3) && concept_ssl_lookup) {
         *  result = concept_ssl_lookup(r->pool, r->server, r->connection, r,
         *                              var + 4);
         * }*/
    } else if (var[4] == ':')   {
        if (var[5]) {
            request_rec *rr;
            const char  *path;

            if (!strncasecmp(var, "HTTP", 4))
                result = lookup_header(var + 5, ctx);
        }
    }
    /* well, do it the hard way */
    else {
        char           *p;
        apr_time_exp_t tm;

        /* can't do this above, because of the getenv call */
        for (p = var; *p; ++p)
            *p = apr_toupper(*p);

        switch (varlen) {
            case  4:
                if (!strcmp(var, "TIME")) {
                    apr_time_exp_lt(&tm, apr_time_now());
                    result = apr_psprintf(r->pool, "%04d%02d%02d%02d%02d%02d",
                                          tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
                                          tm.tm_hour, tm.tm_min, tm.tm_sec);
                    //rewritelog((r, 1, ctx->perdir, "RESULT='%s'", result));
                    return((char *)result);
                }
                break;

            case  5:

                /*if (!strcmp(var, "HTTPS")) {
                 *  int flag = rewrite_is_https && rewrite_is_https(r->connection);
                 *  return apr_pstrdup(r->pool, flag ? "on" : "off");
                 * }*/
                break;

            case  8:
                switch (var[6]) {
                    case 'A':
                        if (!strcmp(var, "TIME_DAY")) {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%02d", tm.tm_mday));
                        }
                        break;

                    case 'E':
                        if (!strcmp(var, "TIME_SEC")) {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%02d", tm.tm_sec));
                        }
                        break;

                    case 'I':
                        if (!strcmp(var, "TIME_MIN")) {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%02d", tm.tm_min));
                        }
                        break;

                    case 'O':
                        if (!strcmp(var, "TIME_MON")) {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%02d", tm.tm_mon + 1));
                        }
                        break;
                }
                break;

            case  9:
                switch (var[7]) {
                    case 'A':
                        if ((var[8] == 'Y') && !strcmp(var, "TIME_WDAY")) {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%d", tm.tm_wday));
                        } else if (!strcmp(var, "TIME_YEAR"))   {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%04d", tm.tm_year + 1900));
                        }
                        break;

                    case 'E':
                        if (!strcmp(var, "IS_SUBREQ"))
                            result = (r->main ? "true" : "false");
                        break;

                    case 'F':
                        if (!strcmp(var, "PATH_INFO")) {
                            if ((r->path_info) && (r->path_info[0]))
                                result = r->path_info;
                            else
                                result = r->uri;
                        }
                        break;

                    case 'P':
                        if (!strcmp(var, "AUTH_TYPE"))
                            result = r->ap_auth_type;
                        break;

                    case 'S':
                        if (!strcmp(var, "HTTP_HOST"))
                            result = lookup_header("Host", ctx);
                        break;

                    case 'U':
                        if (!strcmp(var, "TIME_HOUR")) {
                            apr_time_exp_lt(&tm, apr_time_now());
                            return(apr_psprintf(r->pool, "%02d", tm.tm_hour));
                        }
                        break;
                }
                break;

            case 11:
                switch (var[8]) {
                    case 'A':
                        if (!strcmp(var, "SERVER_NAME"))
                            result = ap_get_server_name(r);
                        break;

                    case 'D':
                        if ((*var == 'R') && !strcmp(var, "REMOTE_ADDR"))
                            result = r->connection->remote_ip;
                        else if (!strcmp(var, "	"))
                            result = r->connection->local_ip;
                        break;

                    case 'E':
                        if ((*var == 'H') && !strcmp(var, "HTTP_ACCEPT"))
                            result = lookup_header("Accept", ctx);
                        else if (!strcmp(var, "THE_REQUEST"))
                            result = r->the_request;
                        break;

                    case 'I':
                        if (!strcmp(var, "API_VERSION"))
                            return(apr_psprintf(r->pool, "%d:%d",
                                                MODULE_MAGIC_NUMBER_MAJOR,
                                                MODULE_MAGIC_NUMBER_MINOR));
                        break;

                    case 'K':
                        if (!strcmp(var, "HTTP_COOKIE"))
                            result = lookup_header("Cookie", ctx);
                        break;

                    case 'O':
                        if ((*var == 'S') && !strcmp(var, "SERVER_PORT"))
                            return(apr_psprintf(r->pool, "%u", ap_get_server_port(r)));
                        else if ((var[7] == 'H') && !strcmp(var, "REMOTE_HOST"))
                            /*result = ap_get_remote_host(r->connection,r->per_dir_config,
                             *                          REMOTE_NAME, NULL);*/
                            result = ap_get_remote_logname(r);
                        else if (!strcmp(var, "REMOTE_PORT"))
                            return(apr_itoa(r->pool, r->connection->remote_addr->port));
                        break;

                    case 'S':
                        if ((*var == 'R') && !strcmp(var, "REMOTE_USER"))
                            result = r->user;
                        else if (!strcmp(var, "SCRIPT_USER")) {
                            result = "<unknown>";
                            if (r->finfo.valid & APR_FINFO_USER)
                                apr_uid_name_get((char **)&result, r->finfo.user,
                                                 r->pool);
                        }
                        break;

                    case 'U':
                        if (!strcmp(var, "REQUEST_URI"))
                            result = r->uri;
                        break;
                }
                break;

            case 12:
                switch (var[3]) {
                    case 'I':
                        if (!strcmp(var, "SCRIPT_GROUP")) {
                            result = "<unknown>";
                            if (r->finfo.valid & APR_FINFO_GROUP)
                                apr_gid_name_get((char **)&result, r->finfo.group,
                                                 r->pool);
                        }
                        break;

                    case 'O':
                        if (!strcmp(var, "REMOTE_IDENT"))
                            result = ap_get_remote_logname(r);
                        break;

                    case 'P':
                        if (!strcmp(var, "HTTP_REFERER"))
                            result = lookup_header("Referer", ctx);
                        break;

                    case 'R':
                        if (!strcmp(var, "QUERY_STRING"))
                            result = r->args;
                        break;

                    case 'V':
                        if (!strcmp(var, "SERVER_ADMIN"))
                            result = r->server->server_admin;
                        break;
                }
                break;

            case 13:
                if (!strcmp(var, "DOCUMENT_ROOT"))
                    result = ap_document_root(r);
                break;

            case 14:
                if ((*var == 'H') && !strcmp(var, "HTTP_FORWARDED"))
                    result = lookup_header("Forwarded", ctx);
                else if (!strcmp(var, "REQUEST_METHOD"))
                    result = r->method;
                break;

            case 15:
                switch (var[7]) {
                    case 'E':
                        if (!strcmp(var, "HTTP_USER_AGENT"))
                            result = lookup_header("User-Agent", ctx);
                        break;

                    case 'F':
                        if (!strcmp(var, "SCRIPT_FILENAME"))
                            result = r->filename; /* same as request_filename (16) */
                        break;

                    case 'P':
                        if (!strcmp(var, "SERVER_PROTOCOL"))
                            result = r->protocol;
                        break;

                    case 'S':
                        if (!strcmp(var, "SERVER_SOFTWARE")) {
                            buffer2  = (char *)ap_get_server_banner();
                            buffer2 += (char *)" ConceptWEB Embedded";
                            //result = ap_get_server_banner();
                            result = (const char *)buffer2.c_str();
                        }
                        break;
                }
                break;

            case 16:
                if (!strcmp(var, "REQUEST_FILENAME"))
                    result = r->filename; /* same as script_filename (15) */
                break;

            case 21:
                if (!strcmp(var, "HTTP_PROXY_CONNECTION"))
                    result = lookup_header("Proxy-Connection", ctx);
                break;
        }
    }

    return(apr_pstrdup(r->pool, result ? result : ""));
}

AnsiString env_buffer;
static char *GetEnvEmulator(const char *name, void *userdata) {
    EnvContainer *ec     = (EnvContainer *)userdata;
    const char   *result = 0;

    if (ec) {
        request_rec *r = ec->r;
        switch (name[0]) {
            case 'H':
            case 'h':
                if (!strcasecmp(name, "HTTP_USER_AGENT"))
                    result = apr_table_get(r->headers_in, "User-Agent");
                else if (!strcasecmp(name, "HTTP_REFERER"))
                    result = apr_table_get(r->headers_in, "Referer");
                else if (!strcasecmp(name, "HTTP_COOKIE"))
                    result = apr_table_get(r->headers_in, "Cookie");
                else if (!strcasecmp(name, "HTTP_FORWARDED"))
                    result = apr_table_get(r->headers_in, "Forwarded");
                else if (!strcasecmp(name, "HTTP_HOST"))
                    result = apr_table_get(r->headers_in, "Host");
                else if (!strcasecmp(name, "HTTP_PROXY_CONNECTION"))
                    result = apr_table_get(r->headers_in, "Proxy-Connection");
                else if (!strcasecmp(name, "HTTP_ACCEPT"))
                    result = apr_table_get(r->headers_in, "Accept");
                else if ((strlen(name) > 5) && !strncasecmp(name, "HTTP:", 5))
                    // all other headers from which we are still not know about
                    result = apr_table_get(r->headers_in, name + 5);
                break;

            case 'R':
            case 'r':
                if (!strcasecmp(name, "REQUEST_METHOD"))
                    result = r->method;
                else if (!strcasecmp(name, "REQUEST_SCHEME"))
                    result = ap_http_scheme(r);
                else if (!strcasecmp(name, "REQUEST_URI"))
                    result = r->uri;
                else if (!strcasecmp(name, "REQUEST_FILENAME"))
                    result = r->filename;

                /*else if (!strcasecmp(name, "REMOTE_HOST"))
                 *      result = ap_get_remote_host(r->connection, r->per_dir_config,
                 *                                                              REMOTE_NOLOOKUP, NULL);*/
                else if (!strcasecmp(name, "REMOTE_ADDR"))
                    result = r->connection->remote_ip;
                else if (!strcasecmp(name, "REMOTE_IDENT"))
                    result = ap_get_remote_logname(r);
                else if (!strcasecmp(name, "REMOTE_USER"))
                    result = r->user;
                break;

            case 'S':
            case 's':
                if (!strncasecmp(name, "SSL", 3)) break;                 //shortcut common case

                if (!strcasecmp(name, "SERVER_ADMIN"))
                    result = r->server->server_admin;
                else if (!strcasecmp(name, "SERVER_NAME"))
                    result = ap_get_server_name(r);
                else if (!strcasecmp(name, "SERVER_PORT")) {
                    env_buffer = (long)ap_get_server_port(r);
                    result     = env_buffer.c_str();
                } else if (!strcasecmp(name, "SERVER_PROTOCOL"))
                    result = r->protocol;
                else if (!strcasecmp(name, "SCRIPT_FILENAME"))
                    result = r->filename;
                else if (!strcmp(name, "SERVER_SOFTWARE")) {
                    buffer2  = (char *)ap_get_server_banner();
                    buffer2 += (char *)" ConceptWEB Embedded";
                    //result = ap_get_server_banner();
                    result = (const char *)buffer2.c_str();
                    //result = ap_get_server_banner();
                }
                break;

            default:
                //if (!strcasecmp(name, "PATH_INFO"))
                //	result = r->path_info;
                //else
                if (!strcasecmp(name, "QUERY_STRING"))
                    result = r->args;
                else if (!strcasecmp(name, "IS_SUBREQ"))
                    result = (r->main != NULL ? "true" : "false");
                else if (!strcasecmp(name, "DOCUMENT_ROOT"))
                    result = ap_document_root(r);
                else if (!strcasecmp(name, "AUTH_TYPE"))
                    result = r->ap_auth_type;
                else if (!strcasecmp(name, "THE_REQUEST"))
                    result = r->the_request;
                else if (!strcmp(name, "PATH_INFO")) {
                    if ((r->path_info) && (r->path_info[0]))
                        result = r->path_info;
                    else
                        result = r->uri;
                }
                break;
        }

        /*if (!result)
         *      result=(char *)lookup_variable((char *)name, ec);*/
        if (!result)
            result = (char *)apr_table_get(ec->r->subprocess_env, name);
        return((char *)result);
    } else
        return(0);
}

int SetHeaders(const char *key, const char *value, void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;

    if (ec) {
        if (!strcasecmp(key, "content-type"))
            ap_set_content_type(ec->r, value);
        else
            apr_table_set(ec->r->headers_out, key, value);
        return(1);
    }
    return(0);
}

static apr_off_t get_content_length(request_rec *r) {
    apr_off_t len = 0;

    if (r->clength > 0)
        return(r->clength);
    else if (r->main == NULL) {
        const char *clp = apr_table_get(r->headers_in, "Content-Length");

        if (clp) {
            char *errp;
            if (apr_strtoff(&len, clp, &errp, 10) || *errp || (len < 0))
                len = 0; /* parse error */
        }
    }

    return(len);
}

int check_type(apr_bucket *bucket, char *what) {
    if (!bucket->type)
        return(0);
    int res = strncmp((char *)bucket->type, what, 3);
    return(!res);
}

/*int getPOST2(EnvContainer *ec) {
 *      request_rec *r=ec->r;
 *      apr_status_t rv;
 *      int seen_eos, child_stopped_reading;
 *  apr_bucket_brigade *bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);
 *      apr_off_t count_bytes=get_content_length(r);
 *
 *  seen_eos = 0;
 *  child_stopped_reading = 0;
 *      fprintf(stderr, "Content length: %i\n", count_bytes);
 *      if (count_bytes) {
 *              ec->post_data=new char [count_bytes+1];
 *              ec->post_data[0]=0;
 *      } else
 *              return 0;
 *
 *      int dbpos = 0;
 *      int tlen=0;
 *  //if (conf->logname) {
 *  //    dbuf = apr_palloc(r->pool, conf->bufbytes + 1);
 *  //    dbpos = 0;
 *  //}
 *  do {
 *      apr_bucket *bucket;
 *
 *      rv = ap_get_brigade(r->input_filters, bb, AP_MODE_READBYTES,
 *                          APR_BLOCK_READ, HUGE_STRING_LEN);
 *
 *      if (rv != APR_SUCCESS) {
 *          ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
 *                        "Error reading request entity data");
 *          return HTTP_INTERNAL_SERVER_ERROR;
 *      }
 *
 *      for (bucket = APR_BRIGADE_FIRST(bb);
 *           bucket != APR_BRIGADE_SENTINEL(bb);
 *           bucket = APR_BUCKET_NEXT(bucket))
 *      {
 *
 *          const char *data;
 *          apr_size_t len;
 *
 *          //if (APR_BUCKET_IS_EOS(bucket)) {
 *                      if (check_type(bucket, "EOS")) {
 *              seen_eos = 1;
 *              break;
 *          }
 *
 *          // We can't do much with this.
 *          //if (APR_BUCKET_IS_FLUSH(bucket)) {
 *                      if (check_type(bucket, "FLUSH")) {
 *              continue;
 *          }
 *
 *          // If the child stopped, we still must read to EOS.
 *          if (child_stopped_reading) {
 *              continue;
 *          }
 *
 *          // read
 *          apr_bucket_read(bucket, &data, &len, APR_BLOCK_READ);
 *                      if (len>0) {
 *                              memcpy(ec->post_data + tlen, data, len);
 *                              tlen+=len;
 *                      } else {
 *                              seen_eos=1;
 *                              break;
 *                      }
 *                      if (tlen>=count_bytes) {
 *                              seen_eos=1;
 *                              break;
 *                      }
 *
 *          //if (dbpos < count_bytes) {
 *          //    int cursize;
 * //
 * //              if ((dbpos + len) > count_bytes) {
 *  //                cursize = count_bytes - dbpos;
 *    //          }
 *      //        else {
 *        //          cursize = len;
 *          //    }
 *              //memcpy(dbuf + dbpos, data, cursize);
 *              //dbpos += cursize;
 *          //}
 *
 *          // Keep writing data to the child until done or too much time
 *          // elapses with no progress or an error occurs.
 *          //
 *          //rv = apr_file_write_full(script_out, data, len, NULL);
 *
 *          //if (rv != APR_SUCCESS) {
 *          //    child_stopped_reading = 1;
 *          //}
 *      }
 *      apr_brigade_cleanup(bb);
 *  }
 *  while (!seen_eos);
 *      ec->post_data_len=tlen;
 *      return tlen;
 * }*/

int getPOST3(EnvContainer *ec) {
    apr_size_t         len, tlen = 0;
    request_rec        *r = ec->r;
    apr_bucket_brigade *brigade;
    apr_off_t          count_bytes = get_content_length(r);
    char *buf = 0;

    /*if (!count_bytes) {// default 64k buffer
     *      count_bytes=0xFFFF;
     *      fprintf(stderr, "mod_concept: For '%s' no content\n", r->filename);
     * }*/

    if (count_bytes) {
        count_bytes += 8192;

        if (ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK) != OK) {
            fprintf(stderr, "mod_concept: Error reading client block\n");
            return(0);
        }

        ec->post_data    = new char [count_bytes + 1];
        ec->post_data[0] = 0;
    } else
        return(0);

    buf = ec->post_data;

    len  = count_bytes;
    tlen = 0;

    if (ap_should_client_block(r))
        while ((len = ap_get_client_block(r, ec->post_data + tlen, count_bytes - tlen)) > 0)
            tlen += len;

    ec->post_data_len   = tlen;
    ec->post_data[tlen] = 0;
    //if (ec->post_data)
    //	fprintf(stderr, "mod_concept: For '%s' post:\n%s\n", r->filename, ec->post_data);
    return(tlen);
}

/*
 * int getPOST(EnvContainer *ec) {
 *      apr_size_t len, tlen=0;
 *      request_rec *r=ec->r;
 *      apr_bucket_brigade *brigade;
 *      apr_off_t count_bytes=get_content_length(r);
 *      char *buf=0;
 *
 *      fprintf(stderr, "Content length: %i\n", count_bytes);
 *      if (count_bytes) {
 *              ec->post_data=new char [count_bytes+1];
 *              ec->post_data[0]=0;
 *      } else
 *              return 0;
 *
 *      buf=ec->post_data;
 *
 *      brigade = apr_brigade_create(r->pool, r->connection->bucket_alloc);
 *      len = count_bytes;
 *
 *      while (ap_get_brigade(r->input_filters, brigade, AP_MODE_READBYTES, APR_BLOCK_READ, len) == APR_SUCCESS) {
 *              apr_status_t rv = APR_SUCCESS;
 *
 *              apr_off_t off;
 *              apr_status_t s = apr_brigade_length(brigade, 0, &off);
 *              fprintf(stderr, "Rep len: %i\n", off);
 *
 *              fprintf(stderr, "Loop: %i\n", len);
 *              rv=apr_brigade_flatten(brigade, buf, &len);
 *              if (rv!=APR_SUCCESS) {
 *                      fprintf(stderr, "RV: %i\n", rv);
 *              }
 *              rv=apr_brigade_cleanup(brigade);
 *              tlen += len;
 *              fprintf(stderr, "Buf: %s\n", buf);
 *              fprintf(stderr, "Len: %i\n", len);
 *              if (tlen == count_bytes || !len) {
 *                      break;
 *              }
 *              buf += len;
 *              len = count_bytes - tlen;
 *      }
 *      apr_brigade_destroy(brigade);
 *
 *      ec->post_data_len=tlen;
 *      ec->post_data[tlen]=0;
 *      return tlen;
 * }*/

int fgetcEmulator(void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;

    if ((ec->post_data) && (ec->post_data_index < ec->post_data_len))
        return((int)(unsigned char)ec->post_data[ec->post_data_index++]);
    return(EOF);
}

char *fgetsEmulator(char *str, int size, void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;
    int          c;
    char         *ptr;

    for (ptr = str; size > 1; size--) {
        c = fgetcEmulator(userdata);
        if (c == EOF) break;
        *ptr++ = (char)c;
        if (c == '\n') break;
    }

    *ptr = '\0';
    if (strlen(str) == 0) return(NULL);

    return(str);

    /*size--;
     * if ((ec->post_data) && ((ec->post_data_index+1) < ec->post_data_len)) {
     *      int rest=ec->post_data_len-ec->post_data_index;
     *      if (rest>size) {
     *              memcpy(str, ec->post_data, size);
     *              ec->post_data_index+=size;
     *              str[size]=0;
     *      } else {
     *              memcpy(str, ec->post_data, rest);
     *              ec->post_data_index=ec->post_data_len;
     *              size=0;
     *      }
     *      return str;
     * }
     * return 0;*/
}

void Print(char *str, int length, void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;

    if ((ec) && (ec->r)) {
        if (length > 0)
            ap_rwrite(str, length, ec->r);
        else
            ap_rputs(str, ec->r);
    }    /* else {
          *     if (length>0)
          *             std::cout.write(str,length);
          *     else
          *             std::cout << str;
          *
          *     std::cout.flush();
          * }*/
}

#ifndef _WIN32
long GetPrivateProfileString(
    const char *lpAppName,
    const char *lpKeyName,
    const char *lpDefault,
    char       *lpReturnedString,
    long       nSize,
    const char *lpFileName) {
    FILE *in         = 0;
    long filesize    = 0;
    char *filebuffer = 0;
    char section_name[4096];
    char key_name[4096];
    int  error = 0;

    in = fopen(lpFileName, "rb");
    if (in) {
        fseek(in, 0L, SEEK_END);
        filesize = ftell(in);
        fseek(in, 0L, SEEK_SET);
        if (filesize) {
            filebuffer = new char[filesize];
            if (filebuffer) {
                if (fread(filebuffer, 1, filesize, in) == filesize) {
                    char in_name           = 0;
                    char in_quote          = 0;
                    char in_looked_section = 0;
                    char in_looked_key     = 0;
                    char in_looked_value   = 0;
                    char in_comment        = 0;
                    char in_key            = 0;
                    int  temp_pos          = 0;
                    int  key_pos           = 0;
                    int  value_pos         = 0;

                    section_name[0] = 0;
                    key_name[0]     = 0;

                    for (int i = 0; i < filesize; i++) {
                        if (in_comment) {
                            if ((filebuffer[i] == '\n') || (filebuffer[i] == '\r'))
                                in_comment = 0;
                            continue;
                        }

                        if (((filebuffer[i] == '\n') || (filebuffer[i] == '\r')) && (in_looked_value)) {
                            lpReturnedString[value_pos++] = 0;
                            break;
                        }

                        if (!in_quote) {
                            if (filebuffer[i] == ';') {
                                if (in_looked_value)
                                    break;
                                in_comment = 1;
                                continue;
                            }

                            if (filebuffer[i] == '[') {
                                // a inceput o sectziune noua, iar sectziunea precedenta este cea cautata ...
                                if (in_looked_section) {
                                    error = 1;
                                    break;
                                }
                                temp_pos = 0;
                                in_name  = 1;
                                continue;
                            }
                            if (in_name) {
                                if (filebuffer[i] == ']') {
                                    in_name = 0;
                                    section_name[temp_pos] = 0;
                                    if (!CMP_FUNC(section_name, lpAppName))
                                        in_looked_section = 1;
                                    continue;
                                }
                                if ((filebuffer[i] == '\n') || (filebuffer[i] == '\r')) {
                                    in_name = 0;
                                    continue;
                                }
                                if (in_name)
                                    section_name[temp_pos++] = filebuffer[i];
                            }
                        }
                        // schimb flag-ul de ghilimele, daca le-am gasit ...
                        if (!in_name) {
                            if (filebuffer[i] == '"') {
                                in_quote = !in_quote;
                                continue;
                            }
                            if (((filebuffer[i] == '\n') || (filebuffer[i] == '\r')) && (in_quote)) {
                                in_quote = 0;
                                continue;
                            }
                        }

                        if (in_looked_key) {
                            if (filebuffer[i] == '=') {
                                in_looked_value = 1;
                                continue;
                            }
                            if (in_looked_value) {
                                if (value_pos < nSize - 1) {
                                    if (value_pos)
                                        lpReturnedString[value_pos++] = filebuffer[i];
                                    else if ((filebuffer[i] != ' ') && (filebuffer[i] != '\t'))
                                        lpReturnedString[value_pos++] = filebuffer[i];
                                } else {
                                    lpReturnedString[value_pos++] = 0;
                                    break;
                                }
                                continue;
                            }
                        }

                        if (in_looked_section) {
                            if (((filebuffer[i] >= '0') && (filebuffer[i] <= '9')) ||
                                ((filebuffer[i] >= 'a') && (filebuffer[i] <= 'z')) ||
                                ((filebuffer[i] >= 'A') && (filebuffer[i] <= 'Z')) ||
                                (filebuffer[i] == '_')) {
                                if (in_key == 0) {
                                    key_pos = 0;
                                    in_key  = 1;
                                }
                                key_name[key_pos++] = filebuffer[i];
                            } else if (in_key)    {
                                key_name[key_pos] = 0;
                                if (!CMP_FUNC(key_name, lpKeyName))
                                    in_looked_key = 1;
                                in_key = 0;
                            }
                            // TO DO ... sunt in sectiunea cautata ...
                        }
                    }
                } else
                    error = 1;
            } else
                error = 1;
        } else
            error = 1;
        fclose(in);
    } else
        error = 1;

    if (error)
        strncpy(lpReturnedString, lpDefault, nSize);
    return((long)in);
}

#endif

AnsiString GetKey(char *ini_name, char *section, char *key, char *def) {
    char value[4096];

    value[0] = 0;
    GetPrivateProfileString(section, key, def, value, 4096, ini_name);
    return(AnsiString(value));
}

#ifdef _WIN32
extern "C" {
extern BOOL WINAPI SetDllDirectoryA(LPCTSTR lpathName);
}
#endif

AnsiString GetDirectory() {
#ifdef _WIN32
    /*char buffer[4096];
     * buffer[0]=0;
     * GetModuleFileName(NULL,buffer,4096);
     * for (int i=strlen(buffer)-1;i>=0;i--) {
     *  if ((buffer[i]=='/') || (buffer[i]=='\\')) {
     *      buffer[i+1]=0;
     *      break;
     *  }
     * }
     * return AnsiString(buffer);*/
    char       tempPath[8192];
    char       file_buffer[8193];
    AnsiString tp_path;

    tempPath[0] = 0;

    if (GetTempPath(8192, tempPath) > 0)
        tp_path = tempPath;

    tp_path += "\\GetConceptDirectory.tmp";



    int res = system(AnsiString("HTTPModHelper.exe > \"") + AnsiString(tp_path) + "\"");
    if (res != 0xAB)
        fprintf(stderr, "mod_concept: cannot determine the Concept instalation path, code %i\n", res);
    char *path = getenv("CONCEPT_BIN_PATH");
    if (!path) {
        FILE *in = fopen(tp_path.c_str(), "rb");
        if (in) {
            int size = fread(file_buffer, 1, 8192, in);
            if (size > 0) {
                file_buffer[size] = 0;
                path = file_buffer;
            }
            fclose(in);
            unlink(tp_path.c_str());
        }
    }

    if (path) {
 #ifdef _WIN32
        SetDllDirectoryA(path);
 #endif
        return(AnsiString(path));
    } else {
        fprintf(stderr, "mod_concept: cannot determine the Concept instalation path (no environment CONCEPT_BIN_PATH is set)\n", path);
        return(AnsiString(""));
    }
#else
    return(LINUX_PATH);
#endif
}

void SetWebDirectory(char *path) {
    char       *buffer;
    AnsiString temp(path);

    //temp=getenv("PATH_TRANSLATED");
    if (temp != (char *)"") {
        buffer = temp.c_str();
        for (int i = strlen(buffer) - 1; i >= 0; i--)
            if ((buffer[i] == '/') || (buffer[i] == '\\')) {
                buffer[i + 1] = 0;
                break;
            }
        chdir(buffer);
    }
}

int need_compilation(char *filename) {
    AnsiString accel_filename = filename;

    accel_filename += ".accel";
    struct stat buf1, buf2;
    // if some error, report no need for compilation
    if (stat(filename, &buf1))
        return(0);

    if (stat(accel_filename.c_str(), &buf2))
        return(1);

    // last modification time don't match ... need recompilation ...
    if (buf1.st_mtime != buf2.st_mtime)
        return(1);

    return(0);
}

int Connect(char *HOST, int PORT) {
    struct sockaddr_in sin;
    struct hostent     *hp;
    int CLIENT_SOCKET = -1;

    if ((hp = gethostbyname(HOST)) == 0) {
        fprintf(stderr, "Unable to resolve specified host (%s)\n", HOST);
        return(-1);
    }

    memset(&sin, 0, sizeof(sin));
    //INADDR_ANY;
    sin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    sin.sin_family      = AF_INET;
    sin.sin_port        = htons(PORT);

    // creez socket-ul
    if ((CLIENT_SOCKET = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == INVALID_SOCKET) {
        fprintf(stderr, "Error creating socket for host (%s)\n", HOST);
        return(-1);
    }

    int flag = 1;
    setsockopt(CLIENT_SOCKET,                     /* socket affected */
               IPPROTO_TCP,                       /* set option at TCP level */
               TCP_NODELAY,                       /* name of option */
               (char *)&flag,                     /* the cast is historical cruft */
               sizeof(int));                      /* length of option value */

    if (connect(CLIENT_SOCKET, (struct sockaddr *)&sin, sizeof(sin)) == INVALID_SOCKET) {
        fprintf(stderr, "Error connecting to specified host (%s)\n", HOST);
        return(-1);
    }

    return(CLIENT_SOCKET);
}

void Done(void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;

    if (!ec)
        return;

    if (ec->client_sock == -1)
        return;
#ifdef _WIN32
    closesocket(ec->client_sock);
#else
    close(ec->client_sock);
    ec->client_sock = -1;
#endif

    if (ec->c_piper > 0) {
        close(ec->c_piper);
        ec->c_piper = 0;
        if (unlink(ec->namer) != 0) {
#ifdef _WIN32
            Sleep(10);
#else
            usleep(10000);
#endif
            unlink(ec->namew);
        }
    }
    if (ec->c_pipew > 0) {
        close(ec->c_pipew);
        ec->c_pipew = 0;
        if (unlink(ec->namew) != 0) {
#ifdef _WIN32
            Sleep(10);
#else
            usleep(10000);
#endif
            unlink(ec->namew);
        }
    }
#ifdef _WIN32
    WSACleanup();
#endif
}

int NotifyServer(int CLIENT_SOCKET, char *key, int *pipew, int *piper, long *_APID, long *_parentAPID, bool *debug, void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;

    if (CLIENT_SOCKET == -1) {
        fprintf(stderr, "Error getting APID for CGI application\n");
        return(-1);
    }

#ifdef _WIN32
    char tmp[8192];
    tmp[0] = 0;
    GetTempPath(8192, tmp);
#else
    char       *tmp = D_TEMP_DIR;
    AnsiString temp(tmp);
    temp += "/";

    AnsiString tplw(temp);
    AnsiString tplr(temp);
#endif


#ifdef _WIN32
    char szTempName[8192];
    GetTempFileName(tmp, TEMPLATE_PIPE, 0, szTempName);
    ec->namew = szTempName;
#else
    tplw += TEMPLATE_PIPE;
    tplr += TEMPLATE_PIPE;

    ec->namew = mktemp(tplw.c_str());
#endif

#ifdef _WIN32
    *pipew = open(ec->namew.c_str(), O_WRONLY | O_BINARY | O_CREAT);
#else
    *pipew = open(ec->namew.c_str(), O_WRONLY | O_CREAT);
#endif

#ifdef _WIN32
    GetTempFileName(tmp, TEMPLATE_PIPE, 0, szTempName);
    ec->namer = szTempName;
#else
    ec->namer = mktemp(tplr.c_str());
#endif


#ifdef _WIN32
    *piper = open(ec->namer.c_str(), O_RDONLY | O_BINARY | O_CREAT);
#else
    *piper = open(ec->namer.c_str(), O_RDONLY | O_CREAT);
#endif


    AnsiString query;

    query  = "piper ";
    query += ec->namew; // write for us is read for server and vice-versa
    query += "\r\npipew ";
    query += ec->namer;
    query += "\r\ngetapid\r\n";

    if ((key) && (key[0])) {
        query += "getparent ";
        query += key;
        query += "\r\n";
    }

    int len = query.Length();

    if (send(CLIENT_SOCKET, query.c_str(), len, 0) != len) {
        fprintf(stderr, "Concept Server query (request APID) failed\n");
        return(-1);
    }

    long APID;
    int  rec_count = 0;
    int  received  = 0;
    do {
        received   = recv(CLIENT_SOCKET, (char *)&APID + rec_count, sizeof(long) - rec_count, 0);
        rec_count += received;
    } while ((rec_count < sizeof(long)) && (received > 0));

    if (received == sizeof(long))
        *_APID = ntohl(APID);
    else {
        fprintf(stderr, "Concept Server query (request APID) failed (APID size is invalid)\n");
        return(-1);
    }


    if ((key) && (key[0])) {
        long parentAPID = 0;
        rec_count = 0;
        received  = 0;
        do {
            received   = recv(CLIENT_SOCKET, (char *)&parentAPID + rec_count, sizeof(long) - rec_count, 0);
            rec_count += received;
        } while ((rec_count < sizeof(long)) && (received > 0));

        if (received == sizeof(long)) {
            long pAPID = ntohl(parentAPID);
            if (pAPID < 0) {
                *debug = true;
                pAPID *= -1;
            }
            *_parentAPID = pAPID;
        } else {
            fprintf(stderr, "Concept Server query (request PARENT APID) failed (APID size is invalid)\n");
            return(-1);
        }
    }

    return(0);
}

AnsiString HOME_BASE;
AnsiString Library;
AnsiString Include;
int        autocompile = 0;

AnsiString host;
int        port          = 2663;
int        handle_cgi    = 0;
int        allow_debug   = 0;
int        allow_parents = 0;
int        seconds       = 10;

#ifndef STATIC_LINK
API_INTERPRETER2 Concept_Execute = 0;
API_INTERPRETER  Concept_Compile = 0;
API_SETNOTIFY    SetNotifyParent = 0;
HMODULE          hDLL            = 0;
#endif

static int lib_init = 0;

#ifdef _WIN32
AnsiString INI_PATH;
#endif


int ModInit() {
    if (lib_init)
        return(0);
    lib_init = 1;
#ifdef _WIN32
    WSADATA data;
    WSAStartup(MAKEWORD(1, 1), &data);
#endif

    HOME_BASE = GetDirectory();
#ifdef _WIN32
    INI_PATH = HOME_BASE + INI_FILE;
#endif

#ifdef _WIN32
    Library = HOME_BASE + GetKey(INI_PATH, "Paths", "Library", "Library") + "\\";
    Include = HOME_BASE + GetKey(INI_PATH, "Paths", "Include", "Include") + "\\";
#else
    Library = HOME_BASE + GetKey(INI_PATH, "Paths", "Library", "Library") + "/";
    Include = HOME_BASE + GetKey(INI_PATH, "Paths", "Include", "Include") + "/";
#endif
    autocompile = GetKey(INI_PATH, "Run", "AutoCompile", "0").ToInt();

    host          = GetKey(INI_PATH, "Server", "APIDHost", "localhost");
    port          = GetKey(INI_PATH, "Server", "CGIPort", "2663").ToInt();
    handle_cgi    = GetKey(INI_PATH, "Server", "HandleCGIRequests", "0").ToInt();
    allow_debug   = GetKey(INI_PATH, "Web", "AllowDebug", "0").ToInt();
    allow_parents = GetKey(INI_PATH, "Web", "AllowCASParent", "0").ToInt();

#ifdef _WIN32
    switch_to_binary();
#endif

#ifndef STATIC_LINK
 #ifdef _WIN32
    hDLL = LoadLibrary(LIBRARY);
 #else
    hDLL = dlopen(/*HOME_BASE+*/ LIBRARY_PATH, RTLD_LAZY);
 #endif

    if (!hDLL) {
        //printf("%s%c%c\n","Content-Type:text/plain;charset=iso-8859-1",13,10);
        fprintf(stderr, "mod_concept: core library '%s' cannot be loaded\n", LIBRARY);
        return(0);
    }

 #ifdef _WIN32
    Concept_Execute = (API_INTERPRETER2)GetProcAddress(hDLL, "Concept_Execute");
    if (!Concept_Execute)
        Concept_Execute = (API_INTERPRETER2)GetProcAddress(hDLL, "_Concept_Execute");
 #else
    Concept_Execute = (API_INTERPRETER2)dlsym(hDLL, "Concept_Execute");
 #endif
    if (!Concept_Execute) {
        fprintf(stderr, "mod_concept: corrupted core library '%s'\n", LIBRARY);
        return(0);
    }

 #ifdef _WIN32
    Concept_Compile = (API_INTERPRETER)GetProcAddress(hDLL, "Concept_Compile");
    if (!Concept_Compile)
        Concept_Compile = (API_INTERPRETER)GetProcAddress(hDLL, "_Concept_Compile");

    SetNotifyParent = (API_SETNOTIFY)GetProcAddress(hDLL, "SetNotifyParent");
    if (!SetNotifyParent)
        SetNotifyParent = (API_SETNOTIFY)GetProcAddress(hDLL, "_SetNotifyParent");
 #else
    Concept_Compile = (API_INTERPRETER)dlsym(hDLL, "Concept_Compile");
    SetNotifyParent = (API_SETNOTIFY)dlsym(hDLL, "SetNotifyParent");
 #endif
#endif

    if (SetNotifyParent)
        SetNotifyParent(RemoteNotify);
    seconds = GetKey(INI_PATH, "Web", "Timeout", "10").ToInt();
    if (seconds <= 0)
        seconds = 10;
}

void ModDone() {
#ifndef STATIC_LINK
    if (hDLL)
 #ifdef _WIN32
        FreeLibrary(hDLL);
 #else
        dlclose(hDLL);
 #endif
#endif

}

//int ModExecuteT(char *filename, void *userdata) {
#ifdef _WIN32
DWORD WINAPI ModExecuteT(LPVOID userdata) {
#else
LPVOID ModExecuteT(LPVOID userdata) {
#endif
    EnvContainer *ec = (EnvContainer *)userdata;
#ifdef _WIN32
    WSADATA data;
    WSAStartup(MAKEWORD(1, 1), &data);
#endif

    //char *filename=getenv("PATH_TRANSLATED");
    int  final_res = 0;
    int  run_debug = 0;
    char *filename = ec->filename;

    if (!filename) {
        fprintf(stderr, "Invalid WEBServer query\n");
        return(0); //-1;
    }

    SetWebDirectory(filename);

    // check for autocompile
    int res = 0;
    if ((autocompile) && (need_compilation(filename)) && (Concept_Compile))
        res = Concept_Compile(filename, Include, Library, (void *)Print, 0, -1, 0, 0, 0, 0, 0, -1, -1, -1, -1);

    long APID       = 0;
    long parentAPID = -1;
    bool debug      = false;
    //char *key=0;
    AnsiString my_key;

    if ((allow_parents) || (allow_debug))
        if ((my_key.LoadFile(AnsiString(filename) + KEY_EXTENSION) != 0) || (!my_key.Length()))
            my_key.LoadFile(KEY_FILE);

    if (handle_cgi) {
        ec->client_sock = Connect(host, port);
        NotifyServer(ec->client_sock, my_key.c_str(), &ec->c_pipew, &ec->c_piper, &APID, &parentAPID, &debug, userdata);
    }
    if (!allow_debug)
        debug = false;
    else
    if (!allow_parents)
        parentAPID = -1;


    if (!res) {
        int pipe_in  = ec->c_piper ? ec->c_piper : -1;
        int pipe_out = ec->c_pipew ? ec->c_pipew : -1;

        int  apid     = APID ? APID : -1;
        void *debuger = 0;

        if (debug) {
            debuger   = InitDebuger(filename, pipe_out, pipe_in, parentAPID);
            run_debug = 1;
            res       = Concept_Execute(filename, Include, Library, (void *)Print, 0, -2, CONCEPT_DEBUGGER_TRAP, debuger, 0, 0, 0, pipe_in, pipe_out, apid, parentAPID, (void *)userdata);
        } else
            res = Concept_Execute(filename, Include, Library, (void *)Print, 0, -1, 0, 0, 0, 0, 0, pipe_in, pipe_out, apid, parentAPID, (void *)userdata);
        semp(ec->sem_lock);
        ec->done_thread = 1;
        semv(ec->sem_lock);
        if (res == -3) {
            printf("Content-Type: text/html\n\n");
            printf("<HEAD><TITLE>404 Not Found</TITLE></HEAD>\n");
            printf("<BODY><H1>404 Not Found</H1>\n");
            printf("The requested Concept script/URL was not found on this server.\n");
            printf("</BODY>\n");
            final_res = -1;
        }
    }

    if (handle_cgi)
        Done(userdata);

    return(0); ///*res*/final_res;
}

int ModExecute(char *filename, void *userdata) {
    EnvContainer *ec = (EnvContainer *)userdata;

    ec->done_thread = 0;
#ifdef _WIN32
    DWORD threadID;
    CreateThread(NULL, 0, ModExecuteT, (LPVOID)ec, 0, &threadID);
#else
    //pthread *threadID;
    pthread_t threadID;
    pthread_create(&threadID, NULL, ModExecuteT, (LPVOID)ec);
#endif
    int marcseconds = seconds * 100;
    for (int i = 0; i < marcseconds; i++) {
        semp(ec->sem_lock);
        if (ec->done_thread)
            return(0);                   //final_res;
        semv(ec->sem_lock);
#ifdef _WIN32
        Sleep(10);
#else
        usleep(10000);
#endif
    }
    Print("<br /><b>ERROR :</b> Script exceeded maximum execution time of ", -1, ec);
    Print(AnsiString(seconds).c_str(), -1, ec);
    Print(" seconds.\n", -1, ec);
    fprintf(stderr, "Script '%s' exeeded maximum exectuion time of %i seconds.\n", filename, seconds);
    ec->done_thread = 1;
    Done(userdata);
    return(0);
}

int check_file(char *filename) {
    FILE *in = 0;

    //char *filename=getenv("PATH_TRANSLATED");
    in = fopen(filename, "rb");
    if (in) {
        fclose(in);
        return(1);
    }
    return(0);
}

#define CONCEPT_MAGIC_TYPE    "application/x-httpd-concept"

static void add_ssi_vars(request_rec *r) {
    apr_table_t *e = r->subprocess_env;

    if (r->path_info && (r->path_info[0] != '\0')) {
        request_rec *pa_req;

        apr_table_setn(e, "PATH_INFO", ap_escape_shell_cmd(r->pool,
                                                           r->path_info));

        pa_req = ap_sub_req_lookup_uri(ap_escape_uri(r->pool, r->path_info),
                                       r, NULL);
        if (pa_req->filename)
            apr_table_setn(e, "PATH_TRANSLATED",
                           apr_pstrcat(r->pool, pa_req->filename,
                                       pa_req->path_info, NULL));
        else {
            apr_table_setn(e, "PATH_TRANSLATED", r->filename);
            apr_table_setn(e, "PATH_INFO", r->uri);
        }
        ap_destroy_sub_req(pa_req);
    } else {
        apr_table_setn(e, "PATH_TRANSLATED", r->filename);
        apr_table_setn(e, "PATH_INFO", r->uri);
    }

    if (r->args) {
        char *arg_copy = apr_pstrdup(r->pool, r->args);

        apr_table_setn(e, "QUERY_STRING", r->args);
        ap_unescape_url(arg_copy);
        apr_table_setn(e, "QUERY_STRING_UNESCAPED",
                       ap_escape_shell_cmd(r->pool, arg_copy));
    }
}

static int mod_concept_method_handler(request_rec *r) {
    // Send a message to stderr (apache redirects this to the error log)
    // We need to flush the stream so that the message appears right away.
    // Performing an fflush() in a production system is not good for
    // performance - don't do this for real.

    if (strcmp(r->handler, CONCEPT_MAGIC_TYPE))
        // not for us to execute
        return(DECLINED);
         /*if (!(r->finfo.protection & APR_UEXECUTE)) {
          *      fprintf(stderr,"mod_concept: File '%s' lacks execution rights.\n", r->uri);
          *      return DECLINED;
          * }*/

    if ((r->finfo.filetype == 0) || (!check_file(r->filename))) {
        fprintf(stderr, "script '%s' not found or unable to stat\n", r->uri);
        return(HTTP_NOT_FOUND);
    }
    if (r->finfo.filetype == APR_DIR) {
        fprintf(stderr, "attempt to invoke directory '%s' as script\n", r->uri);
        return(HTTP_FORBIDDEN);
    }

    ap_add_common_vars(r);
    ap_add_cgi_vars(r);
    add_ssi_vars(r);

    EnvContainer *ec = new EnvContainer();
    ec->r      = r;
    ec->getenv = GetEnvEmulator;
    ec->header = SetHeaders;
    ec->print  = Print;
    ec->fgetc  = fgetcEmulator;
    ec->fgets  = fgetsEmulator;

    ec->client_sock     = -1;
    ec->c_piper         = 0;
    ec->c_pipew         = 0;
    ec->done_thread     = 0;
    ec->filename        = r->filename;
    ec->post_data       = 0;
    ec->post_data_len   = 0;
    ec->post_data_index = 0;
    getPOST3(ec);
    seminit(ec->sem_lock, 1);

    SetHeaders("Content-type", "text/html", ec);
    int res = ModExecute(r->filename, ec);
    semdel(ec->sem_lock);
    if (ec->post_data)
        delete[] ec->post_data;
    delete ec;
    if (res)
        fprintf(stderr, "mod_concept: Error executing script (%s), code: %i\n", r->filename, res);

    return(OK);
}

static int concept_pre_config(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp) {
    ModInit();
    return(OK);
}

/*static int concept_apache_server_startup(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
 * {
 *      return OK;
 * } */

static void mod_concept_register_hooks(apr_pool_t *p) {
    // I think this is the call to make to register a handler for method calls (GET PUT et. al.).
    // We will ask to be last so that the comment has a higher tendency to
    // go at the end.
    ap_hook_pre_config(concept_pre_config, NULL, NULL, APR_HOOK_MIDDLE);
    //ap_hook_post_config(concept_apache_server_startup, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(mod_concept_method_handler, NULL, NULL, APR_HOOK_MIDDLE);
    //ap_hook_child_init(php_apache_child_init, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA concept_module =
{
    // Only one callback function is provided.  Real
    // modules will need to declare callback functions for
    // server/directory configuration, configuration merging
    // and other tasks.
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    mod_concept_register_hooks                          /* callback for registering hooks */
};

#ifdef _WIN32
// DLL Entry Point ... nothing ...
int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void *lpReserved) {
    return(1);
}

#endif
