#ifndef __CONCEPTINTERPRETER_H
#define __CONCEPTINTERPRETER_H

#include "Optimizer.h"
#include "Debugger.h"
#include "Array.h"
#include "ConceptPools.h"

//---------------------------------------------------------
#define FREE_VARIABLE(VARIABLE)                                                                \
    VARIABLE->LINKS--;                                                                         \
    if (VARIABLE->LINKS < 1) {                                                                 \
        if (VARIABLE->CLASS_DATA) {                                                            \
            if ((VARIABLE->TYPE == VARIABLE_CLASS) || (VARIABLE->TYPE == VARIABLE_DELEGATE)) { \
                if (!--((CompiledClass *)VARIABLE->CLASS_DATA)->LINKS) {                       \
                    delete (CompiledClass *)VARIABLE->CLASS_DATA; }                            \
            } else                                                                             \
            if (VARIABLE->TYPE == VARIABLE_ARRAY) {                                            \
                if (!--((Array *)VARIABLE->CLASS_DATA)->LINKS) {                               \
                    delete (Array *)VARIABLE->CLASS_DATA; }                                    \
            }                                                                                  \
        }                                                                                      \
        delete VARIABLE;                                                                       \
    }

#define FREE_VARIABLE_NO_OBJECTS(VARIABLE) \
    VARIABLE->LINKS--;                     \
    if (VARIABLE->LINKS < 1) {             \
        delete VARIABLE; }
//---------------------------------------------------------
#define MAX_STACK_LIST    0x100
typedef struct tsSCStack {
    AnsiString     *cls;
    TinyString     *mem;
    unsigned short line;
    POOLED(tsSCStack)
} SCStack;

#define STACK(_cls, _mem, _caller_line)                                                                          \
    {                                                                                                            \
        if (ConceptInterpreter::CALL_STACK_COUNT < MAX_STACK_LIST) {                                             \
            ConceptInterpreter::CALL_STACK_COUNT++;                                                              \
            if (ConceptInterpreter::CALL_STACK_COUNT > 1) {                                                      \
                ConceptInterpreter::CALL_STACK [ConceptInterpreter::CALL_STACK_COUNT - 2].line = _caller_line; } \
        }                                                                                                        \
        SCStack *SCtarget = &ConceptInterpreter::CALL_STACK [ConceptInterpreter::CALL_STACK_COUNT - 1];          \
        SCtarget->cls  = &_cls;                                                                                  \
        SCtarget->mem  = &_mem;                                                                                  \
        SCtarget->line = 0;                                                                                      \
    }

#define UNSTACK    if (ConceptInterpreter::CALL_STACK_COUNT > 0) { ConceptInterpreter::CALL_STACK_COUNT--; }


class ConceptInterpreter {
    friend class ClassMember;
    friend class ClassCode;
    friend void *GetStaticData();

    friend void SetStaticData(void *);

private:
    Optimizer *OPT;
    //VariableDATA **LOCAL_CONTEXT;
    STATIC_FLAG SimpleStream *out;
    STATIC_FLAG AnsiList     *ClassList;
    //INTEGER ClassID;
    INTEGER      LocalClassID;
    ClassMember  *OWNER;
    VariableDATA *THROW_DATA;
    INTEGER      CATCH_INSTRUCTION_POINTER;
    INTEGER      CATCH_VARIABLE;
    INTEGER      PREVIOUS_TRY;

    ParamList *OPERATOR_PARAM;

    STATIC_FLAG DEBUGGER_CALLBACK DEBUGGER_TRAP;
    STATIC_FLAG void              *DEBUGGER_RESERVED;
    STATIC_FLAG char              *SERVER_PUBLIC_KEY;
    STATIC_FLAG char              *SERVER_PRIVATE_KEY;
    STATIC_FLAG char              *CLIENT_PUBLIC_KEY;
    char static_call_main;
public:
    POOLED(ConceptInterpreter)
    THREADED_FLAG static SCStack CALL_STACK [MAX_STACK_LIST];
    THREADED_FLAG static unsigned short int CALL_STACK_COUNT;
    void AcknoledgeRunTimeError(AnsiException *Exc);

    ConceptInterpreter(Optimizer *O, SimpleStream *Out, AnsiList *CLST, INTEGER LocalClsID, ClassMember *Owner, DEBUGGER_CALLBACK DC, void *DEBUGGER_RESERVED, char *SPubK, char *SPrivK, char *CPubK);
    VariableDATA **CreateEnviroment(VariableDATA *Sender, ParamList *FORMAL_PARAM, VariableDATA **SenderCTX);
    void DestroyEnviroment(VariableDATA **LOCAL_CONTEXT);
    VariableDATA *Interpret(VariableDATA **LOCAL_CONTEXT, long ClassID);

    ~ConceptInterpreter(void);
};

#endif // __CONCEPTINTERPRETER_H
