#ifndef __CONCEPT_POOLS_H
#define __CONCEPT_POOLS_H
#include <stdlib.h>
#include <stdio.h>

#ifdef USE_BOOST_LIBRARY
 #include <boost/pool/object_pool.hpp>

 #define POOLED(PCLS)                                                                       \
    static boost::object_pool<PCLS> POOL;                                                   \
    inline void PCLS::operator delete(void *p, size_t size) { PCLS::POOL.free((PCLS *)p); } \
    inline void *PCLS::operator new(size_t size) { return PCLS::POOL.malloc(); }

 #define POOLED_IMPLEMENTATION(PCLS)    boost::object_pool<PCLS> PCLS::POOL;
#else
 #define POOLED(PCLS)
 #define POOLED_IMPLEMENTATION(PCLS)
#endif


#define concept_fread     cached_fread
#define concept_fwrite    fwrite
#define concept_fseek     cached_fseek
#define concept_fopen     cached_fopen
#define concept_fclose    cached_fclose
#define concept_FILE      CachedFILE

struct CachedFILE {
    FILE *in;

    char sbuf [8192];
    int  buf_fill;
    int  buf_pos;
};

int cached_fread(void *buf, size_t size, size_t len, CachedFILE *in);
int cached_fseek(CachedFILE *in, long int offset, int origin);
CachedFILE *cached_fopen(const char *filename, const char *mode);
int cached_fclose(CachedFILE *in);

#endif // __CONCEPT_POOLS_H
