#define CACHE_MEMBERS

#include "ModuleLink.h"
#include "AnsiException.h"
#include "Array.h"
#include "ConceptInterpreter.h"
#include "library.h"
#include <string.h>

#ifdef _WIN32
 #include <windows.h>    // LoadLibrary
#else
 #include <stdarg.h>
 #include <dlfcn.h>      // dlopen
#endif

#ifdef CACHE_MEMBERS
#include <map>
#include "Codes.h"
static std::map<unsigned int, SYS_INT> FunctionMap;
#endif

INTEGER ImportModule(AnsiString& MODULE_MASK, AnsiList *Errors, INTEGER line, AnsiString FILENAME, AnsiList *TARGET, PIFAlizator *Sender, INTEGER _NO_ERROR_REPORT) {
#ifdef _WIN32
    // windows module
    HANDLE hLIBRARY = LoadLibrary(MODULE_MASK + MODULE_EXTENSION);
#else
    // linux module
    HANDLE hLIBRARY = dlopen(MODULE_MASK + MODULE_EXTENSION, RTLD_LAZY);
#endif
    if (!hLIBRARY) {
        if (!_NO_ERROR_REPORT)
#ifdef _WIN32
        {
            Errors->Add(new AnsiException(ERR670, line, 670, MODULE_MASK, FILENAME), DATA_EXCEPTION);
        }

#else
        {
            Errors->Add(new AnsiException(AnsiString(ERR670) + AnsiString(": ") + (char *)dlerror(), line, 670, MODULE_MASK, FILENAME), DATA_EXCEPTION);
        }
#endif

        else {
            return(0);
        }
    } else {
        /*INTEGER *HMOD=new INTEGER;
         * *HMOD=(INTEGER)hLIBRARY;
         * TARGET->Add(HMOD,DATA_32_BIT);*/
        TARGET->Add(hLIBRARY, DATA_HANDLER);
        // --------------- Module Management functions -----------------------------------------//
        AnsiString PREFIX = (char *)"_";
#ifdef _WIN32
        CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hLIBRARY, ON_CREATE);
#else
        CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hLIBRARY, ON_CREATE);
#endif

        if (!_PROC_ADR) {
#ifdef _WIN32
            _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hLIBRARY, PREFIX + ON_CREATE);
#else
            _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hLIBRARY, PREFIX + ON_CREATE);
#endif
        }

        if (_PROC_ADR) {
            char *LOAD_RESULT = (char *)_PROC_ADR((SYS_INT)hLIBRARY, Sender, Invoke);
            if (LOAD_RESULT) {
                Errors->Add(new AnsiException(ERR880, line, 880, LOAD_RESULT, FILENAME), DATA_EXCEPTION);
            }
        }
        // -------------------------------------------------------------------------------------//
    }
    return(1);
}

INTEGER UnImportMoudule(HMODULE hMODULE) {
    // --------------- Module Management functions -----------------------------------------//
    AnsiString PREFIX = (char *)"_";

#ifdef _WIN32
    CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hMODULE, ON_DESTROY);
#else
    CONCEPT_MANAGEMENT_CALL _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hMODULE, ON_DESTROY);
#endif


    PREFIX += ON_DESTROY;
    if (!_PROC_ADR) {
#ifdef _WIN32
        _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)GetProcAddress((HMODULE)hMODULE, PREFIX.c_str() /*+ON_DESTROY*/);
#else
        _PROC_ADR = (CONCEPT_MANAGEMENT_CALL)dlsym((HMODULE)hMODULE, PREFIX.c_str() /*+ON_DESTROY*/);
#endif
    }

    if (_PROC_ADR) {
        char *LOAD_RESULT = (char *)_PROC_ADR((SYS_INT)hMODULE, 0, Invoke);
    }
    // TO DO : eventual adauga eroarea de la unload in lista de erori ...
    // -------------------------------------------------------------------------------------//

#ifdef _WIN32
    // windows module
    return(FreeLibrary((HMODULE)hMODULE));
#else
    // linux module
    return(dlclose((HMODULE)hMODULE));
#endif
    return(0);
}

INTEGER SetVariable(VariableDATA *VD, INTEGER TYPE, char *STRING_VALUE, NUMBER NUMBER_VALUE) {
    //---------------//
    CLASS_CHECK(VD);
    //---------------//
    char copy_buffer = 0;
    if (TYPE == _LINK__STRING_DATA) {
        TYPE        = VARIABLE_STRING;
        copy_buffer = 1;
        if (!(INTEGER)NUMBER_VALUE) {
            STRING_VALUE = 0;
        }
    }
    VD->TYPE = TYPE;
    if (VD->TYPE == VARIABLE_STRING) {
        if (!NUMBER_VALUE) {
            VD->STRING_DATA = STRING_VALUE;
        } else {
            if (copy_buffer) {
                VD->STRING_DATA.LinkBuffer(STRING_VALUE, (INTEGER)NUMBER_VALUE);
            } else {
                VD->STRING_DATA.LoadBuffer(STRING_VALUE, (INTEGER)NUMBER_VALUE);
            }
        }
        VD->NUMBER_DATA = 0;
    } else
    if ((VD->TYPE == VARIABLE_CLASS) || (VD->TYPE == VARIABLE_ARRAY) || (VD->TYPE == VARIABLE_DELEGATE)) {
        VD->CLASS_DATA  = (void *)STRING_VALUE;
        VD->NUMBER_DATA = NUMBER_VALUE;
        if ((TYPE == VARIABLE_CLASS) || (TYPE == VARIABLE_DELEGATE)) {
            ((CompiledClass *)VD->CLASS_DATA)->LINKS++;
        } else {
            ((Array *)VD->CLASS_DATA)->LINKS++;
        }
    } else {
        VD->NUMBER_DATA = NUMBER_VALUE;
    }
    return(0);
}

INTEGER GetVariable(VariableDATA *VD, INTEGER *TYPE, char **STRING_VALUE, NUMBER *NUMBER_VALUE) {
    *TYPE         = VD->TYPE;
    *STRING_VALUE = VD->STRING_DATA.c_str();
    if (VD->TYPE == VARIABLE_STRING) {
        *NUMBER_VALUE = VD->STRING_DATA.Length();
    } else
    if ((VD->TYPE == VARIABLE_CLASS) || (VD->TYPE == VARIABLE_ARRAY)) {
        *STRING_VALUE = (char *)VD->CLASS_DATA;
        *NUMBER_VALUE = 0;
    } else
    if (VD->TYPE == VARIABLE_DELEGATE) {
        *STRING_VALUE = (char *)VD->CLASS_DATA;
        *NUMBER_VALUE = VD->NUMBER_DATA;
    } else {
        *NUMBER_VALUE = VD->NUMBER_DATA;
    }
    return(0);
}

INTEGER GetClassMember(void *CLASS_PTR, char *class_member_name, INTEGER *TYPE, char **STRING_VALUE, NUMBER *NUMBER_VALUE) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        //if ((CCode) && (CONTEXT)) {
        if (CCode) {
            INTEGER m_type = 0;
            INTEGER index  = CCode->HasMemberInCompiled(class_member_name, &m_type);
            if (index) {
                index--;

                INTEGER index2 = CCode->HasMember(class_member_name);
                index2--;
                ClassMember *CM = (ClassMember *)CCode->Members->Item(index2);

                if ((CM) && (CM->IS_FUNCTION)) {
                    index = CCode->GetAbsoluteMemberID(index + 1) - 1;
                    //VariableDATA
                    ParamList PLIST;
                    PLIST.COUNT = 0;

                    AnalizerElement AE;
                    AE._DEBUG_INFO_LINE = AE.ID = AE.TYPE = 0;

                    //int is_property=CM->IS_FUNCTION==3 ? 1 : 0;


                    VariableDATA *Owner = new VariableDATA;
                    //Owner->ACCESS=ACCESS_LOCAL;
                    Owner->CALL_SET           = 0;
                    Owner->CLASS_DATA         = CLASS_PTR;
                    Owner->EXTRA_DATA         = 0;
                    Owner->IS_PROPERTY_RESULT = 0;
                    Owner->LINKS       = 1;
                    Owner->NUMBER_DATA = 0;
                    Owner->TYPE        = VARIABLE_CLASS;
                    ((CompiledClass *)CLASS_PTR)->LINKS++;

                    try {
                        VariableDATA *VarDATA = CCode->ExecuteMember(
                            index,
                            Owner,
                            &AE,
                            true,
                            CM->IS_FUNCTION == 3 ? 0 : &PLIST,
                            (VariableDATA **)-1,
                            0,
                            CCode->CLSID,
                            CCode->CLSID
                            );
                        int result = GetVariable(VarDATA, TYPE, STRING_VALUE, NUMBER_VALUE);
                        FREE_VARIABLE(VarDATA);
                        FREE_VARIABLE(Owner);
                        return(result);
                    } catch (VariableDATA *THROW_DATA) {
                        FREE_VARIABLE(THROW_DATA);
                        FREE_VARIABLE(Owner);
                        return(-1);
                    }
                }
                index = CCode->RELOCATIONS2 [index] - 1;
                if (index >= 0) {
                    if ((!m_type) && (!CONTEXT [index])) {
                        VariableDATA *RESULT = new VariableDATA;
                        RESULT->CLASS_DATA         = 0;
                        RESULT->LINKS              = 1;
                        RESULT->NUMBER_DATA        = 0;
                        RESULT->TYPE               = VARIABLE_NUMBER;
                        RESULT->IS_PROPERTY_RESULT = 0;
                        RESULT->CALL_SET           = 0;
                        RESULT->EXTRA_DATA         = 0;
                        CONTEXT [index]            = RESULT;
                    }
                    if (CONTEXT [index]) {
                        return(GetVariable(CONTEXT [index], TYPE, STRING_VALUE, NUMBER_VALUE));
                    }
                } else {
                    return(-1);
                }
            }
        }
    }
    return(-1);
}

INTEGER GetClassMemberVariable(void *CLASS_PTR, char *class_member_name, void **ptr) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        if ((CCode) && (CONTEXT)) {
            INTEGER m_type = 0;
            INTEGER index  = CCode->HasMemberInCompiled(class_member_name, &m_type);
            if (index) {
                index--;
                index = CCode->RELOCATIONS2 [index] - 1;
                if ((!m_type) && (!CONTEXT [index])) {
                    VariableDATA *RESULT = new VariableDATA;
                    RESULT->CLASS_DATA         = 0;
                    RESULT->LINKS              = 1;
                    RESULT->NUMBER_DATA        = 0;
                    RESULT->TYPE               = VARIABLE_NUMBER;
                    RESULT->IS_PROPERTY_RESULT = 0;
                    RESULT->CALL_SET           = 0;
                    RESULT->EXTRA_DATA         = 0;
                    CONTEXT [index]            = RESULT;
                }
                if (CONTEXT [index]) {
                    *ptr = CONTEXT [index];
                }
            }
            return(0);
        }
    }
    return(-1);
}

INTEGER SetClassMember(void *CLASS_PTR, char *class_member_name, INTEGER TYPE, char *STRING_VALUE, NUMBER NUMBER_VALUE) {
    if (CLASS_PTR) {
        ClassCode    *CCode    = ((CompiledClass *)CLASS_PTR)->GetClass();
        VariableDATA **CONTEXT = ((CompiledClass *)CLASS_PTR)->GetContext();
        //if ((CCode) && (CONTEXT)) {
        if (CCode) {
            INTEGER m_type = 0;
            INTEGER index  = CCode->HasMemberInCompiled(class_member_name, &m_type);
            if (index) {
                index--;

                INTEGER index2 = CCode->HasMember(class_member_name);
                index2--;
                ClassMember *CM = (ClassMember *)CCode->Members->Item(index2);

                if ((CM) && (CM->IS_FUNCTION == 3) && (CM->MEMBER_SET)) {
                    index = CCode->GetAbsoluteMemberID(index + 1) - 1;

                    AnalizerElement AE;
                    AE._DEBUG_INFO_LINE = AE.ID = AE.TYPE = 0;

                    //int is_property=CM->IS_FUNCTION==3 ? 1 : 0;

                    VariableDATA *Parameter = new VariableDATA;
                    //Parameter->ACCESS=ACCESS_LOCAL;
                    Parameter->NUMBER_DATA        = NUMBER_VALUE;
                    Parameter->CALL_SET           = 0;
                    Parameter->EXTRA_DATA         = 0;
                    Parameter->IS_PROPERTY_RESULT = 0;
                    Parameter->LINKS       = 1;
                    Parameter->NUMBER_DATA = NUMBER_VALUE;
                    Parameter->TYPE        = TYPE;

                    VariableDATA *CONTAINER [1];
                    CONTAINER [0] = Parameter;

                    if ((TYPE == VARIABLE_CLASS) || (TYPE == VARIABLE_DELEGATE)) {
                        Parameter->CLASS_DATA = (void *)STRING_VALUE;
                        ((CompiledClass *)CLASS_PTR)->LINKS++;
                    } else
                    if (TYPE == VARIABLE_ARRAY) {
                        Parameter->CLASS_DATA = (void *)STRING_VALUE;
                        ((Array *)CLASS_PTR)->LINKS++;
                    } else
                    if (TYPE == VARIABLE_STRING) {
                        Parameter->STRING_DATA = STRING_VALUE;
                    }

                    VariableDATA *Owner = new VariableDATA;
                    //Owner->ACCESS=ACCESS_LOCAL;
                    Owner->CALL_SET           = 0;
                    Owner->CLASS_DATA         = CLASS_PTR;
                    Owner->EXTRA_DATA         = 0;
                    Owner->IS_PROPERTY_RESULT = 0;
                    Owner->LINKS       = 1;
                    Owner->NUMBER_DATA = 0;
                    Owner->TYPE        = VARIABLE_CLASS;
                    ((CompiledClass *)CLASS_PTR)->LINKS++;

                    try {
                        CCode->SetProperty(
                            index,
                            Owner,
                            &AE,
                            true,
                            1,
                            CONTAINER,
                            CCode->CLSID,
                            CCode->CLSID
                            );

                        FREE_VARIABLE(Owner);
                        FREE_VARIABLE(Parameter);
                        return(0);
                    } catch (VariableDATA *THROW_DATA) {
                        FREE_VARIABLE(THROW_DATA);
                        FREE_VARIABLE(Owner);
                        FREE_VARIABLE(Parameter);
                        return(-1);
                    }
                }
            }
            // added if (CONTEXT)
            index = CCode->RELOCATIONS2 [index] - 1;
            if (index >= 0) {
                if ((CONTEXT) && (!m_type) && (!CONTEXT [index])) {
                    VariableDATA *RESULT = new VariableDATA;
                    RESULT->CLASS_DATA         = 0;
                    RESULT->LINKS              = 1;
                    RESULT->NUMBER_DATA        = 0;
                    RESULT->TYPE               = VARIABLE_NUMBER;
                    RESULT->IS_PROPERTY_RESULT = 0;
                    RESULT->CALL_SET           = 0;
                    RESULT->EXTRA_DATA         = 0;
                    CONTEXT [index]            = RESULT;
                }
                if (CONTEXT [index]) {
                    return(SetVariable(CONTEXT [index], TYPE, STRING_VALUE, NUMBER_VALUE));
                }
            } else {
                return(-1);
            }
        }
    }
    return(-1);
}

INTEGER Invoke(INTEGER INVOKE_TYPE, ...) {
    va_list ap;

    va_start(ap, INVOKE_TYPE);
    INTEGER result = INVOKE_SUCCESS;
    switch (INVOKE_TYPE) {
        case INVOKE_NEW_BUFFER:
            {
                INTEGER size  = va_arg(ap, INTEGER);
                char    **str = va_arg(ap, char **);
                *str = 0;
                if (size) {
                    *str = (char *)malloc(size);               //new char[size];
                    if (!(*str)) {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                }
            }
            break;

        case INVOKE_DELETE_BUFFER:
            {
                char *str = va_arg(ap, char *);
                if (str) {
                    free(str);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    //delete[] str;
                } else                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        {
                    result = CANNOT_INVOKE_IN_THIS_CASE;
                }
            }
            break;

        case INVOKE_SET_VARIABLE:
            {
                VariableDATA *vdata = va_arg(ap, VariableDATA *);
                INTEGER      type   = va_arg(ap, INTEGER);
                char         *str   = va_arg(ap, char *);
                NUMBER       nvalue = va_arg(ap, NUMBER);
                result = SetVariable(vdata, type, str, nvalue);
            }
            break;

        case INVOKE_GET_VARIABLE:
            {
                VariableDATA *vdata  = va_arg(ap, VariableDATA *);
                INTEGER      *type   = va_arg(ap, INTEGER *);
                char         **str   = va_arg(ap, char **);
                NUMBER       *nvalue = va_arg(ap, NUMBER *);
                result = GetVariable(vdata, type, str, nvalue);
            }
            break;

        case INVOKE_SET_CLASS_MEMBER:
            {
                void    *handle  = va_arg(ap, void *);
                char    *mname   = va_arg(ap, char *);
                INTEGER type     = va_arg(ap, INTEGER);
                char    *szvalue = va_arg(ap, char *);
                NUMBER  nvalue   = va_arg(ap, NUMBER);
                result = SetClassMember(handle, mname, type, szvalue, nvalue);
            }
            break;

        case INVOKE_GET_CLASS_MEMBER:
            {
                void    *handle   = va_arg(ap, void *);
                char    *mname    = va_arg(ap, char *);
                INTEGER *type     = va_arg(ap, INTEGER *);
                char    **szvalue = va_arg(ap, char **);
                NUMBER  *nvalue   = va_arg(ap, NUMBER *);
                result = GetClassMember(handle, mname, type, szvalue, nvalue);
                break;
            }
            break;

        case INVOKE_GET_CLASS_VARIABLE:
            {
                VariableDATA *handle = va_arg(ap, VariableDATA *);
                if ((handle) && (handle->TYPE == VARIABLE_CLASS)) {
                    char *mname = va_arg(ap, char *);
                    void **ptr  = va_arg(ap, void **);
                    result = GetClassMemberVariable(handle->CLASS_DATA, mname, ptr);
                }
                break;
            }
            break;

        case INVOKE_LOCK_VARIABLE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    target->LINKS++;
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_DYNAMIC_LOCK:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    if ((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE)) {
                        ((CompiledClass *)target->CLASS_DATA)->LINKS++;
                    } else
                    if (target->TYPE == VARIABLE_ARRAY) {
                        ((Array *)target->CLASS_DATA)->LINKS++;
                    } else {
                        result = INVALID_INVOKE_PARAMETER;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CREATE_VARIABLE:
            {
                VariableDATA **target = va_arg(ap, VariableDATA * *);
                *target = new VariableDATA;
                //(*target)->ACCESS=ACCESS_LOCAL;
                (*target)->CALL_SET           = 0;
                (*target)->CLASS_DATA         = 0;
                (*target)->EXTRA_DATA         = 0;
                (*target)->IS_PROPERTY_RESULT = 0;
                (*target)->LINKS       = 1;
                (*target)->NUMBER_DATA = 0;
                (*target)->TYPE        = VARIABLE_NUMBER;
            }
            break;

        case INVOKE_FREE_VARIABLE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    FREE_VARIABLE(target);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CREATE_ARRAY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                CLASS_CHECK(target);
                target->TYPE       = VARIABLE_ARRAY;
                target->CLASS_DATA = new Array();
            }
            break;

        case INVOKE_CREATE_OBJECT:
            {
                PIFAlizator  *pif        = va_arg(ap, PIFAlizator *);
                VariableDATA *target     = va_arg(ap, VariableDATA *);
                char         *class_name = va_arg(ap, char *);
                CLASS_CHECK(target);
                target->TYPE = VARIABLE_NUMBER;

                int       class_count = pif->ClassList.Count();
                int       clsid       = -1;
                ClassCode *CC         = 0;
                for (int i = 0; i < class_count; i++) {
                    CC = (ClassCode *)pif->ClassList.Item(i);
                    if (CC->NAME == class_name) {
                        clsid = i;
                        break;
                    }
                }
                if (clsid == -1) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }

                // am clsid ... mai trebuie sa instantziez clasa ...
                ParamList FORMAL_PARAM;
                FORMAL_PARAM.COUNT       = 0;
                FORMAL_PARAM.PARAM_INDEX = 0;

                VariableDATA dummyVD [1];

                target->TYPE       = VARIABLE_CLASS;
                target->CLASS_DATA = CC->CreateInstance(target, 0, &FORMAL_PARAM, (VariableDATA **)&dummyVD);
                if (!target->CLASS_DATA) {
                    target->TYPE = VARIABLE_NUMBER;
                    result       = INVALID_INVOKE_PARAMETER;
                    break;
                }

                target->TYPE = VARIABLE_CLASS;
            }
            break;

        case INVOKE_ARRAY_VARIABLE:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                VariableDATA **res   = va_arg(ap, VariableDATA * *);
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    // TO DO !
                    *res = ((Array *)target->CLASS_DATA)->ModuleGet(index);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_ARRAY_VARIABLE_BY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                VariableDATA **res   = va_arg(ap, VariableDATA * *);
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    // TO DO !
                    *res = ((Array *)target->CLASS_DATA)->ModuleGet(key);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_ELEMENT:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                void         *res    = 0;
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    // TO DO !
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(index);
                    INTEGER *type = va_arg(ap, INTEGER *);
                    char    **str = va_arg(ap, char **);
                    NUMBER  *nr   = va_arg(ap, double *);
                    GetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_ELEMENT_BY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                void         *res    = 0;
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    // TO DO !
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(key);
                    INTEGER *type = va_arg(ap, INTEGER *);
                    char    **str = va_arg(ap, char **);
                    NUMBER  *nr   = va_arg(ap, double *);
                    GetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_SET_ARRAY_ELEMENT:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                void         *res    = 0;
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(index);

                    INTEGER type = va_arg(ap, INTEGER);
                    char    *str = va_arg(ap, char *);
                    NUMBER  nr   = va_arg(ap, double);
                    SetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_SET_ARRAY_ELEMENT_BY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                void         *res    = 0;
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    res = ((Array *)target->CLASS_DATA)->ModuleGet(key);

                    INTEGER type = va_arg(ap, INTEGER);
                    char    *str = va_arg(ap, char *);
                    NUMBER  nr   = va_arg(ap, double);
                    SetVariable((VariableDATA *)res, type, str, nr);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_KEY:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                char         **key   = va_arg(ap, char **);
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    *key = ((Array *)target->CLASS_DATA)->GetKey(index);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_ARRAY_ELEMENT_IS_SET:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                INTEGER      index   = va_arg(ap, INTEGER);
                char         *key    = va_arg(ap, char *);
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    if (index >= 0) {
                        result = ((Array *)target->CLASS_DATA)->FindIndex(index) != -1;
                    } else {
                        AnsiString k(key);
                        result = ((Array *)target->CLASS_DATA)->FindKey(&k) != -1;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_INDEX:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                char         *key    = va_arg(ap, char *);
                INTEGER      *index  = va_arg(ap, INTEGER *);
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    *index = ((Array *)target->CLASS_DATA)->GetIndex(key);
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_ARRAY_COUNT:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if ((target->TYPE == VARIABLE_ARRAY) && (target->CLASS_DATA)) {
                    result = ((Array *)target->CLASS_DATA)->Count();
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CALL_DELEGATE:
            {
                VariableDATA *target            = va_arg(ap, VariableDATA *);
                VariableDATA **SENDER_RESULT    = va_arg(ap, VariableDATA * *);
                VariableDATA **SENDER_EXCEPTION = va_arg(ap, VariableDATA * *);

                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    INTEGER      parameters = va_arg(ap, INTEGER);
                    VariableDATA **CTX      = 0;
                    bool         is_list    = false;
                    if (parameters == -1) {
                        CTX = va_arg(ap, VariableDATA * *);
                        int i = 0;
                        while (CTX [i++]) {
                        }
                        parameters = i - 1;
                        is_list    = true;
                    } else {
                        // atention HERE !!! CTX is -1 to not be similar to the call from Main !!!
                        CTX = parameters ? new VariableDATA * [parameters] : (VariableDATA **)-1;               //0;
                    }
                    ParamList FORMAL_PARAM;
                    FORMAL_PARAM.COUNT       = parameters;
                    FORMAL_PARAM.PARAM_INDEX = parameters ? DELTA_REF(&FORMAL_PARAM, new INTEGER [parameters]) : 0;

                    if (is_list) {
                        for (int i = 0; i < parameters; i++) {
                            DELTA_UNREF(&FORMAL_PARAM, FORMAL_PARAM.PARAM_INDEX) [i] = i + 1;
                            CTX [i]->LINKS++;
                        }
                    } else {
                        for (int i = 0; i < parameters; i++) {
                            DELTA_UNREF(&FORMAL_PARAM, FORMAL_PARAM.PARAM_INDEX) [i] = i + 1;
                            INTEGER type = va_arg(ap, INTEGER);
                            if (type == VARIABLE_UNDEFINED) {
                                CTX [i] = va_arg(ap, VariableDATA *);
                                CTX [i]->LINKS++;
                            } else {
                                char   *str = va_arg(ap, char *);
                                NUMBER nr   = va_arg(ap, double);

                                CTX [i] = new VariableDATA;
                                //CTX[i]->ACCESS=ACCESS_LOCAL;
                                CTX [i]->CALL_SET           = 0;
                                CTX [i]->CLASS_DATA         = 0;
                                CTX [i]->EXTRA_DATA         = 0;
                                CTX [i]->IS_PROPERTY_RESULT = 0;
                                CTX [i]->LINKS       = 1;
                                CTX [i]->NUMBER_DATA = 0;
                                SetVariable(CTX [i], type, str, nr);
                            }
                        }
                    }
                    VariableDATA *RESULT = target;
                    VariableDATA *lOwner = 0;
                    try {
                        lOwner = new VariableDATA;
                        //lOwner->ACCESS=ACCESS_LOCAL;
                        lOwner->CALL_SET           = 0;
                        lOwner->CLASS_DATA         = target->CLASS_DATA;
                        lOwner->EXTRA_DATA         = 0;
                        lOwner->IS_PROPERTY_RESULT = 0;
                        lOwner->LINKS       = 1;
                        lOwner->NUMBER_DATA = 0;
                        lOwner->TYPE        = VARIABLE_CLASS;

                        ((CompiledClass *)RESULT->CLASS_DATA)->LINKS++;

                        *SENDER_RESULT = ((CompiledClass *)RESULT->CLASS_DATA)->_Class->ExecuteMember(
                            (INTEGER)RESULT->NUMBER_DATA,
                            lOwner,
                            0,
                            1,
                            &FORMAL_PARAM,
                            CTX,
                            0,
                            ((CompiledClass *)RESULT->CLASS_DATA)->_Class->CLSID,
                            ((CompiledClass *)RESULT->CLASS_DATA)->_Class->CLSID);
                        FREE_VARIABLE(lOwner);
                    } catch (VariableDATA *LAST_THROW) {
                        FREE_VARIABLE(lOwner);
                        *SENDER_EXCEPTION = LAST_THROW;
                    }

                    if ((CTX) && ((long)CTX != -1)) {
                        for (int i = 0; i < parameters; i++) {
                            FREE_VARIABLE(CTX [i]);
                        }
                        if (!is_list) {
                            delete[] CTX;
                        }
                    }
                    if (FORMAL_PARAM.PARAM_INDEX) {
                        delete[] DELTA_UNREF(&FORMAL_PARAM, FORMAL_PARAM.PARAM_INDEX);
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_COUNT_DELEGATE_PARAMS:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    int mid = ((CompiledClass *)target->CLASS_DATA)->_Class->RELOCATIONS [(long)target->NUMBER_DATA];
                    //ClassMember *CM=((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS[(long)target->NUMBER_DATA];
                    ClassMember *CM = mid ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS [mid - 1] : 0;
                    if (CM) {
                        result = CM->PARAMETERS_COUNT;
                    } else {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_DELEGATE_NAMES:
            {
                VariableDATA *target       = va_arg(ap, VariableDATA *);
                char         **class_name  = va_arg(ap, char **);
                char         **member_name = va_arg(ap, char **);
                if ((target->TYPE == VARIABLE_DELEGATE) && (target->CLASS_DATA)) {
                    int mid = ((CompiledClass *)target->CLASS_DATA)->_Class->RELOCATIONS [(long)target->NUMBER_DATA];
                    //ClassMember *CM=((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS[(long)target->NUMBER_DATA];
                    ClassMember *CM = mid ? ((CompiledClass *)target->CLASS_DATA)->_Class->pMEMBERS [mid - 1] : 0;
                    if (CM) {
                        *class_name  = ((CompiledClass *)target->CLASS_DATA)->_Class->NAME.c_str();
                        *member_name = CM->NAME.c_str();
                    } else {
                        result = CANNOT_INVOKE_IN_THIS_CASE;
                    }
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_DEFINE_CLASS:
            {
                PIFAlizator *pif = va_arg(ap, PIFAlizator *);

                char *classname = va_arg(ap, char *);

                if (pif->ClassExists(classname)) {
                    result = INVALID_INVOKE_PARAMETER;
                    break;
                }

                ClassCode *CC = new ClassCode(classname, pif);

                char *member_name = va_arg(ap, char *);
                char is_array     = 0;
                while (member_name) {
                    int        len = strlen(member_name);
                    AnsiString name;
                    if ((member_name [len - 2] == '[') && (member_name [len - 1] == ']')) {
                        is_array = 1;
                        name.LoadBuffer(member_name, len - 2);
                    } else {
                        name     = member_name;
                        is_array = 0;
                    }

                    ClassMember *CM = CC->AddMember(member_name, 0, "STATIC_MODULE", ACCESS_PUBLIC, true);
                    if (is_array) {
                        CM->VD->TYPE = VARIABLE_ARRAY;
                    }
                    //CM->IS_FUNCTION = 0;
                    pif->CheckMember(/*AnsiString(member_name)*/ name);

                    //member_name=va_arg(ap,char *);
                }
                CC->NEEDED        = 0;
                CC->DEFINED_LEVEL = -1;                //pif->INCLUDE_LEVEL+1;
                pif->ClassList.Add(CC, DATA_CLASS_CODE);
            }
            break;

        case INVOKE_DEFINE_CONSTANT:
            {
                PIFAlizator *pif      = va_arg(ap, PIFAlizator *);
                char        *CONSTANT = va_arg(ap, char *);
                char        *value    = va_arg(ap, char *);

                if ((pif) && (CONSTANT) && (value)) {
                    int ct_count           = pif->ConstantList.Count();
                    VariableDESCRIPTOR *VD = 0;
                    for (int i = 0; i < ct_count; i++) {
                        VariableDESCRIPTOR *VTEMP = (VariableDESCRIPTOR *)pif->ConstantList.Item(i);
                        if (VTEMP->name == CONSTANT) {
                            VD = VTEMP;
                            break;
                        }
                    }
                    if (!VD) {
                        VD         = new VariableDESCRIPTOR;
                        VD->BY_REF = 0;
                        pif->ConstantList.Add(VD, DATA_VAR_DESCRIPTOR);
                    }
                    VD->name  = CONSTANT;
                    VD->value = value;
                    VD->USED  = -1;
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_GET_APPLICATION_INFO:
            {
                PIFAlizator *pif         = va_arg(ap, PIFAlizator *);
                int         *pipein      = va_arg(ap, int *);
                int         *pipeout     = va_arg(ap, int *);
                int         *apid        = va_arg(ap, int *);
                int         *parent_apid = va_arg(ap, int *);

                if ((pif) && (apid) && (pipein) && (pipeout) && (parent_apid)) {
                    *pipein      = pif->pipe_read;
                    *pipeout     = pif->pipe_write;
                    *apid        = pif->apid;
                    *parent_apid = pif->parent_apid;
                }
            }
            break;

        case INVOKE_GET_SERIAL_CLASS:
            {
                CompiledClass *CC             = va_arg(ap, CompiledClass *);
                int           max_members     = va_arg(ap, int);
                char          **class_name    = va_arg(ap, char **);
                char          **pmembers      = va_arg(ap, char **);
                char          *flags          = va_arg(ap, char *);
                char          *access         = va_arg(ap, char *);
                char          *types          = va_arg(ap, char *);
                char          **szValue       = va_arg(ap, char **);
                double        *n_data         = va_arg(ap, double *);
                void          **class_data    = va_arg(ap, void **);
                void          **variable_data = va_arg(ap, void **);

                if (!CC) {
                    result = CANNOT_INVOKE_INTERFACE;
                    break;
                }
                *class_name = CC->_Class->NAME.c_str();
                result      = CC->_Class->GetSerialMembers(CC, max_members, pmembers, flags, access, types, szValue, n_data, class_data, variable_data);
            }
            break;

        case INVOKE_GET_MEMBER_FROM_ID:
            {
                CompiledClass *CC       = va_arg(ap, CompiledClass *);
                long          member_id = va_arg(ap, long);
                char          **member  = va_arg(ap, char **);
                if (!CC) {
                    result = CANNOT_INVOKE_INTERFACE;
                    break;
                }

                if (CC->_Class->pMEMBERS [member_id]) {
                    *member = CC->_Class->pMEMBERS [member_id]->NAME.c_str();
                } else {
                    *member = 0;
                    result  = CANNOT_INVOKE_INTERFACE;
                }
            }
            break;

        case INVOKE_HAS_MEMBER:
            {
                CompiledClass *CC     = va_arg(ap, CompiledClass *);
                char          *member = va_arg(ap, char *);
                if (!CC) {
                    result = CANNOT_INVOKE_INTERFACE;
                    break;
                }

                if (CC->_Class->HasMember(member)) {
                    result = 1;
                } else {
                    result = 0;
                }
            }
            break;

        case INVOKE_OBJECT_LINKS:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target->CLASS_DATA) {
                    if ((target->TYPE == VARIABLE_CLASS) || (target->TYPE == VARIABLE_DELEGATE)) {
                        result = ((CompiledClass *)target->CLASS_DATA)->LINKS;
                    } else
                    if (target->TYPE == VARIABLE_ARRAY) {
                        result = ((Array *)target->CLASS_DATA)->LINKS;
                    }
                }
            }
            break;

        case INVOKE_VAR_LINKS:
            {
                VariableDATA *target = va_arg(ap, VariableDATA *);
                if (target) {
                    result = target->LINKS;
                }
            }
            break;

        case INVOKE_CLI_ARGUMENTS:
            {
                long *argc   = va_arg(ap, long *);
                char ***argv = va_arg(ap, char ***);
                *argv = PIFAlizator::argv;
                *argc = PIFAlizator::argc;
            }
            break;

        case INVOKE_SORT_ARRAY_BY_KEYS:
            {
                Array *target = va_arg(ap, Array *);
                Array **res   = va_arg(ap, Array * *);

                if (target) {
                    *res = target->SortArrayElementsByKey();
                } else {
                    result = INVALID_INVOKE_PARAMETER;
                }
            }
            break;

        case INVOKE_CHECK_POINT:
            {
                int status = va_arg(ap, int);
                if (PIFAlizator::CheckPoint) {
                    PIFAlizator::CheckPoint(status);
                } else {
                    result = CANNOT_INVOKE_INTERFACE;
                }
            }
            break;

        case INVOKE_GET_USERDATA:
            {
                PIFAlizator *pif   = va_arg(ap, PIFAlizator *);
                void        **data = va_arg(ap, void **);
                if (!pif) {
                    result = INVALID_INVOKE_PARAMETER;
                } else {
                    *data = pif->out->userdata;
                }
            }
            break;

        case INVOKE_GET_THREAD_DATA:
            {
                void **data = va_arg(ap, void **);
                *data = GetStaticData();
            }
            break;

        case INVOKE_SET_THREAD_DATA:
            {
                void *data = va_arg(ap, void *);
                SetStaticData(data);
            }
            break;

        case INVOKE_PRINT:
            {
                PIFAlizator *pif  = va_arg(ap, PIFAlizator *);
                char        *what = va_arg(ap, char *);
                INTEGER     len   = va_arg(ap, INTEGER);
                pif->out->Print(what, len);
            }
            break;

        case INVOKE_HIBERNATE:
            {
                PIFAlizator *pif  = va_arg(ap, PIFAlizator *);
                char        *flag = va_arg(ap, char *);
                if (!pif) {
                    result = INVALID_INVOKE_PARAMETER;
                } else {
                    if (*flag) {
                        pif->Hibernate();
                    } else {
                        //pif->HibernateWake();
                    }
                }
            }
            break;

        default:
            return(CANNOT_INVOKE_INTERFACE);
    }
    va_end(ap);
    return(result);
}

SYS_INT LinkFunction(char *FUNCTION_NAME, AnsiList *TARGET, void **CACHED_hDLL) {
#ifdef CACHE_MEMBERS
    unsigned int key=murmur_hash(FUNCTION_NAME, strlen(FUNCTION_NAME));
    SYS_INT val=FunctionMap[key];
    if (val)
        return val;
#endif

    AnsiString PREFIX  = (char *)"CONCEPT__";
    AnsiString PREFIX2 = (char *)"CONCEPT_";
    int        count   = TARGET->Count();

    //return 1;
    if (CACHED_hDLL) {
#ifdef _WIN32
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)GetProcAddress((HMODULE)*CACHED_hDLL, PREFIX2 + FUNCTION_NAME);
#else
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)dlsym((HMODULE)*CACHED_hDLL, PREFIX2 + FUNCTION_NAME);
#endif
        if (!_PROC_ADR) {
#ifdef _WIN32
            _PROC_ADR = (EXTERNAL_CALL)GetProcAddress((HMODULE)*CACHED_hDLL, PREFIX + FUNCTION_NAME);
#else
            _PROC_ADR = (EXTERNAL_CALL)dlsym((HMODULE)*CACHED_hDLL, PREFIX + FUNCTION_NAME);
#endif
        }

        if (_PROC_ADR) {
#ifdef CACHE_MEMBERS
            FunctionMap[key]=(SYS_INT)_PROC_ADR;
#endif
            return((SYS_INT)_PROC_ADR);
        }
    }
    for (INTEGER i = 0; i < count; i++) {
        HMODULE hDLL = (HMODULE)TARGET->Item(i);
#ifdef _WIN32
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)GetProcAddress(hDLL, PREFIX2 + FUNCTION_NAME);
#else
        EXTERNAL_CALL _PROC_ADR = (EXTERNAL_CALL)dlsym(hDLL, PREFIX2 + FUNCTION_NAME);
#endif
        if (!_PROC_ADR) {
#ifdef _WIN32
            _PROC_ADR = (EXTERNAL_CALL)GetProcAddress(hDLL, PREFIX + FUNCTION_NAME);
#else
            _PROC_ADR = (EXTERNAL_CALL)dlsym(hDLL, PREFIX + FUNCTION_NAME);
#endif
        }

        if (_PROC_ADR) {
            *CACHED_hDLL = hDLL;
#ifdef CACHE_MEMBERS
            FunctionMap[key]=(SYS_INT)_PROC_ADR;
#endif
            return((SYS_INT)_PROC_ADR);
        }
    }
    return(0);
}

void DoneLinking() {
#ifdef CACHE_MEMBERS
    FunctionMap.clear();
#endif;
}