#ifndef __OPTIMIZER_H
#define __OPTIMIZER_H

#include "PIFAlizator.h"
#include "ConceptTypes.h"
#include "Codes.h"
#include <stdio.h>
#include "ConceptPools.h"

#define DELTA_REF(base, p)      ((INTEGER *)(((SYS_INT)p) - ((SYS_INT)base)))
#define DELTA_UNREF(base, p)    ((INTEGER *)(p ? ((SYS_INT)base + (SYS_INT)p) : 0))

class Optimizer {
    friend class ConceptInterpreter;
    friend class ClassMember;
    friend class PIFAlizator;
private:
    STATIC_FLAG ClassCode *_CLASS;
    STATIC_FLAG char      *_MEMBER;
    STATIC_FLAG INTEGER   LAST_DEBUG_TRAP;

    STATIC_FLAG AnsiList *PIFList;
    STATIC_FLAG AnsiList *VDList;
    //AnsiList *Errors;
    //AnsiList *Warnings;
    STATIC_FLAG PIFAlizator *PIFOwner;
    /*INTEGERchar DebugOn;*/
    STATIC_FLAG INTEGER PIF_POSITION;
    STATIC_FLAG char    *_DEBUG_INFO_FILENAME;

    STATIC_FLAG char NO_WARNING_EMPTY;
    STATIC_FLAG char NO_WARNING_ATTR;

    AnsiList *OptimizedPIF;

    AnsiList *ParameterList;

    OptimizedElement   *CODE;
    INTEGER            codeCount;
    VariableDESCRIPTOR *DATA;
    INTEGER            dataCount;
    ParamList          *PARAMS;
    INTEGER            paramCount;
    //OptimizedElement    *CONTINUE_Element;
    //OptimizedElement    *BREAK_Element;
    STATIC_FLAG AnsiList *CONTINUE_Elements;
    STATIC_FLAG AnsiList *BREAK_Elements;

    STATIC_FLAG char _clean_condition;

    /*INTEGER*/            //char             ITERATIVE_STRUCTURE;

    // le folosesc pt try in try ...
    STATIC_FLAG INTEGER CATCH_ELEMENT;

    INTEGER OptimizeSwitch();
    INTEGER OptimizeArray();
    INTEGER OptimizeForExpression(bool is_increment = false);
    INTEGER OptimizeExpression(INTEGER ID, INTEGER TYPE, INTEGER IS_PARAM_LIST = 0, INTEGER FIRST_CALL = 1, char FLAGS = 0, AnalizerElement *PREC_METHOD = 0);
    INTEGER OptimizeKeyWord(INTEGER ID, INTEGER TYPE);
    INTEGER OptimizeAny(INTEGER ID = KEY_SEP, INTEGER TYPE = TYPE_SEPARATOR, char FLAGS = 0);

    INTEGER GetFirstOperator(AnsiList *LST, INTEGER start, INTEGER end);
    INTEGER Check(AnalizerElement *Target, AnalizerElement *AE);

    INTEGER Require(AnalizerElement *Element /*, INTEGER POSITION*/);

    void CopyElement(AnalizerElement *SRC, AnalizerElement *DEST);
    void CopyElement(AnalizerElement *SRC, SHORT_AnalizerElement *DEST);
    void CopyElement(AnalizerElement *SRC, PartialAnalizerElement *DEST);
    void BuildParameterList(INTEGER START_POS, AnalizerElement *METHOD);
    bool TryCheckParameters(OptimizedElement *OE, int p_count, int *minp, int *maxp, ClassMember **target_CM = 0, ClassCode *owner_CC = 0);

public:
    POOLED(Optimizer)

    Optimizer(PIFAlizator *P, AnsiList *_PIFList, AnsiList *_VDList, char *Filename, ClassCode *cls, char *member, bool is_unserialized = false);
    int Optimize();
    int OptimizeMemoryUsage();
    AnsiString DEBUG_INFO();

    ~Optimizer();
    int Serialize(FILE *out, bool is_lib = false);
    int Unserialize(concept_FILE *in, AnsiList *ModuleList, bool is_lib = false, int *ClassNames = 0, int *Relocation = 0);
    static int ComputeSharedSize(concept_FILE *in);
    void GenerateIntermediateCode();
};

#endif // __OPTIMIZER_H
