#include "PIFAlizator.h"
#include "AnsiParser.h"
#include "TinyString.h"
#include "ConceptTypes.h"
#include "AnsiException.h"
#include "Optimizer.h"
#include "ConceptInterpreter.h"
#include "SHManager.h"
#include "ModuleLink.h"
#include <stdio.h>
#include <string.h>
#include <time.h>
#ifdef _WIN32
 #include <io.h>
#endif

POOLED_IMPLEMENTATION(PIFAlizator)

//#define AddGeneralMember(MNAME) GeneralMembers.Add(new TinyString(MNAME),DATA_TINYSTRING)
#define AddGeneralMember(MNAME)    GeneralMembers.Add(MNAME)

//using namespace Concept;
int PIFAlizator::argc = 0;
char        **PIFAlizator::argv      = 0;
int         PIFAlizator::last_result = 0;
CHECK_POINT PIFAlizator::CheckPoint  = 0;

void PIFAlizator::AcknoledgeRunTimeError(AnsiException *Exc) {
    // notify the client about the run-time error ...
    // --------------------------------------- //
    AnsiString cstack = "Stack:\n";

    for (int i = 0; i < ConceptInterpreter::CALL_STACK_COUNT; i++) {
        cstack += *ConceptInterpreter::CALL_STACK [i].cls;
        cstack += ".";
        cstack += ConceptInterpreter::CALL_STACK [i].mem->c_str();
        cstack += ":";
        cstack += AnsiString((long)ConceptInterpreter::CALL_STACK [i].line);
        cstack += "\n";
    }
    out->ClientError(Exc->ToString() + cstack);
    // --------------------------------------- //
    Errors.Add(Exc, DATA_EXCEPTION);
}

void PIFAlizator::DefineConstant(char *name, char *value, int is_string) {
    VariableDESCRIPTOR *VD = new VariableDESCRIPTOR;

    VD->name   = name;
    VD->USED   = (char)this->INCLUDE_LEVEL;
    VD->BY_REF = 0;

    AnsiString c_val;

    if (is_string) {
        c_val += "\"";
    }

    int len = strlen(value);

    for (int i = 0; i < len; i++) {
        if (value [i] == '\\') {
            c_val += value [i];
        }
        c_val += value [i];
    }

    //VD->value   +=   value;
    if (is_string) {
        c_val += "\"";
    }

    VD->value.LoadBuffer(c_val.c_str(), c_val.Length());
    ConstantList.Add(VD, DATA_VAR_DESCRIPTOR);
}

PIFAlizator::PIFAlizator(AnsiString INC_DIR, AnsiString LIB_DIR, AnsiString *S, SimpleStream *Out, AnsiString _FileName, INTEGER debug, DEBUGGER_CALLBACK DC, void *DEBUGGER_RESERVED, char *SPubK, char *SPrivK, char *CPubK) {
    out                     = Out;
    FileName                = _FileName;
    InputStream             = S;
    ENTRY_CLASS             = 0;
    DebugOn                 = debug;
    DEBUGGER_TRAP           = DC;
    SERVER_PUBLIC_KEY       = SPubK;
    SERVER_PRIVATE_KEY      = SPrivK;
    CLIENT_PUBLIC_KEY       = CPubK;
    IMPORT_DIR              = LIB_DIR;
    INCLUDE_DIR             = INC_DIR;
    this->DEBUGGER_RESERVED = DEBUGGER_RESERVED;
    USE_WARN                = DEFAULT_USE_WARNINGS;
    USE_EXC                 = DEFAULT_USE_EXCEPTIONS;
    USE_IMPLICIT            = DEFAULT_USE_IMPLICIT;
    INCLUDE_LEVEL           = 0;
    //MAIN_ENTRY_POINT=DEFAULT_ENTRY_POINT;
    // adaugare Constante
    DefineConstant("_C_INCLUDE_DIR", INC_DIR);
    DefineConstant("_C_LIB_DIR", LIB_DIR);
    DefineConstant("_C_FILENAME", _FileName);
    DefineConstant("true", "1", 0);
    DefineConstant("false", "0", 0);
    DefineConstant("null", "0", 0);
    basic_constants_count = this->ConstantList.Count();
    //DefineConstant("_C_DEBUG",          AnsiString(DebugOn));

    // adaugare operatori ...
    AddGeneralMember(MUL);
    AddGeneralMember(DIV);
    AddGeneralMember(REM);
    AddGeneralMember(SUM);
    AddGeneralMember(DIF);
    AddGeneralMember(SHL);
    AddGeneralMember(SHR);
    AddGeneralMember(LES);
    AddGeneralMember(LEQ);
    AddGeneralMember(GRE);
    AddGeneralMember(GEQ);
    AddGeneralMember(EQU);
    AddGeneralMember(NEQ);
    AddGeneralMember(AND);
    AddGeneralMember(XOR);
    AddGeneralMember(OR);
    AddGeneralMember(BAN);
    AddGeneralMember(BOR);
    AddGeneralMember(ASG);
    AddGeneralMember(ASU);
    AddGeneralMember(ADI);
    AddGeneralMember(AMU);
    AddGeneralMember(ADV);
    AddGeneralMember(ARE);
    AddGeneralMember(AAN);
    AddGeneralMember(AXO);
    AddGeneralMember(AOR);
    AddGeneralMember(ASL);
    AddGeneralMember(ASR);
    AddGeneralMember(SELECTOR_I);

    AddGeneralMember(NOT);
    AddGeneralMember(COM);
    AddGeneralMember(INC);
    AddGeneralMember(DEC);
    //basic_members_count=GeneralMembers.Count();

    this->StaticLinks      = 0;
    this->StaticLinksCount = 0;
    this->StaticLinksSize  = 0;
    this->CachedhDLL       = 0;

    //IncludedList.Add(new AnsiString(_FileName),DATA_STRING);
}

//#include <iostream>
void PIFAlizator::SetPipe(int pipein, int pipeout, int apid, int papid) {
    this->apid        = apid;
    this->parent_apid = papid;
    this->pipe_read   = pipein;
    this->pipe_write  = pipeout;

    /*std::cout << "APID:" << apid << "\n";
     * std::cout << "PARENT:" << papid << "\n";
     * std::cout << "pipe_read:" << pipein << "\n";
     * std::cout << "pipe_write:" << pipeout << "\n";*/
}

PIFAlizator::~PIFAlizator(void) {
    int count = ModuleList.Count();

    // inchid toate bibliotecile link-uite
    for (INTEGER i = 0; i < count; i++) {
        //INTEGER *hMOD=(INTEGER *)ModuleList[i];
        //UnImportMoudule(*hMOD);
        HMODULE hLIBRARY = (HMODULE)ModuleList [i];
        UnImportMoudule(hLIBRARY);
    }
    this->Warnings.Clear();
    this->Errors.Clear();
    this->UndefinedMembers.Clear();
    this->UndefinedClasses.Clear();
    this->GeneralMembers.Clear();
    this->IncludedList.Clear();
    this->ModuleList.Clear();
    this->ModuleNamesList.Clear();

    //============================================//
    // bug - adaugat pe 2 august 2010
    // genereaza crash uneori daca sterg entry_class-ul
    if (ENTRY_CLASS) {
        this->ClassList.Remove(ENTRY_CLASS - 1);
    }
    //============================================//

    this->ClassList.Clear();
    this->ConstantList.Clear();
    this->VariableDescriptors.Clear();

    this->PIF.Clear();

    if (this->StaticLinks) {
        free(this->StaticLinks);
    }
}

unsigned int PIFAlizator::LinkStatic(char *funname) {
    unsigned int res = 0;
    SYS_INT      ref = LinkFunction(funname, &ModuleList, &CachedhDLL);

    if (ref) {
        for (int i = 0; i < this->StaticLinksCount; i++) {
            if (this->StaticLinks [i] == ref) {
                return(i + 1);
            }
        }
        this->StaticLinksCount++;
        if (this->StaticLinksCount > this->StaticLinksSize) {
            this->StaticLinksSize += STATIC_LINKS_INCREMENT;
            if (this->StaticLinks) {
                this->StaticLinks = (SYS_INT *)realloc(this->StaticLinks, sizeof(SYS_INT) * this->StaticLinksSize);
            } else {
                this->StaticLinks = (SYS_INT *)malloc(sizeof(SYS_INT) * this->StaticLinksSize);
            }
        }
        this->StaticLinks [this->StaticLinksCount - 1] = ref;
        res = this->StaticLinksCount;
    }
    return(res);
}

INTEGER PIFAlizator::AddUndefinedMember(AnsiString& member, AnsiString& _CLASS, TinyString& _MEMBER, long line) {
    UndefinedMember *UM = new UndefinedMember;

    UM->filename = FileName;
    UM->line     = line;
    UM->name     = member;
    UM->_CLASS   = _CLASS;
    UM->_MEMBER  = _MEMBER.c_str();
    UndefinedMembers.Add(UM, DATA_UNDEFINEDMEMBER);
    AddGeneralMember(member);
    return(GeneralMembers.Count());
}

INTEGER PIFAlizator::AddUndefinedClass(AnsiString& member, AnsiString& _CLASS, TinyString& _MEMBER, long line) {
    UndefinedMember *UM = new UndefinedMember;

    UM->filename = FileName;
    UM->line     = line;
    UM->name     = member;
    UM->_CLASS   = _CLASS;
    UM->_MEMBER  = _MEMBER.c_str();
    UndefinedClasses.Add(UM, DATA_UNDEFINEDMEMBER);
    return(UndefinedClasses.Count());
}

#define DefineMember(MNAME)                                                    \
    {                                                                          \
        INTEGER cnt = UndefinedMembers.Count();                                \
        for (INTEGER i = 0; i < cnt; i++) {                                    \
            UndefinedMember *UM = (UndefinedMember *)UndefinedMembers.Item(i); \
            if (UM->name == MNAME) {                                           \
                UndefinedMembers.Delete(i);                                    \
                break;                                                         \
            }                                                                  \
        }                                                                      \
    }

#define DefineClass(CNAME)                                                     \
    {                                                                          \
        INTEGER cnt = UndefinedClasses.Count();                                \
        for (INTEGER i = 0; i < cnt; i++) {                                    \
            UndefinedMember *UM = (UndefinedMember *)UndefinedClasses.Item(i); \
            if (UM->name == CNAME) {                                           \
                UndefinedClasses.Delete(i);                                    \
                break;                                                         \
            }                                                                  \
        }                                                                      \
    }

/*INTEGER PIFAlizator::AddGeneralMember(AnsiString name) {
 *  GeneralMembers.Add(new AnsiString(name),DATA_STRING);
 *      return 1;
 * }*/

INTEGER PIFAlizator::VariableIsDescribed(AnsiString& S, AnsiList *VDList) {
    INTEGER Count = VDList->Count();

    for (INTEGER i = 0; i < Count; i++) {
        VariableDESCRIPTOR *VD = (VariableDESCRIPTOR *)(*VDList) [i];
        if (VD->name == S) {
            VD->USED = 1;
            return(i + 1);
        }
    }
    return(0);
}

INTEGER PIFAlizator::ListContains(AnsiString& S, AnsiList *VDList, char is_tiny) {
    INTEGER Count = VDList->Count();

    if (is_tiny) {
        for (INTEGER i = 0; i < Count; i++) {
            TinyString *STR = (TinyString *)(*VDList) [i];
            if (*STR == S) {
                return(i + 1);
            }
        }
    } else {
        for (INTEGER i = 0; i < Count; i++) {
            AnsiString *STR = (AnsiString *)(*VDList) [i];
            if (*STR == S) {
                return(i + 1);
            }
        }
    }
    return(0);
}

SYS_INT PIFAlizator::ClassExists(char *name, char by_addr, int *index) {
    INTEGER Count = ClassList.Count();

    for (INTEGER i = 0; i < Count; i++) {
        ClassCode *CC = (ClassCode *)ClassList [i];
        if (CC->NAME == name) {
            if (CC->NEEDED != -1) {
                if (!CC->NEEDED) {
                    CC->UpdateNeeded();
                }
                CC->NEEDED = 1;
            }
            if (by_addr) {
                if (index) {
                    *index = i + 1;
                }
                return((SYS_INT)CC);
            }
            return(i + 1);
        }
    }
    return(0);
}

INTEGER PIFAlizator::BuildVariable(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER ACCESS) {
    AnsiString sPARSE;     //          =   P->NextAtom();

    P->NextAtom(sPARSE);

    char    is_array = 0;
    INTEGER TYPE     = GetType(sPARSE);
    INTEGER _ID      = GetID(sPARSE);
    char    sign     = 1;

    if ((TYPE == TYPE_OPERATOR) && (_ID == KEY_SELECTOR_I)) {
        is_array = 1;
        // declaratie ca array ...
        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);
        TYPE = GetType(sPARSE);
        _ID  = GetID(sPARSE);
    }

    if ((_ID) || (TYPE != TYPE_METHOD)) {
        AnsiException *Exc = new AnsiException(ERR160, on_line ? on_line : P->LastLine(), 160, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(0);
    }

    //if (!ListContains(sPARSE,&GeneralMembers, true))
    if (!GeneralMembers.ContainsString(sPARSE.c_str())) {
        AddGeneralMember(sPARSE);
    } else {
        DefineMember(sPARSE);
    }

    ClassMember *CM = CC->AddMember(sPARSE, on_line ? on_line : P->LastLine(), FileName, ACCESS, true);
    //CM->IS_FUNCTION = 0;
    if (is_array) {
        CM->VD->TYPE = VARIABLE_ARRAY;
    }

    // ";"
    //sPARSE  =   P->NextAtom();
    P->NextAtom(sPARSE);
    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((TYPE == TYPE_OPERATOR) && (_ID == KEY_ASG)) {
        if (is_array) {
            AnsiException *Exc = new AnsiException(ERR1170, on_line ? on_line : P->LastLine(), 1170, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        }

        do {
            //sPARSE  =   P->NextAtom();
            P->NextAtom(sPARSE);
            TYPE = GetType(sPARSE);
            _ID  = GetID(sPARSE);

            if (TYPE == TYPE_OPERATOR) {
                if (_ID == KEY_DIF) {
                    sign *= -1;
                }
            }
        } while ((TYPE == TYPE_OPERATOR) && ((_ID == KEY_SUM) || (_ID == KEY_DIF)));

        if (TYPE == TYPE_HEX_NUMBER) {
            INTEGER TEMP_NR = 0;
            sscanf(sPARSE.c_str() + 2, "%x", &TEMP_NR);
            sPARSE = AnsiString(TEMP_NR);
            TYPE   = TYPE_NUMBER;
        }

        if ((TYPE == TYPE_STRING) || (TYPE == TYPE_NUMBER)) {
            if (TYPE == TYPE_STRING) {
                CM->VD->value = Strip(sPARSE);
            } else {
                CM->VD->nValue = sPARSE.ToFloat() * sign;
            }

            CM->VD->TYPE = TYPE == TYPE_STRING   ?   VARIABLE_STRING :   VARIABLE_NUMBER;
        } else {
            AnsiException *Exc = new AnsiException(ERR1160, on_line ? on_line : P->LastLine(), 1160, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        }

        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);
        TYPE = GetType(sPARSE);
        _ID  = GetID(sPARSE);
    }

    if ((TYPE != TYPE_SEPARATOR) && (_ID != KEY_SEP)) {
        AnsiException *Exc = new AnsiException(ERR340, on_line ? on_line : P->LastLine(), 340, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(1);
    }
    return(0);
}

/*INTEGER PIFAlizator::BuildArray(ClassCode *CC, ClassMember *CM, AnsiParser *P, INTEGER on_line) {
 *      VariableDESCRIPTOR *VD=new VariableDESCRIPTOR;
 *      VD->name      =   "CONSTANT Array";
 *      VD->TYPE      =   VARIABLE_NUMBER;
 *      VD->USED      =   0;
 *      VD->nValue    =   0;
 *      VD->BY_REF    =   0;
 *      VD->TYPE      =   VARIABLE_ARRAY;
 *
 *      VDList->Add(VD,DATA_VAR_DESCRIPTOR);
 *      INTEGER _ID=VDList->Count();
 *      return _ID;
 * }*/

AnsiString PIFAlizator::GetSpecial(AnsiParser *P, ClassCode *CC, ClassMember *CM, AnsiString special) {
    AnsiString result;

    if (special == (char *)"line") {
        result = P->LastLine();
    } else
    if (special == (char *)"filename") {
        result += (char *)"'";
        int len = FileName.Length();
        int pos = 0;
        // len-2 = at least one character for the filename
        for (int i = len - 2; i >= 0; i--) {
            char c = FileName [i];
            if ((c == '/') || (c == '\\')) {
                pos = i + 1;
                break;
            }
        }
        result += AnsiString(FileName.c_str() + pos);
        result += (char *)"'";
    } else
    if (special == (char *)"path") {
        result += (char *)"'";
        int len = FileName.Length();
        for (int i = 0; i < len; i++) {
            char c = FileName [i];
            if (c == '\\') {
                result += '/';
            } else {
                result += c;
            }
        }
        result += (char *)"'";
    } else
    if (special == (char *)"class") {
        result += (char *)"'";
        result += CC->NAME;
        result += (char *)"'";
    } else
    if (special == (char *)"member") {
        result += (char *)"'";
        result += CM->NAME.c_str();
        result += (char *)"'";
    } else
    if (special == (char *)"time") {
        result = (long)time(NULL);
    } else {
        AnsiException *Exc = new AnsiException(ERR1209, P->LastLine(), 1209, special, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        //result=P->NextAtom();
        P->NextAtom(result);
    }
    return(result);
}

INTEGER PIFAlizator::BuildFunction(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER ACCESS, INTEGER OPERATOR, char STATIC, char *prec_parse) {
    INTEGER                      BEGIN_END_LEVEL = 0;
    INTEGER                      ATOMIC_LEVEL    = 0;
    bool                         is_destructor   = false;
    register INTEGER             TYPE            = -1;
    register /*INTEGER*/ SYS_INT _ID             = -1;
    register INTEGER             PREC_TYPE       = -1;
    register /*INTEGER*/ SYS_INT PREC_ID         = -1;
    AnsiString                   sPARSE; //          =   P->NextAtom();

    if (prec_parse) {
        sPARSE = prec_parse;
    } else {
        P->NextAtom(sPARSE);
    }
    char IN_VAR_STATAMENT = 0;

    ClassMember *CM = CC->AddMember(sPARSE, on_line ? on_line : P->LastLine(), FileName, ACCESS, false);
    CM->IS_FUNCTION = 1;
    CM->IS_OPERATOR = OPERATOR;
    if (!OPERATOR) {
        CM->IS_STATIC = STATIC;
    }

    TYPE = GetType(sPARSE);

    if ((TYPE == TYPE_OPERATOR) && (!OPERATOR)) {
        AnsiException *Exc = new AnsiException(ERR300, on_line ? on_line : P->LastLine(), 300, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    } else
    if ((TYPE != TYPE_OPERATOR) && (OPERATOR)) {
        AnsiException *Exc = new AnsiException(ERR310, on_line ? on_line : P->LastLine(), 310, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    }

    if (CM->NAME == CC->NAME) {
        CC->CONSTRUCTOR        = CC->Members->Count();
        CC->CONSTRUCTOR_MEMBER = CM;
    } else
    if (CM->NAME == (char *)C_FINALIZE) {
        sPARSE                = CC->NAME;
        sPARSE               += (char *)"@";
        sPARSE               += C_FINALIZE;
        CM->NAME              = sPARSE;
        CC->DESTRUCTOR        = CC->Members->Count();
        CC->DESTRUCTOR_MEMBER = CM;
        is_destructor         = true;
    }

    //if (!ListContains(sPARSE,&GeneralMembers, true)) {
    if (!GeneralMembers.ContainsString(sPARSE.c_str())) {
        if (OPERATOR) {
            AnsiException *Exc = new AnsiException(ERR320, on_line ? on_line : P->LastLine(), 320, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        } else {
            AddGeneralMember(sPARSE);
        }
    } else {
        DefineMember(sPARSE);
    }

    AnsiList *PIFList = CM->PIF_DATA;
    AnsiList *VDList  = CM->VariableDescriptors;

    // adaugare membru THIS
    VariableDESCRIPTOR *VD = new VariableDESCRIPTOR;
    VD->name   = THIS_CLASS;
    VD->TYPE   = VARIABLE_NUMBER;
    VD->USED   = 1;
    VD->BY_REF = 0;
    VD->nValue = 0;

    VDList->Add(VD, DATA_VAR_DESCRIPTOR);

    // pana aici

    // izolare parametri ...
    /*INTEGER*/ char NEXT_IS_BY_REF = 0;
    char             NEXT_IS_ASG    = 0;
    INTEGER          WANT_PARAM     = 0;
    INTEGER          COUNTS         = 0;
    char             START_DEFAULT  = 0;
    char             sign           = 1;

    char    NEXT_IS_TYPE  = 0;
    INTEGER NEXT_IS_CLSID = 0;

    VariableDESCRIPTOR *VDPARAM = 0;
    do {
        COUNTS++;
        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);

        //if (sPARSE==NULL_STRING)
        if (!sPARSE.Length()) {
            break;
        }

        TYPE = GetType(sPARSE);
        _ID  = GetID(sPARSE);

        if (TYPE == TYPE_HEX_NUMBER) {
            INTEGER TEMP_NR = 0;
            sscanf(sPARSE.c_str() + 2, "%x", &TEMP_NR);
            sPARSE = AnsiString(TEMP_NR);
            TYPE   = TYPE_NUMBER;
        }

        if (((TYPE == TYPE_STRING) || (TYPE == TYPE_NUMBER)) && (VDPARAM)) {
            if (NEXT_IS_ASG) {
                char type = TYPE == TYPE_STRING   ?   VARIABLE_STRING :   VARIABLE_NUMBER;
                if (VDPARAM->TYPE < 0) {
                    if ((VDPARAM->TYPE == -VARIABLE_ARRAY) || (VDPARAM->TYPE == -VARIABLE_CLASS) ||
                        (VDPARAM->TYPE == -VARIABLE_DELEGATE)) {
                        AnsiException *Exc = new AnsiException(ERR1207, on_line ? on_line : P->LastLine(), 1207, sPARSE, FileName, CC->NAME, CM->NAME);
                        Errors.Add(Exc, DATA_EXCEPTION);
                    } else if (VDPARAM->TYPE != -type) {
                        AnsiException *Exc = new AnsiException(ERR1208, on_line ? on_line : P->LastLine(), 1208, sPARSE, FileName, CC->NAME, CM->NAME);
                        Errors.Add(Exc, DATA_EXCEPTION);
                    }
                }

                VDPARAM->value  = Strip(sPARSE);
                VDPARAM->nValue = VDPARAM->value.ToFloat() * sign;
                VDPARAM->TYPE   = type; //TYPE==TYPE_STRING   ?   VARIABLE_STRING :   VARIABLE_NUMBER;
                VDPARAM         = 0;
                NEXT_IS_ASG     = 0;
                sign            = 1;
                continue;
            } else {
                AnsiException *Exc = new AnsiException(ERR1130, on_line ? on_line : P->LastLine(), 1130, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                NEXT_IS_ASG = 0;
                continue;
            }
        } else
        if (NEXT_IS_ASG) {
            if (sPARSE == (char *)DIF) {
                sign *= -1;
                continue;
            } else if (sPARSE == (char *)SUM) {
                continue;
            }


            AnsiException *Exc = new AnsiException(ERR1140, on_line ? on_line : P->LastLine(), 1140, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        }

        if ((TYPE == TYPE_METHOD) && (_ID == KEY_VAR)) {
            // variabila declarata cu VAR;
            NEXT_IS_BY_REF = 1;
            continue;
        } else
        if (TYPE == TYPE_OPERATOR) {
            if (_ID == KEY_P_OPEN) {
                if (ATOMIC_LEVEL) {
                    AnsiException *Exc = new AnsiException(ERR780, on_line ? on_line : P->LastLine(), 780, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                } else {
                    ATOMIC_LEVEL++;
                    WANT_PARAM = 1;
                }
            } else
            if (_ID == KEY_P_CLOSE) {
                ATOMIC_LEVEL--;
                if ((WANT_PARAM) && (COUNTS != 2)) {
                    AnsiException *Exc = new AnsiException(ERR790, on_line ? on_line : P->LastLine(), 790, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                }
            } else
            // verificare atribuire ...
            if ((_ID == KEY_ASG) && (VDPARAM)) {
                NEXT_IS_ASG = 1;
                if (!START_DEFAULT) {
                    START_DEFAULT = 1;
                    CM->MUST_PARAMETERS_COUNT--;
                }
                continue;
            } else
            if (_ID == KEY_COMMA) {                                   // era doar "=": possible incorrect asignment
                if (!WANT_PARAM) {
                    WANT_PARAM = 1;
                } else {
                    AnsiException *Exc = new AnsiException(ERR790, on_line ? on_line : P->LastLine(), 790, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                }
            } else {
                AnsiException *Exc = new AnsiException(ERR780, on_line ? on_line : P->LastLine(), 780, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
        } else
        if ((TYPE == TYPE_METHOD) && (!_ID) && (WANT_PARAM)) {
            if (sPARSE == (char *)"number") {
                NEXT_IS_TYPE  = VARIABLE_NUMBER;
                NEXT_IS_CLSID = 0;
                continue;
            } else
            if (sPARSE == (char *)"string") {
                NEXT_IS_TYPE  = VARIABLE_STRING;
                NEXT_IS_CLSID = 0;
                continue;
            } else
            if (sPARSE == (char *)"array") {
                NEXT_IS_TYPE  = VARIABLE_ARRAY;
                NEXT_IS_CLSID = 0;
                continue;
            } else
            if (sPARSE == (char *)"delegate") {
                NEXT_IS_TYPE  = VARIABLE_DELEGATE;
                NEXT_IS_CLSID = 0;
                continue;
            } else
            if (sPARSE == (char *)"object") {
                NEXT_IS_TYPE  = VARIABLE_CLASS;
                NEXT_IS_CLSID = 0;
                continue;
            } else {
                INTEGER cls_id = this->ClassExists(sPARSE.c_str());
                if (cls_id) {
                    NEXT_IS_TYPE  = VARIABLE_CLASS;
                    NEXT_IS_CLSID = cls_id;
                    continue;
                }
            }

            if (VariableIsDescribed(sPARSE, VDList)) {
                AnsiException *Exc = new AnsiException(ERR800, on_line ? on_line : P->LastLine(), 800, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
            if ((START_DEFAULT) && (VDPARAM)) {
                AnsiException *Exc = new AnsiException(ERR1150, on_line ? on_line : P->LastLine(), 1150, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else if (!START_DEFAULT) {
                CM->MUST_PARAMETERS_COUNT++;
            }

            // adaugare variabila ...
            VDPARAM         = new VariableDESCRIPTOR;
            VDPARAM->BY_REF = NEXT_IS_BY_REF;
            VDPARAM->USED   = 1;
            VDPARAM->name   = sPARSE;
            VDPARAM->TYPE   = VARIABLE_NUMBER;
            VDPARAM->nValue = 0;
            //VDPARAM->value  =   "";

            if (NEXT_IS_TYPE) {
                VDPARAM->TYPE = -NEXT_IS_TYPE;
                if (NEXT_IS_TYPE == VARIABLE_CLASS) {
                    VDPARAM->nValue = NEXT_IS_CLSID;
                }
            }

            //CM->Parameters.Add(VDPARAM,DATA_VAR_DESCRIPTOR,1);
            CM->PARAMETERS_COUNT++;

            VDList->Add(VDPARAM, DATA_VAR_DESCRIPTOR);
            WANT_PARAM = 0;
        } else {
            AnsiException *Exc = new AnsiException(ERR770, on_line ? on_line : P->LastLine(), 770, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
            if (TYPE == TYPE_KEYWORD) {
                if (_ID == KEY_BEGIN) {
                    BEGIN_END_LEVEL++;
                }
                if (_ID == KEY_END) {
                    return(0);
                }
                ATOMIC_LEVEL = 0;
            }
        }
        NEXT_IS_BY_REF = 0;
        NEXT_IS_ASG    = 0;
        NEXT_IS_TYPE   = 0;
        NEXT_IS_CLSID  = 0;
    } while (ATOMIC_LEVEL);
    if ((VDPARAM) && (START_DEFAULT)) {
        AnsiException *Exc = new AnsiException(ERR1150, on_line ? on_line : P->LastLine(), 1150, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    }
    if ((is_destructor) && (CM->PARAMETERS_COUNT)) {
        AnsiException *Exc = new AnsiException(ERR1180, on_line ? on_line : P->LastLine(), 1180, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    }
    // pana aici ...

    int  PREC_IS_CONSTANT = 0;
    int  IS_CONSTANT      = 0;
    char IS_ARRAY         = 0;
    //int PREC_PREC_TYPE;
    //int PREC_PREC_ID;
    AnsiString cached;
    TYPE = -1;
    _ID  = 0;
    do {
        //PREC_PREC_TYPE=PREC_TYPE;
        //PREC_PREC_ID=PREC_ID;

        // tzin minte daca este operator, in caz ca intalnesc + sau - sa stiu sa-l clasific
        // in KEY_SUM sau KEY_POZ, respectiv KEY_DIF sau KEY_NEG

        PREC_TYPE = TYPE;
        PREC_ID   = _ID;

        _ID = 0;

        if (cached.Length()) {
            sPARSE = cached;
            cached = (char *)"";
        } else {
            P->NextAtom(sPARSE);
        }
        //sPARSE  =   P->NextAtom();

        char *ptr_parse = sPARSE.c_str();

        //if (sPARSE==NULL_STRING)
        //if (!sPARSE.Length()) {
        if ((!sPARSE.Length()) && (P->Done())) {
            if (BEGIN_END_LEVEL) {
                AnsiException *Exc = new AnsiException(ERR1210, on_line ? on_line : P->LastLine(), 1210, AnsiString((long)BEGIN_END_LEVEL) + AnsiString(" level(s)"), FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                //eroare !
            }
            break;
        }

        if (ptr_parse [0] == C_SPECIAL) {
            P->NextAtom(sPARSE);
            sPARSE = GetSpecial(P, CC, CM, sPARSE);
            //sPARSE=GetSpecial(P,CC, CM, P->NextAtom());
            if (!sPARSE.Length()) {
                break;
            }
        }

        // iau tipul atomului
        TYPE = GetType(sPARSE);

        if (TYPE == TYPE_HEX_NUMBER) {
            unsigned INTEGER TEMP_NR = 0;
            sscanf(sPARSE.c_str() + 2, "%x", &TEMP_NR);
            sPARSE = AnsiString((double)TEMP_NR);
            TYPE   = TYPE_NUMBER;
        }

        if (TYPE < 0) {
            if ((ptr_parse [0] == '"') || (ptr_parse [0] == '\'')) {
                /*int p1=sPARSE.Pos((char *)"\n");
                 * int p2=sPARSE.Pos((char *)"\r");
                 * int min=p1;
                 * if (p1>p2)
                 *  min=p2;
                 * if (min>1)
                 *  sPARSE.c_str()[min-1]=0;*/
                AnsiException *Exc = new AnsiException(ERR830, on_line ? on_line : P->LastLine(), 830, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else {
                AnsiException *Exc = new AnsiException(ERR050, on_line ? on_line : P->LastLine(), 50, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
            continue;
        }

        // este o variabila declarata pe locul 2 intr-o declaratzie de forma var a,b;
        if ((IN_VAR_STATAMENT) && (!ATOMIC_LEVEL) && (sPARSE == (char *)CONCEPT_COMMA)) {
            TYPE = TYPE_KEYWORD;
            _ID  = KEY_VAR;
            AnalizerElement *AE = new AnalizerElement;
            AE->ID               = KEY_SEP;
            AE->TYPE             = TYPE_SEPARATOR;
            AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
            //AE->_DEBUG_INFO_FILENAME    =   FileName;
            AE->_INFO_OPTIMIZED = 0;
            AE->_PARSE_DATA     = sPARSE;
            AE->_HASH_DATA      = 0;
            PIFList->Add(AE, DATA_ANALIZER_ELEMENT);
            continue;
        }

        if ((PREC_TYPE == TYPE_KEYWORD) && (PREC_ID == KEY_VAR)) {
            // verific daca este declarare cu array []

            if (TYPE == TYPE_METHOD) {
                _ID = GetID(sPARSE);
                if (_ID) {
                    TYPE = TYPE_KEYWORD;
                }
            }

            if ((TYPE == TYPE_OPERATOR) || (TYPE == TYPE_KEYWORD)) {
                _ID = GetID(sPARSE);
                if (_ID == KEY_SELECTOR_I) {
                    TYPE     = PREC_TYPE;
                    _ID      = PREC_ID;
                    IS_ARRAY = 1;
                    continue;
                }
                // daca este cuvant rezervat ...
                AnsiException *Exc = new AnsiException(ERR890, on_line ? on_line : P->LastLine(), 890, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                continue;
            }
            if (VariableIsDescribed(sPARSE, VDList)) {
                AnsiException *Exc = new AnsiException(ERR070, on_line ? on_line : P->LastLine(), 70, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                continue;
            } else {
                TYPE = TYPE_VARIABLE;
            }
        } else
        if (((PREC_TYPE != TYPE_OPERATOR) || (PREC_ID != KEY_SEL)) && (TYPE == TYPE_METHOD) && (VariableIsDescribed(sPARSE, VDList))) {
            TYPE = TYPE_VARIABLE;
        }

        // daca este variabila o pun in lista de descriptori de variabilie ...
        if (TYPE == TYPE_VARIABLE) {
            _ID = VariableIsDescribed(sPARSE, VDList);
            if (!_ID) {
                VariableDESCRIPTOR *VD = new VariableDESCRIPTOR;
                VD->name   = sPARSE;
                VD->TYPE   = VARIABLE_NUMBER;
                VD->USED   = 0;
                VD->nValue = 0;
                VD->BY_REF = 0;
                if (IS_ARRAY) {
                    VD->TYPE = VARIABLE_ARRAY;
                }

                VDList->Add(VD, DATA_VAR_DESCRIPTOR);
                _ID = VDList->Count();
            }
            IS_ARRAY = 0;

            // nu poate fi folosit THIS intr-un membru static !!!
            // THIS este pe prima pozitzie in descriptori ...
            if ((STATIC) && (_ID == 1)) {
                AnsiException *Exc = new AnsiException(ERR930, on_line ? on_line : P->LastLine(), 930, sPARSE, FileName, CC->NAME, CM->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                continue;
            }
        }

        // daca este constanta, la fel, dar fara sa-i mai verific existentza ...
        if ((TYPE == TYPE_STRING) || (TYPE == TYPE_NUMBER)) {
            VariableDESCRIPTOR *VD = new VariableDESCRIPTOR;
            //AnsiString strip=StripString(&sPARSE)
            AnsiString strip;
            StripString(&sPARSE, strip);
            if (strip.Length()) {
                VD->value.LoadBuffer(strip.c_str(), strip.Length());
            }
            VD->nValue = VD->value.ToFloat();
            VD->USED   = 1;
            VD->TYPE   = (TYPE == TYPE_STRING) ? VARIABLE_STRING : VARIABLE_NUMBER;
            VD->BY_REF = 0;

            VDList->Add(VD, DATA_VAR_DESCRIPTOR);
            _ID         = VDList->Count();
            TYPE        = TYPE_VARIABLE;
            IS_CONSTANT = 1;
        } else {
            IS_CONSTANT = 0;
        }

        if (!_ID) {
            // iau ID-ul lui, in caz ca este cuvant rezervat !
            _ID = GetID(sPARSE);

            if (TYPE == TYPE_KEYWORD) {
                if (_ID == KEY_BEGIN) {
                    BEGIN_END_LEVEL++;
                } else if (_ID == KEY_END) {
                    BEGIN_END_LEVEL--;
                }
            }

            // daca arata ca o functie, si are si ID, inseamna ca este cuvant cheie !
            if ((TYPE == TYPE_METHOD) && (_ID)) {
                TYPE = TYPE_KEYWORD;
            }
        }

        if (TYPE == TYPE_KEYWORD) {
            if (_ID == KEY_SUPER) {
                ClassCode *PARENT = (ClassCode *)CC->first_parent;
                if (PARENT) {
                    TYPE   = TYPE_METHOD;
                    _ID    = 0;
                    sPARSE = PARENT->NAME;
                    // NOU ---------------------------------

                    // Pana aici

                    /*if (PARENT->CONSTRUCTOR) {
                     *  if (CC->HasMember(PARENT->CONSTRUCTOR_MEMBER->NAME)) {
                     *      _ID=ListContains(PARENT->CONSTRUCTOR_MEMBER->NAME,&GeneralMembers);
                     *      TYPE=TYPE_METHOD;
                     *
                     *
                     *      // nu poate fi folosit THIS intr-un membru static !!!
                     *      if (STATIC) {
                     *          AnsiException *Exc=new AnsiException(ERR930,on_line ? on_line : P->LastLine(),930,sPARSE,FileName,CC->NAME,CM->NAME);
                     *          Errors.Add(Exc,DATA_EXCEPTION);
                     *          continue;
                     *      }
                     *      // adaug this
                     *      AnalizerElement *AE         =   new AnalizerElement;
                     *      AE->ID                      =   1;
                     *      AE->TYPE                    =   TYPE_VARIABLE;
                     *      AE->_DEBUG_INFO_LINE        =   on_line ? on_line : P->LastLine();
                     *      //AE->_DEBUG_INFO_FILENAME    =   FileName;
                     *      AE->_INFO_OPTIMIZED         =   0;
                     *      AE->_PARSE_DATA             =   (char *)"this";
                     *      AE->_HASH_DATA              =   0;
                     *      PIFList->Add(AE,DATA_ANALIZER_ELEMENT);
                     *
                     *      // adaug operatorul
                     *      AE         =   new AnalizerElement;
                     *      AE->ID                      =   KEY_SEL;
                     *      AE->TYPE                    =   TYPE_OPERATOR;
                     *      AE->_DEBUG_INFO_LINE        =   on_line ? on_line : P->LastLine();
                     *      //AE->_DEBUG_INFO_FILENAME    =   FileName;
                     *      AE->_INFO_OPTIMIZED         =   0;
                     *      AE->_PARSE_DATA             =   (char *)".";
                     *      AE->_HASH_DATA              =   0;
                     *      PIFList->Add(AE,DATA_ANALIZER_ELEMENT);
                     *
                     *  } else {
                     *      _ID=0;
                     *      AnsiException *Exc=new AnsiException(ERR1000,on_line ? on_line : P->LastLine(),1000,sPARSE,FileName,CC->NAME,CM->NAME);
                     *      Errors.Add(Exc,DATA_EXCEPTION);
                     *      continue;
                     *  }
                     * } else {
                     *  AnsiException *Exc=new AnsiException(ERR990,on_line ? on_line : P->LastLine(),990,sPARSE,FileName,CC->NAME,CM->NAME);
                     *  Errors.Add(Exc,DATA_EXCEPTION);
                     *  continue;
                     * }*/
                } else {
                    AnsiException *Exc = new AnsiException(ERR980, on_line ? on_line : P->LastLine(), 980, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                    continue;
                }
            } else
            if (_ID == KEY_PRAGMA) {
                BuildPragma(P);
                continue;
            }
        }

        if ((TYPE == TYPE_OPERATOR) && (PREC_IS_CONSTANT) && (GetPriority(_ID) == OPERATOR_LEVEL_8)) {
            AnsiException *Exc = new AnsiException(ERR760, on_line ? on_line : P->LastLine(), 760, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        }
        PREC_IS_CONSTANT = IS_CONSTANT;

        if (TYPE == TYPE_OPERATOR) {
            if (_ID == KEY_DLL_CALL) {
                int cnt = PIFList->Count() - 1;
                AnalizerElement *AE_PREC = (AnalizerElement *)PIFList->Item(cnt);
                if ((AE_PREC) && (AE_PREC->TYPE == TYPE_METHOD)) {
                    AnalizerElement *AE_PREC2 = (AnalizerElement *)PIFList->Item(cnt - 2);
                    AnalizerElement *AE_PREC1 = (AnalizerElement *)PIFList->Item(cnt - 1);
                    if ((AE_PREC2->TYPE == TYPE_VARIABLE) && (AE_PREC2->ID == 1) &&
                        (AE_PREC1->TYPE == TYPE_OPERATOR) && (AE_PREC1->ID == KEY_SEL)) {
                        PREC_ID = ClassExists(AE_PREC->_PARSE_DATA.c_str());
                        if (PREC_ID) {
                            AE_PREC->ID   = PREC_ID;
                            AE_PREC->TYPE = TYPE_CLASS;
                            PREC_TYPE     = TYPE_CLASS;
                            int cnt = PIFList->Count() - 3;
                            PIFList->Delete(cnt);
                            PIFList->Delete(cnt);
                        }
                    }
                }
            }


            if ((PREC_TYPE == TYPE_OPERATOR) && (PREC_ID == KEY_NEW)) {
                if (_ID == KEY_SELECTOR_I) {
                    _ID  = STATIC_CLASS_ARRAY;
                    TYPE = TYPE_CLASS;
                } else {
                    _ID = 0;
                }
            } else
            if ((PREC_TYPE == TYPE_CLASS) && (_ID == KEY_SEL)) {
                _ID = KEY_DLL_CALL;
            }
        }

        // verificare daca este identificator de clasa
        if ((!_ID) && (PREC_TYPE == TYPE_OPERATOR) && (PREC_ID == KEY_NEW)) {
            _ID = ClassExists(sPARSE.c_str());

            if (!_ID) {
                Warning(WRN10007, on_line ? on_line : P->LastLine(), 10007, sPARSE, FileName);

                ClassCode *CC1 = new ClassCode(sPARSE, this);
                CC1->DEFINED_LEVEL = this->INCLUDE_LEVEL;
                CC1->NEEDED        = -1;
                ClassList.Add(CC1, DATA_CLASS_CODE);
                _ID = ClassList.Count();
                //_ID=AddUndefinedClass(sPARSE,CC->NAME,CM->NAME,on_line ? on_line : P->LastLine())+1;
                AddUndefinedClass(sPARSE, CC->NAME, CM->NAME, on_line ? on_line : P->LastLine()) + 1;


                /*AnsiException *Exc=new AnsiException(ERR100,on_line ? on_line : P->LastLine(),100,sPARSE,FileName,CC->NAME,CM->NAME);
                 * Errors.Add(Exc,DATA_EXCEPTION);
                 * continue;*/
            }
            TYPE = TYPE_CLASS;
        }
        // este metoda ?
        if ((!_ID) &&
            (TYPE == TYPE_METHOD) &&
            (PREC_TYPE == TYPE_OPERATOR) &&
            ((PREC_ID == KEY_SEL) || (PREC_ID == KEY_DLL_CALL))) {
            // nu-l inventariez, pentru ca nu stiu ce tip este variabila din care
            // provine ... verificarea sa va face in run-time !;
            //------ NOU !!! --------------------------------------------//
            if (PREC_ID == KEY_DLL_CALL) {
                INTEGER         cnt = PIFList->Count();
                AnalizerElement *AE = (AnalizerElement *)PIFList->Item(cnt - 2);
                if ((AE) && (AE->TYPE == TYPE_CLASS) && (AE->ID > 0)) {
                    ClassCode *CC_temp = (ClassCode *)ClassList.Item(AE->ID - 1);
                    if (CC_temp) {
                        AnsiString temp_mname = CC_temp->NAME;
                        temp_mname += (char *)"@";
                        temp_mname += sPARSE;
                        _ID         = CC->HasMember(temp_mname);
                        if (_ID) {
                            PIFList->Delete(cnt - 2);
                            PIFList->Delete(cnt - 2);
                            sPARSE = temp_mname;
                            // adaug this
                            AnalizerElement *AE = new AnalizerElement;
                            AE->ID               = 1;
                            AE->TYPE             = TYPE_VARIABLE;
                            AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                            //AE->_DEBUG_INFO_FILENAME    =   FileName;
                            AE->_INFO_OPTIMIZED = 0;
                            AE->_PARSE_DATA     = (char *)"this";
                            AE->_HASH_DATA      = 0;
                            PIFList->Add(AE, DATA_ANALIZER_ELEMENT);

                            // adaug operatorul
                            AE                   = new AnalizerElement;
                            AE->ID               = KEY_SEL;
                            AE->TYPE             = TYPE_OPERATOR;
                            AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                            //AE->_DEBUG_INFO_FILENAME    =   FileName;
                            AE->_INFO_OPTIMIZED = 0;
                            AE->_PARSE_DATA     = (char *)".";
                            AE->_HASH_DATA      = 0;
                            PIFList->Add(AE, DATA_ANALIZER_ELEMENT);
                        }
                    }
                }
            }
            //-----------------------------------------------------------//
            //_ID=ListContains(sPARSE,&GeneralMembers, true);
            _ID = GeneralMembers.ContainsString(sPARSE.c_str());
            if (!_ID) {
                if (PREC_ID == KEY_DLL_CALL) {
                    INTEGER cnt = PIFList->Count();
                    if (cnt >= 2) {
                        AnalizerElement *AE = (AnalizerElement *)PIFList->Item(cnt - 2);
                        if ((AE->TYPE == TYPE_CLASS) && (AE->ID > 0)) {
                            ClassCode *CC_T = (ClassCode *)ClassList.Item(AE->ID - 1);
                            if ((CC_T->NEEDED > -1) && (CC_T != CC)) {
                                cnt -= 2;
                                PIFList->Delete(cnt);
                                PIFList->Delete(cnt);

                                // adaug this
                                AnalizerElement *AE = new AnalizerElement;
                                AE->ID               = 1;
                                AE->TYPE             = TYPE_VARIABLE;
                                AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                                //AE->_DEBUG_INFO_FILENAME    =   FileName;
                                AE->_INFO_OPTIMIZED = 0;
                                AE->_PARSE_DATA     = (char *)"this";
                                AE->_HASH_DATA      = 0;
                                PIFList->Add(AE, DATA_ANALIZER_ELEMENT);

                                // adaug operatorul
                                AE                   = new AnalizerElement;
                                AE->ID               = KEY_SEL;
                                AE->TYPE             = TYPE_OPERATOR;
                                AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                                //AE->_DEBUG_INFO_FILENAME    =   FileName;
                                AE->_INFO_OPTIMIZED = 0;
                                AE->_PARSE_DATA     = (char *)".";
                                AE->_HASH_DATA      = 0;
                                PIFList->Add(AE, DATA_ANALIZER_ELEMENT);

                                /* Adaugat de curand !
                                 * sPARSE=CC_T->NAME;
                                 * sPARSE+=(char *)"@";
                                 * sPARSE+=C_FINALIZE;
                                 */
                                if (sPARSE == (char *)C_FINALIZE) {
                                    sPARSE  = CC_T->NAME;
                                    sPARSE += (char *)"@";
                                    sPARSE += C_FINALIZE;
                                }
                                //_ID=ListContains(sPARSE,&GeneralMembers,true);
                                _ID = GeneralMembers.ContainsString(sPARSE.c_str());
                            }
                        }
                    }
                }

                /*AnsiException *Exc=new AnsiException(ERR140,on_line ? on_line : P->LastLine(),140,sPARSE,FileName,CC->NAME,CM->NAME);
                 * Errors.Add(Exc,DATA_EXCEPTION);
                 * continue;*/
                if (!_ID) {
                    _ID = AddUndefinedMember(sPARSE, CC->NAME, CM->NAME, on_line ? on_line : P->LastLine());
                }
            }
        }

        if ((!_ID) && (TYPE == TYPE_METHOD)) {
            _ID = LinkStatic(sPARSE.c_str());     //LinkFunction(sPARSE,&ModuleList);
            // este apel in DLL ?
            if (_ID) {
                // da, atunci adaug un operator virtual !
                AnalizerElement *AE = new AnalizerElement;
                AE->ID               = STATIC_CLASS_DLL;
                AE->TYPE             = TYPE_CLASS;
                AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                //AE->_DEBUG_INFO_FILENAME    =   FileName;
                AE->_INFO_OPTIMIZED = 0;
                AE->_PARSE_DATA     = (char *)"LIBRARY";
                AE->_HASH_DATA      = 0;
                PIFList->Add(AE, DATA_ANALIZER_ELEMENT);


                // da, atunci adaug un operator virtual !
                //AnalizerElement *AE         =   new AnalizerElement;
                AE                   = new AnalizerElement;
                AE->ID               = KEY_DLL_CALL;
                AE->TYPE             = TYPE_OPERATOR;
                AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                //AE->_DEBUG_INFO_FILENAME    =   FileName;
                AE->_INFO_OPTIMIZED = 0;
                AE->_PARSE_DATA     = (char *)"::";
                AE->_HASH_DATA      = 0;
                PIFList->Add(AE, DATA_ANALIZER_ELEMENT);
            }
        }

        /*if ((!_ID) && (PREC_TYPE!=TYPE_OPERATOR) && (PREC_ID!=KEY_SEL)) {
         *  _ID=ClassExists(sPARSE);
         *  if (_ID)
         *      TYPE=TYPE_CLASS;
         * }*/
        // verific daca face parte din clasa curenta ...
        if ((!STATIC) && (!_ID) && (TYPE == TYPE_METHOD)) {
            if (CC->HasMember(sPARSE)) {
                if (sPARSE == CC->NAME) {
                    P->NextAtom(cached);
                    if ((cached == (char *)".") || (cached == (char *)"->") || (cached == (char *)"::")) {
                        _ID = 0;
                    } else {
                        _ID = GeneralMembers.ContainsString(sPARSE.c_str());
                    }
                } else {
                    P->NextAtom(cached);
                    if (cached != (char *)"::") {
                        _ID = GeneralMembers.ContainsString(sPARSE.c_str());
                    }
                }

                if (_ID) {
                    // nu poate fi folosit THIS intr-un membru static !!!
                    if (STATIC) {
                        AnsiException *Exc = new AnsiException(ERR930, on_line ? on_line : P->LastLine(), 930, sPARSE, FileName, CC->NAME, CM->NAME);
                        Errors.Add(Exc, DATA_EXCEPTION);
                        continue;
                    }
                    // adaug this
                    AnalizerElement *AE = new AnalizerElement;
                    AE->ID               = 1;
                    AE->TYPE             = TYPE_VARIABLE;
                    AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                    //AE->_DEBUG_INFO_FILENAME    =   FileName;
                    AE->_INFO_OPTIMIZED = 0;
                    AE->_PARSE_DATA     = (char *)"this";
                    AE->_HASH_DATA      = 0;
                    PIFList->Add(AE, DATA_ANALIZER_ELEMENT);

                    // adaug operatorul
                    AE                   = new AnalizerElement;
                    AE->ID               = KEY_SEL;
                    AE->TYPE             = TYPE_OPERATOR;
                    AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
                    //AE->_DEBUG_INFO_FILENAME    =   FileName;
                    AE->_INFO_OPTIMIZED = 0;
                    AE->_PARSE_DATA     = (char *)".";
                    AE->_HASH_DATA      = 0;
                    PIFList->Add(AE, DATA_ANALIZER_ELEMENT);
                }
            }
        } else
        if ((STATIC) && (!_ID) && (TYPE == TYPE_METHOD)) {
            if (CC->HasMember(sPARSE)) {
                //cached=P->NextAtom();
                if (!cached.Length()) {
                    P->NextAtom(cached);
                }
                if ((cached == (char *)".") || (cached == (char *)"->") || (cached == (char *)"::")) {
                    //nothing ... //_ID=0;
                } else {
                    //_ID=ListContains(sPARSE,&GeneralMembers, true);
                    _ID = GeneralMembers.ContainsString(sPARSE.c_str());

                    if (_ID) {
                        // nu poate fi folosit THIS intr-un membru static !!!
                        if (STATIC) {
                            AnsiException *Exc = new AnsiException(ERR930, on_line ? on_line : P->LastLine(), 930, sPARSE, FileName, CC->NAME, CM->NAME);
                            Errors.Add(Exc, DATA_EXCEPTION);
                            continue;
                        }
                    }
                }
            }
        }

        if (!_ID) {
            // verific daca nu cumva este clasa ... imi va folosi daca vreau sa folosesc
            // acel operator static ::
            _ID = ClassExists(sPARSE.c_str());
            if (_ID) {
                TYPE = TYPE_CLASS;
            }
        }

        if (!_ID) {
            if ((USE_IMPLICIT) && (TYPE == TYPE_METHOD)) {
                TYPE = TYPE_VARIABLE;
                VariableDESCRIPTOR *VD = new VariableDESCRIPTOR;
                VD->name   = sPARSE;
                VD->TYPE   = VARIABLE_NUMBER;
                VD->USED   = 0;
                VD->nValue = 0;
                VD->BY_REF = 0;
                if (IS_ARRAY) {
                    VD->TYPE = VARIABLE_ARRAY;
                }

                VDList->Add(VD, DATA_VAR_DESCRIPTOR);
                _ID      = VDList->Count();
                IS_ARRAY = 0;
            } else {
                //cached=P->NextAtom();
                if (!cached.Length()) {
                    P->NextAtom(cached);
                }

                if ((cached == (char *)".") || (cached == (char *)"::")) {
                    Warning(WRN10007, on_line ? on_line : P->LastLine(), 10007, sPARSE, FileName);

                    ClassCode *CC1 = new ClassCode(sPARSE, this);
                    CC1->DEFINED_LEVEL = this->INCLUDE_LEVEL;
                    CC1->NEEDED        = -1;
                    ClassList.Add(CC1, DATA_CLASS_CODE);
                    _ID = ClassList.Count();
                    AddUndefinedClass(sPARSE, CC->NAME, CM->NAME, on_line ? on_line : P->LastLine()) + 1;
                    TYPE = TYPE_CLASS;
                } else {
                    AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P->LastLine(), 10, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                    continue;
                }
            }
        }

        if ((TYPE == TYPE_OPERATOR) && ((PREC_TYPE == TYPE_VARIABLE) || (PREC_TYPE == TYPE_METHOD) || ((PREC_TYPE == TYPE_OPERATOR) && (PREC_ID == KEY_INDEX_CLOSE)))) {
            if (_ID == KEY_INC) {
                _ID = KEY_INC_LEFT;
            } else
            if (_ID == KEY_DEC) {
                _ID = KEY_DEC_LEFT;
            }
        }

        if ((TYPE == TYPE_OPERATOR) && (_ID == KEY_P_OPEN) && (PREC_TYPE == TYPE_VARIABLE)) {
            // este o variabila cu parametri => este o functzie delegate !
            INTEGER pos = PIFList->Count();

            // da, atunci adaug un operator virtual !
            AnalizerElement *AE = new AnalizerElement;
            AE->ID               = STATIC_CLASS_DELEGATE;
            AE->TYPE             = TYPE_CLASS;
            AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
            //AE->_DEBUG_INFO_FILENAME    =   FileName;
            AE->_INFO_OPTIMIZED = 0;
            AE->_PARSE_DATA     = (char *)"DELEGATE";
            AE->_HASH_DATA      = 0;
            PIFList->Insert(AE, pos - 1, DATA_ANALIZER_ELEMENT);


            // da, atunci adaug un operator virtual !
            //AnalizerElement *AE         =   new AnalizerElement;
            AE                   = new AnalizerElement;
            AE->ID               = KEY_DLL_CALL;
            AE->TYPE             = TYPE_OPERATOR;
            AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
            //AE->_DEBUG_INFO_FILENAME    =   FileName;
            AE->_INFO_OPTIMIZED = 0;
            AE->_PARSE_DATA     = (char *)"::";
            AE->_HASH_DATA      = 0;
            PIFList->Insert(AE, pos, DATA_ANALIZER_ELEMENT);
        }

        // filtrez daca este SUMA sau semn +,-
        if ((TYPE == TYPE_OPERATOR) && ((PREC_TYPE == TYPE_OPERATOR) || (PREC_TYPE == TYPE_KEYWORD)) && (PREC_ID != KEY_P_CLOSE) && (PREC_ID != KEY_INDEX_CLOSE)) {
            if (_ID == KEY_SUM) {
                _ID = KEY_POZ;
            } else
            if (_ID == KEY_DIF) {
                _ID = KEY_NEG;
            }
        }

        // verific daca nu este cumva vorba de o declaratzie ... de genul
        // "function" sau "var" sau "class"
        if ((TYPE != TYPE_KEYWORD) || (_ID != KEY_VAR)) {
            AnalizerElement *AE = new AnalizerElement;
            AE->ID               = _ID;
            AE->TYPE             = TYPE;
            AE->_DEBUG_INFO_LINE = on_line ? on_line : P->LastLine();
            //AE->_DEBUG_INFO_FILENAME    =   FileName;
            AE->_INFO_OPTIMIZED = 0;
            AE->_PARSE_DATA     = sPARSE;
            AE->_HASH_DATA      = 0;
            PIFList->Add(AE, DATA_ANALIZER_ELEMENT);

            if (TYPE == TYPE_OPERATOR) {
                switch (_ID) {
                    case KEY_P_OPEN:
                    case KEY_INDEX_OPEN:
                        ATOMIC_LEVEL++;
                        break;

                    case KEY_P_CLOSE:
                    case KEY_INDEX_CLOSE:
                        ATOMIC_LEVEL--;
                        break;
                }
            }

            if ((IN_VAR_STATAMENT) && (!ATOMIC_LEVEL)) {
                if ((TYPE == TYPE_SEPARATOR) && (_ID == KEY_SEP)) {
                    IN_VAR_STATAMENT = 0;
                }
            }
        } else {
            IN_VAR_STATAMENT = 1;
        }
    } while (BEGIN_END_LEVEL);

    // verific de variabile nefolosite ...
    for (int i = 0; i < VDList->Count(); i++) {
        VariableDESCRIPTOR *VD = (VariableDESCRIPTOR *)VDList->Item(i);
        if (!VD->USED) {
            //AnsiException *Exc=0;
            if (OPERATOR) {
                Warning(WRN10003, on_line ? on_line : P->LastLine(), 10003, AnsiString("var '") + VD->name.c_str() + AnsiString("' in operator '") + CM->NAME.c_str() + (char *)"'");
            } else {
                Warning(WRN10003, on_line ? on_line : P->LastLine(), 10003, AnsiString("var '") + VD->name.c_str() + AnsiString("' in function '") + CM->NAME.c_str() + (char *)"'");
            }


            /*if (OPERATOR)
             *  Exc=new AnsiException(WRN10003,on_line ? on_line : P->LastLine(),10003,AnsiString("var '") + VD->name + AnsiString("' in operator '")+CM->NAME+(char *)"'",FileName);
             * else
             *  Exc=new AnsiException(WRN10003,on_line ? on_line : P->LastLine(),10003,AnsiString("var '") + VD->name + AnsiString("' in function '")+CM->NAME+(char *)"'",FileName);
             * Warnings.Add(Exc,DATA_EXCEPTION);*/
        }
    }
    return(0);
}

INTEGER PIFAlizator::BuildOverride(ClassCode *CC, AnsiParser *P, INTEGER on_line) {
    INTEGER    TYPE = -1;
    INTEGER    _ID  = -1;
    AnsiString sPARSE;     //          =   P->NextAtom();

    P->NextAtom(sPARSE);

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((_ID) || (TYPE != TYPE_METHOD)) {
        AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P->LastLine(), 010, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    } else {
        CC->RemoveMember(sPARSE, on_line ? on_line : P->LastLine(), FileName);
    }
    return(0);
}

INTEGER PIFAlizator::BuildProperty(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER ACCESS) {
    INTEGER    TYPE = -1;
    INTEGER    _ID  = -1;
    AnsiString sPARSE;     //          =   P->NextAtom();

    P->NextAtom(sPARSE);

    ClassMember *CM = CC->AddMember(sPARSE, on_line ? on_line : P->LastLine(), FileName, ACCESS, true);
    CM->IS_FUNCTION = 3;

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((_ID) || (TYPE != TYPE_METHOD)) {
        AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P->LastLine(), 010, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    }

    //if (!ListContains(sPARSE,&GeneralMembers, true))
    if (!GeneralMembers.ContainsString(sPARSE.c_str())) {
        AddGeneralMember(sPARSE);
    } else {
        DefineMember(sPARSE);
    }

    //sPARSE  =   P->NextAtom();
    P->NextAtom(sPARSE);

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((TYPE != TYPE_KEYWORD) && (_ID != KEY_BEGIN)) {
        AnsiException *Exc = new AnsiException(ERR350, on_line ? on_line : P->LastLine(), 350, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(1);
    }

    INTEGER SET_MEMBER;
    INTEGER _OUT    = 0;
    INTEGER HAS_GET = 0;
    INTEGER HAS_SET = 0;
    do {
        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);

        TYPE = GetType(sPARSE);
        _ID  = GetID(sPARSE);

        if (((_ID != KEY_SET) && (_ID != KEY_GET)) || (TYPE != TYPE_METHOD)) {
            AnsiException *Exc = new AnsiException(ERR380, on_line ? on_line : P->LastLine(), 380, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        } else {
            if (_ID == KEY_SET) {
                if (HAS_SET) {
                    AnsiException *Exc = new AnsiException(ERR740, on_line ? on_line : P->LastLine(), 740, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                } else {
                    HAS_SET = 1;
                }
            } else
            if (_ID == KEY_GET) {
                if (HAS_GET) {
                    AnsiException *Exc = new AnsiException(ERR730, on_line ? on_line : P->LastLine(), 730, sPARSE, FileName, CC->NAME, CM->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                } else {
                    HAS_GET = 1;
                }
            }
        }

        SET_MEMBER = (_ID == KEY_SET);

        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);

        TYPE = GetType(sPARSE);
        _ID  = GetID(sPARSE);

        if ((_ID) || (TYPE != TYPE_METHOD)) {
            AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P->LastLine(), 010, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
        }

        //INTEGER POSITION=ListContains(sPARSE,&GeneralMembers, true);
        INTEGER POSITION = GeneralMembers.ContainsString(sPARSE.c_str());
        if (!POSITION) {
            //AddGeneralMember(sPARSE);
            AddUndefinedMember(sPARSE, CC->NAME, CM->NAME, on_line ? on_line : P->LastLine());
            POSITION = GeneralMembers.Count();
        }

        SET_MEMBER  ?   CM->MEMBER_SET = POSITION    :  CM->MEMBER_GET = POSITION;

        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);

        TYPE = GetType(sPARSE);
        _ID  = GetID(sPARSE);

        if ((TYPE == TYPE_OPERATOR) && (_ID == KEY_COMMA)) {
            _OUT = 0;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_END)) {
            _OUT = 1;
        } else {
            AnsiException *Exc = new AnsiException(ERR360, on_line ? on_line : P->LastLine(), 360, sPARSE, FileName, CC->NAME, CM->NAME);
            Errors.Add(Exc, DATA_EXCEPTION);
            return(1);
        }
    } while (!_OUT);
    // verific daca are set, dar lipseste get ...
    if ((HAS_SET) && (!HAS_GET)) {
        AnsiException *Exc = new AnsiException(ERR720, on_line ? on_line : P->LastLine(), 720, NULL_STRING, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(1);
    }

    return(0);
}

INTEGER PIFAlizator::BuildEvent(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER ACCESS) {
    INTEGER    TYPE = -1;
    INTEGER    _ID  = -1;
    AnsiString sPARSE;     //          =   P->NextAtom();

    P->NextAtom(sPARSE);

    ClassMember *CM = CC->AddMember(sPARSE, on_line ? on_line : P->LastLine(), FileName, ACCESS, true);
    CM->IS_FUNCTION = 2;

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((_ID) || (TYPE != TYPE_METHOD)) {
        AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P->LastLine(), 010, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    }

    //if (!ListContains(sPARSE,&GeneralMembers, true))
    if (!GeneralMembers.ContainsString(sPARSE.c_str())) {
        AddGeneralMember(sPARSE);
    } else {
        DefineMember(sPARSE);
    }


    //sPARSE  =   P->NextAtom();
    P->NextAtom(sPARSE);

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((TYPE != TYPE_METHOD) && (_ID != KEY_TRIGERS)) {
        AnsiException *Exc = new AnsiException(ERR330, on_line ? on_line : P->LastLine(), 330, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(1);
    }

    //sPARSE  =   P->NextAtom();
    P->NextAtom(sPARSE);

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((_ID) || (TYPE != TYPE_METHOD)) {
        AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P->LastLine(), 010, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
    }

    //INTEGER POSITION=ListContains(sPARSE,&GeneralMembers, true);
    INTEGER POSITION = GeneralMembers.ContainsString(sPARSE.c_str());
    if (!POSITION) {
        AddGeneralMember(sPARSE);
        POSITION = GeneralMembers.Count();
    }
    //CM->MEMBER_TRIGERS     =   POSITION;
    CM->MEMBER_GET = POSITION;

    //sPARSE  =   P->NextAtom();
    P->NextAtom(sPARSE);

    TYPE = GetType(sPARSE);
    _ID  = GetID(sPARSE);

    if ((TYPE != TYPE_SEPARATOR) && (_ID != KEY_SEP)) {
        AnsiException *Exc = new AnsiException(ERR340, on_line ? on_line : P->LastLine(), 340, sPARSE, FileName, CC->NAME, CM->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(1);
    }
    return(0);
}

INTEGER PIFAlizator::ExtendClass(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER OWNER_CLSID) {
    AnsiString sPARSE;     //          =   P->NextAtom();

    P->NextAtom(sPARSE);

    INTEGER TYPE = GetType(sPARSE);
    INTEGER _ID  = GetID(sPARSE);

    if ((_ID) || (TYPE != TYPE_METHOD)) {
        AnsiException *Exc = new AnsiException(ERR170, on_line ? on_line : P->LastLine(), 170, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(0);
    }

    _ID = ClassExists(sPARSE.c_str());

    if (!_ID) {
        AnsiException *Exc = new AnsiException(ERR180, on_line ? on_line : P->LastLine(), 180, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(0);
    }
    ClassCode *PARENT = (ClassCode *)ClassList [_ID - 1];

    if (PARENT == CC) {
        AnsiException *Exc = new AnsiException(ERR1240, on_line ? on_line : P->LastLine(), 1240, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(0);
    }

    if (_ID > OWNER_CLSID) {
        AnsiException *Exc = new AnsiException(ERR1230, on_line ? on_line : P->LastLine(), 1230, sPARSE, FileName, CC->NAME);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(0);
    }

    if (!CC->first_parent) {
        CC->first_parent = PARENT;
    }

    INTEGER MemberCount = PARENT->Members->Count();
    for (INTEGER i = 0; i < MemberCount; i++) {
        ClassMember *CM = (ClassMember *)PARENT->Members->Item(i);
        CC->AddLinked(CM);

        /*ClassMember *newCM  =   CC->AddMember(CM->NAME,on_line ? on_line : P->LastLine(),FileName,CM->ACCESS,!CM->IS_FUNCTION);
         *
         * if (!newCM)
         *  return 0;
         *
         * newCM->IS_FUNCTION      =   CM->IS_FUNCTION;
         *
         * newCM->VD.BY_REF        =   CM->VD.BY_REF;
         * newCM->VD.USED          =   CM->VD.USED;
         * newCM->VD.name          =   CM->VD.name;
         * newCM->VD.TYPE          =   CM->VD.TYPE;
         * newCM->VD.value         =   CM->VD.value;
         * newCM->PARAMETERS_COUNT =   CM->PARAMETERS_COUNT;
         * newCM->MEMBER_GET       =   CM->MEMBER_GET;
         * newCM->MEMBER_SET       =   CM->MEMBER_SET;
         * newCM->MEMBER_TRIGERS   =   CM->MEMBER_TRIGERS;
         *
         * new((ClassCode *)(CM->Defined_In))->_DEBUG_INFO_FILENAME = ((ClassCode *)(CM->Defined_In))->_DEBUG_INFO_FILENAME;
         *
         * newCM->Defined_In       =   CM->Defined_In;
         *
         * if (CM->IS_FUNCTION==1) {
         *  INTEGER PIF_COUNT=CM->PIF_DATA->Count();
         *  for (INTEGER j=0;j<PIF_COUNT;j++) {
         *      AnalizerElement *AE     =   (AnalizerElement *)CM->PIF_DATA->Item(j);
         *      AnalizerElement *newAE  =   new AnalizerElement;
         *
         *      //newAE->_DEBUG_INFO_FILENAME =   AE->_DEBUG_INFO_FILENAME;
         *      newAE->_DEBUG_INFO_LINE     =   AE->_DEBUG_INFO_LINE;
         *      newAE->_INFO_OPTIMIZED      =   AE->_INFO_OPTIMIZED;
         *      newAE->_PARSE_DATA          =   AE->_PARSE_DATA;
         *      newAE->ID                   =   AE->ID;
         *      newAE->TYPE                 =   AE->TYPE;
         *      newAE->_HASH_DATA           =   0;
         *
         *      newCM->PIF_DATA->Add(newAE, DATA_ANALIZER_ELEMENT);
         *  }
         *
         *  INTEGER DESC_COUNT=CM->VariableDescriptors->Count();
         *  for (INTEGER k=0;k<DESC_COUNT;k++) {
         *      VariableDESCRIPTOR *_VD     =   (VariableDESCRIPTOR *)CM->VariableDescriptors->Item(k);
         *      VariableDESCRIPTOR *newVD   =   new VariableDESCRIPTOR;
         *
         *      newVD->BY_REF               =   _VD->BY_REF;
         *      newVD->USED                 =   _VD->USED;
         *      newVD->name                 =   _VD->name;
         *      newVD->TYPE                 =   _VD->TYPE;
         *      newVD->value                =   _VD->value;
         *      newVD->nValue               =   _VD->nValue;
         *
         *      newCM->VariableDescriptors->Add(newVD, DATA_VAR_DESCRIPTOR);
         *  }
         * }
         */
    }
    return(0);
}

INTEGER PIFAlizator::BuildClass(AnsiParser *P, INTEGER on_line) {
    INTEGER    BEGIN_END_LEVEL = 0;
    INTEGER    TYPE            = 0;
    INTEGER    _ID             = 0;
    INTEGER    ACCESS          = ACCESS_PUBLIC;
    AnsiString sPARSE;     //          =   P->NextAtom();

    P->NextAtom(sPARSE);
    INTEGER START_CLASS      = 0;
    INTEGER cls_index        = 0xFFFFFFF;
    char    STATIC           = 0;
    char    ALREADY_REPORTED = 0;

    ClassCode *IMPLICIT_PTR = (ClassCode *)ClassExists(sPARSE.c_str(), true, &cls_index);

    //if (ClassExists(sPARSE)) {
    if ((IMPLICIT_PTR) && (IMPLICIT_PTR->NEEDED != -1)) {
        AnsiException *Exc = new AnsiException(ERR750, on_line ? on_line : P->LastLine(), 750, sPARSE, FileName);
        Errors.Add(Exc, DATA_EXCEPTION);
    }

    ClassCode *CC = IMPLICIT_PTR;

    if (!CC) {
        CC = new ClassCode(sPARSE, this);
        ClassList.Add(CC, DATA_CLASS_CODE);
    } else {
        CC->NEEDED = 1;
        DefineClass(sPARSE);
    }

    CC->DEFINED_LEVEL = this->INCLUDE_LEVEL;


    if (CC->NAME == (char *)MAIN_ENTRY_POINT) {
        if (CC->NEEDED != -1) {
            CC->NEEDED = 1;
        }
        ENTRY_CLASS = ClassList.Count();
    }

    do {
        //sPARSE  =   P->NextAtom();
        P->NextAtom(sPARSE);

        if (sPARSE == NULL_STRING) {
            break;
        }

        TYPE = GetType(sPARSE);
        _ID  = 0;

        if (TYPE == TYPE_HEX_NUMBER) {
            INTEGER TEMP_NR = 0;
            sscanf(sPARSE.c_str() + 2, "%x", &TEMP_NR);
            sPARSE = AnsiString(TEMP_NR);
            TYPE   = TYPE_NUMBER;
        }

        if (!_ID) {
            _ID = GetID(sPARSE);
            // daca arata ca o functie, si are si ID, inseamna ca este cuvant cheie !
            if ((TYPE == TYPE_METHOD) && (_ID)) {
                TYPE = TYPE_KEYWORD;
            }

            if (TYPE == TYPE_SEPARATOR) {
                continue;
            }

            if (TYPE == TYPE_KEYWORD) {
                if (_ID == KEY_BEGIN) {
                    BEGIN_END_LEVEL++;
                    START_CLASS = 1;
                    continue;
                } else if (_ID == KEY_END) {
                    BEGIN_END_LEVEL--;
                    continue;
                }
            }
        }

        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_STATIC)) {
            STATIC = 1;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_PUBLIC)) {
            ACCESS = ACCESS_PUBLIC;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_PRIVATE)) {
            ACCESS = ACCESS_PRIVATE;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_PROTECTED)) {
            ACCESS = ACCESS_PROTECTED;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_EXTENDS)) {
            if (!START_CLASS) {
                ExtendClass(CC, P, on_line, cls_index);
            } else {
                AnsiException *Exc = new AnsiException(ERR470, on_line ? on_line : P->LastLine(), 470, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
            ALREADY_REPORTED = 0;
        } else
        if ((TYPE == TYPE_KEYWORD) && ((_ID == KEY_FUNCTION) || (_ID == KEY_OPERATOR))) {
            if ((STATIC) && (_ID != KEY_FUNCTION)) {
                AnsiException *Exc = new AnsiException(ERR940, on_line ? on_line : P->LastLine(), 940, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
            BuildFunction(CC, P, on_line, ACCESS, _ID == KEY_OPERATOR, (_ID != KEY_OPERATOR) ? STATIC : 0);
            ACCESS           = ACCESS_PUBLIC;
            STATIC           = 0;
            ALREADY_REPORTED = 0;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_VAR)) {
            if (STATIC) {
                AnsiException *Exc = new AnsiException(ERR940, on_line ? on_line : P->LastLine(), 940, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                STATIC = 0;
            }
            BuildVariable(CC, P, on_line, ACCESS);
            ACCESS           = ACCESS_PUBLIC;
            ALREADY_REPORTED = 0;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_CEVENT)) {
            if (STATIC) {
                AnsiException *Exc = new AnsiException(ERR940, on_line ? on_line : P->LastLine(), 940, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                STATIC = 0;
            }
            BuildEvent(CC, P, on_line, ACCESS);
            ACCESS           = ACCESS_PUBLIC;
            ALREADY_REPORTED = 0;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_PROPERTY)) {
            if (STATIC) {
                AnsiException *Exc = new AnsiException(ERR940, on_line ? on_line : P->LastLine(), 940, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                STATIC = 0;
            }
            BuildProperty(CC, P, on_line, ACCESS);
            ACCESS           = ACCESS_PUBLIC;
            ALREADY_REPORTED = 0;
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_OVERRIDE)) {
            if (STATIC) {
                AnsiException *Exc = new AnsiException(ERR940, on_line ? on_line : P->LastLine(), 940, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
                STATIC = 0;
            }
            if (ACCESS != ACCESS_PUBLIC) {
                AnsiException *Exc = new AnsiException(ERR710, on_line ? on_line : P->LastLine(), 710, sPARSE, FileName, CC->NAME);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
            BuildOverride(CC, P, on_line);
            ACCESS           = ACCESS_PUBLIC;
            ALREADY_REPORTED = 0;
        } else {
            if (TYPE == TYPE_METHOD) {
                //AnsiException *Exc=new AnsiException(ERR450,on_line ? on_line : P->LastLine(),450,sPARSE,FileName,CC->NAME);
                //Errors.Add(Exc,DATA_EXCEPTION);
                BuildFunction(CC, P, on_line, ACCESS, 0, STATIC, sPARSE.c_str());
                ACCESS           = ACCESS_PUBLIC;
                STATIC           = 0;
                ALREADY_REPORTED = 0;
            } else
            if ((!ALREADY_REPORTED) || (!START_CLASS)) {
                if (TYPE == TYPE_KEYWORD) {
                    AnsiException *Exc = new AnsiException(ERR430, on_line ? on_line : P->LastLine(), 430, sPARSE, FileName, CC->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                } else
                if (TYPE == TYPE_OPERATOR) {
                    AnsiException *Exc = new AnsiException(ERR440, on_line ? on_line : P->LastLine(), 440, sPARSE, FileName, CC->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                } else {
                    AnsiException *Exc = new AnsiException(ERR460, on_line ? on_line : P->LastLine(), 460, sPARSE, FileName, CC->NAME);
                    Errors.Add(Exc, DATA_EXCEPTION);
                }
                ALREADY_REPORTED = 1;
            }
            STATIC = 0;
            ACCESS = ACCESS_PUBLIC;
        }
    } while ((BEGIN_END_LEVEL) || (!START_CLASS));
    return(0);
}

INTEGER PIFAlizator::Warning(char *WRN, int line, int wrn_code, char *extra, char *filename) {
    if (USE_WARN) {
        AnsiException *Exc = new AnsiException(WRN, line, wrn_code, extra, filename ? filename : FileName.c_str());
        Warnings.Add(Exc, DATA_EXCEPTION);
        return(1);
    }
    return(0);
}

INTEGER PIFAlizator::BuildPragma(AnsiParser *P) {
    AnsiString pragma = P->GetConstant();
    AnsiParser PRAGMA_PARSER(&pragma, 0);
    AnsiString first_param;     //=PRAGMA_PARSER.NextAtom(1);

    PRAGMA_PARSER.NextAtom(first_param, 1);

    if (first_param == PRAGMA_WARNINGS) {
        //first_param=PRAGMA_PARSER.NextAtom(1);
        PRAGMA_PARSER.NextAtom(first_param, 1);
        if (first_param == PRAGMA_ON) {
            USE_WARN = 1;
        } else
        if (first_param == PRAGMA_OFF) {
            USE_WARN = 0;
        } else {
            Warning(WRN10006, P->LastLine(), 10006, first_param);
        }
    } else
    if (first_param == PRAGMA_EXCEPTIONS) {
        //first_param=PRAGMA_PARSER.NextAtom(1);
        PRAGMA_PARSER.NextAtom(first_param, 1);
        if (first_param == PRAGMA_ON) {
            USE_EXC = 1;
        } else
        if (first_param == PRAGMA_OFF) {
            USE_EXC = 0;
        } else {
            Warning(WRN10006, P->LastLine(), 10006, first_param);
        }
    } else
    if (first_param == PRAGMA_IMPLICIT) {
        PRAGMA_PARSER.NextAtom(first_param, 1);
        //first_param=PRAGMA_PARSER.NextAtom(1);
        if (first_param == PRAGMA_ON) {
            USE_IMPLICIT = 1;
        } else
        if (first_param == PRAGMA_OFF) {
            USE_IMPLICIT = 0;
        } else {
            Warning(WRN10006, P->LastLine(), 10006, first_param);
        }
    } else {
        Warning(WRN10005, P->LastLine(), 10005, first_param);
    }

    return(0);
}

INTEGER PIFAlizator::IncludePackage(AnsiString filename) {
    // verific daca nu cumva este deja inclus
    // este deja inclus ...
    AnsiString RELATIVE_MODULE_NAME = TEMP_INC_DIR + filename;
    RELATIVE_MODULE_NAME = this->NormalizePath(&RELATIVE_MODULE_NAME);
    // filename with no absolute path in it
    AnsiString clean_filename_dir;

    if (ListContains(/*filename*/ RELATIVE_MODULE_NAME, &IncludedList)) {
        return(1);
    }

    //AnsiString fname=filename;

    //filename+=(char *)DEFAULT_PACK_EXTENSION;
    //RELATIVE_MODULE_NAME+=(char *)DEFAULT_PACK_EXTENSION;
    FILE *in = fopen(/*filename*/ AnsiString(RELATIVE_MODULE_NAME + (char *)DEFAULT_PACK_EXTENSION).c_str(), "rb");

    if (!in) {
        AnsiString MODULE_NAME_DIR = AnsiString(INCLUDE_DIR) + filename;
        if (ListContains(filename /*MODULE_NAME_DIR*/, &IncludedList)) {
            return(1);
        }

        in = fopen(AnsiString(MODULE_NAME_DIR + (char *)DEFAULT_PACK_EXTENSION).c_str(), "rb");
        clean_filename_dir = filename;
        filename           = MODULE_NAME_DIR;
    } else {
        filename           = RELATIVE_MODULE_NAME;
        clean_filename_dir = RELATIVE_MODULE_NAME;
    }
    if (in) {
        fclose(in);
        IncludedList.Add(new AnsiString(clean_filename_dir), DATA_STRING);
        //IncludedList.Add(new AnsiString(/*fname*/filename),DATA_STRING);

        AnsiString OLD_TEMP_INC_DIR = TEMP_INC_DIR;
        TEMP_INC_DIR = GetPath(&clean_filename_dir);
        //TEMP_INC_DIR=GetPath(&filename);

        Unserialize(AnsiString(filename + (char *)DEFAULT_PACK_EXTENSION).c_str(), true);

        TEMP_INC_DIR = OLD_TEMP_INC_DIR;

        return(1);
    }
    return(0);
}

AnsiString PIFAlizator::NormalizePath(AnsiString *MODULE_NAME) {
    char *str=MODULE_NAME->c_str();
    int len=MODULE_NAME->Length();

    int path_pos[0xFF];
    int path_count=0;
    int ignore_next=0;

    // warning: prec_str is in reverse order !
    AnsiString prec_str;
    for (int i=len-1;i>=-1;i--) {
        // force a '/'
        char c= (i >= 0) ? str[i] : '/';
        if (c=='\\')
            c='/';

        if (c=='/') {
            if (prec_str.Length()) {
                if (i>=0)
                    str[i]=0;
                if (prec_str==(char *)"..") {
                    ignore_next++;
                } else
                if (prec_str!=(char *)".") {
                    if (ignore_next)
                        ignore_next--;
                    else {
                        if (path_count<0xFF)
                            path_pos[path_count++]=i+1;
                    }

                }
                prec_str=(char *)"";
            }
        } else
            prec_str+=c;
    }


    AnsiString result;
    for (int i=0;i<ignore_next;i++)
        result+="../";

    char not_first=0;
    for (int i=path_count-1;i>=0;i--) {
        int pos=path_pos[i];
        if (pos>=0) {
            if (not_first)
                result+=(char *)"/";
            else
                not_first=1;
            result+=str+pos;
        }
    }
    return result;
}

INTEGER PIFAlizator::IncludeFile(AnsiString MODULE_NAME, INTEGER on_line) {
    INCLUDE_LEVEL++;
    AnsiString FILENAME             = MODULE_NAME;
    AnsiString RELATIVE_MODULE_NAME = TEMP_INC_DIR + MODULE_NAME;
    AnsiString clean_filename_dir;

    RELATIVE_MODULE_NAME            = this->NormalizePath(&RELATIVE_MODULE_NAME);

    if (!IncludePackage(MODULE_NAME)) {
        AnsiString ModuleStream;
        // incerc sa incarc modulul ...
        if (ModuleStream.LoadFile(/*MODULE_NAME*/ RELATIVE_MODULE_NAME.c_str())) {
            AnsiString MODULE_NAME_DIR = AnsiString(INCLUDE_DIR) + MODULE_NAME;
            if (ModuleStream.LoadFile(MODULE_NAME_DIR.c_str())) {
                AnsiException *Exc = new AnsiException(ERR270, on_line, 270, MODULE_NAME, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
                INCLUDE_LEVEL--;
                return(0);
            }
            clean_filename_dir = MODULE_NAME;
            MODULE_NAME        = MODULE_NAME_DIR;
        } else {
            MODULE_NAME        = RELATIVE_MODULE_NAME;
            clean_filename_dir = RELATIVE_MODULE_NAME;
        }
        if (!ListContains(/*FILENAME*/ /*MODULE_NAME*/ clean_filename_dir, &IncludedList)) {
            IncludedList.Add(new AnsiString(/*FILENAME*/ /*MODULE_NAME*/ clean_filename_dir), DATA_STRING);
            AnsiString OldFile = FileName;
            FileName = MODULE_NAME;

            AnsiString OLD_TEMP_INC_DIR = TEMP_INC_DIR;
            TEMP_INC_DIR = GetPath(&clean_filename_dir /*MODULE_NAME*/ /*FILENAME*/);
            Execute(&ModuleStream, /*on_line*/ 0, USE_WARN, USE_EXC, USE_IMPLICIT);
            TEMP_INC_DIR = OLD_TEMP_INC_DIR;

            FileName = OldFile;
        }
    }
    INCLUDE_LEVEL--;
    return(1);
}

AnsiString PIFAlizator::GetPath(AnsiString *S) {
    int        len = S->Length();
    AnsiString result;
    AnsiString tmp;

    char *s = S->c_str();

    for (int i = len - 1; i >= 0; i--) {
        if ((s [i] == '/') || (s [i] == '\\')) {
            tmp = s;
            char *tmp2 = tmp.c_str();
            if (i < len - 1) {
                tmp2 [i + 1] = 0;
            }
            result = tmp2;
            break;
        }
    }
    return(result);
}

INTEGER PIFAlizator::Execute(AnsiString *Stream, INTEGER on_line, char _USE_WARN, char _USE_EXC, char _USE_IMPLICIT) {
    AnsiParser P(Stream, &ConstantList);
    AnsiString sPARSE;

    INTEGER _ID;
    INTEGER TYPE;

    P.Atomize();

    do {
        _ID = 0;

        //sPARSE  =   P.NextAtom();
        P.NextAtom(sPARSE);

        //if (sPARSE==NULL_STRING)
        if (!sPARSE.Length()) {
            break;
        }

        // iau tipul atomului
        TYPE = GetType(sPARSE);

        if (TYPE == TYPE_HEX_NUMBER) {
            INTEGER TEMP_NR = 0;
            sscanf(sPARSE.c_str() + 2, "%x", &TEMP_NR);
            sPARSE = AnsiString(TEMP_NR);
            TYPE   = TYPE_NUMBER;
        }

        if (TYPE < 0) {
            AnsiException *Exc = new AnsiException(ERR050, on_line ? on_line : P.LastLine(), 50, sPARSE, FileName);
            Errors.Add(Exc, DATA_EXCEPTION);
            continue;
        }

        if (!_ID) {
            // iau ID-ul lui, in caz ca este cuvant rezervat !
            _ID = GetID(sPARSE);

            // daca arata ca o functie, si are si ID, inseamna ca este cuvant cheie !
            if ((TYPE == TYPE_METHOD) && (_ID)) {
                TYPE = TYPE_KEYWORD;
            }
        }

        if (!_ID) {
            AnsiException *Exc = new AnsiException(ERR010, on_line ? on_line : P.LastLine(), 10, sPARSE, FileName);
            Errors.Add(Exc, DATA_EXCEPTION);
            continue;
        }

        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_CLASS)) {
            BuildClass(&P, on_line);
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_INCLUDE)) {
            //AnsiString MODULE_NAME=P.GetConstant();//P.NextAtom();
            //MODULE_NAME=Strip(MODULE_NAME);
            AnsiString MODULE_NAME = Strip2(P.GetConstant());
            if (!IncludeFile(MODULE_NAME, on_line ? on_line : P.LastLine())) {
                continue;
            }
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_DEFINE)) {
            // nu se dezvolta constantele !
            //sPARSE=P.NextAtom(1);
            P.NextAtom(sPARSE, 1);
            if (GetType(sPARSE) != TYPE_METHOD) {
                AnsiException *Exc = new AnsiException(ERR820, on_line ? on_line : P.LastLine(), 820, sPARSE, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else {
                AnsiString         value = P.GetConstant();            //P.NextAtom();
                VariableDESCRIPTOR *VD   = new VariableDESCRIPTOR;
                VD->BY_REF = 0;
                VD->name   = sPARSE;
                VD->value  = value;
                VD->USED   = this->INCLUDE_LEVEL;

                INTEGER POS = VariableIsDescribed(sPARSE, &ConstantList);

                if (POS) {
                    ConstantList.Delete(POS - 1);

                    /*AnsiException *Exc=new AnsiException(WRN10004,on_line ? on_line : P.LastLine(),10004,sPARSE,FileName);
                     * Warnings.Add(Exc,DATA_EXCEPTION);*/
                    Warning(WRN10004, on_line ? on_line : P.LastLine(), 10004, sPARSE);
                }

                ConstantList.Add(VD, DATA_VAR_DESCRIPTOR);
            }
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_IMPORT)) {
            // import DLL form !
            sPARSE = Strip2(P.GetConstant() /*P.NextAtom()*/);
            sPARSE = this->NormalizePath(&sPARSE);
            int  li_len           = ModuleNamesList.Count();
            bool already_imported = false;
            for (int il = 0; il < li_len; il++) {
                AnsiString *mod = (AnsiString *)ModuleNamesList.Item(il);
                if ((mod) && (*mod == sPARSE)) {
                    already_imported = true;
                    break;
                }
            }
            if (!already_imported) {
                ModuleNamesList.Add(new AnsiString(sPARSE), DATA_STRING);
                if (!ImportModule(AnsiString(IMPORT_DIR) + sPARSE, &Errors, on_line ? on_line : P.LastLine(), FileName, &ModuleList, this, 1))
                    ImportModule(sPARSE, &Errors, on_line ? on_line : P.LastLine(), FileName, &ModuleList, this);
            }
        } else
        if ((TYPE == TYPE_KEYWORD) && (_ID == KEY_PRAGMA)) {
            BuildPragma(&P);
        } else {
            if (TYPE == TYPE_KEYWORD) {
                AnsiException *Exc = new AnsiException(ERR430, on_line ? on_line : P.LastLine(), 430, sPARSE, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else
            if (TYPE == TYPE_OPERATOR) {
                AnsiException *Exc = new AnsiException(ERR440, on_line ? on_line : P.LastLine(), 440, sPARSE, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else
            if (TYPE == TYPE_METHOD) {
                AnsiException *Exc = new AnsiException(ERR450, on_line ? on_line : P.LastLine(), 450, sPARSE, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else
            if (TYPE == TYPE_SEPARATOR) {
                AnsiException *Exc = new AnsiException(ERR480, on_line ? on_line : P.LastLine(), 480, sPARSE, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
            } else {
                AnsiException *Exc = new AnsiException(ERR460, on_line ? on_line : P.LastLine(), 460, sPARSE, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
            }
        }
    } while (sPARSE != NULL_STRING);

    //USE_WARN=_USE_WARN;
    //USE_EXC=_USE_EXC;
    USE_IMPLICIT = _USE_IMPLICIT;
    return(0);
}

INTEGER PIFAlizator::CheckUndefinedMembers() {
    int cnt = this->UndefinedMembers.Count();

    for (int i = 0; i < cnt; i++) {
        UndefinedMember *UM  = (UndefinedMember *)UndefinedMembers.Item(i);
        AnsiException   *Exc = new AnsiException(ERR140, UM->line, 140, UM->name, UM->filename, UM->_CLASS, UM->_MEMBER);
        Errors.Add(Exc, DATA_EXCEPTION);
    }
    return(cnt);
}

INTEGER PIFAlizator::CheckUndefinedClasses() {
    int cnt = this->UndefinedClasses.Count();

    for (int i = 0; i < cnt; i++) {
        UndefinedMember *UM  = (UndefinedMember *)UndefinedClasses.Item(i);
        AnsiException   *Exc = new AnsiException(ERR100, UM->line, 100, UM->name, UM->filename, UM->_CLASS, UM->_MEMBER);
        Errors.Add(Exc, DATA_EXCEPTION);
    }
    return(cnt);
}

INTEGER PIFAlizator::Execute() {
    int result = Execute(InputStream);

    CheckUndefinedMembers();
    CheckUndefinedClasses();
    return(result);
}

INTEGER PIFAlizator::ErrorCount() {
    return(Errors.Count());
}

INTEGER PIFAlizator::WarningCount() {
    return(Warnings.Count());
}

#define COMPILED_FILE_ID    "CONC0100"
#define PACKAGE_FILE_ID     "PONC0100"
int PIFAlizator::Serialize(char *filename, bool is_lib) {
    FILE *out = fopen(filename, "wb");

    if (out) {
        // scriu id-ul fisierului ...
        if (is_lib) {
            concept_fwrite(PACKAGE_FILE_ID, strlen(PACKAGE_FILE_ID), 1, out);
        } else {
            concept_fwrite(COMPILED_FILE_ID, strlen(COMPILED_FILE_ID), 1, out);
        }
        char PRAGMA_SET = USE_WARN | (USE_EXC * 2);
        fputc(PRAGMA_SET, out);

        // salvez dependintzele
        INTEGER module_list = this->ModuleNamesList.Count();
        concept_fwrite(&module_list, sizeof(module_list), 1, out);
        for (INTEGER i = 0; i < module_list; i++) {
            AnsiString *member_name = (AnsiString *)ModuleNamesList [i];
            member_name->Serialize(out, SERIALIZE_16BIT_LENGTH);
        }

        if (is_lib) {
            // salvez constantele
            INTEGER constant_list = this->ConstantList.Count();

            INTEGER temp = 0;

            for (INTEGER i = basic_constants_count; i < constant_list; i++) {
                VariableDESCRIPTOR *VD = (VariableDESCRIPTOR *)ConstantList [i];
                if (!VD->USED) {
                    temp++;
                }
            }

            concept_fwrite(&temp, sizeof(temp), 1, out);
            for (INTEGER i = basic_constants_count; i < constant_list; i++) {
                VariableDESCRIPTOR *VD = (VariableDESCRIPTOR *)ConstantList [i];
                if (!VD->USED) {
                    VD->name.Serialize(out, SERIALIZE_8BIT_LENGTH);
                    VD->value.Serialize(out, SERIALIZE_16BIT_LENGTH);
                }
            }

            // salvez dependintzele
            INTEGER include_list = this->IncludedList.Count();
            concept_fwrite(&include_list, sizeof(include_list), 1, out);
            for (INTEGER i = 0; i < include_list; i++) {
                AnsiString *member_name = (AnsiString *)IncludedList [i];
                member_name->Serialize(out, SERIALIZE_16BIT_LENGTH);
            }

            // salvez dependintzele - > clasele
            INTEGER class_list = this->ClassList.Count();
            concept_fwrite(&class_list, sizeof(class_list), 1, out);
            for (INTEGER i = 0; i < class_list; i++) {
                ClassCode *CC1 = (ClassCode *)ClassList.Item(i);
                CC1->NAME.Serialize(out, SERIALIZE_16BIT_LENGTH);
            }
        }

        // salvez GeneralMembers
        INTEGER general_members = this->GeneralMembers.Count();
        concept_fwrite(&general_members, sizeof(general_members), 1, out);
        for (INTEGER i = 0; i < general_members; i++) {
            TinyString member_name(GeneralMembers [i]);
            member_name.Serialize(out, SERIALIZE_16BIT_LENGTH);
        }

        INTEGER class_count = ClassList.Count();
        if (is_lib) {
            INTEGER class_count2 = 0;
            for (INTEGER ii = 0; ii < class_count; ii++) {
                ClassCode *CC = (ClassCode *)ClassList [ii];
                if (CC->DEFINED_LEVEL == 0) {
                    class_count2++;
                }
            }
            concept_fwrite(&class_count2, sizeof(class_count2), 1, out);
        } else {
            // retin numarul de clase ...
            //concept_fwrite(&class_count,sizeof(class_count),1,out);
            //---------------------------------------------------//
            INTEGER class_count2 = 0;
            for (INTEGER ii = 0; ii < class_count; ii++) {
                ClassCode *CC = (ClassCode *)ClassList [ii];
                if (CC->DEFINED_LEVEL != -1) {
                    class_count2++;
                }
            }
            concept_fwrite(&class_count2, sizeof(class_count2), 1, out);
            //---------------------------------------------------//
        }
        for (INTEGER ii = 0; ii < class_count; ii++) {
            ClassCode *CC = (ClassCode *)ClassList [ii];
            if (is_lib) {
                if (CC->DEFINED_LEVEL == 0) {
                    CC->Serialize(out, is_lib);
                }
            } else
            if (CC->DEFINED_LEVEL != -1) {                                                                                                                                                                                                                                                                                                                                                                                                                                                           // is defined in a static library
                CC->Serialize(out, is_lib);
            }
        }
        if (!is_lib) {
            concept_fwrite(&ENTRY_CLASS, sizeof(ENTRY_CLASS), 1, out);
        }

        fclose(out);
        return(1);
    }
    return(-1);
}

int PIFAlizator::ComputeSharedSize(char *filename) {
    //int total_size=0;
    char         static_buffer [0xFF];
    concept_FILE *in = concept_fopen(filename, "rb");

    if (!in) {
        return(0);
    }

    char *cftype = COMPILED_FILE_ID;
    int  size    = (int)strlen(cftype);

    FREAD_FAIL(static_buffer, size, 1, in);
    static_buffer [size] = 0;
    if (strcmp(static_buffer, cftype)) {
        concept_fclose(in);
        return(0);
    }

    // pragma !
    SKIP(1, in);
    INTEGER module_list;
    FREAD_FAIL(&module_list, sizeof(module_list), 1, in);

    for (INTEGER i = 0; i < module_list; i++) {
        //size+=sizeof(AnsiString);
        //size+=AnsiString::ComputeSharedSize(in, SERIALIZE_16BIT_LENGTH);
        AnsiString::ComputeSharedSize(in, SERIALIZE_16BIT_LENGTH);
    }

    INTEGER general_members;
    FREAD_FAIL(&general_members, sizeof(general_members), 1, in);
    for (INTEGER i = 0; i < general_members; i++) {
        //size+=sizeof(AnsiString);
        //size+=AnsiString::ComputeSharedSize(in, SERIALIZE_16BIT_LENGTH);
        size += AnsiString::ComputeSharedSize(in, SERIALIZE_16BIT_LENGTH);
    }

    INTEGER class_count;
    FREAD_FAIL(&class_count, sizeof(class_count), 1, in);
    for (INTEGER ii = 0; ii < class_count; ii++) {
        //size+=sizeof(ClassCode);
        size += ClassCode::ComputeSharedSize(in, general_members);
    }

    INTEGER Cached_ENTRY_CLASS = -1;
    FREAD_FAIL(&Cached_ENTRY_CLASS, sizeof(Cached_ENTRY_CLASS), 1, in);

    concept_fclose(in);

    return(size);
}

int PIFAlizator::Unserialize(char *filename, bool is_lib) {
    concept_FILE *in = concept_fopen(filename, "rb");
    char         static_buffer [0xFF];
    int          *Classes            = 0;
    int          *Members_Relocation = 0;

    if (in) {
        // verific id-ul fisierului ...
        char *cftype = COMPILED_FILE_ID;
        if (is_lib) {
            cftype = PACKAGE_FILE_ID;
        }

        int size = (int)strlen(cftype);
        if (!concept_fread(static_buffer, size, 1, in)) {
            concept_fclose(in);
            AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
            Errors.Add(Exc, DATA_EXCEPTION);
            return(-3);
        }
        static_buffer [size] = 0;

        if (strcmp(static_buffer, cftype)) {
            concept_fclose(in);
            AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
            Errors.Add(Exc, DATA_EXCEPTION);
            return(-3);
        }
        //char PRAGMA_SET=fgetc(in);
        char PRAGMA_SET = 0;
        concept_fread(&PRAGMA_SET, 1, 1, in);

        USE_WARN = PRAGMA_SET % 2;
        USE_EXC  = PRAGMA_SET / 2;


        // incarc
        INTEGER module_list;
        if (!concept_fread(&module_list, sizeof(module_list), 1, in)) {
            concept_fclose(in);
            AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
            Errors.Add(Exc, DATA_EXCEPTION);
            return(-3);
        }
        for (INTEGER i = 0; i < module_list; i++) {
            AnsiString *member_name = new AnsiString();
            member_name->Unserialize(in, SERIALIZE_16BIT_LENGTH);
            if ((is_lib) && (ListContains(*member_name, &ModuleNamesList))) {
                delete member_name;
                continue;
            }
            ModuleNamesList.Add(member_name, DATA_STRING);

            
            if (!ImportModule(AnsiString(IMPORT_DIR) + (*member_name), &Errors, 0, "", &ModuleList, this, 1))
                ImportModule(*member_name, &Errors, 0, "", &ModuleList, this);
        }

        if (is_lib) {
            // salvez constantele
            INTEGER constant_list;
            if (!concept_fread(&constant_list, sizeof(constant_list), 1, in)) {
                concept_fclose(in);
                AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
                return(-3);
            }

            AnsiString ct_name;
            AnsiString ct_value;
            for (INTEGER i = 0; i < constant_list; i++) {
                ct_name.Unserialize(in, SERIALIZE_8BIT_LENGTH);
                ct_value.Unserialize(in, SERIALIZE_16BIT_LENGTH);
                this->DefineConstant(ct_name.c_str(), ct_value.c_str(), false);
            }

            INTEGER include_list;
            if (!concept_fread(&include_list, sizeof(include_list), 1, in)) {
                concept_fclose(in);
                AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
                return(-3);
            }

            AnsiString include_name;
            for (INTEGER i = 0; i < include_list; i++) {
                include_name.Unserialize(in, SERIALIZE_16BIT_LENGTH);
                IncludeFile(include_name, 0);
            }

            INTEGER class_list;
            if (!concept_fread(&class_list, sizeof(class_list), 1, in)) {
                concept_fclose(in);
                AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
                return(-3);
            }
            if (class_list) {
                Classes = new int [class_list];
            }

            AnsiString class_name;
            for (INTEGER i = 0; i < class_list; i++) {
                class_name.Unserialize(in, SERIALIZE_16BIT_LENGTH);
                Classes [i] = ClassExists(class_name.c_str()) - 1;
            }

            // incarc GeneralMembers - in library !
            INTEGER general_members;
            if (!concept_fread(&general_members, sizeof(general_members), 1, in)) {
                concept_fclose(in);
                AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
                if (Classes) {
                    delete[] Classes;
                }
                return(-3);
            }

            if (general_members) {
                Members_Relocation = new int [general_members];
            }

            AnsiString member_name;
            for (INTEGER i = 0; i < general_members; i++) {
                member_name.Unserialize(in, SERIALIZE_16BIT_LENGTH);
                //GeneralMembers.Add(member_name,DATA_STRING);
                //int reloc=ListContains(member_name,&GeneralMembers, true);
                int reloc = GeneralMembers.ContainsString(member_name.c_str());
                if (!reloc) {
                    AddGeneralMember(member_name);
                    reloc = GeneralMembers.Count();
                }
                Members_Relocation [i] = reloc - 1;
            }
        } else {
            // incarc GeneralMembers
            INTEGER general_members;
            if (!concept_fread(&general_members, sizeof(general_members), 1, in)) {
                concept_fclose(in);
                AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
                Errors.Add(Exc, DATA_EXCEPTION);
                return(-3);
            }
            //GeneralMembers.Clear();
            char is_pooled  = (char)SHIsPooled();
            bool is_created = SHIsCreated();

            GeneralMembers.ReInit(general_members, is_pooled);
            TinyString member_name;
            if ((is_pooled) && (!is_created)) {
                for (INTEGER i = 0; i < general_members; i++) {
                    GeneralMembers.PoolMap(AnsiString::ComputeSharedSize(in, SERIALIZE_16BIT_LENGTH));
                }
            } else {
                for (INTEGER i = 0; i < general_members; i++) {
                    /*TinyString *member_name=new TinyString();
                     * member_name->Unserialize(in,SERIALIZE_16BIT_LENGTH);
                     * GeneralMembers.Add(member_name);*/
                    member_name.Unserialize(in, SERIALIZE_16BIT_LENGTH);
                    AddGeneralMember(member_name);
                }
            }
        }

        // retzin numarul de clase ...
        INTEGER class_count;
        if (!concept_fread(&class_count, sizeof(class_count), 1, in)) {
            concept_fclose(in);
            AnsiException *Exc = new AnsiException(ERR1120, 0, 1120, filename, FileName);
            Errors.Add(Exc, DATA_EXCEPTION);
            if (Classes) {
                delete[] Classes;
            }
            if (Members_Relocation) {
                delete[] Members_Relocation;
            }
            return(-3);
        }

        for (INTEGER ii = 0; ii < class_count; ii++) {
            ClassCode *CC = new ClassCode(NULL_STRING, this);
            ClassList.Add(CC, DATA_CLASS_CODE);
            CC->Unserialize(in, &ClassList, is_lib, Classes, Members_Relocation);
            //ClassCode *CC=(ClassCode *)ClassList[ii];
            //CC->Serialize(in);
        }

        if (!is_lib) {
            concept_fread(&ENTRY_CLASS, sizeof(ENTRY_CLASS), 1, in);
        }
        concept_fclose(in);

        if (Classes) {
            delete[] Classes;
        }
        if (Members_Relocation) {
            delete[] Members_Relocation;
        }
        DoneLinking();
        return(0);
    }
    return(-3);
}

void PIFAlizator::OptimizeMemoryUsage() {
    int class_count = ClassList.Count();

    for (register INTEGER ii = 0; ii < class_count; ii++) {
        ClassCode *CC = (ClassCode *)ClassList [ii];

        int members_count = CC->Members->Count();

        for (register INTEGER jj = 0; jj < members_count; jj++) {
            ClassMember *CM = (ClassMember *)CC->Members->Item(jj);

            if ((CM->IS_FUNCTION == 1) && (CM->Defined_In == CC) && (CM->OPTIMIZER) && (CM->PIF_DATA)) {
                ((Optimizer *)CM->OPTIMIZER)->OptimizeMemoryUsage();
            }
        }
    }
}

void PIFAlizator::Optimize(char use_compiled_code) {
    int class_count = ClassList.Count();

    for (register INTEGER ii = 0; ii < class_count; ii++) {
        ClassCode *CC = (ClassCode *)ClassList [ii];
        if (!use_compiled_code) {
            CC->CLSID = ii;
        }
        int members_count = CC->Members->Count();

        for (register INTEGER jj = 0; jj < members_count; jj++) {
            ClassMember *CM = (ClassMember *)CC->Members->Item(jj);

            /*if (CC->CONSTRUCTOR==(jj+1))
             *  CC->CONSTRUCTOR=ListContains(CM->NAME,&GeneralMembers);*/

            INTEGER LOCAL_CLASSID = ((ClassCode *)CM->Defined_In)->CLSID;
            if (!use_compiled_code) {
                // a doua conditie e pusa de curand, dupa ce am scos copierea
                // membrilor ...
                if ((CM->IS_FUNCTION == 1) && (CM->Defined_In == CC) && (!CM->OPTIMIZER)) {
                    CM->OPTIMIZER = new Optimizer(this, CM->PIF_DATA, CM->VariableDescriptors, ((ClassCode *)(CM->Defined_In))->_DEBUG_INFO_FILENAME.c_str(), CC, CM->NAME.c_str());
                    ((Optimizer *)CM->OPTIMIZER)->Optimize();
                    ((Optimizer *)CM->OPTIMIZER)->GenerateIntermediateCode();
                    //this->out->Print(((Optimizer *)CM->OPTIMIZER)->DEBUG_INFO());
                }
            }
            //if (CM->IS_FUNCTION==1)
            //    CM->INTERPRETER=new ConceptInterpreter((Optimizer *)CM->OPTIMIZER,out,&ClassList,/*CC->CLSID,*/LOCAL_CLASSID,CM,DEBUGGER_TRAP,DEBUGGER_RESERVED,SERVER_PUBLIC_KEY,SERVER_PRIVATE_KEY,CLIENT_PUBLIC_KEY);
        }
        if (/*(!use_compiled_code) &&*/ (!CC->pMEMBERS)) {
            // optimizare clasa ...
            CC->GenerateCode(&GeneralMembers);
        }
    }
    PIF.Clear();
    DoneLinking();
}

bool PIFAlizator::CheckRunable() {
    if (!ENTRY_CLASS) {
        AnsiException *Exc = new AnsiException(ERR080, 0, 80, MAIN_ENTRY_POINT, FileName);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(false);
    }
    ClassCode *CC = (ClassCode *)ClassList [ENTRY_CLASS - 1];
    if (!CC->CONSTRUCTOR) {
        AnsiException *Exc = new AnsiException(ERR090, 0, 90, MAIN_ENTRY_POINT, FileName);
        Errors.Add(Exc, DATA_EXCEPTION);
        return(false);
    }
    return(true);
}

void *PIFAlizator::GetStartingPoint() {
    // TO DO !
    ClassCode *CC = (ClassCode *)ClassList [ENTRY_CLASS - 1];

    //ClassMember *CM=(ClassMember *)CC->Members[CC->CONSTRUCTOR-1];
    //return CM->OPTIMIZER;
    return(CC);
}

AnsiString PIFAlizator::DEBUG_INFO() {
    AnsiString res;

    for (INTEGER ii = 0; ii < ClassList.Count(); ii++) {
        ClassCode *CC = (ClassCode *)ClassList [ii];
        res += AnsiString("In class \"") + CC->NAME + AnsiString("\" :\n");
        for (INTEGER jj = 0; jj < CC->Members->Count(); jj++) {
            ClassMember *CM = (ClassMember *)CC->Members->Item(jj);

            if (CM->OPTIMIZER) {
                res += ((Optimizer *)CM->OPTIMIZER)->DEBUG_INFO();
            }

            AnsiList *PIFList = CM->PIF_DATA;
            AnsiList *VDList  = CM->VariableDescriptors;

            res += AnsiString("In class member \"") + CM->NAME.c_str() + AnsiString("\" :\n");
            res += "------------ Program internal form ----------------------------------------\n";
            for (INTEGER i = 0; i < PIFList->Count(); i++) {
                AnalizerElement *AE = (AnalizerElement *)(*PIFList) [i];
                AnsiString      strOperator;
                if ((AE->TYPE == TYPE_KEYWORD) || (AE->TYPE == TYPE_OPERATOR)) {
                    strOperator = GetKeyWord(AE->ID);
                }

                res += AnsiString(i) + AnsiString(". ") +
                       AnsiString("\tID: ") + AnsiString(AE->ID) +
                       AnsiString(" '") + strOperator +
                       AnsiString("'\t\tTYPE:") + AnsiString(AE->TYPE) +
                       AnsiString("\t\tLINE:") + AnsiString(AE->_DEBUG_INFO_LINE) +
                       //AnsiString("\tFILE:")   +   AE->_DEBUG_INFO_FILENAME+
                       AnsiString("\n");
            }
            res += "------------ Variables/CT/DATA --------------------------------------------\n";
            for (INTEGER j = 0; j < VDList->Count(); j++) {
                VariableDESCRIPTOR *VD = (VariableDESCRIPTOR *)VDList->Item(j);
                res += AnsiString(j) + AnsiString(". ") +
                       AnsiString("\tName: ") + ((VD->name == NULL_STRING) ? AnsiString("(ct)") : VD->name) +
                       AnsiString("\t\tType: ") + AnsiString(VD->TYPE) +
                       AnsiString("\t\tValue: ") + ((VD->value.Length()) ? AnsiString("(null)") : AnsiString(VD->value.c_str())) +
                       AnsiString("\n");
            }
        }
    }
    res += PRINT_ERRORS();
    return(res);
}

AnsiString PIFAlizator::SerializeWarningsErrors(int ser_warnings) {
    AnsiString result;
    AnsiList   *Target = &Errors;

    if (ser_warnings) {
        Target = &Warnings;
    }

    int count = Target->Count();
    for (INTEGER k = 0; k < count; k++) {
        AnsiException *Ex = (AnsiException *)Target->Item(k);
        result += Ex->GetFileName();
        result += (char *)"<:>";
        result += AnsiString((long)Ex->GetLine());
        result += "<:>";
        result += (char *)Ex->GetClass();
        result += (char *)"<:>";
        result += Ex->GetMember();
        result += (char *)"<:>";
        result += AnsiString((long)Ex->GetID());
        result += (char *)"<:>";
        result += Ex->GetText();
        result += (char *)"<:>";
        result += Ex->GetExtra();
        if (k != count - 1) {
            result += (char *)"<:>";
        }
    }
    return(result);
}

AnsiString PIFAlizator::PRINT_ERRORS(int html) {
    AnsiString res;

    if (html == 1) {
        if (!this->out->userdata) {
            res = (char *)"Content-Type: text/plain\r\n\r\n";
        }
    }
    res += "\n------------ Errors -------------------------------------------------------\n";
    AnsiString filename;
    AnsiString _CLASS;
    AnsiString _MEMBER;
    int        count = Errors.Count();
    for (INTEGER k = 0; k < count; k++) {
        AnsiException *Ex = (AnsiException *)Errors [k];

        /*if (html)
         *  res+="<br>";*/
        if (filename != Ex->GetFileName()) {
            filename = Ex->GetFileName();
            res     += "In ";
            res     += filename;
            res     += ":\n";
        }
        if (_CLASS != Ex->GetClass()) {
            _CLASS = Ex->GetClass();
            if (_CLASS != (char *)"") {
                res += "  In class ";
                res += _CLASS;
                res += ":\n";
            } else {
                res += "  Outside class:\n";
            }
        }

        if (_MEMBER != Ex->GetMember()) {
            _MEMBER = Ex->GetMember();
            if (_MEMBER != (char *)"") {
                res += (char *)"    In member ";
                res += _CLASS;
                res += (char *)".";
                res += (char *)_MEMBER;
                res += (char *)":\n";
            } else {
                res += "    Outside class member:\n";
            }
        }
        res += "      ";
        res += AnsiString(k) + AnsiString(". ") +
               AnsiString("E") + AnsiString(Ex->GetID()) +
               AnsiString(" on line ") + AnsiString(Ex->GetLine()) +
               AnsiString(": ") + Ex->GetText();
        if (Ex->GetExtra() != (char *)"") {
            res += AnsiString(" (\'") + Ex->GetExtra() + AnsiString("\') ");
        }

        res += AnsiString("\n");
    }
    return(res);
}

AnsiString PIFAlizator::PRINT_WARNINGS(int html) {
    AnsiString res;

    if (html == 1) {
        if (!this->out->userdata) {
            res = (char *)"Content-Type: text/plain\r\n\r\n";
        }
    }
    res += (char *)"------------ Warnings -----------------------------------------------------\n";
    for (INTEGER k = 0; k < Warnings.Count(); k++) {
        /*if (html)
         *  res+="<br>";*/
        AnsiException *Ex = (AnsiException *)Warnings [k];
        res += AnsiString(k) + AnsiString(". ") +
               AnsiString("\tID: ") + AnsiString(Ex->GetID()) +
               AnsiString("\tText: ") + Ex->GetText() +
               AnsiString("\tLine: ") + AnsiString(Ex->GetLine()) +
               AnsiString("\tExtra: ") + Ex->GetExtra() +
               AnsiString("\tFile name: ") + Ex->GetFileName() +
               AnsiString("\n");
    }
    return(res);
}

AnsiString PIFAlizator::DEBUG_CLASS_CONFIGURATION() {
    AnsiString res;

    res += (char *)"------------ Class Configuration ------------------------------------------\n";
    INTEGER CLASS_COUNT = ClassList.Count();
    for (INTEGER i = 0; i < CLASS_COUNT; i++) {
        ClassCode *CC           = (ClassCode *)ClassList [i];
        INTEGER   MEMBERS_COUNT = CC->Members->Count();
        res += (char *)"Class '";
        res += CC->NAME;
        res += (char *)"' class ID :";
        res += (char *)AnsiString(CC->CLSID);
        res += (char *)"\n";
        for (INTEGER j = 0; j < MEMBERS_COUNT; j++) {
            ClassMember *CM = (ClassMember *)CC->Members->Item(j);
            res += (char *)"\t::";
            res += CM->NAME.c_str();
            if (CM->IS_FUNCTION == 1) {
                res += (char *)" is a function\n\t\tDATA:\n";
                for (INTEGER k = 0; k < ((Optimizer *)CM->OPTIMIZER)->dataCount; k++) {
                    res += (char *)"\t\t\tTYPE:";
                    res += AnsiString((long)((Optimizer *)CM->OPTIMIZER)->DATA [k].TYPE);
                    res += (char *)"\n";
                }
                res += (char *)"\t\tCODE:\n";
                for (INTEGER k = 0; k < ((Optimizer *)CM->OPTIMIZER)->codeCount; k++) {
                    res += (char *)"\t\t\tOPERATOR TYPE:";
                    res += AnsiString((long)((Optimizer *)CM->OPTIMIZER)->CODE [k].Operator.TYPE);
                    res += (char *)"\tOPERATOR ID:";
                    res += AnsiString(((Optimizer *)CM->OPTIMIZER)->CODE [k].Operator.ID);
                    res += (char *)"\n";
                }
            } else {
                res += (char *)"\n";
            }
        }
        res += (char *)"\n\tDYNAMIC MEMBERS:\n";
        for (INTEGER j = 0; j < CC->pMEMBERS_COUNT; j++) {
            ClassMember *CM = (ClassMember *)CC->pMEMBERS [j];
            res += (char *)"\t\t[+]\t";
            res += AnsiString(j);
            res += (char *)"\t";
            if (CM) {
                res += CM->NAME.c_str();
            } else {
                res += (char *)"___NADA___";
            }
            res += (char *)"\n";
        }
    }
    return(res);
}

void PIFAlizator::CheckMember(AnsiString& member_name) {
    //if (!ListContains(member_name,&GeneralMembers, true)) {
    if (!GeneralMembers.ContainsString(member_name.c_str())) {
        AddGeneralMember(member_name);
    } else {
        DefineMember(member_name);
    }
}

void PIFAlizator::Hibernate() {
    int len = this->ClassList.Count();

    for (int i = 0; i < len; i++) {
        ClassCode *CC = (ClassCode *)this->ClassList.Item(i);
        if (CC) {
            CC->__gc_obj.Foreach(ClassCode::Hibernate);
        }
    }
}
