#ifndef __PIFALIZATOR_H
#define __PIFALIZATOR_H

#include "AnsiString.h"
#include "AnsiList.h"
#include "Codes.h"
#include "AnsiParser.h"
#include "ClassCode.h"
#include "Debugger.h"
#include "ConceptPools.h"
#include "TinyString.h"
#include "StaticList.h"

#define PRAGMA_WARNINGS       (char *)"warnings"
#define PRAGMA_EXCEPTIONS     (char *)"lib_exceptions"
#define PRAGMA_IMPLICIT       (char *)"implicit"
#define PRAGMA_ENTRY_POINT    (char *)"entry_point"

#define PRAGMA_ON             (char *)"on"
#define PRAGMA_OFF            (char *)"off"


#define MAIN_ENTRY_POINT      "Main"
//#define DEFAULT_ENTRY_POINT           "Main"

//#define INCLUDE_DIR                 "Include/"
//#define IMPORT_DIR                  "Library/"
#define DEFAULT_USE_WARNINGS      1
#define DEFAULT_USE_EXCEPTIONS    0
#define DEFAULT_USE_IMPLICIT      0

#define DEFAULT_BIN_EXTENSION     ".accel"
#define DEFAULT_PACK_EXTENSION    ".package"

#define STATIC_LINKS_INCREMENT    0x10
class PIFAlizator {
    friend class Optimizer;
    friend class ClassCode;
    friend class ClassMember;
    friend class ConceptInterpreter;
    friend INTEGER Invoke(INTEGER INVOKE_TYPE, ...);

    StaticList GeneralMembers;
    int        pipe_read;
    int        pipe_write;
    int        apid;
    int        parent_apid;

    SimpleStream *out;

    INTEGER DebugOn;

    AnsiString FileName;
    AnsiString *InputStream;

    AnsiList ClassList;

    AnsiList IncludedList;
    AnsiList ModuleList;
    AnsiList ModuleNamesList;
    AnsiList ConstantList;
    AnsiList UndefinedMembers;
    AnsiList UndefinedClasses;

    SYS_INT      *StaticLinks;
    unsigned int StaticLinksCount;
    unsigned int StaticLinksSize;

    unsigned int LinkStatic(char *funname);
    INTEGER AddUndefinedMember(AnsiString& member, AnsiString& _CLASS, TinyString& _MEMBER, long line);

    //INTEGER AddGeneralMember(AnsiString name);
    //INTEGER DefineMember(AnsiString name);
    INTEGER CheckUndefinedMembers();

    INTEGER AddUndefinedClass(AnsiString& member, AnsiString& _CLASS, TinyString& _MEMBER, long line);
    INTEGER CheckUndefinedClasses();

    INTEGER VariableIsDescribed(AnsiString& S, AnsiList *VDList);

    AnsiString GetSpecial(AnsiParser *P, ClassCode *CC, ClassMember *CM, AnsiString special);
    INTEGER BuildArray(ClassCode *CC, ClassMember *CM, AnsiParser *P, INTEGER on_line);
    INTEGER BuildFunction(ClassCode *CC, AnsiParser *P, INTEGER on_line = 0, INTEGER ACCESS = ACCESS_PUBLIC, INTEGER OPERATOR = 0, char STATIC = 0, char *prec_parse = 0);
    INTEGER BuildProperty(ClassCode *CC, AnsiParser *P, INTEGER on_line = 0, INTEGER ACCESS = ACCESS_PUBLIC);
    INTEGER BuildEvent(ClassCode *CC, AnsiParser *P, INTEGER on_line = 0, INTEGER ACCESS = ACCESS_PUBLIC);
    INTEGER BuildOverride(ClassCode *CC, AnsiParser *P, INTEGER on_line);
    INTEGER IncludePackage(AnsiString filename);
    INTEGER IncludeFile(AnsiString MODULE_NAME, INTEGER on_line);

    AnsiString NormalizePath(AnsiString *MODULE_NAME);

    INTEGER BuildClass(AnsiParser *P, INTEGER on_line = 0);

    INTEGER BuildPragma(AnsiParser *P);

    INTEGER Execute(AnsiString *Stream, INTEGER on_line = 0, char _USE_WARN = DEFAULT_USE_WARNINGS, char _USE_EXC = DEFAULT_USE_EXCEPTIONS, char _USE_IMPLICIT = DEFAULT_USE_IMPLICIT);
    SYS_INT ClassExists(char *name, char by_addr = 0, int *index = 0);
    INTEGER ListContains(AnsiString& S, AnsiList *VDList, char is_tiny = false);
    INTEGER BuildVariable(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER ACCESS);
    INTEGER ExtendClass(ClassCode *CC, AnsiParser *P, INTEGER on_line, INTEGER OWNER_CLSID);

    INTEGER ENTRY_CLASS;

    DEBUGGER_CALLBACK DEBUGGER_TRAP;
    void              *DEBUGGER_RESERVED;
    char              *SERVER_PUBLIC_KEY;
    char              *SERVER_PRIVATE_KEY;
    char              *CLIENT_PUBLIC_KEY;

    AnsiString INCLUDE_DIR;
    AnsiString IMPORT_DIR;
    AnsiString TEMP_INC_DIR;

    char USE_WARN;
    char USE_EXC;
    char USE_IMPLICIT;
    int  INCLUDE_LEVEL;

    //AnsiString MAIN_ENTRY_POINT;

    INTEGER Warning(char *WRN, int line, int wrn_code, char *extra, char *filename = 0);
    void DefineConstant(char *name, char *value, int is_string = 1);

    void CheckMember(AnsiString& member_name);

    //char basic_members_count;
    char basic_constants_count;
public:
    POOLED(PIFAlizator)
    void *CachedhDLL;
    static int         argc;
    static char        **argv;
    static int         last_result;
    static CHECK_POINT CheckPoint;
    static AnsiString GetPath(AnsiString *S);

    void AcknoledgeRunTimeError(AnsiException *Exc);

    AnsiList PIF;
    AnsiList VariableDescriptors;
    AnsiList Errors;
    AnsiList Warnings;

    AnsiString DEBUG_CLASS_CONFIGURATION();
    AnsiString DEBUG_INFO();
    AnsiString PRINT_ERRORS(int html = 0);
    AnsiString PRINT_WARNINGS(int html = 0);
    AnsiString SerializeWarningsErrors(int ser_warnings = 0);

    INTEGER ErrorCount();
    INTEGER WarningCount();

    PIFAlizator(AnsiString INC_DIR, AnsiString LIB_DIR, AnsiString *S, SimpleStream *Out, AnsiString _FileName = "", INTEGER debug = 0, DEBUGGER_CALLBACK DC = 0, void *DEBUGGER_RESERVED = 0, char *SPubK = 0, char *SPrivK = 0, char *CPubK = 0);
    void SetPipe(int pipein, int pipeout, int apid, int papid);
    INTEGER Execute();
    void Optimize(char use_compiled_code = 0);
    void OptimizeMemoryUsage();
    bool CheckRunable();
    void *GetStartingPoint();
    int Serialize(char *filename, bool is_lib = false);
    int Unserialize(char *filename, bool is_lib = false);
    static int ComputeSharedSize(char *filename);
    void Hibernate();

    ~PIFAlizator(void);
};

#endif //__PIFALIZATOR_H
