#include "StaticList.h"
#include <string.h>
#include <time.h>
#include "SHManager.h"

#define STATIC_GET(data)    char **data = allocated ? (char **)((long)this + DataOffset) : 0;
#define STATIC_SET(data)    DataOffset  = (long)data - (long)this;

StaticList::StaticList() {
    count         = 0;
    allocated     = 0;
    DataOffset    = 0;
    pooled_marker = 0;
}

int StaticList::EnsureSpace() {
    if (count >= allocated) {
        int new_size = allocated + STATICLIST_BLOCKSIZE;
        STATIC_GET(data)
        data = (char **)realloc(data, new_size * sizeof(char *));
        if (!data) {
            return(0);
        }
        allocated += STATICLIST_BLOCKSIZE;
        //data[count]=0;
        STATIC_SET(data)
        return(1);
    }
    return(1);
}

void StaticList::ReInit(int i, bool pooled) {
    this->Clear();
    char **data = (char **)malloc(i * sizeof(char *));
    if (data) {
        //memset(data, 0, sizeof(data));

        allocated = i;
        count     = 0;
        STATIC_SET(data)
        if (pooled) {
            pooled_marker = i;
        }
    }
}

void StaticList::PoolMap(int len) {
    if (len <= 0) {
        return;
    }
    if (count < pooled_marker) {
        STATIC_GET(data)
        char *s      = (char *)SHAlloc(len);
        data [count] = s;
        count++;
    }
}

void StaticList::Add(char *str, int len) {
    if ((!str) || (!len)) {
        return;
    }
    if (len < 0) {
        len = strlen(str);
    }
    if (!len) {
        return;
    }
    if (this->EnsureSpace()) {
        STATIC_GET(data)
        char *s = 0;
        if (count < pooled_marker) {
            s = (char *)SHAlloc(len + 1);
        } else {
            s = (char *)malloc(len + 1);
        }
        memcpy(s, str, len);
        s [len]      = 0;
        data [count] = s;
        count++;
    }
}

void StaticList::Add(AnsiString& str) {
    this->Add(str.c_str(), str.Length());
}

void StaticList::Add(TinyString& str) {
    this->Add(str.c_str(), str.Length());
}

char *StaticList::operator[](int i) {
    return(this->Item(i));
}

int StaticList::ContainsString(char *str) {
    if (!str) {
        return(0);
    }
    STATIC_GET(data)

    for (int i = 0; i < count; i++) {
        char *elem = data [i];
        if ((*elem == *str) && (!strcmp(elem, str))) {
            return(i + 1);
        }
    }
    return(0);
}

char *StaticList::Item(int i) {
    if ((i >= 0) && (i < count)) {
        STATIC_GET(data)
        return(data [i]);
    }
    return(0);
}

void StaticList::Clear() {
    if (count) {
        STATIC_GET(data)
        for (int i = pooled_marker; i < count; i++) {
            // if data[i] is null, free does nothing
            // so no pre-check
            free(data [i]);
        }
        free(data);
        count         = 0;
        allocated     = 0;
        DataOffset    = 0;
        pooled_marker = 0;
    }
}

int StaticList::Count() {
    return(this->count);
}

StaticList::~StaticList(void) {
    this->Clear();
}
