#include "TinyString.h"
#include "SHManager.h"

#ifdef USE_DEPRECATED
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
#else
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
#endif

POOLED_IMPLEMENTATION(TinyString)

#define breakeven_point    12

#define fast_memcpy(d, s, n)                                                                 \
    { register size_t nn = (size_t)(n);                                                      \
      if (nn >= breakeven_point) { memcpy((d), (s), nn); }                                   \
      else if (nn > 0) {             /* proc call overhead is worth only for large strings*/ \
          register char *dd; register const char *ss;                                        \
          for (ss = (s), dd = (d); nn > 0; nn--) { *dd++ = *ss++; } } }

//#define MEMCPY      memcpy
#define MEMCPY    fast_memcpy



/*#define HAVE_MMX
 * #include "fastmemcpy.h"
 *
 #define MEMCPY      fast_memcpy*/

/*#include "dmemcopy.h"
 *
 #define MEMCPY      dmemcpy*/

/*void memcpy32(void *dest, void *src, int bcount) {
 *  if (bcount<4) {
 *      memcpy(dest,src,bcount);
 *      return;
 *  }
 *  int blocks=bcount/4;
 *  bcount%=4;
 *
 *  long *ldest=(long *)dest;
 *  long *lsrc=(long *)src;
 *
 *  void *end=lsrc+blocks;
 *
 * label1:
 * *ldest++=*lsrc++;
 *  if (lsrc!=end)
 *      goto label1;
 *
 *  if (bcount)
 *      memcpy(ldest,lsrc,bcount);
 * }*/

TinyString::TinyString() {
    DataOffset = 0;
}

TinyString::TinyString(char *value) {
    DataOffset = 0;
    operator=(value);
}

TinyString::TinyString(long i) {
    DataOffset = 0;
    operator=(i);
}

TinyString::TinyString(int i) {
    DataOffset = 0;
    operator=((long)i);
}

TinyString::TinyString(char c) {
    DataOffset = 0;
    operator=(c);
}

TinyString::TinyString(double d) {
    DataOffset = 0;
    operator=(d);
}

TinyString::TinyString(const TinyString& S) {
    DataOffset = 0;
    //if (S.DataOffset) {
    if (S.DataOffset) {
        operator=(((char *)&S) + (S.DataOffset));
    } else {
        operator=((char *)0);
    }
    //}
}

void TinyString::operator=(char *value) {
    size_t len   = 0;
    char   *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + (DataOffset);
    }
    if (Data) {
        free(Data);
    }
    DataOffset = 0;
    Data       = 0;
    if (value) {
        len  = strlen(value);
        Data = (char *)malloc(len + 1);
        MEMCPY(Data, value, len);
        Data [len] = 0;
        DataOffset = (long)Data - (long)this;
    }
}

void TinyString::operator=(TinyString& S) {
    operator=(S.c_str());
}

void TinyString::operator=(AnsiString& S) {
    operator=(S.c_str());
}

void TinyString::operator=(char c) {
    char buffer [2];

    buffer [0] = c;
    buffer [1] = 0;
    operator=(buffer);
}

void TinyString::operator=(long i) {
    char buffer [MAX_DECIMALS];

    //operator=(ltoa(i,buffer,10));

    // pt linux !
    sprintf(buffer, "%li", i);
    operator=(buffer);
}

void TinyString::operator=(double d) {
    char buffer [MAX_DECIMALS];

    sprintf(buffer, "%.30g", d);
    //_gcvt(d,MAX_DECIMALS-10,buffer);
    int len = strlen(buffer);
    if (((len > 1) && ((buffer [len - 1] == '.') || (buffer [len - 1] == ',')))) {
        buffer [len - 1] = 0;
    }
    operator=(buffer);
}

void TinyString::operator=(int i) {
    operator=((long)i);
}

int TinyString::operator <(char *str) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if ((Data) && (str)) {
        return(strcmp(Data, str) < 0);
    }
    return(0);
}

int TinyString::operator >(char *str) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if ((Data) && (str)) {
        return(strcmp(Data, str) > 0);
    }
    return(0);
}

int TinyString::operator >=(char *str) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if ((Data) && (str)) {
        return(strcmp(Data, str) >= 0);
    }
    return(0);
}

int TinyString::operator <=(char *str) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    if ((Data) && (str)) {
        return(strcmp(Data, str) <= 0);
    }
    return(0);
}

int TinyString::operator==(char *str) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if ((Data) && (str)) {
        if (strcmp(Data, str)) {
            return(0);
        } else {
            return(1);
        }
    }
    if ((!Data) && (str) && (!str [0])) {
        return(1);
    }

    if ((!str) && (Data) && (!Data [0])) {
        return(1);
    }

    if ((!Data) && (!str)) {
        return(1);
    }
    return(0);
}

int TinyString::operator !=(char *str) {
    return(!operator==(str));
}

int TinyString::operator==(TinyString& S) {
    return(operator==(S.c_str()));
}

int TinyString::operator==(AnsiString& S) {
    return(operator==(S.c_str()));
}

int TinyString::operator!=(TinyString& S) {
    return(operator!=(S.c_str()));
}

int TinyString::operator!=(AnsiString& S) {
    return(operator!=(S.c_str()));
}

int TinyString::operator >(TinyString& S) {
    return(operator >(S.c_str()));
}

int TinyString::operator <(TinyString& S) {
    return(operator <(S.c_str()));
}

int TinyString::operator >=(TinyString& S) {
    return(operator >=(S.c_str()));
}

int TinyString::operator <=(TinyString& S) {
    return(operator <=(S.c_str()));
}

TinyString::~TinyString(void) {
    /*if (Data)
     *  delete[] Data;*/
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if (Data) {
        free(Data);
    }
    DataOffset = 0;
}

TinyString::operator char *() {
    return(c_str());
}

TinyString::operator AnsiString() {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    return(AnsiString(Data));
}

char *TinyString::c_str() {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    if (!Data) {
        return("");
    }
    return(Data);
}

void TinyString::operator +=(char *value) {
    size_t len;
    size_t len_value;

    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    if ((Data) && (value)) {
        //len=strlen(Data);
        len       = strlen(Data);
        len_value = strlen(value);
        size_t new_length = len + len_value;

        Data = (char *)realloc(Data, new_length + 1);
        MEMCPY(Data + len, value, len_value);

        Data [new_length] = 0;

        DataOffset = (long)Data - (long)this;
    } else {
        operator=(value);
    }
}

void TinyString::operator +=(TinyString& S) {
    operator+=(S.c_str());
}

void TinyString::operator +=(char c) {
    char buffer [2];

    buffer [0] = c;
    buffer [1] = 0;
    operator+=(buffer);
}

char TinyString::operator [](unsigned long index) {
    size_t len;
    char   *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if (Data) {
        len = strlen(Data);
        if (index < len) {
            return(Data [index]);
        }
    }
    return(0);
}

char TinyString::operator [](int index) {
    size_t len;
    char   *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if (Data) {
        len = strlen(Data);
        if ((index < len) && (index >= 0)) {
            return(Data [index]);
        }
    }
    return(0);
}

long TinyString::ToInt() {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    if (Data) {
        return(atol(Data));
    }

    return(0);
}

double TinyString::ToFloat() {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }
    if (Data) {
        return(atof(Data));
    }
    return(0);
}

TinyString TinyString::operator +(TinyString& S) {
    //TinyString ret(Data);
    // bug corectat de pe 15 ianuarie 2006
    // problema : Daca data arata de forma "abcde\0\0\0\0" si concatenam un sir ...
    // nu era bine, ca-l punea imediat dupa e, in loc sa-l puna dupa al 4-lea 0
    TinyString ret     = *this;
    char       *S_Data = 0;

    if (S.DataOffset) {
        S_Data = ((char *)(void *)&S) + S.DataOffset;
    }
    if (S_Data) {
        ret += S;
    }
    return(ret);
}

long TinyString::Length() {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    if (!Data) {
        return(0);
    }
    return(strlen(Data));
}

long TinyString::Pos(TinyString substr) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    long result = -1;
    char *ptr   = strstr(Data, substr.c_str());
    if (ptr) {
        result = ptr - Data + 1;
    }
    return(result);
}

void TinyString::LoadBuffer(char *buffer, int size) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    free(Data);

    Data = (char *)malloc(size + 1);
    MEMCPY(Data, buffer, size);
    Data [size] = 0;

    DataOffset = (long)Data - (long)this;
}

int TinyString::Serialize(FILE *out, int type) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    size_t         _LENGTH = Length();
    unsigned char  uClen   = (unsigned char)_LENGTH;
    unsigned short uSlen   = (unsigned short)_LENGTH;
    switch (type) {
        case SERIALIZE_8BIT_LENGTH:
            concept_fwrite(&uClen, sizeof(uClen), 1, out);
            break;

        case SERIALIZE_16BIT_LENGTH:
            concept_fwrite(&uSlen, sizeof(uSlen), 1, out);
            break;

        default:
            concept_fwrite(&_LENGTH, sizeof(_LENGTH), 1, out);
            break;
    }
    concept_fwrite(Data, _LENGTH, 1, out);
    return(1);
}

int TinyString::Unserialize(concept_FILE *in, int type, char use_pool) {
    char *Data = 0;

    if (DataOffset) {
        Data = ((char *)(void *)this) + DataOffset;
    }

    long           len;
    unsigned char  uClen = (unsigned char)0;
    unsigned short uSlen = (unsigned short)0;

    switch (type) {
        case SERIALIZE_8BIT_LENGTH:
            if (!concept_fread(&uClen, sizeof(uClen), 1, in)) {
                return(-1);
            }
            len = uClen;
            break;

        case SERIALIZE_16BIT_LENGTH:
            if (!concept_fread(&uSlen, sizeof(uSlen), 1, in)) {
                return(-1);
            }
            len = uSlen;
            break;

        default:
            if (!concept_fread(&len, sizeof(len), 1, in)) {
                return(-1);
            }
            break;
    }

    // moving the top pointer
    if ((len) && (use_pool == -1)) {
        concept_fseek(in, len, SEEK_CUR);
        SHAlloc(len + 1);
        // just move the stack pointer and return (already in shared memory)
        return(1);
    }

    if (Data) {
        free(Data);
    }
    Data       = 0;
    DataOffset = 0;

    if (len) {
        if (use_pool == 1) {
            Data = (char *)SHAlloc(len + 1);
            concept_fread(Data, len, 1, in);
            Data [len] = 0;
        } else {
            Data = (char *)malloc(len + 1);
            concept_fread(Data, len, 1, in);
            Data [len] = 0;
        }
        DataOffset = (long)Data - (long)this;
    }
    return(1);
}
