//#define DEBUG
//#define TIMMING
#ifdef _WIN32
 #include <windows.h>
#endif

#include "SHManager.h"
#include "AnsiString.h"
#include "PIFAlizator.h"
#include "library.h"
#include "Debugger.h"
#include "ConceptTypes.h"
#include "ConceptInterpreter.h"

#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32
 #include <io.h>

//#include <utime.h>
 #include <sys/utime.h>
#else
 #include <unistd.h>
 #include <utime.h>
#endif
#include <stdio.h>
#include <string>
#include <stdlib.h>

#ifdef TIMMING
 #include <time.h>
#endif

POOLED_IMPLEMENTATION(tsVariableDESCRIPTOR);
POOLED_IMPLEMENTATION(tsVariableDATA);
POOLED_IMPLEMENTATION(tsUndefinedMember)
POOLED_IMPLEMENTATION(tsAnalizerElement)
POOLED_IMPLEMENTATION(tsSHORT_AnalizerElement)
POOLED_IMPLEMENTATION(tsPartialAnalizerElement)
POOLED_IMPLEMENTATION(tsOptimizedElement)
POOLED_IMPLEMENTATION(tsParamList)

THREADED_FLAG int static_pipe_out = 0;
THREADED_FLAG int           static_parent      = 0;
THREADED_FLAG NOTIFY_PARENT RemoteNotifyParent = 0;
THREADED_FLAG void          *cache_buffer      = 0;

/*void write_debug(char *what) {
 *      FILE *f=fopen("c:\\debug.txt","rt+");
 *      if (f) {
 *              concept_fseek(f, 0, SEEK_END);
 *              concept_fwrite(what, 1, strlen(what), f);
 *              fclose(f);
 *      }
 * }*/

int last_result = 0;

CachedFILE *cached_fopen(const char *filename, const char *mode) {
    CachedFILE *fin = 0;
    FILE       *in  = fopen(filename, mode);

    if (in) {
        fin = (CachedFILE *)malloc(sizeof(CachedFILE));
        if (fin) {
            fin->buf_fill = 0;
            fin->buf_pos  = 0;
            fin->in       = in;
            return(fin);
        } else {
            fclose(in);
        }
    }
    return(0);
}

int cached_fclose(CachedFILE *in) {
    int res = 0;

    if (in) {
        res = fclose(in->in);
        free(in);
    }
    return(res);
}

int cached_fseek(CachedFILE *in, long int offset, int origin) {
    switch (origin) {
        case SEEK_CUR:
            {
                int delta = in->buf_fill - in->buf_pos;
                if (offset < delta) {
                    in->buf_pos += offset;
                } else {
                    offset -= delta;
                    if (offset) {
                        fseek(in->in, offset, SEEK_CUR);
                    }
                    in->buf_fill = 0;
                    in->buf_pos  = 0;
                }
            }
            return(0);
            break;

        case SEEK_SET:
        case SEEK_END:
        default:
            in->buf_fill = 0;
            in->buf_pos  = 0;
            return(fseek(in->in, offset, origin));
            break;
    }
    return(-1);
}

int cached_fread(void *buf, size_t size, size_t len, CachedFILE *in) {
    int total_size = size * len;
    int total      = 0;

    if ((!in->buf_fill) || (in->buf_pos >= in->buf_fill)) {
        in->buf_fill = fread(in->sbuf, 1, sizeof(in->sbuf), in->in);
        in->buf_pos  = 0;
    }
    if (in->buf_fill > 0) {
        int remaining = in->buf_fill - in->buf_pos;
        if (total_size <= remaining) {
            memcpy(buf, in->sbuf + in->buf_pos, total_size);
            in->buf_pos += total_size;
            total        = total_size;
        } else {
            if (remaining) {
                memcpy(buf, in->sbuf + in->buf_pos, remaining);
                total      += remaining;
                total_size -= remaining;
            }
            do {
                in->buf_fill = fread(in->sbuf, 1, sizeof(in->sbuf), in->in);
                in->buf_pos  = 0;
                if (in->buf_fill > 0) {
                    if (total_size <= in->buf_fill) {
                        memcpy(((char *)buf) + total, in->sbuf, total_size);
                        in->buf_pos = total_size;
                        total      += total_size;
                        break;
                    } else {
                        memcpy(((char *)buf) + total, in->sbuf, in->buf_fill);
                        total      += in->buf_fill; //total_size;
                        total_size -= in->buf_fill;
                        in->buf_pos = in->buf_fill;
                        //in->buf_fill=0;
                        //in->buf_pos=0;
                    }
                }
            } while (in->buf_fill > 0);
        }
    }
    return(total / size);
}

int NotifyParent(int pipe_out, int apid, int msg_id, AnsiString DATA) {
    if (RemoteNotifyParent) {
        return(RemoteNotifyParent(pipe_out, apid, msg_id, DATA.Length(), DATA.c_str()));
    }

    if ((apid <= 0) || (pipe_out <= 0)) {
        return(0);
    }

    int params [3];
    params [0] = apid;
    params [1] = msg_id;
    params [2] = DATA.Length();

    /*write_debug(">>> Using pipe:");
     * write_debug(AnsiString((long)pipe_out));
     * write_debug(" : apid : ");
     * write_debug(AnsiString((long)apid));
     * write_debug(" : msgid : ");
     * write_debug(AnsiString((long)msg_id));
     * write_debug("\n");*/

    write(pipe_out, params, sizeof(int) * 3);
    if (params [2]) {
        write(pipe_out, DATA.c_str(), params [2]);
    }

    return(1);
}

void DeturnatedPrint(char *text, int len, void *userdata) {
    AnsiString data;

    if (len > -1) {
        data.LoadBuffer(text, len);
    } else {
        data = text;
    }
    NotifyParent(static_pipe_out, static_parent, -10, data);
}

CONCEPT_DLL_API SetNotifyParent(NOTIFY_PARENT np) {
    RemoteNotifyParent = np;
    return(0);
}

CONCEPT_DLL_API SetArguments(int argc, char **argv) {
    PIFAlizator::argc = argc;
    PIFAlizator::argv = argv;
    return(0);
}

struct ThreadData {
    int               static_pipe_out;
    int               static_parent;
    NOTIFY_PARENT     RemoteNotifyParent;
    int               argc;
    char              **argv;
    SimpleStream      *out;
    AnsiList          *ClassList;
    DEBUGGER_CALLBACK DEBUGGER_TRAP;
    void              *DEBUGGER_RESERVED;
    char              *SERVER_PUBLIC_KEY;
    char              *SERVER_PRIVATE_KEY;
    char              *CLIENT_PUBLIC_KEY;
    GarbageCollector  *Array__gc_obj;
};

void *GetStaticData() {
#ifndef SINGLE_PROCESS_DLL
    ThreadData *td = new ThreadData();
    td->static_pipe_out    = static_pipe_out;
    td->static_parent      = static_parent;
    td->RemoteNotifyParent = RemoteNotifyParent;
    td->argc = PIFAlizator::argc;
    td->argv = PIFAlizator::argv;

    td->out                = ConceptInterpreter::out;
    td->ClassList          = ConceptInterpreter::ClassList;
    td->DEBUGGER_TRAP      = ConceptInterpreter::DEBUGGER_TRAP;
    td->DEBUGGER_RESERVED  = ConceptInterpreter::DEBUGGER_RESERVED;
    td->SERVER_PUBLIC_KEY  = ConceptInterpreter::SERVER_PUBLIC_KEY;
    td->SERVER_PRIVATE_KEY = ConceptInterpreter::SERVER_PRIVATE_KEY;
    td->CLIENT_PUBLIC_KEY  = ConceptInterpreter::CLIENT_PUBLIC_KEY;
    td->Array__gc_obj      = Array::__gc_obj;

    return(td);
#else
    return(0);
#endif
}

void SetStaticData(void *data) {
#ifndef SINGLE_PROCESS_DLL
    ThreadData *td = (ThreadData *)data;
    static_pipe_out    = td->static_pipe_out;
    static_parent      = td->static_parent;
    RemoteNotifyParent = td->RemoteNotifyParent;
    PIFAlizator::argc  = td->argc;
    PIFAlizator::argv  = td->argv;

    ConceptInterpreter::out                = td->out;
    ConceptInterpreter::ClassList          = td->ClassList;
    ConceptInterpreter::DEBUGGER_TRAP      = td->DEBUGGER_TRAP;
    ConceptInterpreter::DEBUGGER_RESERVED  = td->DEBUGGER_RESERVED;
    ConceptInterpreter::SERVER_PUBLIC_KEY  = td->SERVER_PUBLIC_KEY;
    ConceptInterpreter::SERVER_PRIVATE_KEY = td->SERVER_PRIVATE_KEY;
    ConceptInterpreter::CLIENT_PUBLIC_KEY  = td->CLIENT_PUBLIC_KEY;
    Array::__gc_obj = td->Array__gc_obj;

    delete td;
#endif
}

CONCEPT_DLL_API Concept_CachedInit(char *filename) {
    // 1. test if compiled !
    struct stat buf;
    AnsiString  accel_filename(filename);

    accel_filename += (char *)DEFAULT_BIN_EXTENSION;

    /*#ifdef _WIN32
     *  AnsiString key("Global\\");
     *  int len=strlen(filename);
     *  for (int i=0;i<len;i++) {
     *      char c=filename[i];
     *      if (c=='\\')
     *          c='/';
     *      key+=c;
     *  }
     #else*/
    AnsiString key(filename);
    //#endif
    FILE *in = fopen(accel_filename.c_str(), "rb");
    if (!in) {
        //file does not exists
        return(-2);
    }
    fclose(in);
#ifdef _WIN32
    if (!stat(accel_filename.c_str(), &buf)) {
        key += (char *)":";
        key += AnsiString((long)buf.st_size);
        key += (char *)":";
        key += AnsiString((long)buf.st_mtime);
    }
#endif

    SHLockWait(key.c_str());
    cache_buffer = SHPoolOpen(key.c_str());
    if (cache_buffer) {
        return(0xFFFF);
    }
    if (!cache_buffer) {
        // 2. create buffer, compute size !
        int size = PIFAlizator::ComputeSharedSize(accel_filename.c_str());
        if (size > 0) {
            // 255 bytes reserved ...
            size += 0xFF;
            SHLock(key.c_str());
            cache_buffer = SHPoolAlloc(key.c_str(), size);
            if (cache_buffer) {
                return(size);
            }
        }
    }
    return(-1);
}

CONCEPT_DLL_API Concept_CachedDone(char *filename) {
    if (cache_buffer) {
        SHPoolFree(cache_buffer);
        cache_buffer = 0;
        SHUnLock();
    }
    return(0);
}

CONCEPT_DLL_API Concept_Compile(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent) {
    // programm internal form
    AnsiString S;

    if (S.LoadFile(filename)) {
        return(-3);
    }

    SimpleStream SS(fp, sock);

    char is_cgi = 0;
    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }
#ifdef TIMMING
    long start = clock();
    SS.Print("PIF...");
#endif
    PIFAlizator PIF(inc_dir, lib_dir, &S, &SS, filename, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY);
    PIF.SetPipe(pipe_in, pipe_out, apid, parent);
    PIF.Execute();
#ifdef TIMMING
    long total = clock() - start;
    SS.Print("done\n");
    SS.Print((double)total);
    SS.Print("OPTIMIZE...");
    start = clock();
#endif
    bool build_package = false;
    if (!PIF.ErrorCount()) {
        PIF.Optimize();
        build_package = !PIF.CheckRunable();

        if ((PIF.Errors.Count() == 1) && (build_package)) {
            PIF.Errors.Delete(PIF.Errors.Count() - 1);
        }

#ifdef OPTIMIZE_MEMORY
        if (!PIF.Errors.Count()) {
            SS.Print("Optimising memory usage ...\n");
            PIF.OptimizeMemoryUsage();
        }
#endif
        if (build_package) {
            //PIF.Errors.Clear();
            SS.Print("Building package ...\n");
        }
    }
#ifdef TIMMING
    total = clock() - start;
    SS.Print("done\n");
    SS.Print((double)total);
#endif
    if (!PIF.ErrorCount()) {
        if ((PIF.WarningCount()) && (!is_cgi)) {
            SS.Print(PIF.PRINT_WARNINGS());
            SS.Print("\n\nThere are warnings.\n");
        }
        AnsiString accel_filename = AnsiString(filename);
        accel_filename += build_package ? (char *)DEFAULT_PACK_EXTENSION : (char *)DEFAULT_BIN_EXTENSION;
        PIF.Serialize(accel_filename, build_package);
        // set file times ...
        struct stat buf;
        if (!stat(filename, &buf)) {
            struct utimbuf timebuf;
            timebuf.actime  = buf.st_mtime;
            timebuf.modtime = buf.st_mtime;

            // set the time for the accel file
            utime(accel_filename.c_str(), &timebuf);
        }
    } else {
        SS.Print(PIF.PRINT_ERRORS(is_cgi));
        return(-2);
    }
    return(0);
}

CONCEPT_DLL_API LastResult() {
    return(last_result);
}

CONCEPT_DLL_API Concept_Execute(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, void *userdata) {
    //Concept_CachedInit(filename);
    ForeignPrint fp_notify = 0;

    if ((parent > 0) && (pipe_out > 0)) {
        static_pipe_out = pipe_out;
        static_parent   = parent;
        fp_notify       = DeturnatedPrint;
    }

    SimpleStream SS(fp, sock, fp_notify, userdata);
    AnsiString   S;
    if (S.LoadFile(filename)) {
        NotifyParent(pipe_out, parent, -6, "File not found");
        SS.Print(AnsiString("File '") + AnsiString(filename) + "' is not available\n", -1, -2);
        return(-3);
    }

    char is_cgi = 0;

    if (debug == -2) {
        is_cgi = 1;
        debug  = 1;
    }

    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }
    PIFAlizator PIF(inc_dir, lib_dir, &S, &SS, filename, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY);
    PIF.SetPipe(pipe_in, pipe_out, apid, parent);
    NotifyParent(pipe_out, parent, -1, "Unserializing");
    //int start=clock();
    if ((!debug) && (!PIF.Unserialize(AnsiString(filename) + DEFAULT_BIN_EXTENSION))) {
        NotifyParent(pipe_out, parent, -1, "Optimizing");
        PIF.Optimize(true);
        //printf("Total time: %i\n", clock()-start);
        // in case we have some cache locking
        SHUnLock();
    } else {
        NotifyParent(pipe_out, parent, -1, "Compiling");
        PIF.Execute();
        if (!PIF.ErrorCount()) {
            NotifyParent(pipe_out, parent, -1, "Optimizing");
            PIF.Optimize();
#ifdef OPTIMIZE_MEMORY
            NotifyParent(pipe_out, parent, -1, "Optimizing memory usage");
            PIF.OptimizeMemoryUsage();
#endif
        }
    }
    if (!PIF.ErrorCount()) {
        PIF.CheckRunable();
    }
    if (!PIF.ErrorCount()) {
        if ((PIF.WarningCount()) && (!is_cgi)) {
            SS.Print(PIF.PRINT_WARNINGS());
            SS.Print("\n\nThere are warnings.\n");
            NotifyParent(pipe_out, parent, -5, PIF.SerializeWarningsErrors(1));
        }
        NotifyParent(pipe_out, parent, -4, "Running");

        ClassCode *CC = (ClassCode *)PIF.GetStartingPoint();
        //SS.Print(PIF.DEBUG_CLASS_CONFIGURATION());
        //PIF.Serialize(AnsiString(filename)+DEFAULT_BIN_EXTENSION);

        VariableDATA *STATIC_ENTRY = new VariableDATA;
        //STATIC_ENTRY->ACCESS         =   0;
        STATIC_ENTRY->CLASS_DATA         = 0;
        STATIC_ENTRY->LINKS              = 2;              // never to be deleted
        STATIC_ENTRY->NUMBER_DATA        = 0;
        STATIC_ENTRY->TYPE               = VARIABLE_CLASS; //TYPE_CLASS;
        STATIC_ENTRY->CALL_SET           = 0;
        STATIC_ENTRY->IS_PROPERTY_RESULT = 0;

        // nu mai e nevoie, pt ca se sterge singur ...
        STATIC_ENTRY->CLASS_DATA = CC->CreateInstance(STATIC_ENTRY, 0, 0, 0);
        STATIC_ENTRY->CLASS_DATA = 0;

        // already deleted
        //if (STATIC_ENTRY->CLASS_DATA)
        //    delete (CompiledClass *)STATIC_ENTRY->CLASS_DATA;

        delete STATIC_ENTRY;

        //delete (CompiledClass *)STATIC_ENTRY.CLASS_DATA;
        NotifyParent(pipe_out, parent, -6, "Done");
        last_result = PIFAlizator::last_result;
        if (PIF.ErrorCount()) {
            if (SS.AlreadyPrinted()) {
                SS.Print(PIF.PRINT_ERRORS(is_cgi ? 2 : 0));
            } else {
                SS.Print(PIF.PRINT_ERRORS(is_cgi));
            }
            if (!is_cgi) {
                SS.Print("There are RUN-TIME errors.", -1, -2);
            } else {
                SS.Print("\n\nThere are RUN-TIME errors.\n");
            }

            //printf("%s", PIF.PRINT_ERRORS());
            NotifyParent(pipe_out, parent, -2, PIF.SerializeWarningsErrors());
            return(-1);
        }
    } else {
        if (SS.AlreadyPrinted()) {
            SS.Print(PIF.PRINT_ERRORS(is_cgi ? 2 : 0));
        } else {
            SS.Print(PIF.PRINT_ERRORS(is_cgi));
        }

        if (!is_cgi) {
            SS.Print("There are compile-time errors.", -1, -2);
        }

        NotifyParent(pipe_out, parent, -3, PIF.SerializeWarningsErrors());
        //while (true);
        return(-2);
    }
    return(0);
}

//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_Execute2(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, CHECK_POINT cp, void *userdata) {
    PIFAlizator::CheckPoint = cp;
    return(Concept_Execute(filename, inc_dir, lib_dir, fp, sock, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY, pipe_in, pipe_out, apid, parent, userdata));
}

//-----------------------------------------------------------------------------------
#ifdef _WIN32
//-----------------------------------------------------------------------------------
// DLL Entry Point ... nothing ...
int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void *lpReserved) {
    return(1);
}

//-----------------------------------------------------------------------------------
#endif
