//#define GO_DEBUG

//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include "AnsiString.h"

#include <libxml/xmlreader.h>

#define MAX_OBJECTS     0xFFFF

void *BACK_REFERENCES[MAX_OBJECTS];
char BACK_TYPES[MAX_OBJECTS];
int BACK_REF_COUNT;
INVOKE_CALL LocalInvoker;
void *err_ser;
AnsiString full_error;
void *top_variable;

void *ConceptHandler=0;

void Serialize(void *pData, xmlNodePtr parent, char is_simple=0);
void SerializeVariable(char *member, int type, char *szData, NUMBER nData, void *class_data, void *variable_data, xmlNodePtr parent, char is_simple=0);
void SerializeArray(void *pData, void *arr_id, xmlNodePtr parent, char is_simple=0);

#ifdef GO_DEBUG
    #include <iostream>
    #define DEBUG(data) std::cout << (data) << "\n"; std::cout.flush();
    #define DEBUG2(data,data2) std::cout << (data) << ": " << (data2) << "\n"; std::cout.flush();
    #define DEBUG3(data,data2,data3) std::cout << (data) << ": " << (data2) << ", " << (data3) << "\n"; std::cout.flush();
#else
    #define DEBUG(data)
    #define DEBUG2(data,data1)
    #define DEBUG3(data,data1,data2)
#endif

#define RESET_ERROR     if (xml_error.Length()) xml_error.LoadBuffer(0,0);

xmlDocPtr doc=0;
//---------------------------------------------------------------------------
static AnsiString xml_error;
void MyGenericErrorFunc(void * ctx, const char * msg, ...) {
    va_list args;
    va_start(args, msg);
    char buffer[0xFFF];
    buffer[0]=0;
    vsprintf(buffer, msg, args);
    va_end(args);
    xml_error+=buffer;
}
//---------------------------------------------------------------------------
#ifdef _WIN32
xmlFreeFunc xmlFree=0;
#endif
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
#ifdef _WIN32
    xmlMallocFunc xmlMalloc=0;
    xmlReallocFunc xmlRealloc=0;
    if(!xmlFree)
        xmlMemGet(&xmlFree,&xmlMalloc,&xmlRealloc,NULL);
#endif
    LocalInvoker=Invoke;
    ConceptHandler=HANDLER;
	return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API ON_DESTROY_CONTEXT MANAGEMENT_PARAMETERS {
	return 0;
}
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(SerializeError,0)
    INTEGER len=xml_error.Length();
    if (len>0) {
        AnsiString res="<pre><![CDATA[\n";
        res+=xml_error;
        res+="\n]]></pre>";
        RETURN_BUFFER(xml_error.c_str(), xml_error.Length())
    } else {
        RETURN_STRING("");
    }
END_IMPL
//------------------------------------------------------------------------
int CheckBack(void *pData) {
    // check if a reference is already serialized (anti-cyclic reference)
    for (int i=0;i<BACK_REF_COUNT;i++) {
        if (BACK_REFERENCES[i]==pData)
            return i+1;
    }
    if (BACK_REF_COUNT<MAX_OBJECTS)
        BACK_REFERENCES[BACK_REF_COUNT++]=pData;
    return 0;
}
//---------------------------------------------------------------------------
void SerializeArray(void *pData, void *arr_id, xmlNodePtr parent, char is_simple) {
    void *newpData;
    char *key;
    xmlNodePtr node;
    xmlNodePtr node2;

    if (!is_simple) {
        int ref_ptr=CheckBack(arr_id);
        if (ref_ptr) {
            DEBUG("...backreferenced array");
            node = xmlNewNode(NULL, BAD_CAST "cyclic_reference");
            xmlNewProp(node, BAD_CAST "refID", BAD_CAST AnsiString((long)ref_ptr).c_str());

            xmlAddChild(parent, node);
            return;
        }
    }


    DEBUG2("           array",arr_id);
    int count=LocalInvoker(INVOKE_GET_ARRAY_COUNT,pData);

    if (is_simple){
        node=parent;
        if (!node) {
            node = xmlNewNode(NULL, BAD_CAST "array");
            xmlDocSetRootElement(doc, node);
        }
    } else {
        node = xmlNewNode(NULL, BAD_CAST "array");
        xmlNewProp(node, BAD_CAST "cycid", BAD_CAST AnsiString((long)BACK_REF_COUNT).c_str());
        xmlAddChild(parent, node);
    }

    for (int i=0;i<count;i++) {
        newpData=0;
        key=0;
        LocalInvoker(INVOKE_ARRAY_VARIABLE,pData,i,&newpData);
        LocalInvoker(INVOKE_GET_ARRAY_KEY,pData,i,&key);
        if (newpData) {
            node2 = xmlNewNode(NULL, BAD_CAST "element");
            xmlAddChild(node,node2);

            if (key)
                xmlNewProp(node2, BAD_CAST "key", BAD_CAST key);
            xmlNewProp(node2, BAD_CAST "index", BAD_CAST AnsiString((long)i).c_str());

            char *szData;
            INTEGER type;
            NUMBER nData;
            
            LocalInvoker(INVOKE_GET_VARIABLE,newpData,&type,&szData,&nData);

            if (type==VARIABLE_DELEGATE) {
                char *deleg_name=0;
                //xmlNewProp(node2, BAD_CAST "type", BAD_CAST "delegate");
                LocalInvoker(INVOKE_GET_MEMBER_FROM_ID,(void *)szData,(long)nData,&deleg_name);
                nData=(long)deleg_name;
                char *class_name=0;
                LocalInvoker(INVOKE_GET_SERIAL_CLASS,(void *)szData,(int)0,&class_name,(char **)0,(char *)0,(char *)0,(char *)0,(char **)0,(NUMBER *)0,(char *)0,(char *)0);
                SerializeVariable(/*key*/0,type,class_name,nData,/*0*/(void *)szData,newpData,node2,is_simple);
            } else
            if (type!=VARIABLE_CLASS) {
                SerializeVariable(/*key*/0,type,szData,nData,/*0*/(void *)szData,newpData,node2,is_simple);
            } else {
                if (!is_simple)
                    xmlNewProp(node2, BAD_CAST "type", BAD_CAST "object");
                Serialize(szData,node2,is_simple);
            }
        }
    }
}
//---------------------------------------------------------------------------

void SerializeVariable(char *member, int type, char *szData, NUMBER nData, void *class_data, void *variable_data, xmlNodePtr parent, char is_simple) {
    DEBUG2("      -> variable", member ? member : "array element");
    xmlNodePtr node;
    xmlNodePtr node2;

    if (member) {
        if (is_simple) {
            node = xmlNewNode(NULL, BAD_CAST (member ? member : "%NO"));
            xmlAddChild(parent, node);
        } else {
            node = xmlNewNode(NULL, BAD_CAST "member");
            xmlAddChild(parent, node);
            xmlNewProp(node, BAD_CAST "name", BAD_CAST (member ? member : "%NO"));
        }
    } else {
        node=parent;
    }

    switch (type) {
        case VARIABLE_STRING:
                DEBUG2("           string",(void *)szData);
                DEBUG2("           length",(long)nData);
                if (nData>0)
                    DEBUG2("           content[0]",szData[0]);
                xmlNewProp(node, BAD_CAST "type", BAD_CAST "string");
                if ((szData) && (nData>0)) {
                   node2 = xmlNewText(BAD_CAST szData);
                   xmlAddChild(node, node2);
                }
                break;
        case VARIABLE_NUMBER:
                DEBUG("           number");
                xmlNewProp(node, BAD_CAST "type", BAD_CAST "number");
                node2 = xmlNewText(BAD_CAST AnsiString(nData).c_str());
                xmlAddChild(node, node2);
                break;
        case VARIABLE_CLASS:
                DEBUG("           class");
                xmlNewProp(node, BAD_CAST "type", BAD_CAST "class");
                Serialize(class_data,node,is_simple);
                break;
        case VARIABLE_ARRAY:
                DEBUG("           array");
                xmlNewProp(node, BAD_CAST "type", BAD_CAST "array");
                SerializeArray(variable_data,class_data,node,is_simple);
                break;
        case VARIABLE_DELEGATE:
                DEBUG("           delegate");
                xmlNewProp(node, BAD_CAST "type", BAD_CAST "delegate");
                DEBUG2("              class",(void *)szData);
                xmlNewProp(node, BAD_CAST "class", BAD_CAST (szData ? szData : "%ERROR"));
                DEBUG2("              member",(void *)(long)nData);
                xmlNewProp(node, BAD_CAST "member", BAD_CAST (nData ? (char *)(long)nData : "%ERROR"));
                break;
    }
    DEBUG("           serialized");
}
//---------------------------------------------------------------------------
void Serialize(void *pData, xmlNodePtr parent, char is_simple) {
    xmlNodePtr node;

    char *class_name=0;
    if (!is_simple) {
        int ref_ptr=CheckBack(pData);
        if (ref_ptr) {
            DEBUG("...backreference");
            node = xmlNewNode(NULL, BAD_CAST "cyclic_reference");
            xmlNewProp(node, BAD_CAST "refID", BAD_CAST AnsiString((long)ref_ptr).c_str());

            xmlAddChild(parent, node);
            return;
        }
    }

    DEBUG("Getting members count, classname");
    int members_count=LocalInvoker(INVOKE_GET_SERIAL_CLASS,pData,(int)0,&class_name,(char **)0,(char *)0,(char *)0,(char *)0,(char **)0,(NUMBER *)0,(char *)0,(char *)0);
    DEBUG("... success");

    if (is_simple) {
        node=parent;
        if (!node) {
            node = xmlNewNode(NULL, BAD_CAST class_name);
            xmlDocSetRootElement(doc, node);
        }
    } else {
        node = xmlNewNode(NULL, BAD_CAST "object");

        if (!parent) {
            xmlDocSetRootElement(doc, node);
        } else
            xmlAddChild(parent, node);

        xmlNewProp(node, BAD_CAST "cycid", BAD_CAST AnsiString((long)BACK_REF_COUNT).c_str());

        DEBUG2(class_name,"start");
        xmlNewProp(node, BAD_CAST "class", BAD_CAST (class_name ? class_name : "%ERROR%"));
    }

    if (members_count>0) {
        char **members=new char * [members_count];
        char *flags=new char[members_count];
        char *access=new char[members_count];
        char *types=new char[members_count];
        char **szValues=new char * [members_count];
        NUMBER *nValues=new NUMBER[members_count];
        void **class_data=new void * [members_count];
        void **variable_data=new void * [members_count];

        //memset(szValues,0,members_count * sizeof(char *));
        //memset(nValues,0,members_count * sizeof(NUMBER));

        DEBUG2(class_name,"invoke");
        int result=LocalInvoker(INVOKE_GET_SERIAL_CLASS,pData,members_count,&class_name,members,flags,access,types,(const char **)szValues,nValues,class_data,variable_data);
        DEBUG2(class_name,"done invoke");

        if (result!=0) {
            err_ser=(void *)"Error in serialization (bug ?)";
            return;
        }
        DEBUG2(class_name,"members loop");

#ifdef GO_DEBUG
        for (int i=0;i<members_count;i++) {
            if ((flags[i]==0) && (types[i]==VARIABLE_STRING)) {
                DEBUG3(class_name,"string member",i);
                DEBUG3(class_name,"string member",(void *)szValues[i]);
            }
        }
#endif

        for (int i=0;i<members_count;i++) {
            if (flags[i]==0) {
                DEBUG3(class_name,"serialize",members[i]);
                SerializeVariable(members[i], types[i], szValues[i], nValues[i], class_data[i], variable_data[i],node,is_simple);
                if (err_ser)
                    return;
            }
        }
        DEBUG2(class_name,"done members loop");

        delete[] members;
        delete[] flags;
        delete[] access;
        delete[] types;
        delete[] szValues;
        delete[] nValues;
        delete[] class_data;
        delete[] variable_data;

        DEBUG2(class_name,"delete");
    }
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_SerializeObject CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK_MIN_MAX(1,3,"SerializeObject: SerializeObject(object_to_serialize,filename=\"-\",encoding=\"UTF-8\")");

    LOCAL_INIT;
    char *pData;
    char *filename="-";
    char *encoding="UTF-8";

    GET_ARRAY(0,pData);

    CHECK_CLASS("SerializeObject parameter error: object expected (first parameter)");
    if (PARAMETERS_COUNT>1) {
        GET_CHECK_STRING(1,filename,"SerializeObject parameter error: filename must be of string type(STATIC STRING)");
    }
    if (PARAMETERS_COUNT>2) {
        GET_CHECK_STRING(2,encoding,"SerializeObject parameter error: encoding must be of string type(STATIC STRING)");
    }
    
    BACK_REF_COUNT=0;
    err_ser=0;

    void *old_ctx=xmlGenericErrorContext;
    xmlSetGenericErrorFunc(0, MyGenericErrorFunc);

    doc = xmlNewDoc(BAD_CAST "1.0");
    //xmlDtdPtr dtd = xmlCreateIntSubset(doc, BAD_CAST "root", NULL, BAD_CAST "tree2.dtd");

    Serialize(pData,0);
    DEBUG("All done");
    int is_file=false;

    if ((filename) && (filename[0])) {
        is_file=true;
        DEBUG("XML Save file ...");
        xmlSaveFormatFileEnc(filename, doc, encoding, 1);
        DEBUG("done");
    } else {
        xmlChar *mem;
        int size;

        xmlDocDumpMemoryEnc(doc,&mem,&size,encoding);

        if (size<0)
	        size=0;

        RETURN_BUFFER((char *)mem,size);

        xmlFree(mem);
        //free(mem);
    }


    DEBUG("Clean ...");
    xmlFreeDoc(doc);
    //xmlCleanupParser();
    xmlMemoryDump();
    DEBUG("done");

    if (is_file) {
        RETURN_NUMBER(0);
    }
    xmlSetGenericErrorFunc(old_ctx, xmlGenericError);
    return err_ser;
}
//---------------------------------------------------------------------------
int DoNode(xmlNodePtr node, void *OwnerPTR=0, void *obj_ptr=0) {
    if (!node)
        return 0;

    xmlNode *cur_node = NULL;

    AnsiString prop;
    AnsiString node_type;

    for (cur_node = node; cur_node; cur_node = cur_node->next) {
        if (cur_node->type == XML_ELEMENT_NODE) {
            node_type=(char *)cur_node->name;
            if (node_type==(char *)"member") {
                char *prop_ptr=(char *)xmlGetProp(cur_node, BAD_CAST "type");
                char *name_ptr=(char *)xmlGetProp(cur_node, BAD_CAST "name");
                if ((prop_ptr) && (name_ptr)) {
                    prop=(char *)prop_ptr;
                    if (prop==(char*)"number") {
                        AnsiString value=(cur_node->children ? (char *)cur_node->children->content : (char *)"" );
                        NUMBER nValue=value.ToFloat();
                        void *handler=0;
                        if (!IS_OK(LocalInvoker(INVOKE_GET_CLASS_VARIABLE, OwnerPTR, (char *)name_ptr,&handler))) {
                            err_ser=(char *)"Error in member serialization. Member not defined ?";
                            return -1;
                        }
                        LocalInvoker(INVOKE_SET_CLASS_MEMBER, obj_ptr, (char *)name_ptr, (INTEGER)VARIABLE_NUMBER, 0, nValue);
                    } else
                    if (prop==(char*)"string") {
                        AnsiString value=(cur_node->children ? (char *)cur_node->children->content : (char *)"" );
                        void *handler=0;
                        DEBUG(name_ptr);
                        if (!IS_OK(LocalInvoker(INVOKE_GET_CLASS_VARIABLE, OwnerPTR, (char *)name_ptr,&handler))) {
                            err_ser=(char *)"Error in member serialization. Member not defined ?";
                            return -1;
                        }
                        DEBUG2(name_ptr, "ok, set value :");
                        DEBUG(value.c_str());
                        LocalInvoker(INVOKE_SET_CLASS_MEMBER, obj_ptr, (char *)name_ptr, (INTEGER)VARIABLE_STRING, value.c_str(), (NUMBER)value.Length());
                        DEBUG2(name_ptr, "done value");
                    } else
                    if (prop==(char*)"class") {
                        void *handler=0;
                        DEBUG("Expanding class member ...");
                        if (!IS_OK(LocalInvoker(INVOKE_GET_CLASS_VARIABLE, OwnerPTR, (char *)name_ptr,&handler))) {
                            err_ser=(char *)"Error in member serialization. Member not defined ?";
                            return -1;
                        }
                        DoNode(cur_node->children, handler);
                    } else
                    if (prop==(char*)"array") {
                        void *handler=0;
                        DEBUG("Expanding array member ...");
                        if (!IS_OK(LocalInvoker(INVOKE_GET_CLASS_VARIABLE, OwnerPTR, (char *)name_ptr,&handler))) {
                            err_ser=(char *)"Error in member serialization. Member not defined ?";
                            return -1;
                        }
                        DoNode(cur_node->children, handler);
                    }
                }
                if (prop_ptr)
                    xmlFree(prop_ptr);
                if (name_ptr)
                    xmlFree(name_ptr);
            } else
            if (node_type==(char *)"element") {
                char *key=0;
                void *var_data=OwnerPTR;
                void *elem_data=0;
                char *prop_ptr=(char *)xmlGetProp(cur_node, BAD_CAST "type");


                key=(char *)xmlGetProp(cur_node, BAD_CAST "key");
                if (key) {
                    LocalInvoker(INVOKE_ARRAY_VARIABLE_BY_KEY,var_data,key,&elem_data);
                    //DoNode(cur_node->children, var_data, elem_data);
                } else {
                    INTEGER index=AnsiString((char *)xmlGetProp(cur_node, BAD_CAST "index")).ToInt();
                    if (index<0)
                        index=0;
                    LocalInvoker(INVOKE_ARRAY_VARIABLE,var_data,index,&elem_data);
                    //DoNode(cur_node->children, elem_data);
                }
                if (prop_ptr) {
                    prop=prop_ptr;
                    if (prop==(char *)"number") {
                        AnsiString value=(cur_node->children ? (char *)cur_node->children->content : (char *)"" );
                        NUMBER nValue=value.ToFloat();
                        LocalInvoker(INVOKE_SET_VARIABLE, elem_data, (INTEGER)VARIABLE_NUMBER, 0, nValue);
                    } else
                    if (prop==(char *)"string") {
                        AnsiString value=(cur_node->children ? (char *)cur_node->children->content : (char *)"" );
                        LocalInvoker(INVOKE_SET_VARIABLE, elem_data, (INTEGER)VARIABLE_STRING, value.c_str(), (NUMBER)value.Length());
                    } else
                    if (prop==(char *)"object") {
                        DEBUG("Object in array ...");
                        DoNode(cur_node->children, elem_data);
                    } else
                    if (prop==(char *)"array") {
                        DoNode(cur_node->children, elem_data);
                    }
                    xmlFree(prop_ptr);
                }
            } else
            if (node_type==(char *)"object") {
                char *class_name=0;
                void *var_data=OwnerPTR;

                class_name=(char *)xmlGetProp(cur_node, BAD_CAST "class");
                DEBUG2("Building for ", class_name);

                if (!IS_OK(LocalInvoker(INVOKE_CREATE_OBJECT,ConceptHandler,var_data,class_name))) {
                    full_error=(char *)"Unable to create class. Class ";
                    full_error+=class_name;
                    full_error+=(char *)" undefined.";
                    err_ser=(void *)full_error.c_str();
                    return -1;
                }
                INTEGER type;
                char *str=0;
                NUMBER nvalue;

                LocalInvoker(INVOKE_GET_VARIABLE,var_data,&type,&str,&nvalue);

                char *cyc_id_ptr=(char *)xmlGetProp(cur_node, BAD_CAST "cycid");
                int cyc_id=AnsiString(cyc_id_ptr).ToInt();
                if ((cyc_id>0) && (cyc_id<=MAX_OBJECTS)) {
                    cyc_id--;
                    BACK_REFERENCES[cyc_id]=str;
                    BACK_REF_COUNT++;
                    BACK_TYPES[cyc_id]=VARIABLE_CLASS;
                }

                DoNode(cur_node->children, var_data, (void *)str);
                if (class_name)
                    xmlFree(class_name);
                if (cyc_id_ptr)
                    xmlFree(cyc_id_ptr);
            } else
            if (node_type==(char *)"array") {
                void *var_data=OwnerPTR;
                INTEGER type;
                char *str=0;
                NUMBER nvalue;


                LocalInvoker(INVOKE_CREATE_ARRAY, var_data, 0);

                LocalInvoker(INVOKE_GET_VARIABLE,var_data,&type,&str,&nvalue);

                char *cyc_id_ptr=(char *)xmlGetProp(cur_node, BAD_CAST "cycid");
                int cyc_id=AnsiString(cyc_id_ptr).ToInt();
                if ((cyc_id>0) && (cyc_id<=MAX_OBJECTS)) {
                    cyc_id--;
                    BACK_REFERENCES[cyc_id]=str;
                    BACK_REF_COUNT++;

                    BACK_TYPES[cyc_id]=VARIABLE_ARRAY;
                }

                DoNode(cur_node->children, var_data, (void *)str);
                if (cyc_id_ptr)
                    xmlFree(cyc_id_ptr);
            } else
            if (node_type==(char *)"cyclic_reference") {
                char *ref_id_ptr=(char *)xmlGetProp(cur_node, BAD_CAST "refID");
                int ref_id=AnsiString(ref_id_ptr).ToInt()-1;
                if ((ref_id>=0) && (ref_id<BACK_REF_COUNT)) {
                    DEBUG("Valid cyclic reference !");
                    LocalInvoker(INVOKE_SET_VARIABLE,OwnerPTR,(INTEGER)BACK_TYPES[ref_id],BACK_REFERENCES[ref_id],(NUMBER)0);
                }
                if (ref_id_ptr)
                    xmlFree(ref_id_ptr);
            }
        }
    }
    return 1;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_UnSerializeObject CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK_MIN_MAX(1, 3,"UnSerializeObject: UnSerializeObject(filename|buffer, is_raw_buffer=false, var error)");

    LOCAL_INIT;
    char *filename="-";
    NUMBER _is_raw;
    NUMBER len;
    bool is_raw=false;

    GET_CHECK_BUFFER(0,filename, len,"UnSerializeObject parameter error: filename must be of string type(STATIC STRING)");

    if (PARAMETERS_COUNT>=2) {
        GET_CHECK_NUMBER(1,_is_raw,"UnSerializeObject parameter error: is_raw_buffer must be a number");
        is_raw=(bool)(int)_is_raw;
    }
    
    // setting the return to number, 0
    RETURN_NUMBER(0);
    top_variable=RESULT;

    LIBXML_TEST_VERSION

    void *old_ctx=xmlGenericErrorContext;
    xmlSetGenericErrorFunc(0, MyGenericErrorFunc);
    if (is_raw) {
        doc = xmlReadMemory(filename, (int)len, "include.xml", NULL, 0);
        if (!doc)
            doc = xmlReadMemory(filename, (int)len, "include.xml", "UTF-8", 0);
    } else {
        doc = xmlReadFile(filename, NULL, 0);
        if (!doc)
            doc = xmlReadFile(filename, "UTF-8", 0);
    }

    if (!doc) {
        if (PARAMETERS_COUNT>2) {
            SET_STRING(2, "UnSerializeObject failed to load the XML document");
        }
        xmlSetGenericErrorFunc(old_ctx, xmlGenericError);
        return 0;//return (void *)"UnSerializeObject failed to load the XML document";
    }

    err_ser=0;
    xmlNodePtr node = xmlDocGetRootElement(doc);
    DoNode(node, RESULT);

    xmlFreeDoc(doc);
    //xmlCleanupParser();

    if (PARAMETERS_COUNT>2) {
        if (err_ser) {
            SET_STRING(2, err_ser);
        } else {
            SET_STRING(2, "");
        }
    }
    xmlSetGenericErrorFunc(old_ctx, xmlGenericError);
    return 0;
    //return err_ser;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_HasMember CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(2,"HasMember: HasMember(Object,szMemberName)");

    LOCAL_INIT;
    char *membername=0;
    char *pData;

    GET_ARRAY(0,pData);

    if (TYPE!=VARIABLE_CLASS) {
        RETURN_NUMBER(0);
        return 0;
    }

    GET_CHECK_STRING(1,membername,"HasMember parameter error: szMemberName must be of string type(STATIC STRING)");

    int result=(Invoke(INVOKE_HAS_MEMBER, pData, membername)!=0);

    RETURN_NUMBER(result);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_SetMember CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK(3,"SetMember: SetMember(Object,szMemberName,Value)");
    LOCAL_INIT;
    char *membername=0;
    char *pData;

    char *szData;
    NUMBER nData;

    GET_ARRAY(0,pData);

    if (TYPE!=VARIABLE_CLASS) {
        RETURN_NUMBER(0);
        return 0;
    }

    GET_CHECK_STRING(1,membername,"SetMember parameter error: szMemberName must be of string type(STATIC STRING)");
    
    GET_BUFFER(2,szData,nData);

    RETURN_NUMBER(IS_OK(Invoke(INVOKE_SET_CLASS_MEMBER, pData, membername, TYPE, szData, nData)));
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetMember CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK(3,"GetMember: GetMember(Object,szMemberName,Value)");
    LOCAL_INIT;
    char *membername=0;
    char *pData;

    char *szData=0;
    NUMBER nData=0;

    GET_ARRAY(0,pData);

    if (TYPE!=VARIABLE_CLASS) {
        RETURN_NUMBER(0);
        return 0;
    }

    GET_CHECK_STRING(1,membername,"SetMember parameter error: szMemberName must be of string type(STATIC STRING)");
    
    TYPE=-1;
    int is_ok=IS_OK(Invoke(INVOKE_GET_CLASS_MEMBER, pData, membername, &TYPE, &szData, &nData));
    RETURN_NUMBER(is_ok);
    if (is_ok)
        SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],TYPE,(char *)szData,nData);
    else {
        SET_NUMBER(2,0);
    }
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_ObjectLinks CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK(1,"ObjectLinks: ObjectLinks(object|array|delegate)");
    LOCAL_INIT;

    RETURN_NUMBER((double)Invoke(INVOKE_OBJECT_LINKS, PARAMETER(0)));
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_StaticLinks CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK(1,"StaticLinks: StaticLinks(any_variable_type)");
    LOCAL_INIT;

    RETURN_NUMBER((double)Invoke(INVOKE_VAR_LINKS, PARAMETER(0)));
    return 0;
}
//---------------------------------------------------------------------------
static AnsiString as_result;
CONCEPT_DLL_API CONCEPT__ASSERT CONCEPT_API_PARAMETERS {
    RESET_ERROR

    PARAMETERS_CHECK_MIN_MAX(1,2,"_ASSERT: _ASSERT(condition[, szErrMessage=\"\"])");
    LOCAL_INIT;
    
    as_result="Assertion failed";

    char *pData;
    NUMBER nLength;
    char *err_data=0;

    if (PARAMETERS_COUNT==2) {
        GET_CHECK_STRING(1,err_data, "_ASSERT: szErrMessage should be of STATIC STRING type");
    }

    GET_BUFFER(0,pData,nLength);
    
    if (TYPE==VARIABLE_NUMBER) {
        if ((TYPE==VARIABLE_NUMBER) && (!nLength)) {
            if (err_data) {
                as_result+=": ";
                as_result+=err_data;
            }
            RETURN_NUMBER(0);
            return (void *)as_result.c_str();
        } else {
            RETURN_NUMBER(1);
            return 0;
        }
    }

    RETURN_NUMBER(1);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT__ClsPtr CONCEPT_API_PARAMETERS {
    RESET_ERROR
    PARAMETERS_CHECK(1,"_ClsPtr: _ClsPtr(object)");

    LOCAL_INIT;

    char *pData;
    NUMBER nLength;

    GET_BUFFER(0,pData,nLength);

    RETURN_NUMBER((long)pData);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT__VarPtr CONCEPT_API_PARAMETERS {
    RESET_ERROR
    PARAMETERS_CHECK(1,"_VarPtr: _VarPtr(variable)");

    LOCAL_INIT;

    RETURN_NUMBER((long)PARAMETER(0));
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT__KeySorted CONCEPT_API_PARAMETERS {
    RESET_ERROR
    PARAMETERS_CHECK(1,"_KeySorted: array _KeySorted(array)");

    LOCAL_INIT;

    char *arr=0;
    char *arr2=0;

    GET_CHECK_ARRAY(0,arr,"_KeySorted: paramter should be an array");

    if (IS_OK(Invoke(INVOKE_SORT_ARRAY_BY_KEYS,arr,&arr2))) {
        RETURN_ARRAY(arr2);
    } else
        return (void *)"_KeySorted: failed";

    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT__IsSet CONCEPT_API_PARAMETERS {
    RESET_ERROR
    PARAMETERS_CHECK(2,"_IsSet: boolean _IsSet(array, nIndex|szKey)");

    LOCAL_INIT;

    char *arr=0;
    char *arr2=0;

	NUMBER len;

	INTEGER index;
	char *key;

    GET_CHECK_ARRAY(0,arr,"_IsSet: paramter 1 should be an array");
	GET_BUFFER(1, key, len);

	index=(INTEGER)len;

	if (TYPE==VARIABLE_STRING)
		index=-1;
	else
	if (TYPE==VARIABLE_NUMBER)
		key=0;
	else
		return (void *)"_IsSet: parameter 2 should be a number or a string";

	int result=Invoke(INVOKE_ARRAY_ELEMENT_IS_SET,PARAMETER(0),index,key);

    if (IS_OK(result)) {
        RETURN_NUMBER(result);
    } else
        return (void *)"_IsSet: failed (may be not supported on your version. Please update.)";

    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT__GetKeys CONCEPT_API_PARAMETERS {
    RESET_ERROR
    PARAMETERS_CHECK(1,"_GetKeys: array _GetKeys(array)");

    LOCAL_INIT;

    char *arr=0;
    char *arr2=0;

    GET_CHECK_ARRAY(0,arr,"_GetKeys: paramter should be an array");

    char *key;
    int count=Invoke(INVOKE_GET_ARRAY_COUNT,PARAMETER(0));

	Invoke(INVOKE_CREATE_ARRAY, RESULT, 0);

	for (int i=0;i<count;i++) {
        key=0;
        Invoke(INVOKE_GET_ARRAY_KEY,PARAMETER(0),i,&key);
		if (key) {
			Invoke(INVOKE_SET_ARRAY_ELEMENT, RESULT, (INTEGER)i, VARIABLE_STRING, key, (NUMBER)0);
		} else {
			Invoke(INVOKE_SET_ARRAY_ELEMENT, RESULT, (INTEGER)i, VARIABLE_NUMBER, key, (NUMBER)0);
		}
	}
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(ToXML,1,2)
    RESET_ERROR
    T_OBJECT(0)
    char *pData=PARAM(0);
    char *filename="-";
    char *encoding="UTF-8";

    if (PARAMETERS_COUNT>1) {
        T_STRING(1);
        encoding=PARAM(1);
    }
    
    BACK_REF_COUNT=0;
    err_ser=0;

    void *old_ctx=xmlGenericErrorContext;
    xmlSetGenericErrorFunc(0, MyGenericErrorFunc);

    doc = xmlNewDoc(BAD_CAST "1.0");
    Serialize(pData,0,1);
    DEBUG("All done");
    int is_file=false;


    xmlChar *mem=0;
    int size;

    xmlDocDumpMemoryEnc(doc,&mem,&size,encoding);

    if (size<0)
        size=0;

    RETURN_BUFFER((char *)mem,size);

    xmlFree(mem);
    //free(mem);


    DEBUG("Clean ...");
    xmlFreeDoc(doc);
    //xmlCleanupParser();
    xmlMemoryDump();
    DEBUG("done");

    if (is_file) {
        RETURN_NUMBER(0);
    }
    xmlSetGenericErrorFunc(old_ctx, xmlGenericError);
    return err_ser;
END_IMPL
