//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include "AnsiString.h"
#include <string.h>

#include <poppler.h>

//------------------------------------------------------------------------
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
    g_type_init ();

    return 0;
}
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(PDFLoadBuffer, 1,3)
    T_STRING(0)
    GError *docerr=0;
    char *password=NULL;

    if (PARAMETERS_COUNT>1) {
        T_STRING(1)
        password=PARAM(1);
    }

    if (PARAMETERS_COUNT>2) {
        SET_STRING(2, "");
    }

    PopplerDocument *pdoc=poppler_document_new_from_data(PARAM(0),PARAM_LEN(0), password, &docerr);
    if (docerr) {
        if (PARAMETERS_COUNT>2) {
            SET_STRING(2, docerr->message);
        }
    }
    RETURN_NUMBER((NUMBER)(INTEGER)pdoc)
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(PDFLoad, 1,3)
    T_STRING(0)
    GError *docerr=0;
    char *password=NULL;

    if (PARAMETERS_COUNT>1) {
        T_STRING(1)
        password=PARAM(1);
    }
    if (PARAMETERS_COUNT>2) {
        SET_STRING(2, "");
    }

    PopplerDocument *pdoc=poppler_document_new_from_file(PARAM(0), password, &docerr);
    if (docerr) {
        if (PARAMETERS_COUNT>2) {
            SET_STRING(2, docerr->message);
        }
    }
    RETURN_NUMBER((NUMBER)(INTEGER)pdoc)
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(PDFClose, 1)
    T_NUMBER(0)
    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);

    if (pdoc)
        g_object_unref(pdoc);
    RETURN_NUMBER(0)
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(PDFPageCount, 1)
    T_HANDLE(0)
    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);

    RETURN_NUMBER(poppler_document_get_n_pages(pdoc))
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(PDFPageText, 2)
    T_HANDLE(0)
    T_NUMBER(1)

    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);

    PopplerPage *page=poppler_document_get_page(pdoc, PARAM_INT(1));
    if (page) {
        PopplerRectangle rect;
        gdouble          width, height;

        poppler_page_get_size (page, &width, &height);
        rect.x1 = rect.y1 = 0;
        rect.x2 = width;
        rect.y2 = height;

        RETURN_STRING(poppler_page_get_text(page, POPPLER_SELECTION_GLYPH, &rect))

        g_object_unref (G_OBJECT (page));
    } else {
        RETURN_STRING("")
    }
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(PDFFindText, 3)
    T_HANDLE(0)
    T_NUMBER(1)
    T_STRING(2)

    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);

    PopplerPage *page=poppler_document_get_page(pdoc, PARAM_INT(1));
    if (page) {
        GList *lst=poppler_page_find_text(page, PARAM(2)); 

        int index=0;
        Invoke(INVOKE_CREATE_ARRAY,RESULT);
        for (GList * b = lst; b; b = b->next) {
            PopplerRectangle *rect = (PopplerRectangle *)b->data;
            if (rect) {
                void *newpData=0;
                Invoke(INVOKE_ARRAY_VARIABLE,RESULT,index,&newpData);
                Invoke(INVOKE_CREATE_ARRAY,newpData);

                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,newpData,"x1",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)rect->x1);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,newpData,"x2",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)rect->x2);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,newpData,"y1",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)rect->y1);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,newpData,"y2",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)rect->y2);
                index++;
            }
        }
        g_object_unref (G_OBJECT (page));
        /*if (lst) {
            RETURN_NUMBER(1)
        } else {
            RETURN_NUMBER(0)
        }*/
    }/* else {
        RETURN_NUMBER(0)
    }*/
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(PDFPageImageBuffer, 2, 5)
    T_HANDLE(0)
    T_NUMBER(1)

    const char *type="png";    
    if (PARAMETERS_COUNT>2) {
        T_STRING(2);
        type=PARAM(2);
    }

    if (PARAMETERS_COUNT>4)
        SET_STRING(4, "");

    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);

    PopplerPage *page=poppler_document_get_page(pdoc, PARAM_INT(1));
    if (page) {
        double src_width=0;
        double src_height=0;
        double thumb_zoom=1;
        poppler_page_get_size(page,&src_width,&src_height);

        if (PARAMETERS_COUNT>3) {
            T_NUMBER(3);
            thumb_zoom=PARAM(3);
        }

        if (src_width==0)
            src_width=1;
        if (src_height==0)
            src_height=1;

        GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*thumb_zoom, (int)src_height*thumb_zoom);
        poppler_page_render_to_pixbuf(page,0,0,(int)src_width*thumb_zoom,(int)src_height*thumb_zoom,thumb_zoom,0,pixbuf); 

        GError *docerr=0;
        gchar *buffer=0;
        gsize buffer_size=0;

        if (!gdk_pixbuf_save_to_buffer(pixbuf, &buffer, &buffer_size, type, &docerr, NULL)) {
            if ((PARAMETERS_COUNT>4) && (docerr)) {
                SET_STRING(4, docerr->message);
            }
            RETURN_STRING("");
        } else {
            if (buffer_size) {
                RETURN_BUFFER((char *)buffer, (INTEGER)buffer_size)
            } else {
                RETURN_STRING("");
            }
        }
        if (pixbuf)
            g_object_unref(G_OBJECT(pixbuf));
        g_object_unref (G_OBJECT (page));
    } else {
        RETURN_STRING("");
    }
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(PDFPageImage, 3, 6)
    T_HANDLE(0)
    T_NUMBER(1)
    T_STRING(2)

    const char *type="png";    
    if (PARAMETERS_COUNT>3) {
        T_STRING(3);
        type=PARAM(3);
    }

    if (PARAMETERS_COUNT>5)
        SET_STRING(5, "");

    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);

    PopplerPage *page=poppler_document_get_page(pdoc, PARAM_INT(1));
    if (page) {
        double src_width=0;
        double src_height=0;
        double thumb_zoom=1;
        poppler_page_get_size(page,&src_width,&src_height);

        if (PARAMETERS_COUNT>4) {
            T_NUMBER(4);
            thumb_zoom=PARAM(4);
        }

        if (src_width==0)
            src_width=1;
        if (src_height==0)
            src_height=1;

        GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*thumb_zoom, (int)src_height*thumb_zoom);
        poppler_page_render_to_pixbuf(page,0,0,(int)src_width*thumb_zoom,(int)src_height*thumb_zoom,thumb_zoom,0,pixbuf); 

        GError *docerr=0;

        if (!gdk_pixbuf_save(pixbuf, PARAM(2), type, &docerr, NULL)) {
            if ((PARAMETERS_COUNT>5) && (docerr)) {
                SET_STRING(5, docerr->message);
            }
            RETURN_NUMBER(0)
        } else {
            RETURN_NUMBER(1)
        }
        g_object_unref (G_OBJECT (page));
        if (pixbuf)
            g_object_unref(G_OBJECT(pixbuf));
    } else {
        RETURN_NUMBER(0)
    }
END_IMPL
//------------------------------------------------------------------------
gboolean save_data(const gchar *buf, gsize count, gpointer data, GError **error) {
    AnsiString *S=(AnsiString *)data;
    if ((buf) && (count>0))
        S->AddBuffer((char *)buf, count);
    return TRUE;                  
}
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(PDFAttachments, 1)
    T_HANDLE(0)

    GList *list, *l;
    PopplerDocument *pdoc=(PopplerDocument *)PARAM_INT(0);
    CREATE_ARRAY(RESULT);
    if (!poppler_document_has_attachments(pdoc))
        return 0;
    list = poppler_document_get_attachments(pdoc);
    INTEGER index=0;
    for (l = list; l && l->data; l = g_list_next (l)) {
        PopplerAttachment *attachment = POPPLER_ATTACHMENT (l->data);
        if (attachment) {
            void *ARR=0;
            Invoke(INVOKE_ARRAY_VARIABLE,RESULT,(INTEGER)index++,&ARR);
            if (ARR) {
                CREATE_ARRAY(ARR);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"name",(INTEGER)VARIABLE_STRING,(char *)attachment->name ? attachment->name : "Unknown",(NUMBER)0);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"description",(INTEGER)VARIABLE_STRING,(char *)attachment->description ? attachment->description : "",(NUMBER)0);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"size",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)attachment->size);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"ctime",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)attachment->ctime);
                Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"mtime",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)attachment->mtime);
                AnsiString *S=new AnsiString();
                poppler_attachment_save_to_callback(attachment, save_data, S, NULL);
                if (S->Length())
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"content",(INTEGER)VARIABLE_STRING,(char *)S->c_str(),(NUMBER)S->Length());
                else
                    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,ARR,(INTEGER)"content",(INTEGER)VARIABLE_STRING,(char *)"",(NUMBER)attachment->mtime);
                delete S;
            }
        }
    }
    if (list)
        g_list_free(list);
END_IMPL
//------------------------------------------------------------------------
