//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include <string.h>
#include "AnsiString.h"

#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>
#include <libxml/HTMLtree.h>
#include <libxml/xmlIO.h>
#include <libxml/xinclude.h>
#include <libxml/catalog.h>
#include <libxml/xmlreader.h>
#include <libxml/HTMLparser.h>
#include <libxml/xpath.h>
#include <libxml/uri.h>

#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>

#ifdef _WIN32
xmlFreeFunc xmlFree=0;
#endif

INVOKE_CALL InvokePtr=0;

char **GetCharList(void *arr) {
    INTEGER type=0;
    char *str=0;
    NUMBER nr=0;
    void *newpData=0;
    char **ret=0;
    
    int count=InvokePtr(INVOKE_GET_ARRAY_COUNT,arr);
    ret=new char * [count+1];
    ret[count]=0;

    for (int i=0;i<count;i++) {
        InvokePtr(INVOKE_ARRAY_VARIABLE,arr,i,&newpData);
        if (newpData) {
            char *szData;
            INTEGER type;
            NUMBER nData;

            InvokePtr(INVOKE_GET_VARIABLE,newpData,&type,&szData,&nData);
            if (type==VARIABLE_STRING) {
                ret[i]=szData;
            } else
                ret[i]=0;
        }
    }
    return ret;
}
//-----------------------------------------------------------------------------------
static AnsiString xslt2_error;
void MyGenericErrorFunc(void * ctx, const char * msg, ...) {
    va_list args;
    va_start(args, msg);
    char buffer[0xFFF];
    buffer[0]=0;
    vsprintf(buffer, msg, args);
    va_end(args);
    xslt2_error+=buffer;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
    InvokePtr=Invoke;
    xsltSetGenericErrorFunc(0, MyGenericErrorFunc);
	return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API ON_DESTROY_CONTEXT MANAGEMENT_PARAMETERS {
	return 0;
}
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(XSLTError,0)
    INTEGER len=xslt2_error.Length();
    if (len>0) {
        AnsiString res="<pre><![CDATA[\n";
        res+=xslt2_error;
        res+="\n]]></pre>";
        RETURN_BUFFER(res.c_str(), res.Length())
    } else {
        RETURN_STRING("");
    }
END_IMPL
//------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL_MINMAX_PARAMS(XSLTProcess,2,3)
    T_STRING(0)
    T_STRING(1)

    xsltStylesheetPtr cur = NULL;
    xmlDocPtr doc, res, docstyle;

    int i;
    const char *params[1];
    params[0] = NULL;

    const char **params2=0;
    if (PARAMETERS_COUNT>2) {
        T_ARRAY(2);
        params2=(const char **)GetCharList(PARAMETER(2));
    }

    if (xslt2_error.Length())
        xslt2_error.LoadBuffer(0,0);

    void *old_ctx=xmlGenericErrorContext;
    xmlSetGenericErrorFunc(0, MyGenericErrorFunc);

    xmlSubstituteEntitiesDefault(1);
    xmlLoadExtDtdDefaultValue = 1;

    doc = xmlReadMemory(PARAM(0), PARAM_LEN(0), "/xml/in.xml", 0, XML_PARSE_RECOVER);
    docstyle = xmlReadMemory(PARAM(1), PARAM_LEN(1), "/xslt/in.xslt", 0, XML_PARSE_RECOVER);

    if ((doc) && (docstyle)) {
        cur = xsltParseStylesheetDoc(docstyle);

        if (cur) {
            res = xsltApplyStylesheet(cur, doc, params2 ? params2 : params);

            xmlChar *doc_txt_ptr=0;
            int doc_txt_len=0;
            xsltSaveResultToString(&doc_txt_ptr, &doc_txt_len, res, cur);
            if (doc_txt_len>0) {
                RETURN_BUFFER((char *)doc_txt_ptr, doc_txt_len);
            } else {
                RETURN_STRING("")
            }

            xsltFreeStylesheet(cur);
            xmlFreeDoc(res);
            xmlFreeDoc(doc);

            xsltCleanupGlobals();
            xmlCleanupParser();
        } else {
	        RETURN_STRING("");
        }
    } else {
	    RETURN_STRING("");
    }
    if (params2)
        delete[] params2;
    xmlSetGenericErrorFunc(old_ctx, xmlGenericError);
END_IMPL
//------------------------------------------------------------------------
