//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include "AnsiString.h"

#ifdef _WIN32
    #include <io.h>
    #include <winsock.h>
#else
    #include <unistd.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <string.h>
    #include <arpa/inet.h>


    #include <sys/select.h>
    #define INVALID_SOCKET	-1

    int eof(int stream) {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 10;
        fd_set s;
        FD_ZERO(&s);
        FD_SET(stream,&s);

        int val=select(stream+1, &s, 0, 0, &timeout);

        if (val) {
           if (FD_ISSET(stream, &s))
               return 0;
        }
	return 1;
    }
#endif

//=====================================================================================//
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
#ifdef _WIN32
    WSADATA wsaData;
    WSAStartup( MAKEWORD( 2, 2 ), &wsaData);
#endif
	Invoke(INVOKE_DEFINE_CONSTANT, HANDLER, "PROTOCOL_TCP","0");
	Invoke(INVOKE_DEFINE_CONSTANT, HANDLER, "PROTOCOL_UDP","1");
	return 0;
}
//=====================================================================================//
CONCEPT_DLL_API ON_DESTROY_CONTEXT MANAGEMENT_PARAMETERS {
#ifdef _WIN32
    WSACleanup();
#endif
	return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketCreate CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(0,1,"SocketCreate: SocketCreate(nTYPE=PROTOCOL_TCP)");

    LOCAL_INIT;
    
    NUMBER type=0;
    int sock=INVALID_SOCKET;

    if (PARAMETERS_COUNT>0) {
        GET_CHECK_NUMBER(0,type,"SocketCreate: parameter 1 should be a number");
    }

    if ((int)type==0) {
        sock=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    } else
    if ((int)type==1) {
        sock=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    }


    RETURN_NUMBER(sock);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketSetOption CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(4,"SocketSetOption: SocketSetOptions(nSocket, nLevel, nOption, nValue)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nLevel=0;
    NUMBER nOption=0;
    NUMBER nValue=0;

    GET_CHECK_NUMBER(0,sock,"SocketSetOption: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nLevel,"SocketSetOption: parameter 2 should be a number");
    GET_CHECK_NUMBER(2,nOption,"SocketSetOption: parameter 3 should be a number");
    GET_CHECK_NUMBER(3,nValue,"SocketSetOption: parameter 4 should be a number");

#ifdef _WIN32
    RETURN_NUMBER(setsockopt((int)sock,(int)nLevel,(int)nOption,(char *)(int)nValue, sizeof(int)));
#else
    RETURN_NUMBER(setsockopt((int)sock,(int)nLevel,(int)nOption,(void *)(int)nValue, sizeof(int)));
#endif

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketGetOption CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(4,"SocketGetOption: SocketSetOptions(nSocket, nLevel, nOption, var nValue)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nLevel=0;
    NUMBER nOption=0;
    int nValue=0;

    GET_CHECK_NUMBER(0,sock,"SocketGetOption: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nLevel,"SocketGetOption: parameter 2 should be a number");
    GET_CHECK_NUMBER(2,nOption,"SocketGetOption: parameter 3 should be a number");
    //GET_CHECK_NUMBER(3,nValue,"SocketSetOption: parameter 4 should be a number");



#ifdef _WIN32
    int len = sizeof(nValue);
    RETURN_NUMBER(getsockopt((int)sock,(int)nLevel,(int)nOption,(char *)(int)nValue, &len));
#else
    socklen_t len = sizeof(nValue);
    RETURN_NUMBER(getsockopt((int)sock,(int)nLevel,(int)nOption,(void *)(int)nValue, &len));
#endif

    SET_NUMBER(3,(int)len);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketError CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketError: SocketError(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    GET_CHECK_NUMBER(0,sock,"SocketSetOption: parameter 1 should be a number");

    int err=0;

#ifdef _WIN32
    int errlen = sizeof(err);
    int res=getsockopt((int)sock, SOL_SOCKET, SO_ERROR, (char *)&err, &errlen);
#else
    socklen_t errlen = sizeof(err);
    int res=getsockopt((int)sock, SOL_SOCKET, SO_ERROR, &err, &errlen);
#endif
    if (res<0)
        err=-2;

    RETURN_NUMBER(err);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketConnect CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(3,"SocketConnect: SocketConnect(nSocket, szHostname, nPort)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    char *hostname=0;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketConnect: parameter 1 should be a number");
    GET_CHECK_STRING(1,hostname,"SocketConnect: parameter 2 should be a string");
    GET_CHECK_NUMBER(2,nPort,"SocketConnect: parameter 3 should be a number");


    struct sockaddr_in      sin;
    struct hostent          *hp;

    if ((hp=gethostbyname(hostname))==0) {
        RETURN_NUMBER(-2);
        return 0;
    }

    memset(&sin,0,sizeof(sin));
                                  //INADDR_ANY;  
    sin.sin_addr.s_addr         = ((struct in_addr *)(hp->h_addr))->s_addr;
    sin.sin_family              = AF_INET;
    sin.sin_port                = htons((int)nPort);


    RETURN_NUMBER(connect((int)sock,(struct sockaddr *)  &sin, sizeof(sin)));

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketListen CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(2,3,"SocketListen: SocketListen(nSocket, nPort, nMaxConnections=0x100)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nMaxCon=0x100;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketListen: parameter 1 should be a number");
    GET_CHECK_NUMBER(1,nPort,"SocketListen: parameter 2 should be a number");

    if (PARAMETERS_COUNT>2) {
        GET_CHECK_NUMBER(1,nMaxCon,"SocketListen: parameter 3 should be a number");
    }



    struct sockaddr_in sin;
    memset(&sin, 0, sizeof(sin)); 
    sin.sin_addr.s_addr         = INADDR_ANY;  
    sin.sin_family              = AF_INET;
    sin.sin_port                = htons((int)nPort);

    if (bind((int)sock,(struct sockaddr *)  &sin, sizeof(sin)) == INVALID_SOCKET) {
        RETURN_NUMBER((NUMBER)INVALID_SOCKET);
    }


    RETURN_NUMBER(listen((int)sock,(int)nMaxCon));

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketAccept CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketAccept: SocketAccept(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nPort=0;

    GET_CHECK_NUMBER(0,sock,"SocketAccept: parameter 1 should be a number");

    struct sockaddr_in sin;

    memset(&sin, 0, sizeof(sin)); 
    sin.sin_addr.s_addr         = INADDR_ANY;  
    sin.sin_family              = AF_INET;
    //sin.sin_port                = htons((int)nPort);

#ifdef _WIN32
	int addr_length =   sizeof(sin);
#else
	socklen_t addr_length =   sizeof(sin);
#endif

    RETURN_NUMBER(accept((int)sock,(sockaddr *)  &sin, &addr_length));

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketClose CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketClose: SocketClose(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    GET_CHECK_NUMBER(0,sock,"SocketClose: parameter 1 should be a number");

#ifdef _WIN32
    RETURN_NUMBER(closesocket((int)sock));
#else
    RETURN_NUMBER(close((int)sock));
#endif


    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketRead CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(3,5,"SocketRead: SocketRead(nSocket, var anyBuffer, nSize[, bSocketIsUDP=false[, bReceiveExactSize=false]])");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    NUMBER nSize=0;
    NUMBER isUdp=0;
    NUMBER exactSize=0;

    GET_CHECK_NUMBER(0,sock,"SocketRead: parameter 1 should be a number");
    GET_CHECK_NUMBER(2,nSize,"SocketRead: parameter 3 should be a number");

    if (PARAMETERS_COUNT>3) {
        GET_CHECK_NUMBER(3,isUdp,"SocketRead: parameter 4 should be a number");

        if (PARAMETERS_COUNT>4) {
            GET_CHECK_NUMBER(4,exactSize,"SocketRead: parameter 5 should be a number");
        }
    }

    if ((int)nSize<=0) {
        RETURN_NUMBER(0);
        return 0;
    }

    char *buf=new char[(int)nSize+1];
    buf[0]=0;
    buf[(int)nSize]=0;

    int rec_size=0;

    if (isUdp) {
    #ifdef _WIN32
        struct sockaddr host_address;
	    int addr_length =   sizeof(host_address);
        rec_size=recvfrom((int)sock, buf, (int)nSize,0,&host_address,&addr_length);
    #else
        struct sockaddr host_address;
	    socklen_t addr_length =   sizeof(host_address);
        rec_size=recvfrom((int)sock, buf, (int)nSize,0,&host_address,&addr_length);
    #endif
    } else {
        int len=0;
        do {
            len=recv((int)sock, buf+rec_size, (int)nSize-rec_size,0);

            if (len>0)
                rec_size+=len;

        } while (((int)exactSize) && (len>0) && (rec_size<(int)nSize));
    }

    SET_BUFFER(1, buf, rec_size);

    delete buf;

    RETURN_NUMBER(rec_size);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketWrite CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK_MIN_MAX(2,5,"SocketWrite: SocketWrite(nSocket, anyBuffer[, bSocketIsUDP=false, szUDPHost, nUDPPort])");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;
    char *buffer;
    NUMBER isUdp=0;
    NUMBER nSize=0;
    NUMBER nPort=0;
    char *hostname=0;

    GET_CHECK_NUMBER(0,sock,"SocketWrite: parameter 1 should be a number");
    GET_CHECK_BUFFER(1,buffer,nSize,"SocketWrite: parameter 2 should be a string buffer");

    struct hostent          *hp;
    if (PARAMETERS_COUNT==5) {
        GET_CHECK_NUMBER(2,isUdp,"SocketWrite: parameter 3 should be a number");

        GET_CHECK_STRING(3,hostname,"SocketWirte: parameter 4 should be a string");
        GET_CHECK_NUMBER(4,nPort,"SocketWrite: parameter 5 should be a number");

        if ((hp=gethostbyname(hostname))==0) {
            RETURN_NUMBER(-2);
            return 0;
        }
    }

    int sent_size=0;
    if (isUdp) {
    #ifdef _WIN32
        struct sockaddr_in host_address;
	    int addr_length =   sizeof(host_address);
    #else
        struct sockaddr_in host_address;
	    socklen_t addr_length =   sizeof(host_address);
    #endif
        host_address.sin_family=AF_INET;
        host_address.sin_port=htons((int)nPort);
        host_address.sin_addr.s_addr         = ((struct in_addr *)(hp->h_addr))->s_addr;


        sent_size=sendto((int)sock, buffer, (int)nSize,0,(struct sockaddr *)&host_address,addr_length);
    } else {
        sent_size=send((int)sock, buffer, (int)nSize,0);
    }

    RETURN_NUMBER(sent_size);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketInfo CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketInfo: SocketInfo(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    GET_CHECK_NUMBER(0,sock,"SocketInfo: parameter 1 should be a number");


    struct sockaddr_in peer;
#ifdef _WIN32
    int peerlen = sizeof(struct sockaddr);
#else
    socklen_t peerlen = sizeof(struct sockaddr);
#endif
    getpeername((int)sock,(sockaddr *)&peer,&peerlen);

    Invoke(INVOKE_CREATE_ARRAY, RESULT, 0);

    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"address",(INTEGER)VARIABLE_STRING,(char *)inet_ntoa(peer.sin_addr),(NUMBER)0);
    Invoke(INVOKE_SET_ARRAY_ELEMENT_BY_KEY,RESULT,"port",(INTEGER)VARIABLE_NUMBER,(char *)"",(NUMBER)peer.sin_port);

    return 0;
}
//=====================================================================================//
CONCEPT_DLL_API CONCEPT_SocketHasData CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(1,"SocketHasData: SocketHasData(nSocket)");

    LOCAL_INIT;
    
    NUMBER sock=INVALID_SOCKET;

    int has_data=0;
    GET_CHECK_NUMBER(0,sock,"SocketHasData: parameter 1 should be a number");

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 10000;
    fd_set socks;

    FD_ZERO(&socks);
    FD_SET((int)sock,&socks);

    int sel_val=select(FD_SETSIZE, &socks, 0, 0, &timeout);

    has_data=(sel_val!=0);

    RETURN_NUMBER(has_data);
    return 0;
}
//=====================================================================================//
