//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include "library.h"
#include "AnsiString.h"
#include "AnsiList.h"
//#include "xmlParser.h"
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include "semhh.h"

#define CACHE_SIZE      0xFFFF
#define CHUNK_SIZE      0x80000

#ifdef _WIN32
    #include <io.h>
    #include <winsock.h>
#else
    #include <unistd.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netinet/in.h>

    #include <sys/select.h>
    #include <time.h> // sleep

    int eof(int stream) {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 10;
        fd_set s;
        FD_ZERO(&s);
        FD_SET(stream,&s);

        int val=select(stream+1, &s, 0, 0, &timeout);

        if (val) {
           if (FD_ISSET(stream, &s))
               return 0;
        }
		return 1;
    }
#endif

int sock_eof(int stream) {
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 10;
    fd_set s;
    FD_ZERO(&s);
    FD_SET(stream,&s);

    int val=select(stream+1, &s, 0, 0, &timeout);

    if (val<0)
        return 1;
    if (val) {
       if (FD_ISSET(stream, &s))
           return 0;
    }
	return 1;
}

#include "INTERFACE_TO_AES.h"

AnsiList  BufferedMessages;

#ifdef _WIN32
	#define pipe(phandles)	_pipe (phandles, 4096, _O_BINARY)
#endif

char has_debug=0;
int DebugPipes[2];

int pipe_in;
int pipe_out;
int apid;
int parent_apid;

int is_cached=0;
char *cached_buffer=0;
int cached_size=0;
int alloc_cached_size=0;
void *MAIN_HANDLER=0;

static HHSEM sem_send;
static HHSEM sem_recv;

int deturnated_send(int CLIENT_SOCKET,char *buffer, int size, int extra=0) {
    int ssize=0;
    int orig_size=size;
    semp(sem_send);
    do {
        int to_send=size;
        if (size>CHUNK_SIZE)
            to_send=CHUNK_SIZE;
        ssize=send(CLIENT_SOCKET, buffer, to_send, extra);

        if (ssize<=0) {
            semv(sem_send);
            return -1;
        }

        buffer+=ssize;
        size-=ssize;

    } while (size);

    semv(sem_send);
    return orig_size;
}
//-----------------------------------------------------------------------------------
/*int Cache(char *buffer,long size) {
    int old_size=cached_size;
    cached_size+=size+sizeof(long);
    cached_buffer=(char *)realloc(cached_buffer,cached_size);
    *(long *)(cached_buffer+old_size)=htonl(size);
    memcpy(cached_buffer+old_size+sizeof(long),buffer,size);
    return 1;
}*/
//-----------------------------------------------------------------------------------
int Cache(char *buffer,int size) {
    int old_size=cached_size;
    cached_size+=size+sizeof(int);
    if (alloc_cached_size<cached_size) {
        alloc_cached_size+=(cached_size/CACHE_SIZE + 1)*CACHE_SIZE;
        cached_buffer=(char *)realloc(cached_buffer,/*cached_size*/alloc_cached_size);
    }
    *(int *)(cached_buffer+old_size)=htonl(size);
    memcpy(cached_buffer+old_size+sizeof(int),buffer,size);
    return 1;
}
//-----------------------------------------------------------------------------------
int FlushCache(int CLIENT_SOCKET) {
    int res=-1;
    if ((cached_size) && (cached_buffer)) {
        res=deturnated_send(CLIENT_SOCKET,cached_buffer, cached_size, 0);
        free(cached_buffer);
        cached_size=0;
        alloc_cached_size=0;
        cached_buffer=0;
    }
    return res;
}
//-----------------------------------------------------------------------------------
static char *empty_char=(char *)"\0";
int DeSerializeBuffer2(char *buffer, int size, char **Owner, int *owner_len, int *message, char **Target, int *len_target, char **Value, int *len_value) {
    // OWNER poate avea maxim 256 octeti
    char *ptr=buffer;
    unsigned char owner_size=*(unsigned char *)buffer;
    buffer++;
    *owner_len=owner_size;
    if (owner_size) {
        *Owner=buffer;
        buffer+=owner_size;
    } else
        *Owner=empty_char;

    *message=ntohl(*(int *)buffer);
    buffer+=sizeof(int);
    unsigned short target_size=ntohs(*(unsigned short *)buffer);
    buffer+=sizeof(short);
    *len_target=target_size;
    if (target_size) {
        *Target=buffer;
        buffer+=target_size;
    } else
        *Target=empty_char;

    *len_value=size-((unsigned long)buffer-(unsigned long)ptr);
    if (*len_value > 0)
        *Value=buffer;
    else {
        *Value=empty_char;
        *len_value=0;
    }
    return 1;
}
//-----------------------------------------------------------------------------------
int DeSerializeBuffer(char *buffer, int size, AnsiString *Owner, int *message, AnsiString *Target, AnsiString *Value) {
    // OWNER poate avea maxim 256 octeti
    char *ptr=buffer;
    unsigned char owner_size=*(unsigned char *)buffer;
    buffer++;
    if (owner_size) {
        Owner->LoadBuffer(buffer, owner_size);
        buffer+=owner_size;
    } else
        *Owner=(char *)"";

    *message=ntohl(*(int *)buffer);
    buffer+=sizeof(int);
    unsigned short target_size=ntohs(*(unsigned short *)buffer);
    buffer+=sizeof(short);
    if (target_size) {
        Target->LoadBuffer(buffer, target_size);
        buffer+=target_size;
    } else
        *Target=(char *)"";

    long last_size=size-((unsigned long)buffer-(unsigned long)ptr);
    if (last_size>0)
        Value->LoadBuffer(buffer,last_size);
    else
        *Value=(char *)"";
    return 1;
}
//-----------------------------------------------------------------------------------
char *SerializeBuffer(char **buffer, int *size, AnsiString *Owner, int message, AnsiString *Target, AnsiString *Value) {
    // OWNER poate avea maxim 256 octeti
    *size=sizeof(char)+Owner->Length()+sizeof(int)+sizeof(short)+Target->Length()+Value->Length();
    char *ptr=new char[*size+sizeof(int)];
    *(int *)ptr=htonl(*size);
    char *buf=ptr+sizeof(int);//new char[*size];
    *buffer=buf;

    *(unsigned char *)buf=(unsigned char)Owner->Length();
    buf+=sizeof(char);

    memcpy(buf, Owner->c_str(), Owner->Length());
    buf+=Owner->Length();

    *(int *)buf=htonl(message);
    buf+=sizeof(int);

    *(unsigned short *)buf=htons((unsigned short)Target->Length());
    buf+=sizeof(short);

    memcpy(buf, Target->c_str(), Target->Length());
    buf+=Target->Length();

    memcpy(buf, Value->c_str(), Value->Length());
    return ptr;
}
//-----------------------------------------------------------------------------------
unsigned int AES_encrypt(char *in_buffer, unsigned int in_buffer_size, char *out_buffer, unsigned int out_buffer_size, char *key, unsigned int key_size) {
	if (!En_inited)
		encrypt_init(key,key_size);

	return encrypt(in_buffer,in_buffer_size,out_buffer,out_buffer_size);
}
//-----------------------------------------------------------------------------------
unsigned int AES_decrypt(char *in_buffer, unsigned int in_buffer_size, char *out_buffer, unsigned int out_buffer_size, char *key, unsigned int key_size) {
	if (!Dec_inited)
		decrypt_init(key,key_size);

	return decrypt(in_buffer,in_buffer_size,out_buffer,out_buffer_size);
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API ON_CREATE_CONTEXT MANAGEMENT_PARAMETERS {
	Invoke(INVOKE_GET_APPLICATION_INFO, HANDLER, &pipe_in, &pipe_out, &apid, &parent_apid);
	MAIN_HANDLER=HANDLER;
	seminit(sem_send, 1);
    seminit(sem_recv, 1);
	return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetAPID CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"GetAPID doesn't take any parameters");
    LOCAL_INIT;

    RETURN_NUMBER(apid);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetParentAPID CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"GetParentAPID doesn't take any parameters");
    LOCAL_INIT;

    RETURN_NUMBER(parent_apid);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_PeekAPMessage CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"PeekAPMessage doesn't take any parameters");
    LOCAL_INIT;

    if (pipe_in<=0) {
        RETURN_NUMBER(0);
        return 0;
    }

    int res=eof(pipe_in);
    // nothing to read
    if (res<0)
        res=1;
    RETURN_NUMBER(!res);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_GetAPMessage CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(3,"GetAPMessage parameters error: number GetAPMessage(var SenderAPID, var nMSG_ID, var szData)");
    LOCAL_INIT;

    CHECK_STATIC_PARAM(0,"GetAPMessage: parameter 0 should be static (number or string)");
    CHECK_STATIC_PARAM(1,"GetAPMessage: parameter 1 should be static (number or string)");
    CHECK_STATIC_PARAM(2,"GetAPMessage: parameter 2 should be static (number or string)");

    if (pipe_in<=0) {
        RETURN_NUMBER(0);
        return 0;
    }

    int params[3];
    int total=0;

    do {
        total=read(pipe_in,(void *)params, sizeof(int)*3);
        if (!total) {
#ifdef _WIN32
            Sleep(1);
#else
            usleep(1000);
#endif
        }
    } while (!total);

    if (total!=sizeof(int)*3) {
        RETURN_NUMBER(0);
        return 0;
    }

    char *data=0;
    if (params[2]>0) {
        data=new char[params[2]];
        total+=read(pipe_in,(void *)data, params[2]);
    }

    SET_NUMBER(0,params[0]);
    SET_NUMBER(1,params[1]);

    if (data) {
        SET_BUFFER(2,data,params[2]);
        delete[] data;
    } else {
        SET_STRING(2,"");
    }

    RETURN_NUMBER(total==sizeof(int)*3+params[2]);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_SendAPMessage CONCEPT_API_PARAMETERS {

    PARAMETERS_CHECK(3,"SendAPMessage parameters error: number SendAPMessage(nToAppID, nMSG_ID, szData)");
    LOCAL_INIT;

    if ((pipe_out<=0) || (apid<=0)) {
        RETURN_NUMBER(0);
        return 0;
    }

    NUMBER nMSG_ID;
    NUMBER nTo;
    char *szData;
    NUMBER len;

    GET_CHECK_NUMBER(0,nTo,"SendAPMessage: parameter 0 should be a number (STATIC NUMBER)");
    GET_CHECK_NUMBER(1,nMSG_ID,"SendAPMessage: parameter 1 should be a number (STATIC NUMBER)");
    GET_CHECK_BUFFER(2,szData,len,"SendAPMessage : parameter 2 should be a string (STATIC STRING)");


    int params[3];
    params[0]=(int)nTo;
    params[1]=(int)nMSG_ID;
    params[2]=(int)len;


    if ((params[0]<0) && (params[0]!=-1)) {
        RETURN_NUMBER(0);
        return 0;
    }

    int total=write(pipe_out,(void *)params, sizeof(int)*3);
    if (params[2]>0) {
        total+=write(pipe_out,(void *)szData, params[2]);
    }
    

    RETURN_NUMBER(total==sizeof(int)*3+params[2]);
    return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_CreateDebugPipe CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=0)
            return (void *)"CreateDebugPipe it's a reserved function. You should not use this function.";

        if (has_debug) {
            close(DebugPipes[0]);
            close(DebugPipes[1]);
            has_debug=0;
        }

        if (pipe(DebugPipes)==-1) {
            SetVariable(RESULT,VARIABLE_NUMBER,"",-1);
            return 0;
        }
        write(DebugPipes[1],(char *)&DebugPipes[1],sizeof(int));
        has_debug=0;
        SetVariable(RESULT,VARIABLE_NUMBER,"",DebugPipes[0]);
	    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_DestroyDebugPipe CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=0)
            return (void *)"DestroyDebugPipe it's a reserved function. You should not use this function.";
        if (has_debug) {
            close(DebugPipes[0]);
            close(DebugPipes[1]);
            has_debug=0;
            SetVariable(RESULT,VARIABLE_NUMBER,"",1);
        } else
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
	    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_BeginUpdateSession CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"BeginUpdateSession takes no parameters");
    is_cached++;
    RETURN_NUMBER(0);
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_EndUpdateSession CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"EndUpdateSession takes no parameters");
    if (is_cached)
        is_cached--;
    if (!is_cached) {
        RETURN_NUMBER(FlushCache(CLIENT_SOCKET));
    } else {
        RETURN_NUMBER(0);
    }
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_EndUpdateSession2 CONCEPT_API_PARAMETERS {
    PARAMETERS_CHECK(0,"EndUpdateSession2 takes no parameters");
    if (is_cached)
        is_cached=0;
    if (!is_cached) {
        RETURN_NUMBER(FlushCache(CLIENT_SOCKET));
    } else {
        RETURN_NUMBER(0);
    }
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_send_message CONCEPT_API_PARAMETERS {
        // send_message( TargetName, MESSAGE_ID, MESSAGE_RESERVED, MESSAGE_DATA)
        if (PARAMETERS->COUNT!=4)
            return (void *)"send_message: function takes 4 parameters(usage: send_message [static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA)]";
        
        char    *SENDER_NAME;
        char    *MESSAGE_DATA;
        char    *MESSAGE_TARGET;
        NUMBER  MESSAGE_ID;
        NUMBER  nDUMMY_FILL;
        char    *szDUMMY_FILL;
        INTEGER TYPE;

        AnsiString Owner;
        AnsiString Target;
        AnsiString message_data;

        // get SENDER_NAME
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],&TYPE,&SENDER_NAME,&nDUMMY_FILL);
        if (TYPE!=VARIABLE_STRING)
            return (void *)"send_message: 1st parameter should be of STATIC STRING type";
        else
            Owner.LoadBuffer(SENDER_NAME, (INTEGER)nDUMMY_FILL);

        // message ID
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],&TYPE,&szDUMMY_FILL,&MESSAGE_ID);
        if (TYPE!=VARIABLE_NUMBER)
            return (void *)"send_message: 2nd parameter should be of STATIC NUMBER type";

        // get target
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],&TYPE,&MESSAGE_TARGET,&nDUMMY_FILL);
        if (TYPE!=VARIABLE_STRING)
            return (void *)"send_message: 3rd parameter should be of STATIC STRING type";
        else
            Target.LoadBuffer(MESSAGE_TARGET, (INTEGER)nDUMMY_FILL);

        // get messsage
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],&TYPE,&MESSAGE_DATA,&nDUMMY_FILL);
        if (TYPE==VARIABLE_STRING) {
            //message_data=MESSAGE_DATA;
            message_data.LoadBuffer(MESSAGE_DATA, (INTEGER)nDUMMY_FILL);
        } else
        if (TYPE==VARIABLE_NUMBER)
            message_data=AnsiString(nDUMMY_FILL);
        else
            return (void *)"send_message: 4th parameter should be of STATIC STRING or STATIC_NUMBER type";

        INTEGER res=0;
        char *buffer=0;
        int in_content_size=0;
        int message=(int)MESSAGE_ID;

        SerializeBuffer(&buffer, &in_content_size, &Owner, message, &Target, &message_data);

        if (REMOTE_PUBLIC_KEY)  {
        	// criptez query-ul ...
        	int out_content_size=(in_content_size/16+1)*16+5;//(((in_content_size / 48) + 1) * 64) + 256;
            char *ptr=new char[out_content_size+sizeof(int)];

        	char *out_content=ptr+sizeof(int);

        	int encrypt_result = AES_encrypt(   buffer, in_content_size,
					                            out_content, out_content_size,
                                                REMOTE_PUBLIC_KEY, 16 );
        	if (!encrypt_result) {
            		delete[] ptr;//out_content;
                    delete[] (buffer-sizeof(int));
            		return (void *)"send_message: <<AES_encrypt>> error encrypting message.";
        	}

            int size=encrypt_result;
            *(int *)ptr=htonl(size);

            if (is_cached) {
                Cache(out_content,size);
            } else {
                //deturnated_send(CLIENT_SOCKET,(char *)&size, sizeof(long), 0);
                //res=deturnated_send(CLIENT_SOCKET,out_content, size, 0);
                res=deturnated_send(CLIENT_SOCKET,ptr, size+sizeof(int), 0);
            }
            delete[] ptr;//out_content;
        } else {

        	// trimit dimensiunea mai intai ...
        	//long size=SERVER_QUERY.Length()+1;
            if (is_cached) {
                Cache(buffer,in_content_size);
            } else {
        	    /*deturnated_send(CLIENT_SOCKET,(char *)&in_content_size, sizeof(long), 0);
        	    res=deturnated_send(CLIENT_SOCKET,buffer, in_content_size, 0);*/

                res=deturnated_send(CLIENT_SOCKET,buffer-sizeof(int), in_content_size+sizeof(int), 0);
            } 
	}
        delete[] (buffer-sizeof(int));

        SetVariable(RESULT,VARIABLE_NUMBER,"",0);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_wait_message_ID CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=4) {
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return (void *)"wait_message_ID: function takes 4 parameters(usage: wait_message_ID [static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA)]";
        }
        
        NUMBER  nDUMMY_FILL;
        char    *szDUMMY_FILL;
        INTEGER TYPE;
        char *buffer=0;//[BUFFER_SIZE];
        INTEGER rec_count=0;
        INTEGER MSG_ID=0;
        INTEGER OWNER=0;
        AnsiString temp;

        if (is_cached)
            FlushCache(CLIENT_SOCKET);

        for (int i=0;i<4;i++) {
            GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[i]-1],&TYPE,&szDUMMY_FILL,&nDUMMY_FILL);
            if ((TYPE!=VARIABLE_STRING) && (TYPE!=VARIABLE_NUMBER)) {
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"wait_message_ID: all parameters are passed by reference to this function. All parameters should be of STATIC type.";
            }
            if (i==0) {
                if ((TYPE==VARIABLE_NUMBER) && (nDUMMY_FILL>0))
                    OWNER=(INTEGER)nDUMMY_FILL;
            } else
            if (i==1) {
                if (TYPE==VARIABLE_NUMBER)
                    MSG_ID=(INTEGER)nDUMMY_FILL;
                else {
                    SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                    return (void *)"wait_message_ID: 2nd parameter shoult be an integer (STATIC NUMBER)";
                }
            }
        }


        SetVariable(RESULT,VARIABLE_NUMBER,"",0);

        int len=BufferedMessages.Count();
        if (len) {
            for (int i=0;i<len;i++) {
                TParameters *PARAM=(TParameters *)BufferedMessages.Item(i);
                if ((PARAM) && (PARAM->ID==MSG_ID) && ((OWNER<=0) || (OWNER==PARAM->Sender.ToInt()))) {
                    BufferedMessages.Remove(i);
                    if (OWNER<=0) {
                        SetVariable(
                                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                                    VARIABLE_STRING,
                                    PARAM->Sender.c_str(),
                                    PARAM->Sender.Length());//0
                    }
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                                VARIABLE_NUMBER,
                                "",
                                PARAM->ID);

                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                                VARIABLE_STRING,
                                PARAM->Target.c_str(),
                                PARAM->Target.Length());//0

                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                                VARIABLE_STRING,
                                PARAM->Value.c_str(),
                                PARAM->Value.Length()); //0
                    delete PARAM;

                    // setez cu true (non-0) rezultatul
                    SetVariable(RESULT,VARIABLE_NUMBER,"",i+1);
                    return 0;
                }
            }
        }
        // iau marimea ...
        int size=0;
        int received=0;
        do {
            size=0;
            received=0;
            rec_count=0;
            do {
                received=recv(CLIENT_SOCKET,(char *)&size+rec_count,sizeof(int)-rec_count,0);
                rec_count+=received;
            } while ((rec_count<sizeof(int)) && (received>0));
            size=ntohl(size);

            if (received<=0) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"wait_message_ID: socket disconnected or host down";
                //return 0;
            }

            //if (size > BUFFER_SIZE)
            //   return (void *)"wait_message_ID: received data is too big";
            if (size) {
                buffer=new char[size+1];
                buffer[size]=0;
            }

            // iau xml-ul ...
            rec_count=0;
            do {
                int to_get=size-rec_count;
                if (to_get>CHUNK_SIZE)
                    to_get=CHUNK_SIZE;
                received=recv(CLIENT_SOCKET,buffer+rec_count,to_get,0);
                rec_count+=received;
            } while ((received>0) && (rec_count<size));

            if (received<=0) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                if (buffer)
                    delete[] buffer;
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"wait_message_ID: socket disconnected or host down";
                //return 0;
            }

            char *out_content=0;

            if (LOCAL_PRIVATE_KEY)  {
                out_content=new char[size*2];
                buffer[size]=0;
                int decrypt_result = AES_decrypt(   buffer, size,
                                                    out_content, size,
                                                    LOCAL_PRIVATE_KEY, /*strlen(LOCAL_PRIVATE_KEY)*/16 );
                out_content[decrypt_result]=0;
                if (!decrypt_result) {
                    delete[] out_content;
                    out_content=0;
                } else {
                    size=decrypt_result;
                    if (buffer) {
                        delete[] buffer;
                        buffer=0;
                    }
                }
            }

            int message_id;
            char *Owner;
            int OwnerLen;
            char *Target;
            int TargetLen;
            char *Value;
            int ValueLen;

            int is_looked=1;

            DeSerializeBuffer2(out_content ? out_content : buffer , size, &Owner, &OwnerLen, &message_id, &Target, &TargetLen, &Value, &ValueLen);
            if ((message_id==MSG_ID) && (OWNER>0)) {
                if (OwnerLen) {
                    temp.LoadBuffer(Owner, OwnerLen);
                    if (OWNER!=temp.ToInt())
                        is_looked=0;
                } else
                    is_looked=0;
            }
            if ((message_id==MSG_ID) && (is_looked)) {
                if (OWNER<=0) {
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                                VARIABLE_STRING,
                                Owner,
                                0);
                }
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                            VARIABLE_NUMBER,
                            "",
                            message_id);

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                            VARIABLE_STRING,
                            Target,
                            TargetLen);

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                            VARIABLE_STRING,
                            Value,
                            ValueLen);

                if (out_content) {
                    delete[] out_content;
                    out_content=0;
                }

                if (buffer) {
                    delete[] buffer;
                    buffer=0;
                }
                SetVariable(RESULT,VARIABLE_NUMBER,"",size);
                return 0;
            } else {
                // nu este mesajul cautat, trec mai departe ...
                TParameters *MESSAGE=new TParameters;
                MESSAGE->ID=message_id;
                MESSAGE->Sender=Owner;
                if (TargetLen)
                    MESSAGE->Target.LoadBuffer(Target, TargetLen);
                if (ValueLen)
                    MESSAGE->Value.LoadBuffer(Value, ValueLen);
                BufferedMessages.Add(MESSAGE,DATA_MESSAGE);
            }
            if (out_content) {
                delete[] out_content;
                out_content=0;
            }

            if (buffer) {
	            delete[] buffer;
                buffer=0;
            }
        } while (1);
        SetVariable(RESULT,VARIABLE_NUMBER,"",size);
        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_get_message CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=4) {
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return (void *)"get_message: function takes 4 parameters(usage: send_message [static string OWNER_NAME, static number MESSAGE_ID, string TARGET, static any MESSAGE_DATA)]";
        }

        NUMBER  nDUMMY_FILL;
        char    *szDUMMY_FILL;
        INTEGER TYPE;
        char *buffer=0;//[BUFFER_SIZE];
        INTEGER rec_count=0;

        if (is_cached)
            FlushCache(CLIENT_SOCKET);

        // get SENDER_NAME
        for (int i=0;i<4;i++) {
            GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[i]-1],&TYPE,&szDUMMY_FILL,&nDUMMY_FILL);
            if ((TYPE!=VARIABLE_STRING) && (TYPE!=VARIABLE_NUMBER)) {
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return (void *)"get_message: all parameters are passed by reference to this function. All parameters should be of STATIC type.";
            }
        }


        // verificare Buffered messages (daca am chemat wait_message_ID, este posibil sa am ceva in BuffereMessages
        if (BufferedMessages.Count()) {
            // setez cu true (non-0) rezultatul
            SetVariable(RESULT,VARIABLE_NUMBER,"",1);
            TParameters *PARAM=(TParameters *)BufferedMessages.Remove(0);
            // pregatesc setarile ...
            SetVariable(
                        LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                        VARIABLE_STRING,
                        PARAM->Sender.c_str(),
                        PARAM->Sender.Length());//0
            SetVariable(
                        LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                        VARIABLE_NUMBER,
                        "",
                        PARAM->ID);

            SetVariable(
                        LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                        VARIABLE_STRING,
                        PARAM->Target.c_str(),
                        PARAM->Target.Length());//0

            SetVariable(
                        LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                        VARIABLE_STRING,
                        PARAM->Value.c_str(),
                        PARAM->Value.Length()); //0
            delete PARAM;
            return 0;
        }

        SetVariable(RESULT,VARIABLE_NUMBER,"",0);
        // iau marimea ...
        int size=0;
        int received=0;

        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 100000;
        fd_set socks;

        int sel_val;
        do {
            FD_ZERO(&socks);
            FD_SET(CLIENT_SOCKET,&socks);

            if (pipe_in<0)
                break;
            int res_eof=eof(pipe_in);
            // nothing to read
            if (res_eof<0)
                res_eof=1;

            if (!res_eof) {
                // to do ... preluare mesaj inter-application
                int params[3];
                int total=read(pipe_in,(void *)params, sizeof(int)*3);

#ifndef _WIN32
                if (total<=0) {
                   sel_val=select(FD_SETSIZE, &socks, 0, 0, &timeout);
                   if (sel_val!=0)
                      break;
                }
#endif

                if (total!=sizeof(int)*3) {
                    RETURN_NUMBER(0);
                    return 0;
                }

                char *data=0;
                if (params[2]>0) {
                    data=new char[params[2]+1];
                    total=read(pipe_in,(void *)data, params[2]);
                    data[params[2]]=0;
                }

                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                            VARIABLE_NUMBER,
                            "",
                            params[0]);
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                            VARIABLE_NUMBER,
                            "",
                            -0x100);
                SetVariable(
                            LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                            VARIABLE_NUMBER,
                            "",
                            params[1]);

                if (data)
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                                VARIABLE_STRING,
                                data,
                                params[2]);
                else
                    SetVariable(
                                LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                                VARIABLE_STRING,
                                "",
                                0);

                if (data)
                    delete[] data;
                SetVariable(RESULT,VARIABLE_NUMBER,"",total);
                return 0;
            }
            sel_val=select(FD_SETSIZE, &socks, 0, 0, &timeout);
            /*
                unsigned long pending=0;
                int err=ioctlsocket(CLIENT_SOCKET, FIONREAD, &pending);
                if ((err!=0) || (pending<0)) {
                    SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                    return 0;
                }
                //return (void *)"get_message: socket disconnected or host down";
            */
        } while ((sel_val==0)/* || (sel_val==-1)*/);

        do {
            received=recv(CLIENT_SOCKET,(char *)&size+rec_count,sizeof(int)-rec_count,0);
            if (received<0) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
            rec_count+=received;
        } while ((rec_count<sizeof(int)) && (received>/*=*/0)); // received > 0
        size=ntohl(size);

        if ((received<=0) || (size<=0)) {
            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            return 0;
            //return (void *)"get_message: socket disconnected or host down";
        }

        if (size) {
            buffer=(char *)malloc(size+1);
            if (!buffer) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
            buffer[size]=0;
        }

        rec_count=0;
        do {
            int to_get=size-rec_count;
            if (to_get>CHUNK_SIZE)
                to_get=CHUNK_SIZE;
            received=recv(CLIENT_SOCKET,buffer+rec_count,to_get,0);
            if (received<0) {
                SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
                SetVariable(RESULT,VARIABLE_NUMBER,"",0);
                return 0;
            }
            rec_count+=received;
        } while ((received>/*=*/0) && (rec_count<size));

        if (received<=0) {
            SetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],VARIABLE_NUMBER,"",0x500);
            SetVariable(RESULT,VARIABLE_NUMBER,"",0);
            if (buffer)
                free(buffer);
            return 0;//(void *)"get_message: socket disconnected or host down";
        }

        char *out_content=0;

        if (LOCAL_PRIVATE_KEY)  {
            out_content=(char *)malloc(size*2);
            buffer[size]=0;
            int decrypt_result = AES_decrypt(   buffer, size,
                                                out_content, size,
                                                LOCAL_PRIVATE_KEY, /*strlen(LOCAL_PRIVATE_KEY)*/16 );
            out_content[decrypt_result]=0;
            if (!decrypt_result) {
                free(out_content);
                //return (void *)"get_message: <<RSA_decrypt>> error decrypting message.";
                out_content=0;
                //out_content=buffer;
            } else
                size=decrypt_result;
        }/* else
            out_content=buffer;*/

        AnsiString Owner;
        AnsiString Target;
        AnsiString Value;
        int message_id;

        DeSerializeBuffer(out_content ? out_content : buffer , size, &Owner, &message_id, &Target, &Value);

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],
                    VARIABLE_STRING,
                    Owner.c_str(),
                    Owner.Length());//0
        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[1]-1],
                    VARIABLE_NUMBER,
                    "",
                    message_id);

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[2]-1],
                    VARIABLE_STRING,
                    Target.c_str(),
                    Target.Length());//0

        SetVariable(
                    LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[3]-1],
                    VARIABLE_STRING,
                    Value.c_str(),
                    Value.Length());

        if (out_content)
            free(out_content);

        if (buffer)
	        free(buffer);

        SetVariable(RESULT,VARIABLE_NUMBER,"",size);
        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_confirm_message CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=1)
            return (void *)"confirm_message: function takes 1 parameter. [usage: confirm_message(VALUE)]";

        char    *STR_DATA;
        NUMBER  NUMBER_DATA;
        INTEGER TYPE;

        // get SENDER_NAME
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],&TYPE,&STR_DATA,&NUMBER_DATA);
        if (TYPE!=VARIABLE_NUMBER)
            return (void *)"confirm_message: 1st parameter should be of STATIC NUMBER type";

        deturnated_send(CLIENT_SOCKET,(char *)&NUMBER_DATA, sizeof(NUMBER), 0);

        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_raw_deturnated_send CONCEPT_API_PARAMETERS {
        // send_message( TargetName, MESSAGE_ID, MESSAGE_RESERVED, MESSAGE_DATA)
        if (PARAMETERS->COUNT!=1)
            return (void *)"raw_deturnated_send: function takes 1 parameter. [usage: raw_deturnated_send(DATA)]";
        
        char    *DATA;
        NUMBER  nDUMMY_FILL;
        INTEGER TYPE;

        // get SENDER_NAME
        GetVariable(LOCAL_CONTEXT[PARAMETERS->PARAM_INDEX[0]-1],&TYPE,&DATA,&nDUMMY_FILL);
        if (TYPE!=VARIABLE_STRING)
            return (void *)"raw_deturnated_send: 1st parameter should be of STATIC STRING type";

        AnsiString SERVER_QUERY(DATA);

        int size=htonl(SERVER_QUERY.Length()+1);
        deturnated_send(CLIENT_SOCKET,(char *)&size, sizeof(int), 0);
        INTEGER res=deturnated_send(CLIENT_SOCKET,SERVER_QUERY, size, 0);
        SetVariable(RESULT,VARIABLE_NUMBER,"",res);
        return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_raw_receive CONCEPT_API_PARAMETERS {
        // send_message( TargetName, MESSAGE_ID, MESSAGE_RESERVED, MESSAGE_DATA)
        if (PARAMETERS->COUNT)
            return (void *)"raw_receive: doesn't take any parameters. [usage: raw_receive()]";
        
        //char    *DATA;
        //NUMBER  nDUMMY_FILL;
        //INTEGER TYPE;
        char buffer[BUFFER_SIZE];
        int rec_count=0;

        // iau marimea ...
        int size=0;
        do {
            rec_count+=recv(CLIENT_SOCKET,(char *)&size+rec_count,sizeof(int)-rec_count,0);
        } while ((rec_count<sizeof(int)) && (rec_count>-1));
        size=ntohl(size);

        if (size >= BUFFER_SIZE)
           return (void *)"raw_receive: received data is too big";
           
        // iau xml-ul ...
        rec_count=0;
        int received=0;
        do {
            received=recv(CLIENT_SOCKET,buffer+rec_count,size-rec_count,0);
            rec_count+=received;
        } while ((received>-1) && (rec_count<size));

        buffer[rec_count]=0;

        SetVariable(RESULT,VARIABLE_STRING,buffer,0);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT___DEBUG_PIPEW CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT)
            return (void *)"__DEBUG_PIPEW: doesn't take any parameters.";
        RETURN_NUMBER(pipe_out);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT___DEBUG_PIPER CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT)
            return (void *)"__DEBUG_PIPEW: doesn't take any parameters.";
        RETURN_NUMBER(pipe_in);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_DLL_API CONCEPT_CheckPoint CONCEPT_API_PARAMETERS {
        if (PARAMETERS->COUNT!=1)
            return (void *)"CheckPoint: function takes 1 parameter: the status of checkpoint (0 do not use check points, any other values: time to check point).";

        LOCAL_INIT;
        NUMBER nvariable;

        int status;

        GET_CHECK_NUMBER(0,nvariable,"CheckPoint: status should be a number");

        status=(int)nvariable;

        Invoke(INVOKE_CHECK_POINT, status);

        RETURN_NUMBER(0);
        return 0;
}
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(HaveMessage, 0)
    semp(sem_recv);
    RETURN_NUMBER((BufferedMessages.Count()!=0) || !sock_eof(CLIENT_SOCKET))
    semv(sem_recv);
END_IMPL
//---------------------------------------------------------------------------
CONCEPT_FUNCTION_IMPL(Hibernate, 1)
        T_NUMBER(0)

        char c=(char)PARAM_INT(0);

        Invoke(INVOKE_HIBERNATE, MAIN_HANDLER, &c);

		RETURN_NUMBER(0)
END_IMPL
//---------------------------------------------------------------------------
