<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method="xhtml" indent="yes"/>

<xsl:template match="/">
<xsl:variable name="redirect" select="HTMLContainer/Redirect"/>
<xsl:variable name="op" select="HTMLContainer/Op"/>
<xsl:variable name="message" select="HTMLContainer/Message"/>
<xsl:variable name="template" select="HTMLContainer/Template"/>

<html>
<head>
<title>{title}</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
{extra_head}
</head>
<body>
	<xsl:variable name="sort" select="HTMLContainer/Sort"/>
	<xsl:variable name="desc" select="HTMLContainer/Desc"/>
	<xsl:variable name="criteria" select="HTMLContainer/Criteria"/>
	<xsl:variable name="format" select="HTMLContainer/Format"/>
	<xsl:variable name="username" select="HTMLContainer/Username"/>
	<xsl:variable name="uid" select="HTMLContainer/UID"/>
	<xsl:variable name="scriptid" select="HTMLContainer/ScriptID"/>
	<xsl:variable name="prev_link" select="HTMLContainer/PrevLink"/>
	<xsl:variable name="next_link" select="HTMLContainer/NextLink"/>
	<xsl:variable name="current_page" select="HTMLContainer/Page"/>
	<xsl:variable name="total_pages" select="HTMLContainer/TotalPages"/>
	<xsl:variable name="parent" select="HTMLContainer/IsChildOf"/>

	<xsl:if test="$parent = ''">
		<div class="ui-widget">
			<div class="ui-state-highlight ui-corner-all" style="margin-top: 20px; padding: 0 .7em;"> 
				<p>
					<span class="ui-icon ui-icon-info" style="float: left; margin-right: .3em;"></span>
					<xsl:choose>
						<xsl:when test="HTMLContainer/Username!=''">
							You are logged in as <xsl:value-of disable-output-escaping="yes" select="HTMLContainer/Username"/>
							| <a href="LogIn.csp?logout=1">Sign out</a>
						</xsl:when>
						<xsl:otherwise>
							Please <a href="LogIn.csp?__REDIRECT=index.html">Sign in</a> or <a href="SignUp.csp?__REDIRECT=index.html">sign up</a>
						</xsl:otherwise>
					</xsl:choose>
				</p>
			</div>
		</div>

		<xsl:if test="HTMLContainer/CanAdd='1'">
			<a href="{public_name}.csp?xslt&#x3D;tpls/{public_name}.edit.xslt&amp;op&#x3D;new&amp;__REDIRECT&#x3D;{ldelim}$redirect{rdelim}">New</a>| <br/>
		</xsl:if>
		<form id="search_{public_name}_form" action="{ldelim}$scriptid{rdelim}.csp">
			<input type="text" name="q" id="search_{public_name}" value="{ldelim}$criteria{rdelim}" />
			<input type="submit" value="Go" />
{if advanced_member}
			<fieldset class="ui-state-highlight ui-corner-all">
			<legend>Advanced search</legend>
			<table>
{foreach advanced_member}
			<tr>
			<td><b>{advanced_member.column_name}</b>
			</td>
			<td><input type="checkbox" name="filter{advanced_member.index}" id="filter{advanced_member.index}" value="{advanced_member.member_name}" />
			</td>
			<td>
			<select name="op{advanced_member.index}" id="op{advanced_member.index}">
				<option value="&#x3D;">&#x3D;</option>
				<option value="&lt;&gt;">&lt;&gt;</option>
				<option value="&lt;&#x3D;">&lt;&#x3D;</option>
				<option value="&gt;&#x3D;" selected="selected">&gt;&#x3D;</option>
				<option value="&lt;">&lt;</option>
				<option value="&gt;">&gt;</option>
			</select>
			</td>
			<td>
			<input type="text" name="val{advanced_member.index}" id="val{advanced_member.index}" value="" />
			</td>
			<td>
			<input type="checkbox" name="filter{advanced_member.index2}" id="filter{advanced_member.index2}" value="{advanced_member.member_name}" />
			</td>
			<td>
			<select name="op{advanced_member.index2}" id="op{advanced_member.index2}">
				<option value="&#x3D;">&#x3D;</option>
				<option value="&lt;&gt;">&lt;&gt;</option>
				<option value="&lt;&#x3D;" selected="selected">&lt;&#x3D;</option>
				<option value="&gt;&#x3D;">&gt;&#x3D;</option>
				<option value="&lt;">&lt;</option>
				<option value="&gt;">&gt;</option>
			</select>
			</td>
			<td>
			<input type="text" name="val{advanced_member.index2}" id="val{advanced_member.index2}" value="" />
			</td>
			</tr>
{endfor}
			</table>
			</fieldset>
{endif} 
		</form>
	</xsl:if>

	<xsl:if test="HTMLContainer/LoggedIn='1'">
	<xsl:if test="$parent != ''">
		<a href="{public_name}.csp?p&#x3D;{ldelim}$parent{rdelim}&amp;w&#x3D;{ldelim}$scriptid{rdelim}&amp;xslt&#x3D;tpls/{public_name}.edit.xslt&amp;op&#x3D;new&amp;__REDIRECT&#x3D;{ldelim}$redirect{rdelim}">Add {public_name}</a> <br/>
	</xsl:if>
	</xsl:if>

	<xsl:if test="HTMLContainer/Message!=''">
		<div class="ui-widget">
			<div class="ui-state-error ui-corner-all" style="padding: 0 .7em;"> 
				<p>
					<span class="ui-icon ui-icon-alert" style="float: left; margin-right: .3em;"></span> 
					<xsl:value-of disable-output-escaping="yes" select="HTMLContainer/Message"/>
				</p>
			</div>
		</div>
	</xsl:if>


<xsl:choose>
	<xsl:when test="HTMLContainer/Count &gt; 0">
	<table class="ui-widget ui-widget-content">
		<thead>
			<tr class="ui-widget-header">
				<th scope="col" align="right">#</th>
{foreach column}
{if column.align_right}
				<th scope="col" align="right">{column.column_name}{if column.can_sort} <a rel="pagination" href="{ldelim}$scriptid{rdelim}?q&#x3D;{ldelim}$criteria{rdelim}&amp;sort&#x3D;{column.member_name}&amp;desc&#x3D;1&amp;p&#x3D;{ldelim}$parent{rdelim}">&#x2191;</a> <a rel="pagination" href="{ldelim}$scriptid{rdelim}.csp?q&#x3D;{ldelim}$criteria{rdelim}&amp;sort&#x3D;{column.member_name}&amp;desc&#x3D;0&amp;p&#x3D;{ldelim}$parent{rdelim}&amp;xslt&#x3D;{ldelim}$template{rdelim}">&#x2193;</a>{endif}</th>
{else}
				<th scope="col" align="left">{column.column_name}{if column.can_sort}  <a rel="pagination" href="{ldelim}$scriptid{rdelim}.csp?q&#x3D;{ldelim}$criteria{rdelim}&amp;sort&#x3D;{column.member_name}&amp;desc&#x3D;1&amp;p&#x3D;{ldelim}$parent{rdelim}">&#x2191;</a> <a rel="pagination" href="{ldelim}$scriptid{rdelim}.csp?q&#x3D;{ldelim}$criteria{rdelim}&amp;sort&#x3D;{column.member_name}&amp;desc&#x3D;0&amp;p&#x3D;{ldelim}$parent{rdelim}&amp;xslt&#x3D;{ldelim}$template{rdelim}">&#x2193;</a>{endif}</th>
{endif}
{endfor}
			</tr>
		</thead>
		<tfoot>
			<tr>
				<th scope="row">Page</th>
				<td colspan="{len_column}">
				<xsl:if test="$prev_link != ''">
					<a rel="pagination" href="{ldelim}$scriptid{rdelim}.csp?q&#x3D;{ldelim}$criteria{rdelim}&amp;page&#x3D;{ldelim}$prev_link{rdelim}&amp;sort&#x3D;{ldelim}$sort{rdelim}&amp;desc&#x3D;{ldelim}$desc{rdelim}&amp;p&#x3D;{ldelim}$parent{rdelim}&amp;xslt&#x3D;{ldelim}$template{rdelim}">&lt;&lt;</a>
				</xsl:if>
				<xsl:for-each select="HTMLContainer/PageLink/element">
					<xsl:variable name="Page" select="text()"/>
					<xsl:choose>
						<xsl:when test="text() != '0'">
							<a rel="pagination" href="{ldelim}$scriptid{rdelim}.csp?q&#x3D;{ldelim}$criteria{rdelim}&amp;page&#x3D;{ldelim}$Page{rdelim}&amp;sort&#x3D;{ldelim}$sort{rdelim}&amp;desc&#x3D;{ldelim}$desc{rdelim}&amp;p&#x3D;{ldelim}$parent{rdelim}&amp;xslt&#x3D;{ldelim}$template{rdelim}"><xsl:value-of select="$Page"/></a>
						</xsl:when>
						<xsl:otherwise>
							<b><xsl:value-of select="$current_page"/></b>
						</xsl:otherwise>
					</xsl:choose>
				</xsl:for-each>
				<xsl:if test="$next_link != ''">
					<a rel="pagination" href="{ldelim}$scriptid{rdelim}.csp?q&#x3D;{ldelim}$criteria{rdelim}&amp;page&#x3D;{ldelim}$next_link{rdelim}&amp;sort&#x3D;{ldelim}$sort{rdelim}&amp;desc&#x3D;{ldelim}$desc{rdelim}&amp;p&#x3D;{ldelim}$parent{rdelim}&amp;xslt&#x3D;{ldelim}$template{rdelim}">&gt;&gt;</a>
				</xsl:if>
				</td>
			</tr>
			<tr>
				<th scope="row">Total</th>
				<td colspan="{len_column}"><xsl:value-of select="HTMLContainer/Count"/> {friendly_name}</td>
			</tr>
		</tfoot>
		<tbody>
			<xsl:for-each select="HTMLContainer/Data/element">
			<xsl:variable name="__DBID" select="__DBID"/>
			<tr>
				<th scope="row" align="right"><a href="{public_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}"><xsl:value-of select="__VIEWINDEX"/></a></th>
{foreach column}
{if column.align_right}
				<td scope="col" align="right"><a href="{public_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}"><xsl:value-of disable-output-escaping="yes" select="{column.member_name}"/></a></td>
{else}
{if column.is_picture}
				<td scope="col" align="left">
					<xsl:if test="{column.member_name}_filename != ''">
						<a rel="img_group_{public_name}" title="{ldelim}row.{master_field}{rdelim}" href="get{public_name}{column.member_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}&amp;e&#x3D;.jpg"><img border="0" src="get{public_name}{column.member_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}&amp;thumb&#x3D;1" alt=""/></a> 
					</xsl:if>
				</td>
{else}
{if column.is_video}
				<td scope="col" align="left">
					<xsl:if test="{column.member_name}_filename != ''">
						<a rel="movie_group_{public_name}" href="Show{public_name}{column.member_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}"><img border="0" src="get{public_name}{column.member_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}&amp;thumb&#x3D;1" alt=""/></a> 
					</xsl:if>
				</td>
{else}
				<td scope="col" align="left"><a href="{public_name}.csp?id&#x3D;{ldelim}$__DBID{rdelim}"><xsl:value-of disable-output-escaping="yes" select="{column.member_name}"/></a> </td>
{endif}
{endif}
{endif}
{endfor}
			</tr>
			</xsl:for-each>
		</tbody>	
	</table>
	</xsl:when>
	<xsl:otherwise>
		<p>No records</p>
	</xsl:otherwise>
</xsl:choose>
</body>
</html>
</xsl:template>
</xsl:stylesheet>