#include "ConceptServer.h"
#include <fcntl.h>
#include <iostream>
#include <fstream>
#include <time.h>
#include <sys/types.h>

#ifndef NOLIST
 #include <dirent.h>
#endif

#include <errno.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>

#ifdef _WIN32
 #ifndef IPV6_V6ONLY
  #define IPV6_V6ONLY    27
 #endif
#endif

#define CONCEPT_PIPE_BUFFER    2048

#ifndef _S_IFMT
 #define _S_IFMT               S_IFMT
#endif

#define FAST_TCP
#define SAFE_RECV_TIMEOUTS


#define DSAY(str)         fprintf(stderr, "%s\n", str); fflush(stderr);
#define DSAY2(str, ii)    fprintf(stderr, "%s: %i\n", str, ii); fflush(stderr);

/*
 * int my_eof(int fd) {
 *      fd_set	readfds, excepfds;
 *
 *      FD_ZERO(&readfds);
 *      FD_ZERO(&excepfds);
 *      FD_SET(fd, &readfds);
 *      FD_SET(fd, &excepfds);
 *      if (select(fd+1, &readfds, NULL, &excepfds, NULL) < 0) {
 *      return -1;
 *      }
 *
 *      if (FD_ISSET(fd, &excepfds))
 *              return -1;
 *
 *      if (FD_ISSET(fd, &readfds))
 *              return 1;
 *
 *  return 0;
 * }
 */

#ifdef _WIN32
 #include <io.h>
 #include <ws2tcpip.h>
#else
 #include <sys/select.h>
 #include <netinet/tcp.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <poll.h>


int eof2(int fd) {
    struct pollfd fdinfo[1];

    fdinfo[0].fd     = fd;
    fdinfo[0].events = POLLIN | POLLPRI;
    int ret = poll(fdinfo, 1, 100);
    if (ret < 0) {
        printf("\n\r error");
        return(-1);
    }
    if (((fdinfo[0].revents & POLLIN) == POLLIN) || ((fdinfo[0].revents & POLLPRI) == POLLPRI))
        return(0);
    return(1);
}

int eof(int fd) {
    fd_set         readfds, excepfds;
    struct timeval timeout;

    timeout.tv_sec  = 0;
    timeout.tv_usec = 0;

    FD_ZERO(&readfds);
    FD_ZERO(&excepfds);
    FD_SET(fd, &readfds);
    FD_SET(fd, &excepfds);

    if (select(fd + 1, &readfds, NULL, &excepfds, &timeout) < 0) {
        perror("select");
        return(-1);
    }

    if (FD_ISSET(fd, &excepfds))
        return(-1);

    if (FD_ISSET(fd, &readfds))
        return(0);

    return(1);
}

#endif

#ifdef _WIN32
 #define pipe(phandles)    _pipe(phandles, 4096, _O_BINARY)
#endif

#ifndef O_BINARY
 #define O_BINARY    0
#endif


#define MAX_DEBUG_REQUEST    8192

extern "C" {
#include "rsa/rsa.h"
}

#ifdef _WIN32
 #define LIBRARY    "concept-core.2.0.dll"
#else
 #define LIBRARY    "libconceptcore.2.0.so"
#endif

//typedef int (*DEBUGGER_CALLBACK)(void *VDESC, void *CONTEXT, int Depth, int line, char *filename);

extern "C" {
//    typedef int (*API_INTERPRETER)(char *filename, char *inc_dir, char *lib_dir, void *fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipein, int pipeout, int apid, int parent_apid);
typedef int (*API_INTERPRETER2)(char *filename, char *inc_dir, char *lib_dir, void *fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipein, int pipeout, int apid, int parent_apid, void *userdata);
}

#define SEMP(s)    semp(s)
#define SEMV(s)    semv(s)

HHSEM            ConceptServer::sem_lock;
HHSEM            ConceptServer::sem_web_lock;
AnsiString       ConceptServer::Library;
AnsiString       ConceptServer::Include;
AnsiString       ConceptServer::CertificatePath;
AnsiString       ConceptServer::SafeModePath;
AnsiString       ConceptServer::Service;
AnsiString       ConceptServer::ServicePath = "./";
AnsiString       ConceptServer::LogPath;
AnsiString       ConceptServer::CGITrustedIP;
HostCallBackPath ConceptServer::hcbp = 0;
int ConceptServer::interapp_buffer   = DEFAULT_MAX_INTERAPP_BUFFER;

int  ConceptServer::SAFE_MODE;
int  ConceptServer::max_connections = MAX_CONNECTIONS;
int  ConceptServer::max_idle        = MAX_IDLE;
int  ConceptServer::web_timeout     = 10;
void *ConceptServer::this_reference = 0;

char **ConceptServer::enstr      = 0;
int  ConceptServer::environ_top  = 0;
int  ConceptServer::environ_base = 0;

int ConceptServer::ipv6 = 0;

typedef struct {
    SOCKET socket;
    bool   iscgi;
} ClientParamList;

typedef struct {
    SOCKET             socket;
    struct sockaddr_in sin;
} MyParamList;

#define     DEBUG_RUN         0x01
#define     DEBUG_STEPINTO    0x02
#define     DEBUG_RUNTO       0x03
#define     DEBUG_STEPOVER    0x04

typedef struct {
    long       CLIENT_NOTICE; // VARIABILA SETATA DE INTERPRETOR !!!

    int        last_line;
    int        debug_type;
    int        run_to_line;
    AnsiString FileName;
    AnsiString run_to_filename;
    void       *context_id;
    SOCKET     sock;
    SOCKET     DEBUG_SOCKET;
} ClientDebugInfo;

#ifdef _WIN32
DWORD _GetProcessId(HANDLE hProcess) {
    HANDLE hThread;
    DWORD  Pid;

    /*GetCurrentProcessId = GetProcAddress(GetModuleHandle("Kernel32.dll"),
     *                                   "GetCurrentProcessId");*/

    hThread = CreateRemoteThread(hProcess,
                                 NULL,
                                 0,
                                 (LPTHREAD_START_ROUTINE)GetCurrentProcessId,
                                 NULL,
                                 0,
                                 NULL);
    if (!hThread)
        return(0);
    else {
        WaitForSingleObject(hThread, INFINITE);
        // Exit code of remote thread is the PID
        GetExitCodeThread(hThread, &Pid);
        CloseHandle(hThread);
        return(Pid);
    }
}

#endif

int full_read(int fid, void *buf, int size) {
    int  r_size = 0;
    int  ret    = 0;
    char *pbuf  = (char *)buf;

    if (size <= 0)
        return(0);
    do {
        ret = read(fid, pbuf + r_size, size);
        if (ret > 0) {
            r_size += ret;
            size   -= r_size;
        } else
#ifdef _WIN32
            Sleep(10);
#else
            usleep(10000);
#endif
    } while ((ret >= 0) && (r_size < size));
    return(r_size);
}

unsigned int hash(const void *key) {
    if (!key)
        return(0);
    const unsigned int  m     = 0x5bd1e995;
    const int           r     = 24;
    unsigned int        seed  = 0x7870AAFF;
    const unsigned char *data = (const unsigned char *)key;
    int len = strlen((const char *)data);
    if (!len)
        return(0);

    unsigned int h = seed ^ len;

    while (len >= 4) {
        unsigned int k = *(unsigned int *)data;

        k *= m;
        k ^= k >> r;
        k *= m;

        h *= m;
        h ^= k;

        data += 4;
        len  -= 4;
    }

    switch (len) {
        case 3:
            h ^= data[2] << 16;

        case 2:
            h ^= data[1] << 8;

        case 1:
            h ^= data[0];
            h *= m;
    }

    h ^= h >> 13;
    h *= m;
    h ^= h >> 15;
    return(h);
}

AnsiString ConceptServer::NormalizePath(AnsiString path) {
    AnsiString paths[MAX_PATHS];
    int        paths_len = 0;
    int        len       = path.Length();
    char       *ptr      = path.c_str();
    char       c;
    AnsiString ret;
    AnsiString temp;

    for (int i = 0; i < len; i++) {
        c = ptr[i];
        if (((c == '/') || (c == '\\'))) {
            if (!temp.Length())
                continue;

            if (temp != (char *)".") {
                if (temp == (char *)"..") {
                    if (paths_len)
                        paths_len--;
                    temp             = (char *)"";
                    paths[paths_len] = temp;
                } else {
                    paths[paths_len++] = temp;
                    if (paths_len >= MAX_PATHS)
                        break;
                    temp = (char *)"";
                }
            } else
                temp = (char *)"";
        } else
            temp += c;
    }
    paths[paths_len++] = temp;

    for (int i = 0; i < paths_len; i++)
        if (paths[i].Length()) {
            if (ret.Length())
                ret += (char *)"/";
            ret += paths[i];
        }

    return(ret);
}

void ConceptServer::Print(char *str, int length) {
    if (length > 0)
        std::cout.write(str, length);
    else
        std::cout << str;
}

#ifdef _WIN32
const char *inet_ntop(int af, const void *src, char *dst, socklen_t cnt) {
    if (af == AF_INET) {
        struct sockaddr_in in;
        memset(&in, 0, sizeof(in));
        in.sin_family = AF_INET;
        memcpy(&in.sin_addr, src, sizeof(struct in_addr));
        getnameinfo((struct sockaddr *)&in, sizeof(struct sockaddr_in), dst, cnt, NULL, 0, NI_NUMERICHOST);
        return(dst);
    } else if (af == AF_INET6)   {
        struct sockaddr_in6 in;
        memset(&in, 0, sizeof(in));
        in.sin6_family = AF_INET6;
        memcpy(&in.sin6_addr, src, sizeof(struct in_addr6));
        getnameinfo((struct sockaddr *)&in, sizeof(struct sockaddr_in6), dst, cnt, NULL, 0, NI_NUMERICHOST);
        return(dst);
    }
    return(NULL);
}

#endif

AnsiString getip(int socket, int *port) {
    struct sockaddr_storage addr;
    char       ipstr[INET6_ADDRSTRLEN];
    AnsiString res = (char *)"unknown";

#ifdef _WIN32
    int peerlen = sizeof(addr);
#else
    socklen_t peerlen = sizeof(addr);
#endif
    *port = 0;
    if (socket)
        getpeername(socket, (sockaddr *)&addr, &peerlen);
    const char *ip = 0;
    if (addr.ss_family == AF_INET) {
        struct sockaddr_in *s = (struct sockaddr_in *)&addr;
        ip    = inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
        *port = ntohs(s->sin_port);
    } else { // AF_INET6
        struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
        ip    = inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
        *port = ntohs(s->sin6_port);
    }
    if (ip)
        res = (char *)ip;
    return(res);
}

void SocketSay(int socket, char *what, int what2 = -0xFFFF) {
    struct sockaddr_storage addr;
    char        ipstr[INET6_ADDRSTRLEN];
    int         port     = 0;
    static char *unknown = "unknown";

#ifdef _WIN32
    int peerlen = sizeof(addr);
#else
    socklen_t peerlen = sizeof(addr);
#endif
    if (socket)
        getpeername(socket, (sockaddr *)&addr, &peerlen);
    time_t now;
    time(&now);
    AnsiString sztime = ctime(&now);
    char       *ptr   = sztime.c_str();
    int        len    = sztime.Length();
    if (ptr[len - 1] == '\n')
        ptr[len - 1] = 0;
    const char *ip = 0;
    if (addr.ss_family == AF_INET) {
        struct sockaddr_in *s = (struct sockaddr_in *)&addr;
        port = ntohs(s->sin_port);
        ip   = inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr);
    } else
    if (addr.ss_family == AF_INET6) {
        struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
        port = ntohs(s->sin6_port);
        ip   = inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr);
    }
    if (!ip)
        ip = unknown;
    if (what2 != -0xFFFF)
        std::cout << ptr << ": #" << socket << " " << ip << ":" << port << " > " << what << what2 << "\n";
    else
        std::cout << ptr << ": #" << socket << " " << ip << ":" << port << " > " << what << "\n";
    if (ConceptServer::LogPath.Length()) {
        std::ofstream out;
        out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
        if (what2 != -0xFFFF)
            out << ptr << ": #" << socket << " " << ip << ":" << port << " > " << what << what2 << "\n";
        else
            out << ptr << ": #" << socket << " " << ip << ":" << port << " > " << what << "\n";
        out.flush();
        out.close();
    }
}

int ConceptServer::TALK_TO_CLIENT(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED) {
    ClientDebugInfo *DINFO = (ClientDebugInfo *)DEBUGGER_RESERVED;

    AnsiString SERVER_QUERY = filename;

    /*SERVER_QUERY    =   (char *)"<Message _OWNER=\"";
     * SERVER_QUERY    +=  "%DEBUGGER";
     * SERVER_QUERY    +=  (char *)"\" _ID=-10 _TARGET=\"";
     * SERVER_QUERY    +=  filename;
     * SERVER_QUERY    +=  "\" _DATA=\"";
     * SERVER_QUERY    +=  AnsiString(line);
     * SERVER_QUERY    +=  (char *)"\" />";*/

    int size = SERVER_QUERY.Length() + 1 + sizeof(int);

    send(DINFO->DEBUG_SOCKET, (char *)&size, sizeof(int), 0);
    send(DINFO->DEBUG_SOCKET, (char *)&line, sizeof(int), 0);
    send(DINFO->DEBUG_SOCKET, SERVER_QUERY, SERVER_QUERY.Length() + 1, 0);

    while (1) {
        int size      = 0;
        int received  = 0;
        int rec_count = 0;
        do {
            received   = recv(DINFO->DEBUG_SOCKET, (char *)&size + rec_count, sizeof(int) - rec_count, 0);
            rec_count += received;
        } while ((rec_count < sizeof(int)) && (received > 0));
        size = ntohl(size);

        // host is down ...
        if ((received <= 0) || (size > MAX_DEBUG_REQUEST) || (size <= 0))
            return(1);

        char *buffer = new char[size + 1];
        buffer[size] = 0;
        rec_count    = 0;
        do {
            received   = recv(DINFO->DEBUG_SOCKET, buffer + rec_count, size - rec_count, 0);
            rec_count += received;
        } while ((received > 0) && (rec_count < size));
        AnsiString command = buffer;
        delete[] buffer;

        if (received <= 0)
            return(1);

        if (command == (char *)"stepinto") {
            DINFO->debug_type    = DEBUG_STEPINTO;
            DINFO->CLIENT_NOTICE = 0;
            break;
        } else
        if (command == (char *)"stepover") {
            DINFO->context_id    = CONTEXT;
            DINFO->debug_type    = DEBUG_STEPOVER;
            DINFO->CLIENT_NOTICE = (long)CONTEXT;
            break;
        } else
        if (command == (char *)"run") {
            DINFO->debug_type    = DEBUG_RUN;
            DINFO->CLIENT_NOTICE = 0;
            break;
        } else
        if (command == (char *)"break") {
            DINFO->debug_type    = DEBUG_RUN;
            DINFO->CLIENT_NOTICE = 0;
            return(1);
            break;
        } else
        if (command.Pos((char *)"runto ") == 1) {
            AnsiString line = (char *)command.c_str() + 6;
            DINFO->debug_type      = DEBUG_RUNTO;
            DINFO->run_to_line     = line.ToInt();
            DINFO->run_to_filename = filename;
            DINFO->CLIENT_NOTICE   = 0;
            break;
        } else
        if (command.Pos((char *)"watch ") == 1) {
            AnsiString varname = (char *)command.c_str() + 6;
            char       var_buffer[MAX_DEBUG_REQUEST + 1];
            strcpy(var_buffer, "(Variable not defined)");
            var_buffer[MAX_DEBUG_REQUEST] = 0;
            if (GVP(VDESC, CONTEXT, Depth, varname, var_buffer, MAX_DEBUG_REQUEST)) {
                // to do ... variable undefined ...
            }

            int res_size = strlen(var_buffer) + 1;
            send(DINFO->DEBUG_SOCKET, (char *)&res_size, sizeof(int), 0);
            send(DINFO->DEBUG_SOCKET, var_buffer, res_size, 0);
        }
    }
    return(0);
}

int ConceptServer::DEBUGGER_TRAP(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED) {
    ClientDebugInfo *DINFO = (ClientDebugInfo *)DEBUGGER_RESERVED;
    // to do !
    // Debugger is now trapped !
    //if (DINFO->FileName!=filename)
    //   return 0;
    int result = 0;

    switch (DINFO->debug_type) {
        case DEBUG_RUN:
            break;

        case DEBUG_RUNTO:
            if ((DINFO->run_to_line == line) && (DINFO->run_to_filename == filename))
                // Notify Client
                result = ConceptServer::TALK_TO_CLIENT(VDESC, CONTEXT, Depth, line, filename, GVP, DEBUGGER_RESERVED);
            break;

        case DEBUG_STEPINTO:
            if ((DINFO->last_line != line) || (DINFO->FileName != filename))
                // Notify Client
                // s-a schimbat linia ...
                result = ConceptServer::TALK_TO_CLIENT(VDESC, CONTEXT, Depth, line, filename, GVP, DEBUGGER_RESERVED);
            break;

        case DEBUG_STEPOVER:
            if ((DINFO->last_line != line) || (DINFO->FileName != filename))
                if ((CONTEXT == DINFO->context_id) || (DINFO->CLIENT_NOTICE == -1)) {
                    result = ConceptServer::TALK_TO_CLIENT(VDESC, CONTEXT, Depth, line, filename, GVP, DEBUGGER_RESERVED);
                    DINFO->CLIENT_NOTICE = 0;
                }
            break;
    }
    DINFO->FileName  = filename;
    DINFO->last_line = line;
    return(result);
}

ConceptServer::ConceptServer(AnsiString Cert, AnsiString Lib, AnsiString Inc, AnsiString Safe, int port, int cgiport, HostCallBackPath cb, int max_buffer) {
    this->APPLICATIONS = 0;

    /*this->APPLICATIONS=new INTERCOM_ELEMENT[MAX_APPLICATIONS];
     * for (int i=0;i<MAX_APPLICATIONS;i++) {
     *  target->pipe_in[0]=0;
     *  target->pipe_in[1]=0;
     *  target->pipe_out[0]=0;
     *  target->pipe_out[1]=0;
     *  target->APID=0;
     *  target->done=0;
     *  target->threadID=0;
     *  target->hash=0;
     * }*/

    this->PORT            = port;
    this->PORT_CGI        = cgiport;
    this->ListenSocket    = INVALID_SOCKET;
    this->CGIListenSocket = INVALID_SOCKET;

    if (max_buffer > 0)
        ConceptServer::interapp_buffer = max_buffer * 1024;
    else
        ConceptServer::interapp_buffer = 0;
    ConceptServer::SAFE_MODE       = 1;
    ConceptServer::CertificatePath = Cert;
    ConceptServer::Library         = Lib;
    ConceptServer::Include         = Inc;
    ConceptServer::SafeModePath    = Safe;
    ConceptServer::hcbp            = cb;

    time_t now;
    time(&now);
    AnsiString sztime = ctime(&now);
    char       *ptr   = sztime.c_str();
    int        len    = sztime.Length();
    if (ptr[len - 1] == '\n')
        ptr[len - 1] = 0;

    if (!Safe.Length()) {
        ConceptServer::SAFE_MODE = 0;
        if (ConceptServer::LogPath.Length()) {
            std::ofstream out;
            out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
            out << ptr << " Warning: not running in safe mode!\n";
            out.flush();
            out.close();
        }
    }
    std::ofstream out;
    out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
    out << ptr << " Paths:\n";
    out << ptr << " -> Include  : " << Inc.c_str() << "\n";
    out << ptr << " -> Library  : " << Lib.c_str() << "\n";
    out << ptr << " -> Safe mode: " << Safe.c_str() << "\n";
    out.flush();
    out.close();

    WEBPARENTS_COUNT = 0;
    // application IDs must start with 1
    NEXT_APID       = 1;
    FIRST_SAFE_APID = 1;

    INTER_LOCKED   = 0;
    this_reference = this;
    seminit(sem_lock, 1);
    seminit(sem_web_lock, 1);
}

int ConceptServer::Init() {
    int err = 0;

#ifdef _WIN32
    WSADATA data;
    // starting up windows sockets
    err = WSAStartup(MAKEWORD(2, 2), &data);

    if (err) {
        WSACleanup();
        return(ERR_WSA_STARTUP);
    }
#endif

    // Create Socket
    if (ConceptServer::ipv6)
        this->ListenSocket = socket(AF_INET6, SOCK_STREAM, IPPROTO_TCP);
    else
        this->ListenSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (ListenSocket == INVALID_SOCKET) {    // Fail To Create Socket
#ifdef _WIN32
        WSACleanup();
#endif
        return(ERR_CREATE_SOCKET);
    }

    int flag = 1;
#ifdef FAST_TCP
    setsockopt(ListenSocket,                      /* socket affected */
               IPPROTO_TCP,                       /* set option at TCP level */
               TCP_NODELAY,                       /* name of option */
               (char *)&flag,                     /* the cast is historical cruft */
               sizeof(int));                      /* length of option value */
#endif
    setsockopt(ListenSocket,
               SOL_SOCKET,
               SO_KEEPALIVE,
               (char *)&flag,
               sizeof(int));

#ifdef IPV6_V6ONLY
    if (ConceptServer::ipv6) {
        int ipv6only = 0;
        setsockopt(ListenSocket, IPPROTO_IPV6, IPV6_V6ONLY, (char *)&ipv6only, sizeof(ipv6only));
        SocketSay(ListenSocket, "Listening in IPv6 mode");
    }
#endif

    memset(&sin, 0, sizeof(sin));
    int sin_size = 0;
    if (ConceptServer::ipv6) {
        struct sockaddr_in6 *psin = (sockaddr_in6 *)&sin;
        psin->sin6_addr   = in6addr_any;
        psin->sin6_family = AF_INET6;
        psin->sin6_port   = htons(PORT);
        sin_size          = sizeof(sockaddr_in6);
    } else {
        struct sockaddr_in *psin = (sockaddr_in *)&sin;
        psin->sin_addr.s_addr = INADDR_ANY;
        psin->sin_family      = AF_INET;
        psin->sin_port        = htons(PORT);
        sin_size = sizeof(sockaddr_in);
    }

    if (bind(ListenSocket, (struct sockaddr *)&sin, sin_size) == INVALID_SOCKET) {
#ifdef _WIN32
        closesocket(ListenSocket);
        WSACleanup();
#else
        close(ListenSocket);
#endif
        std::cerr << "Error binding socket\n";
        LOG_DATA("Error binding socket");
        return(ERR_BIND_SOCKET);
    }
    return(NO_WEB_ERROR);
}

int ConceptServer::RouteMessage(int *parameters, char *data, INTERCOM_ELEMENT *IEL, int by_hash) {
    AnsiString buffer;

    if (by_hash) {
        if (!IEL->hash)
            return(-1);
        parameters[0] = IEL->APID;
        buffer.LoadBuffer((char *)parameters, sizeof(int) * 3);
        if (parameters[2] > 0)
            buffer.AddBuffer(data, parameters[2]);
        for (INTERCOM_ELEMENT *target = this->APPLICATIONS; target; target = (INTERCOM_ELEMENT *)target->NEXT)
            if ((target->APID) && (target->hash == IEL->hash) && (target->APID != IEL->APID) && (!target->done))
                this->inter_write(target, buffer.c_str(), buffer.Length());
        return(1);
    }

    for (INTERCOM_ELEMENT *target = this->APPLICATIONS; target; target = (INTERCOM_ELEMENT *)target->NEXT)
        if (target->APID == parameters[0]) {
            if (target->done)
                return(-1);
            parameters[0] = IEL->APID;
            buffer.LoadBuffer((char *)parameters, sizeof(int) * 3);
            if (parameters[2] > 0)
                buffer.AddBuffer(data, parameters[2]);
            this->inter_write(target, buffer.c_str(), buffer.Length(), target->APID != IEL->APID);
            return(1);
        }

    return(0);
}

INTERCOM_ELEMENT *ConceptServer::AddAPID(int *pipein, int *pipeout, unsigned int checksum, unsigned int **checksum_ptr) {
    INTERCOM_ELEMENT *target = new INTERCOM_ELEMENT;

    //if (!this->APPLICATIONS)
    //    NEXT_APID=1;
    if (NEXT_APID < FIRST_SAFE_APID)
        NEXT_APID = FIRST_SAFE_APID;

    target->APID     = NEXT_APID++;
    target->done     = 0;
    target->threadID = 0;
    target->NEXT     = 0;


    if (pipe(target->pipe_in) < 0)
        SocketSay(-1, "Pipe operation failed (pipe_in)");
    if (pipe(target->pipe_out) < 0)
        SocketSay(-1, "Pipe operation failed (pipe_out)");

    *pipein  = target->pipe_out[0];
    *pipeout = target->pipe_in[1];

    seminit(target->lock, 1);
    SEMP(target->lock);

    target->hash = checksum;
    if (checksum_ptr)
        *checksum_ptr = &target->hash;

#ifdef _WIN32
    SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
    DWORD tid;
    target->threadID = CreateThread(&sa, 0, ConceptServer::ClientAPM, (LPVOID)target, 0, &tid);
#else
    pthread_create(&target->threadID, NULL, ConceptServer::ClientAPM, (LPVOID)target);
#endif
    SEMV(target->lock);

    SEMP(sem_lock);
    if (APPLICATIONS)
        target->NEXT = APPLICATIONS;
    else {
        NEXT_APID    = FIRST_SAFE_APID;
        target->APID = NEXT_APID++;
    }
    APPLICATIONS = target;

    SEMV(sem_lock)
    return(target);
}

INTERCOM_ELEMENT *ConceptServer::AddCGIAPID(int pipein, int pipeout, unsigned int checksum) {
    INTERCOM_ELEMENT *target = new INTERCOM_ELEMENT;

    //if (!this->APPLICATIONS)
    //    NEXT_APID=1;
    if (NEXT_APID < FIRST_SAFE_APID)
        NEXT_APID = FIRST_SAFE_APID;

    target->APID     = NEXT_APID++;
    target->done     = 0;
    target->threadID = 0;
    target->NEXT     = 0;

    if (pipe(target->pipe_in) < 0)
        SocketSay(-1, "Pipe operation failed (pipe_in)");
    if (pipe(target->pipe_out) < 0)
        SocketSay(-1, "Pipe operation failed (pipe_out)");

    seminit(target->lock, 1);
    SEMP(target->lock);

    target->pipe_out[0] = pipeout;
    target->pipe_out[1] = pipeout;

    target->pipe_in[0] = pipein;
    target->pipe_in[1] = pipein;

    target->threadID  = 0;
    target->processID = 0;
    target->hash      = checksum;

    target->done = 0;

    SEMV(sem_lock);
#ifdef _WIN32
    SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
    DWORD tid;
    target->threadID = CreateThread(&sa, 0, ConceptServer::ClientAPM, (LPVOID)target, 0, &tid);
#else
    pthread_create(&target->threadID, NULL, ConceptServer::ClientAPM, (LPVOID)target);
#endif

    SEMV(target->lock);

    SEMP(sem_lock);
    if (APPLICATIONS)
        target->NEXT = APPLICATIONS;
    else {
        NEXT_APID    = FIRST_SAFE_APID;
        target->APID = NEXT_APID++;
    }

    APPLICATIONS = target;
    SEMV(sem_lock)

    return(target);
}

int ConceptServer::RemoveAPID(INTERCOM_ELEMENT *target_APID) {
    if (!target_APID)
        return(0);

    int APID = target_APID->APID;
    SEMP(sem_lock);
    SEMP(target_APID->lock);
    target_APID->done = 1;
    target_APID->APID = 0;
    SEMV(target_APID->lock);

    INTERCOM_ELEMENT *prev   = 0;
    INTERCOM_ELEMENT *target = 0;
    for (target = this->APPLICATIONS; target; target = (INTERCOM_ELEMENT *)target->NEXT) {
        if (target == target_APID) {
            if (prev)
                prev->NEXT = target->NEXT;

            if (target == this->APPLICATIONS)
                this->APPLICATIONS = (INTERCOM_ELEMENT *)target->NEXT;
            break;
        }
        prev = target;
    }
    SEMV(target_APID->lock);
    SEMV(sem_lock);

    if (target == target_APID) {
        int timeout = 1000;

        // astept deblocare, in caz ca este blocat
        SEMP(sem_lock);

        SEMP(target->lock);
        while ((this->INTER_LOCKED == APID) && (timeout)) {
#ifdef _WIN32
            Sleep(1);
#else
            usleep(1000);
#endif
            timeout--;
        }

        UnHost(APID);
        SEMV(target->lock);
        SEMV(sem_lock);

        // wait 10 seconds
        timeout = 10000;
        while (true) {
            SEMP(target->lock);
            if (!eof(target->pipe_in[0]))
#ifdef _WIN32
                Sleep(1);
#else
                usleep(1000);
#endif
            timeout--;
            SEMV(target->lock);
            if (!timeout)
                break;
        }

        SEMP(target->lock);
        target->inter_buf.LoadBuffer(0, 0);
        SEMV(target->lock);

        // wait max 3 seconds
        timeout = 3000;
        char done = 0;
        while (true) {
            SEMP(target->lock);
            done = target->done;
            SEMV(target->lock);
            if (done == -1)
                break;
#ifdef _WIN32
            Sleep(1);
#else
            usleep(1000);
#endif
            timeout--;
            if (!timeout) {
                LOG_DATA("Forced APID close (waited more than 3 seconds)");
                break;
            }
        }


        if ((done != -1) && (target->processID)) {
#ifdef _WIN32
            TerminateProcess((HANDLE)target->processID, 0x80);
#else
            kill(target->processID, SIGKILL);
#endif
            target->processID = 0;
            LOG_DATA("APID process terminated (forced)");
        }

        if ((done != -1) && (target->threadID)) {
#ifdef _WIN32
            TerminateThread(target->threadID, 0);
#else
            pthread_cancel(target->threadID);
#endif
            LOG_DATA("APID thread terminated (forced)");

            target->threadID = 0;
        }

        SEMP(target->lock);
        close(target->pipe_in[0]);
        if (target->pipe_in[0] != target->pipe_in[1])
            close(target->pipe_in[1]);

        close(target->pipe_out[0]);
        if (target->pipe_out[0] != target->pipe_out[1])
            close(target->pipe_out[1]);

        target->pipe_in[0]  = 0;
        target->pipe_in[1]  = 0;
        target->pipe_out[0] = 0;
        target->pipe_out[1] = 0;

        target->hash = 0;
        target->APID = 0;
        SEMV(target->lock);

        // de curand adaugat ...
        SEMP(sem_lock);
        semdel(target->lock);

        SEMV(sem_lock);

        delete target;
        return(1);
    }
    return(0);
}

int ConceptServer::InterMessage(INTERCOM_ELEMENT *IEL, int id, int msg_id, AnsiString message, int lock) {
    int params[3];

    params[0] = id;
    params[1] = msg_id;
    params[2] = message.Length();

    AnsiString buffer;
    buffer.LoadBuffer((char *)params, sizeof(int) * 3);
    buffer += message;

    this->inter_write(IEL, buffer.c_str(), buffer.Length(), lock);
    return(0);
}

int ConceptServer::Host(int APID, char *data, bool debug) {
    SEMP(sem_web_lock);
    int where = WEBPARENTS_COUNT;
    int count = 0;

    for (int i = 0; i < MAX_PARENTS; i++) {
        if (count >= WEBPARENTS_COUNT)
            break;

        if (WEBPARENTS[i].flag == FLAG_DELETED) {
            where = i;
            break;
        } else
            count++;
    }
    if (WEBPARENTS_COUNT == MAX_PARENTS) {
        SEMV(sem_web_lock);
        return(-1);
    }

    WEBPARENTS[where].APID = APID;
    WEBPARENTS[where].flag = debug ? FLAG_DEBUG : FLAG_PARENT;
    WEBPARENTS[where].key  = hash(data);

    WEBPARENTS_COUNT++;
    SEMV(sem_web_lock);

    return(0);
}

int ConceptServer::UnHost(int APID, char *data) {
    int count = 0;

    SEMP(sem_web_lock);
    unsigned int data_key = data ? hash(data) : 0;
    for (int i = 0; i < MAX_PARENTS; i++) {
        if (count >= WEBPARENTS_COUNT)
            break;

        if (WEBPARENTS[i].flag != FLAG_DELETED) {
            if (data) {
                if ((WEBPARENTS[i].APID == APID) && (WEBPARENTS[i].key == data_key)) {
                    WEBPARENTS[i].flag = FLAG_DELETED;
                    WEBPARENTS_COUNT--;
                    SEMV(sem_web_lock);
                    return(0);
                }
            } else
            // if no data, remove all the references (RemoveAPID)
            if (WEBPARENTS[i].APID == APID) {
                WEBPARENTS[i].flag = FLAG_DELETED;
                WEBPARENTS_COUNT--;
            }
        } else
            count++;
    }
    SEMV(sem_web_lock);
    return(-1);
}

int ConceptServer::GetHost(char *data, bool& debug) {
    int          count    = 0;
    unsigned int data_key = data ? hash(data) : 0;

    for (int i = 0; i < MAX_PARENTS; i++) {
        if (count >= WEBPARENTS_COUNT)
            break;

        if ((WEBPARENTS[i].flag != FLAG_DELETED) && (WEBPARENTS[i].key == data_key)) {
            debug = (WEBPARENTS[i].flag == FLAG_DEBUG);
            return(WEBPARENTS[i].APID);
        } else
            count++;
    }
    return(0);
}

#ifdef _WIN32
DWORD WINAPI ConceptServer::InterApplicationProcessor(LPVOID server) {
#else
LPVOID ConceptServer::InterApplicationProcessor(LPVOID server) {
#endif
    ConceptServer *THIS_PTR = (ConceptServer *)server;

    bool no_messages;
    while (1) {
        SEMP(sem_lock);
        INTERCOM_ELEMENT *first = THIS_PTR->APPLICATIONS;
        if (!first) {
            SEMV(sem_lock);
#ifdef _WIN32
            Sleep(/*250*/ 100);
#else
            usleep(/*250000*/ 100000);
#endif
            continue;
        }
        no_messages = true;
        int index = 0;
        for (INTERCOM_ELEMENT *target = first; target; target = (INTERCOM_ELEMENT *)target->NEXT) {
            SEMP(target->lock);
            long apid = target->APID;
            SEMV(target->lock);

            if (apid) {
                int params[3]; // app_id, msg_id, size
                int is_eof = eof(target->pipe_in[0]);
                if (is_eof != 0) {
                    if (is_eof < 0)
                        SocketSay(-1, "PIPE EOF error");
                    SEMV(sem_lock);
                    continue;
                } else
                    no_messages = false;
                //SEMP(sem_lock);
                SEMP(target->lock);
                THIS_PTR->INTER_LOCKED = target->APID;
                int how_much = full_read(target->pipe_in[0], (void *)params, sizeof(int) * 3);
                if ((how_much == sizeof(int) * 3) && ((params[0] >= 0) || (params[0] == -1))) {
                    // returneaza 1 daca mesajul a fost trimis
                    char *data = 0;
                    char all   = 1;
                    if (params[2]) {
                        data            = new char[params[2] + 1];
                        data[params[2]] = 0;
                        how_much        = full_read(target->pipe_in[0], (void *)data, params[2]);
                        if (how_much != params[2])
                            all = 0;
                    }
                    if (!all) {
                        THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "Error in routing message. Invalid APID ?", 0);
                        LOG_DATA("Error in routing message. Invalid APID ?");
                    } else
                    if (params[0] != -1) {
                        if (!THIS_PTR->RouteMessage(params, data, target, params[0] == 0)) {
                            THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "Error in routing message. Invalid APID ?", 0);
                            LOG_DATA("Error in routing message. Invalid APID ?");
                        }
                    } else {
                        switch (params[1]) {
                            // register parent for web application
                            case -10:
                                if (data)
                                    THIS_PTR->Host(target->APID, data);
                                else
                                    THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "CAS: No file specified.", 0);
                                break;

                            // unregister parent for web application
                            case -11:
                                if (data)
                                    THIS_PTR->UnHost(target->APID, data);
                                else
                                    THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "CAS: No file specified.", 0);
                                break;

                            // debug for web application
                            case -12:
                                if (data)
                                    THIS_PTR->Host(target->APID, data, true);
                                else
                                    THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "CAS: No file specified.", 0);
                                break;

                            default:
                                THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "CAS: Invalid server message.", 0);
                        }
                    }
                    if (data)
                        delete[] data;
                } else {
                    THIS_PTR->InterMessage(target /*.pipe_out[1]*/, -1, -2, "Error in message. Invalid size ?", 0);
                    LOG_DATA("Error in message. Invalid size ?");
                }
                THIS_PTR->INTER_LOCKED = 0;
                SEMV(target->lock);
            }
        }
        SEMV(sem_lock);
        if (no_messages) {
#ifdef _WIN32
            Sleep(100);
#else
            usleep(100000);
#endif
        }

#ifdef _WIN32
        Sleep(1);
#else
        usleep(1000);
#endif
    }
}

#ifdef _WIN32
DWORD WINAPI ConceptServer::CGIThread(LPVOID server) {
#else
LPVOID ConceptServer::CGIThread(LPVOID server) {
#endif
    ConceptServer *THIS_PTR = (ConceptServer *)server;

#ifdef _WIN32
    Sleep(3000);
#else
    usleep(3000000);
#endif

    if (listen(THIS_PTR->CGIListenSocket, ConceptServer::max_connections) == SOCKET_ERROR) {
        LOG_DATA("CGI Server failed: listen failed. Concept server will continue to run, but will not handle CGI.");
#ifdef _WIN32
        return(ERR_LISTEN_SOCKET);
#else
        return((void *)ERR_LISTEN_SOCKET);
#endif
    }

#ifdef _WIN32
    int   addr_length = sizeof(THIS_PTR->cgisin);
    DWORD threadID;
#else
    socklen_t addr_length = sizeof(THIS_PTR->cgisin);
    pthread_t threadID;
#endif

    struct sockaddr_in peer;
#ifdef _WIN32
    int peerlen = sizeof(struct sockaddr);
#else
    socklen_t peerlen = sizeof(struct sockaddr);
#endif

    while (1) {
        SOCKET CGIClientSocket = accept(THIS_PTR->CGIListenSocket, (sockaddr *)&THIS_PTR->cgisin, &addr_length);
        if (CGIClientSocket == INVALID_SOCKET) {
            LOG_DATA("CGI Server failed: failed on accept");
            return(0);
        }

        getpeername(CGIClientSocket, (sockaddr *)&peer, &peerlen);
        if (ConceptServer::CGITrustedIP != inet_ntoa(peer.sin_addr)) {
#ifdef _WIN32
            closesocket(CGIClientSocket);
#else
            close(CGIClientSocket);
#endif
            LOG_DATA("CGI Server: Unauthorized access from ip: ");
            LOG_DATA(inet_ntoa(peer.sin_addr));
        } else {
            ClientParamList *cpl = new ClientParamList;
            cpl->socket = CGIClientSocket;
            cpl->iscgi  = true;
            // creez threadul clientului
#ifdef _WIN32
            SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
            CreateThread(&sa, 0, ConceptServer::ClientThread, (LPVOID)/*CGIClientSocket*/ cpl, 0, &threadID);
#else
            pthread_create(&threadID, NULL, ConceptServer::ClientThread, (LPVOID)/*CGIClientSocket*/ cpl);
#endif
            //ClientThread((LPVOID)newSOCK);
#ifdef _WIN32
            Sleep(10);
#else
            usleep(10000);
#endif
        }
    }
}

int ConceptServer::StartCGIServer() {
    int err = 0;

    // Create Socket
    this->CGIListenSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (CGIListenSocket == INVALID_SOCKET) {    // Fail To Create Socket
        LOG_DATA("CGI Server failed: socket creation failed. Concept server will continue to run, but will not handle CGI.");
        return(ERR_CREATE_SOCKET);
    }

    int flag = 1;
#ifdef FAST_TCP
    setsockopt(CGIListenSocket,                   /* socket affected */
               IPPROTO_TCP,                       /* set option at TCP level */
               TCP_NODELAY,                       /* name of option */
               (char *)&flag,                     /* the cast is historical cruft */
               sizeof(int));                      /* length of option value */
#endif
    setsockopt(CGIListenSocket,
               SOL_SOCKET,
               SO_KEEPALIVE,
               (char *)&flag,
               sizeof(int));

    memset(&cgisin, 0, sizeof(cgisin));
    cgisin.sin_addr.s_addr = INADDR_ANY;
    cgisin.sin_family      = AF_INET;
    cgisin.sin_port        = htons(PORT_CGI);

    if (bind(CGIListenSocket, (struct sockaddr *)&cgisin, sizeof(cgisin)) == INVALID_SOCKET) {
#ifdef _WIN32
        closesocket(CGIListenSocket);
#else
        close(CGIListenSocket);
#endif
        LOG_DATA("CGI Server failed: bind failed. Concept server will continue to run, but will not handle CGI.");
        return(ERR_BIND_SOCKET);
    }

    LOG_DATA("Concept Server will handle CGI on port: ");
    LOG_DATA(PORT_CGI);
    LOG_DATA("Concept ALEONE salutes you !");
    LOG_DATA("InterApp messages buffer size:");
    LOG_DATA(ConceptServer::interapp_buffer);

#ifdef _WIN32
    DWORD threadID_messages;
    SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
    CreateThread(&sa, 0, ConceptServer::CGIThread, (LPVOID) this, 0, &threadID_messages);
#else
    pthread_t threadID_messages;
    pthread_create(&threadID_messages, NULL, ConceptServer::CGIThread, (LPVOID) this);
#endif

    return(NO_WEB_ERROR);
}

int ConceptServer::Start(ServiceLoader sl) {
#ifdef _WIN32
    DWORD threadID_messages;
    SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
    CreateThread(&sa, 0, ConceptServer::InterApplicationProcessor, (LPVOID) this, 0, &threadID_messages);
#else
    pthread_t threadID_messages;
    pthread_create(&threadID_messages, NULL, ConceptServer::InterApplicationProcessor, (LPVOID) this);
#endif

    threadID = 0;
    // trec pe listen ... astept conexiuni
    if (listen(ListenSocket, ConceptServer::max_connections) == SOCKET_ERROR)
        return(ERR_LISTEN_SOCKET);

    if (sl) {
        int index = 0;
        do {
            AnsiString serv = sl(++index);
            if (serv.Length())
                RunService(serv.c_str());
            else
                break;
        } while (true);
    }

    SOCKET sock = this->ListenSocket;
#ifdef _WIN32
    int addr_length = sizeof(sin);
#else
    socklen_t addr_length = sizeof(sin);
#endif

#ifdef _WIN32
    DWORD threadID;
#else
    //pthread *threadID;
    pthread_t threadID;
#endif
    while (1) {
        SOCKET ClientSocket = accept(sock, (sockaddr *)&sin, &addr_length);
        //SOCKET *newSOCK=new SOCKET;
        //*newSOCK=ClientSocket;

/*#ifdef _WIN32
 *      std::cout    << "Client "
 *                      << (long)ClientSocket
 *                      << " connected ("
 *                      << (int)sin.sin_addr.S_un.S_un_b.s_b1 << "."
 *                      << (int)sin.sin_addr.S_un.S_un_b.s_b2 << "."
 *                      << (int)sin.sin_addr.S_un.S_un_b.s_b3 << "."
 *                      << (int)sin.sin_addr.S_un.S_un_b.s_b4
 *                      << ")\n";
 * #else
 *      // TODO !!
 * #endif
 */
        // creez threadul clientului
        ClientParamList *cpl = new ClientParamList;
        cpl->socket = ClientSocket;
        cpl->iscgi  = false;

#ifdef _WIN32
        SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
        CreateThread(&sa, 0, ConceptServer::ClientThread, (LPVOID)/*ClientSocket*/ cpl, 0, &threadID);
#else
        pthread_create(&threadID, NULL, ConceptServer::ClientThread, (LPVOID)/*ClientSocket*/ cpl);
#endif
        //ClientThread((LPVOID)newSOCK);
#ifdef _WIN32
        Sleep(100);
#else
        usleep(100000);
#endif
    }
    return(NO_WEB_ERROR);
}

int ConceptServer::RunConceptApplication(char *filename, char *arguments, SOCKET sock, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipein, int pipeout, INTERCOM_ELEMENT *apid, int parent_apid, char *host_root, char *server_name) {
    HMODULE          hDLL;
    API_INTERPRETER2 Concept_Execute;
    char             *def_arg = "0";

    if ((!arguments) || (!arguments[0]))
        arguments = def_arg;

    if (ConceptServer::SAFE_MODE) {
        AnsiString buf2 = "CONCEPT_FILENAME=";
        buf2 += filename;
        AnsiString buf3 = "CONCEPT_HOSTROOT=";
        buf3 += host_root;
        AnsiString buf = "CONCEPT_SERVERNAME=";
        if (server_name)
            buf += server_name;
        else
            buf += "unknown";
        AnsiString buf4 = "REMOTE_IP=";
        int        port = 0;
        buf4 += getip(sock, &port);
        AnsiString buf5 = "REMOTE_PORT=";
        buf5 += AnsiString((long)port);
        char **env = (char **)malloc((environ_top + 6) * sizeof(char *));
        if (env) {
            memcpy(env, enstr, environ_top * sizeof(char *));
            env[environ_top]     = buf.c_str();
            env[environ_top + 1] = buf2.c_str();
            env[environ_top + 2] = buf3.c_str();
            env[environ_top + 3] = buf4.c_str();
            env[environ_top + 4] = buf5.c_str();
            env[environ_top + 5] = 0;
        }

        AnsiString temp1;
#ifdef _WIN32
        temp1 = "\"";
#endif
        temp1 += AnsiString(SERVER_PUBLIC_KEY).Replace(" ", "%20");

#ifdef _WIN32
        temp1 += "\"";
#endif

        AnsiString temp3;
#ifdef _WIN32
        temp3 = "\"";
#endif
        temp3 += AnsiString(SERVER_PRIVATE_KEY).Replace(" ", "%20");
#ifdef _WIN32
        temp3 += "\"";
#endif

        AnsiString temp4;
#ifdef _WIN32
        temp4 = "\"";
#endif
        temp4 += AnsiString(CLIENT_PUBLIC_KEY).Replace(" ", "%20");
#ifdef _WIN32
        temp4 += "\"";
#endif

        int result = 0;

#ifdef _WIN32
        AnsiString fname = "\"";
        fname += host_root;
        fname += filename;
        fname += "\"";

        AnsiString INCLUDE = "\"";
        INCLUDE += ConceptServer::Include.c_str();
        if (INCLUDE[(int)INCLUDE.Length() - 1] == '\\')
            INCLUDE += "\\";
        INCLUDE += "\"";

        AnsiString LIBR = "\"";
        LIBR += ConceptServer::Library.c_str();
        if (LIBR[(int)LIBR.Length() - 1] == '\\')
            LIBR += "\\";
        LIBR += "\"";

        /*STARTUPINFO si;
         * PROCESS_INFORMATION pi;
         *
         *  memset(&pi, 0, sizeof(pi));
         *  memset(&si, 0, sizeof(si));
         *  si.cb = sizeof(si);
         *
         *
         * AnsiString command_line="0 ";//="\"";
         *
         * int direct_pipe[2];
         * pipe(direct_pipe);
         *
         *
         * //command_line+=ConceptServer::SafeModePath;
         * //command_line+=(char *)"\" 0 ";
         * if (!SetHandleInformation((HANDLE)sock, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT)) {
         *  SocketSay("Error in setting 'inherit' flag for client socket");
         * }
         *
         * if (!SetHandleInformation((HANDLE)direct_pipe[0], HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT)) {
         *  SocketSay("Error in setting 'inherit' flag");
         * }
         *
         * command_line+=AnsiString((long)direct_pipe[0]).c_str();
         * command_line+=(char *)" ";
         * command_line+=AnsiString((long)pipein).c_str();
         * command_line+=(char *)" ";
         * command_line+=AnsiString((long)pipeout).c_str();
         * command_line+=(char *)" ";
         * command_line+=AnsiString((long)apid->APID).c_str();
         * command_line+=(char *)" ";
         * command_line+=AnsiString((long)parent_apid).c_str();
         * command_line+=(char *)" ";
         * command_line+=(AnsiString("./")+fname).c_str();
         * command_line+=(char *)" ";
         * command_line+=AnsiString((long)sock).c_str();
         * command_line+=(char *)" ";
         * command_line+=INCLUDE.c_str();
         * command_line+=(char *)" ";
         * command_line+=LIBR.c_str();
         * if (temp1.Length()) {
         *  command_line+=(char *)" ";
         *  command_line+=temp1.c_str();
         *  if (temp3.Length()) {
         *      command_line+=(char *)" ";
         *      command_line+=temp3.c_str();//SERVER_PRIVATE_KEY,
         *      if (temp4.Length()) {
         *          command_line+=(char *)" ";
         *          command_line+=temp4.c_str();//CLIENT_PUBLIC_KEY
         *      }
         *  }
         * }
         *
         * if (!CreateProcess(ConceptServer::SafeModePath.c_str(),command_line.c_str(),NULL,NULL,
         *      TRUE, // inherit handles
         *      CREATE_SUSPENDED,NULL,NULL,&si,&pi) ){
         *          fprintf(stderr,"createprocess failed %s: %d\n",ConceptServer::SafeModePath.c_str(),GetLastError());
         *  return -1;
         * }
         * WSAPROTOCOL_INFO wp;
         *
         *
         * int err_ret;
         * if (err_ret=WSADuplicateSocket(sock,pi.dwProcessId,&wp)) {
         *  fprintf(stderr,"duplicate socket error %d\n",err_ret);
         *  return -1;
         * } else {
         *  if (write(direct_pipe[1],&wp, sizeof(wp))!=sizeof(wp)) {
         *      SocketSay("Error in IPC: sent WSAPROTOCOL_INFO size is wrong");
         *  }
         *  std::cerr << "Direct pipe[0]: " << direct_pipe[0] << "\n";
         * }
         *
         * if (ResumeThread(pi.hThread)==-1) {
         *  fprintf(stderr,"error resuming process %d\n",GetLastError());
         *  return -1;
         * }
         * //
         * // On Windows 95, the parent needs to wait until the child
         * // is done with the duplicated handle before closing it.
         * //
         *
         * WaitForSingleObject(pi.hProcess, INFINITE);
         * DWORD res;
         * close(direct_pipe[0]);
         * close(direct_pipe[1]);
         *
         * GetExitCodeProcess(pi.hProcess,&res);
         * result=res;
         *
         * CloseHandle(pi.hProcess);
         * CloseHandle(pi.hThread);*/

        // 128 bit AES
        int direct_pipe[2];
        pipe(direct_pipe);

        int direct_pipe_w[2];
        pipe(direct_pipe_w);

        result = spawnle(_P_NOWAIT,
                         ConceptServer::SafeModePath.c_str(),
                         "0",
                         AnsiString((long)direct_pipe[0]).c_str(),
                         AnsiString((long)pipein).c_str(),
                         AnsiString((long)pipeout).c_str(),
                         AnsiString((long)apid->APID).c_str(),
                         AnsiString((long)parent_apid).c_str(),
                         (/*AnsiString(host_root)+*/ fname).c_str(),
                         AnsiString((long)sock).c_str(), //sock
                         INCLUDE.c_str(),
                         LIBR.c_str(),
                         temp1.c_str(),
                         temp3.c_str(), //SERVER_PRIVATE_KEY,
                         temp4.c_str(), //CLIENT_PUBLIC_KEY
                         arguments,
                         0,
                         env
                         );

        if (write(direct_pipe[1], &direct_pipe_w[1], sizeof(int)) != sizeof(int))
            SocketSay(sock, "Error in IPC, write:  'writeable pipe' size is wrong");
        int procid = (int)GetCurrentProcessId();
        if (read(direct_pipe_w[0], &procid, sizeof(int)) != sizeof(int))
            SocketSay(sock, "Error in IPC, read:  'process id' size is wrong");
        close(direct_pipe_w[0]);
        close(direct_pipe_w[1]);

        if (result) {
            int err_ret;
            WSAPROTOCOL_INFO wp;
            apid->processID = result;

            if (err_ret = WSADuplicateSocket(sock, procid, &wp)) {
                fprintf(stderr, "duplicate socket error %d, process: %d\n", err_ret, procid);
                if (env)
                    free(env);
                return(-1);
            } else if (write(direct_pipe[1], &wp, sizeof(wp)) != sizeof(wp))
                SocketSay(sock, "Error in IPC: sent WSAPROTOCOL_INFO size is wrong");


            WaitForSingleObject((HANDLE)result, INFINITE);
            apid->processID = 0;
            DWORD res;
            GetExitCodeProcess((HANDLE)result, &res);
            result = res;

            if (result == -3)
                SocketSay(sock, AnsiString("File not found or insufficient rights : ") + filename);

            if (result == 101)
                SocketSay(sock, "Application failed to pass a check point. Infinite loop ?");
        }
        close(direct_pipe[0]);
        close(direct_pipe[1]);

        //std::cout << "File not found or insufficient rights : " << filename << "\n";
#else
        AnsiString fname;
        fname += host_root;
        fname += filename;

        pid_t pid = fork();
        if (pid < 0) {
            if (env)
                free(env);
            SocketSay(sock, "Error in forking");
            return(-1);
        }
        if (!pid)
            //result =
            execle(ConceptServer::SafeModePath.c_str(),
                   "0",
                   "no_need_for_pluming_on_penguin",
                   AnsiString((long)pipein).c_str(),
                   AnsiString((long)pipeout).c_str(),
                   AnsiString((long)apid->APID).c_str(),
                   AnsiString((long)parent_apid).c_str(),
                   //(/*AnsiString(host_root)+*/filename).c_str(),
                   fname.c_str(),
                   AnsiString((long)sock).c_str(),
                   ConceptServer::Include.c_str(),
                   ConceptServer::Library.c_str(),
                   temp1.c_str(),
                   temp3.c_str(),
                   temp4.c_str(),
                   arguments,
                   0,
                   env
                   //SERVER_PRIVATE_KEY,
                   //CLIENT_PUBLIC_KEY
                   );
        else {
            //waitpid (pid, NULL, 0);
            int status;

            apid->processID = pid;
            waitpid(pid, &status, 0);
            apid->processID = 0;
            result          = (char)WEXITSTATUS(status);

            /*if ((waitpid(pid, &status, 0) == pid) && (WIFEXITED(status))) {
             *  result = (char)WEXITSTATUS(status);
             * } else {
             *  result = -1;
             * }*/
        }
#endif
        if (env) {
            free(env);
            env = 0;
        }

        return(result);
    }

#ifdef _WIN32
    hDLL = LoadLibrary(LIBRARY);
#else
    // linux module
    hDLL = dlopen(LIBRARY_PATH, RTLD_LAZY);
#endif

    if (!hDLL) {
        SocketSay(sock, AnsiString("Error loading core library : ") + LIBRARY);
        //std::cout << "Error loading core library '" << LIBRARY << "'\r\n";
        return(-1);
    }


#ifdef _WIN32
    Concept_Execute = (API_INTERPRETER2)GetProcAddress(hDLL, "Concept_Execute");
#else
    Concept_Execute = (API_INTERPRETER2)dlsym(hDLL, "Concept_Execute");
#endif
    if (!Concept_Execute) {
        SocketSay(sock, AnsiString("Corrupted library : ") + LIBRARY);
        //std::cout << "Corrupted library : '" << LIBRARY << "'\r\n";
#ifdef _WIN32
        FreeLibrary(hDLL);
#else
        dlclose(hDLL);
#endif
        return(-1);
    }

    int res = Concept_Execute(AnsiString(host_root) + filename, ConceptServer::Include, ConceptServer::Library, (void *)ConceptServer::Print, sock, 0, 0, 0, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY, pipein, pipeout, apid->APID, parent_apid, 0);
    if (res == -3)
        SocketSay(sock, AnsiString("File not found or insufficient rights : ") + filename);
    //std::cout << "File not found or insufficient rights : " << filename << "\n";
#ifdef _WIN32
    FreeLibrary(hDLL);
#else
    dlclose(hDLL);
#endif
    return(res);
}

SOCKET ConceptServer::CONTACT_DEBUGER(char *hostname, int port) {
    struct sockaddr_in sin;
    struct hostent     *hp;
    SOCKET             DEBUG_SOCKET = 0;

    if ((hp = gethostbyname(hostname)) == 0)
        return(0);

    memset(&sin, 0, sizeof(sin));
    //INADDR_ANY;
    sin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    sin.sin_family      = AF_INET;
    sin.sin_port        = htons(port);

    // creez socket-ul
    if ((DEBUG_SOCKET = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == INVALID_SOCKET)
        return(0);

    if (connect(DEBUG_SOCKET, (struct sockaddr *)&sin, sizeof(sin)) == INVALID_SOCKET)
        return(0);

    return(DEBUG_SOCKET);
}

int ConceptServer::RunService(char *filename, int pipein, int pipeout, INTERCOM_ELEMENT *apid, int parent_apid, char *host_root) {
    HMODULE          hDLL;
    API_INTERPRETER2 Concept_Execute;

    if (ConceptServer::Service) {
        int result = 0;

#ifdef _WIN32
        AnsiString fname = "\"";
        fname += host_root;
        fname += filename;
        fname += "\"";

        AnsiString INCLUDE = "\"";
        INCLUDE += ConceptServer::Include.c_str();
        if (INCLUDE[(int)INCLUDE.Length() - 1] == '\\')
            INCLUDE += "\\";
        INCLUDE += "\"";

        AnsiString LIBR = "\"";
        LIBR += ConceptServer::Library.c_str();
        if (LIBR[(int)LIBR.Length() - 1] == '\\')
            LIBR += "\\";
        LIBR += "\"";

        int direct_pipe[2];
        pipe(direct_pipe);

        /*int direct_pipe_w[2];
         * pipe(direct_pipe_w);*/

        result = spawnl(_P_NOWAIT,
                        ConceptServer::Service.c_str(),
                        fname.c_str(),
                        INCLUDE.c_str(),
                        LIBR.c_str(),
                        "0", //AnsiString((long)direct_pipe[0]).c_str(),
                        AnsiString((long)pipein).c_str(),
                        AnsiString((long)pipeout).c_str(),
                        AnsiString((long)apid->APID).c_str(),
                        AnsiString((long)parent_apid).c_str(),
                        0
                        );

        /*if (write(direct_pipe[1],&direct_pipe_w[1], sizeof(int))!=sizeof(int)) {
         *  SocketSay(0,"Error in IPC, write:  'writeable pipe' size is wrong");
         * }
         * int procid=(int)GetCurrentProcessId();
         * if (read(direct_pipe_w[0],&procid, sizeof(int))!=sizeof(int)) {
         *  SocketSay(0,"Error in IPC, read:  'process id' size is wrong");
         * }
         * close(direct_pipe_w[0]);
         * close(direct_pipe_w[1]);*/

        if (result) {
            int err_ret;

            apid->processID = result;
            WaitForSingleObject((HANDLE)result, INFINITE);
            apid->processID = 0;
            DWORD res;
            GetExitCodeProcess((HANDLE)result, &res);
            result = res;
            SocketSay(0, AnsiString("Service ended : ") + filename);

            if (result == -3)
                SocketSay(0, AnsiString("Service not found or insufficient rights : ") + fname);

            if (result == 101)
                SocketSay(0, "Service failed to pass a check point. Infinite loop ?");
        }
        close(direct_pipe[0]);
        close(direct_pipe[1]);

        //std::cout << "File not found or insufficient rights : " << filename << "\n";
#else
        int direct_pipe[2];
        pipe(direct_pipe);

        AnsiString fname;
        fname += host_root;
        fname += filename;

        pid_t pid = fork();
        if (pid < 0)
            return(-1);
        if (!pid)
            //result =
            execl(ConceptServer::Service.c_str(),
                  fname.c_str(),
                  ConceptServer::Include.c_str(),
                  ConceptServer::Library.c_str(),
                  AnsiString((long)direct_pipe[0]).c_str(),
                  AnsiString((long)pipein).c_str(),
                  AnsiString((long)pipeout).c_str(),
                  AnsiString((long)apid->APID).c_str(),
                  AnsiString((long)parent_apid).c_str(),
                  0
                  );
        else {
            //waitpid (pid, NULL, 0);
            int status;

            apid->processID = pid;
            waitpid(pid, &status, 0);
            apid->processID = 0;
            result          = (char)WEXITSTATUS(status);
            SocketSay(0, AnsiString("Service ended : ") + filename);

            /*if ((waitpid(pid, &status, 0) == pid) && (WIFEXITED(status))) {
             *  result = (char)WEXITSTATUS(status);
             * } else {
             *  result = -1;
             * }*/
        }
        close(direct_pipe[0]);
        close(direct_pipe[1]);
#endif
        return(result);
    }

#ifdef _WIN32
    hDLL = LoadLibrary(LIBRARY);
#else
    // linux module
    hDLL = dlopen(LIBRARY_PATH, RTLD_LAZY);
#endif

    if (!hDLL) {
        SocketSay(0, AnsiString("Error loading core library : ") + LIBRARY);
        //std::cout << "Error loading core library '" << LIBRARY << "'\r\n";
        return(-1);
    }


#ifdef _WIN32
    Concept_Execute = (API_INTERPRETER2)GetProcAddress(hDLL, "Concept_Execute");
#else
    Concept_Execute = (API_INTERPRETER2)dlsym(hDLL, "Concept_Execute");
#endif
    if (!Concept_Execute) {
        SocketSay(0, AnsiString("Corrupted library : ") + LIBRARY);
        //std::cout << "Corrupted library : '" << LIBRARY << "'\r\n";
#ifdef _WIN32
        FreeLibrary(hDLL);
#else
        dlclose(hDLL);
#endif
        return(-1);
    }
}

#ifndef _WIN32
extern char **environ;
#endif

void ConceptServer::InheritEnv() {
#ifdef _WIN32
    char **next = _environ;
#else
    char **next = environ;
#endif
    while (*next) {
        environ_top++;
        environ_base++;
        enstr = (char **)realloc(enstr, (environ_top + 1) * sizeof(char *));
        if (enstr) {
            enstr[environ_top - 1] = *next;
            //std::cerr << "IENV: " << *next << "\n";
            enstr[environ_top] = 0;
        }
        next++;
    }
}

void ConceptServer::SetEnvVar(char *varname, char *value) {
    AnsiString my_str = varname;

    my_str += (char *)"=";
    my_str += value;

    environ_top++;
    enstr = (char **)realloc(enstr, (environ_top + 1) * sizeof(char *));
    if (enstr) {
        enstr[environ_top] = 0;
        int  len  = my_str.Length();
        char *buf = (char *)malloc(len + 1);
        if (buf) {
            buf[len] = 0;
            memcpy(buf, my_str.c_str(), len);
            enstr[environ_top - 1] = buf;
            //std::cerr << "ENV: " << buf << "\n";
        } else
            enstr[environ_top - 1] = 0;
    }

/*#ifdef _WIN32
 *  AnsiString my_str=varname;
 *  my_str+=(char *)"=";
 *  my_str+=value;
 *  _putenv(my_str.c_str());
 * #else
 *  setenv(varname, value, 1);
 #endif*/
}

int ConceptServer::RunDebugConceptApplication(char *filename, char *arguments, SOCKET sock, int pipein, int pipeout, INTERCOM_ELEMENT *apid, int parent_apid, char *host_root, char *server_name) {
    HMODULE          hDLL;
    API_INTERPRETER2 Concept_Execute;
    char             *def_arg = "0";

    if ((!arguments) || (!arguments[0]))
        arguments = def_arg;

    if (ConceptServer::SAFE_MODE) {
        AnsiString buf2 = "CONCEPT_FILENAME=";
        buf2 += filename;
        AnsiString buf3 = "CONCEPT_HOSTROOT=";
        buf3 += host_root;
        AnsiString buf = "CONCEPT_SERVERNAME=";
        if (server_name)
            buf += server_name;
        else
            buf += "unknown";
        AnsiString buf4 = "REMOTE_IP=";
        int        port = 0;
        buf4 += getip(sock, &port);
        AnsiString buf5 = "REMOTE_PORT=";
        buf5 += AnsiString((long)port);
        char **env = (char **)malloc((environ_top + 6) * sizeof(char *));
        if (env) {
            memcpy(env, enstr, environ_top * sizeof(char *));
            env[environ_top]     = buf.c_str();
            env[environ_top + 1] = buf2.c_str();
            env[environ_top + 2] = buf3.c_str();
            env[environ_top + 3] = buf4.c_str();
            env[environ_top + 4] = buf5.c_str();
            env[environ_top + 5] = 0;
        }

        int result = 0;

#ifdef _WIN32
        AnsiString fname = "\"";
        fname += host_root;
        fname += filename;
        fname += "\"";

        /*AnsiString fname="\"";
         * fname+=filename;
         * fname+="\"";*/

        AnsiString INCLUDE = "\"";
        INCLUDE += ConceptServer::Include.c_str();
        if (INCLUDE[(int)INCLUDE.Length() - 1] == '\\')
            INCLUDE += "\\";
        INCLUDE += "\"";

        AnsiString LIBR = "\"";
        LIBR += ConceptServer::Library.c_str();
        if (LIBR[(int)LIBR.Length() - 1] == '\\')
            LIBR += "\\";
        LIBR += "\"";

        int direct_pipe[2];
        pipe(direct_pipe);

        int direct_pipe_w[2];
        pipe(direct_pipe_w);

        result = spawnle(_P_NOWAIT,
                         ConceptServer::SafeModePath.c_str(),
                         "1",
                         AnsiString((long)direct_pipe[0]).c_str(),
                         AnsiString((long)pipein).c_str(),
                         AnsiString((long)pipeout).c_str(),
                         AnsiString((long)apid->APID).c_str(),
                         AnsiString((long)parent_apid).c_str(),
                         //(AnsiString(host_root)+fname).c_str(),
                         fname.c_str(),
                         AnsiString((long)sock).c_str(),
                         INCLUDE.c_str(),
                         LIBR.c_str(),
                         arguments,
                         0,
                         env
                         );

        // 128 bit AES

        if (write(direct_pipe[1], &direct_pipe_w[1], sizeof(int)) != sizeof(int))
            SocketSay(sock, "Error in IPC, write:  'writeable pipe' size is wrong");
        int procid = (int)GetCurrentProcessId();
        if (read(direct_pipe_w[0], &procid, sizeof(int)) != sizeof(int))
            SocketSay(sock, "Error in IPC, read:  'process id' size is wrong");
        close(direct_pipe_w[0]);
        close(direct_pipe_w[1]);

        if (result) {
            int err_ret;
            WSAPROTOCOL_INFO wp;
            apid->processID = result;

            if (err_ret = WSADuplicateSocket(sock, procid, &wp)) {
                fprintf(stderr, "duplicate socket error %d, process: %d\n", err_ret, procid);
                if (env)
                    free(env);
                return(-1);
            } else if (write(direct_pipe[1], &wp, sizeof(wp)) != sizeof(wp))
                SocketSay(sock, "Error in IPC: sent WSAPROTOCOL_INFO size is wrong");

            WaitForSingleObject((HANDLE)result, INFINITE);
            apid->processID = 0;
            DWORD res;
            GetExitCodeProcess((HANDLE)result, &res);
            result = res;
            if (result == -3)
                SocketSay(sock, AnsiString("File not found or insufficient rights : ") + filename);

            if (result == 101)
                SocketSay(sock, "Application failed to pass a check point. Infinite loop ?");
        }
        close(direct_pipe[0]);
        close(direct_pipe[1]);

#else
        AnsiString fname;
        fname += host_root;
        fname += filename;

        pid_t pid = fork();
        if (pid < 0) {
            if (env)
                free(env);
            SocketSay(sock, "Error in forking");
            return(-1);
        }

        if (!pid)
            //result=
            execl(ConceptServer::SafeModePath.c_str(),
                  "1",
                  "no_need_for_pluming_on_penguin",
                  AnsiString((long)pipein).c_str(),
                  AnsiString((long)pipeout).c_str(),
                  AnsiString((long)apid->APID).c_str(),
                  AnsiString((long)parent_apid).c_str(),
                  //(AnsiString(host_root)+filename).c_str(), //fname.c_str(),
                  fname.c_str(),
                  AnsiString((long)sock).c_str(),
                  ConceptServer::Include.c_str(),
                  ConceptServer::Library.c_str(),
                  arguments,
                  0,
                  env
                  );
        else {
            int status;

            apid->processID = pid;
            waitpid(pid, &status, 0);
            result          = (char)WEXITSTATUS(status);
            apid->processID = 0;
        }
#endif
        if (env) {
            free(env);
            env = 0;
        }
        return(result);
    }

/*
 * #ifdef _WIN32
 *  hDLL=LoadLibrary(LIBRARY);
 * #else
 *  // linux module
 *  hDLL=dlopen(LIBRARY_PATH, RTLD_LAZY);
 * #endif
 *  if (!hDLL) {
 *     std::cout << "Error loading core library '" << LIBRARY << "'\r\n";
 *     return -1;
 *  }
 *
 * #ifdef _WIN32
 *  Concept_Execute=(API_INTERPRETER)GetProcAddress(hDLL,"Concept_Execute");
 * #else
 *  Concept_Execute=(API_INTERPRETER)dlsym(hDLL,"Concept_Execute");
 * #endif
 *  if (!Concept_Execute) {
 *     std::cout << "Corrupted library : '" << LIBRARY << "'\r\n";
 * #ifdef _WIN32
 *     FreeLibrary(hDLL);
 * #else
 *     dlclose(hDLL);
 * #endif
 *     return -1;
 *  }
 *
 *  SOCKET DEBUG_SOCKET    =   ConceptServer::CONTACT_DEBUGER(DEBUG_HOST,DEBUG_PORT);
 *
 *  if ((!DEBUG_SOCKET) || (DEBUG_SOCKET==INVALID_SOCKET))
 *      return ConceptServer::RunConceptApplication(filename,sock,0,0,0,pipein,pipeout,apid,parent_apid);
 *
 *  ClientDebugInfo DINFO;
 *  DINFO.CLIENT_NOTICE    =   0;
 *  DINFO.last_line        =   0;
 *  DINFO.sock             =   sock;
 *  DINFO.debug_type       =   DEBUG_STEPINTO;
 *  DINFO.run_to_line      =   0;
 *  DINFO.FileName         =   filename;
 *  DINFO.context_id       =   0;
 *  DINFO.DEBUG_SOCKET     =   DEBUG_SOCKET;
 *
 *  int res=Concept_Execute(filename, ConceptServer::Include, ConceptServer::Library, (void *)ConceptServer::Print,sock,1,ConceptServer::DEBUGGER_TRAP,&DINFO, 0, 0, 0, pipein, pipeout, apid, parent_apid);
 *  // inchid conexiunea cu debugger-ul
 * #ifdef _WIN32
 *  closesocket(DEBUG_SOCKET);
 * #else
 *  close(DEBUG_SOCKET);
 * #endif
 *  if (res==-3)
 *     std::cout << "File not found or insufficient rights : " << filename << "\r\n";
 * #ifdef _WIN32
 *  FreeLibrary(hDLL);
 * #else
 *  dlclose(hDLL);
 * #endif
 *  return res;*/
}

#ifdef _WIN32
DWORD WINAPI ConceptServer::RunService_Thread(LPVOID data) {
#else
LPVOID ConceptServer::RunService_Thread(LPVOID data) {
#endif

    char *service          = (char *)data;
    int  parent_apid       = -1;
    INTERCOM_ELEMENT *apid = 0;
    int pipein             = 0;
    int pipeout            = 0;

    apid = ((ConceptServer *)this_reference)->AddAPID(&pipein, &pipeout, hash(service));
    if (!apid)
        return(0);
    if (((ConceptServer *)this_reference)->FIRST_SAFE_APID <= apid->APID)
        ((ConceptServer *)this_reference)->FIRST_SAFE_APID = apid->APID + 1;

    RunService(service, pipein, pipeout, apid, parent_apid, ConceptServer::ServicePath.c_str());
    delete[] service;
    // adaugat de curand
    ((ConceptServer *)this_reference)->RemoveAPID(apid);
}

int ConceptServer::RunService(char *service) {
    int  len            = strlen(service);
    char *service_clone = new char[len + 1];

    service_clone[len] = 0;
    strncpy(service_clone, service, len);

#ifdef _WIN32
    DWORD threadID;
    SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
    CreateThread(&sa, 0, ConceptServer::RunService_Thread, (LPVOID)service_clone, 0, &threadID);
#else
    pthread_t threadID;
    pthread_create(&threadID, NULL, ConceptServer::RunService_Thread, (LPVOID)service_clone);
#endif
    return(0);
}

#ifdef _WIN32
DWORD WINAPI ConceptServer::ClientThread(LPVOID socket) {
#else
LPVOID ConceptServer::ClientThread(LPVOID socket) {
#endif
    int parent_apid        = -1;
    INTERCOM_ELEMENT *apid = 0;
    int pipein             = 0;
    int pipeout            = 0;

    char *buffer = new char[MAX_BUFFER + 1];
    buffer[MAX_BUFFER] = 0;

    AnsiString      SERVER_PUBLIC_KEY;
    AnsiString      SERVER_PRIVATE_KEY;
    AnsiString      CLIENT_PUBLIC_KEY;
    ClientParamList *cpl    = (ClientParamList *)socket;
    SOCKET          _socket = (SOCKET)cpl->socket;
    bool            is_cgi  = cpl->iscgi;
    delete cpl;

    if (is_cgi)
        SocketSay(_socket, "CGI Connected");
    else
        SocketSay(_socket, "Connected");

    int flag = 1;

#ifdef SAFE_RECV_TIMEOUTS
 #ifdef _WIN32
    int tv_infinite = 0;
    //if (!is_cgi) {
    int max_idle = (is_cgi ? ConceptServer::web_timeout : ConceptServer::max_idle) * 1000;
    setsockopt(_socket, SOL_SOCKET, SO_RCVTIMEO, (char *)&max_idle, sizeof(max_idle));
    //}
 #else
    struct timeval tv_infinite;
    //if (!is_cgi) {
    socklen_t optLen = 0;

    if (getsockopt(_socket, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv_infinite, &optLen) == -1)
        tv_infinite.tv_sec = 0x7FFFFFFF;

    //tv_infinite.tv_sec=0;
    //tv_infinite.tv_usec=0;

    struct timeval tv;

    tv.tv_sec  = is_cgi ? ConceptServer::web_timeout : ConceptServer::max_idle;
    tv.tv_usec = 0;

    setsockopt(_socket, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv, sizeof(tv));
    //}
 #endif
#endif

#ifdef FAST_TCP
    setsockopt(_socket,                           /* socket affected */
               IPPROTO_TCP,                       /* set option at TCP level */
               TCP_NODELAY,                       /* name of option */
               (char *)&flag,                     /* the cast is historical cruft */
               sizeof(int));                      /* length of option value */
#endif
    setsockopt(_socket,
               SOL_SOCKET,
               SO_KEEPALIVE,
               (char *)&flag,
               sizeof(int));
    SimpleStream SS(_socket);

    if (!is_cgi) {
        SERVER_PUBLIC_KEY.LoadFile(ConceptServer::CertificatePath + PUBLIC_KEY_FILE);
        SERVER_PRIVATE_KEY.LoadFile(ConceptServer::CertificatePath + PRIVATE_KEY_FILE);

        // trimit cheia publica a server-ului
        int cert_size = SERVER_PUBLIC_KEY.Length();
        if (cert_size) {
            int cert_size_n = htonl(cert_size);
            send(_socket, (char *)&cert_size_n, sizeof(int), 0);
            send(_socket, SERVER_PUBLIC_KEY.c_str(), cert_size, 0);
        } else {
            SocketSay(_socket, "Broken RSA Certificate. Disconnected.");
#ifdef _WIN32
            closesocket(_socket);
#else
            close(_socket);
#endif
            return(0);
        }

        // clientul o foloseste sa cripteze cheia lui AES
        cert_size = 0;
        int received = 0;
        int rec_tot  = 0;
        do {
            received = recv(_socket, (char *)&cert_size + rec_tot, sizeof(int) - rec_tot, 0);
            rec_tot += received;
        } while ((received > 0) && (rec_tot < sizeof(int)));
        cert_size = ntohl(cert_size);

        if ((received < 1) || (rec_tot != sizeof(int)) || (cert_size >= MAX_BUFFER)) {
            SocketSay(_socket, "Broken Client AES Certificate (invalid certificate size). Disconnected.");
#ifdef _WIN32
            closesocket(_socket);
#else
            close(_socket);
#endif

            return(0);
        }

        received = 0;
        rec_tot  = 0;
        do {
            received = recv(_socket, buffer + rec_tot, cert_size - rec_tot, 0);
            rec_tot += received;
        } while ((received > 0) && (rec_tot < cert_size));

        if ((received < 1) || (rec_tot != cert_size)) {
            SocketSay(_socket, "Broken Client AES Certificate (received less bytes than expected). Disconnected.");
#ifdef _WIN32
            closesocket(_socket);
#else
            close(_socket);
#endif
            return(0);
        }

        buffer[rec_tot] = 0;

        int  size         = rec_tot;
        char *out_content = new char[size];
        buffer[size] = 0;
        int decrypt_result = RSA_decrypt(buffer, size,
                                         out_content, size,
                                         SERVER_PRIVATE_KEY.c_str(), SERVER_PRIVATE_KEY.Length());
        out_content[decrypt_result] = 0;
        if (!decrypt_result) {
            delete[] out_content;
            SocketSay(_socket, "Broken Client AES Certificate (error on decrypt). Disconnected.");
#ifdef _WIN32
            closesocket(_socket);
#else
            close(_socket);
#endif
            return(0);
        }

        CLIENT_PUBLIC_KEY  = out_content;
        SERVER_PRIVATE_KEY = CLIENT_PUBLIC_KEY;
        delete[] out_content;
    }

    int          res;
    unsigned int *checksum_ptr = 0;

    //SocketSay(_socket,"Handshake OK");
    if ((this_reference) && (!is_cgi)) {
        apid = ((ConceptServer *)this_reference)->AddAPID(&pipein, &pipeout, 0, &checksum_ptr);
        SocketSay(_socket, "Running using APID: ", apid->APID);
    }

    AnsiString host_path;
    AnsiString req_host;

    if (!is_cgi)
        host_path = "./";

    int total_size;

    // for cgi only (named pipes)
    int        pipe_w = 0;
    AnsiString pipew_name;
    int        pipe_r = 0;
    AnsiString piper_name;

    while (1) {
        //send(_socket,COMMAND_WAIT,strlen(COMMAND_WAIT),0);
        char *baseptr = buffer;
        int  out      = 0;
        // initzializez buffer-ul cu 0;
        memset(buffer, 0, MAX_BUFFER);
        int rcsize = 1;
        total_size = 0;
        while ((rcsize > 0) && (!out)) {
            rcsize = recv(_socket, baseptr, 1, 0);
            //send(_socket,baseptr,rcsize,0);
            // mut pointer-ul cu rcsize octetzi
            if (rcsize > 0) {
                baseptr    += rcsize;
                total_size += rcsize;
            }
            if (total_size >= MAX_BUFFER)
                break;
             // am primit o comanda !!!
            if (strstr(buffer, "\r\n"))
                out = true;
            else
            if (strstr(buffer, "\n"))
                break;
        }
        if (rcsize < 1)
            break;
        // verific daca am GET pe prima pozitzie
        if (out) {
            if (is_cgi) {
                if (strstr(buffer, "pipew ") == buffer) { // set option value
                    char *option = buffer + 6;
                    if (!option)
                        break;
                    char *ptr = strstr(option, "\r");
                    if (ptr)
                        ptr[0] = 0;
                    pipe_w     = open(option, O_WRONLY | O_BINARY | O_CREAT, 0666);
                    pipew_name = option;
                    if (pipe_w == -1) {
                        LOG_DATA("CGI Write pipe failed to open.");
                        LOG_DATA(option);
                        pipe_w = 0;
                        break;
                    }
                } else
                if (strstr(buffer, "piper ") == buffer) { // set option value
                    char *option = buffer + 6;
                    if (!option)
                        break;
                    char *ptr = strstr(option, "\r");
                    if (ptr)
                        ptr[0] = 0;
                    pipe_r     = open(option, O_RDONLY | O_BINARY);
                    piper_name = option;
                    if (pipe_r == -1) {
                        LOG_DATA("CGI Read pipe failed to open.");
                        LOG_DATA(option);
                        pipe_r = 0;
                        break;
                    }
                } else
                if (strstr(buffer, "getapid") == buffer) { // set option value
                    if ((pipe_r) && (pipe_w)) {
                        apid = ((ConceptServer *)this_reference)->AddCGIAPID(pipe_r, pipe_w);
                        int netAPID = htonl((int)apid->APID);
                        send(_socket, (char *)&netAPID, sizeof(netAPID), 0);
                    } else {
                        if (!pipe_r)
                            LOG_DATA("Can't run if CGI Read pipe is not set");

                        if (!pipe_w)
                            LOG_DATA("Can't run if CGI Write pipe is not set");
                        break;
                    }
                } else
                if (strstr(buffer, "getparent ") == buffer) {
                    char *option = buffer + 10;
                    if (!option)
                        break;
                    char *ptr = strstr(option, "\r");
                    if (ptr)
                        ptr[0] = 0;

                    bool debug      = false;
                    int  parentAPID = ((ConceptServer *)this_reference)->GetHost(option, debug);
                    if (debug)
                        parentAPID *= -1;

                    int netAPID = htonl((int)parentAPID);
                    send(_socket, (char *)&netAPID, sizeof(netAPID), 0);
                } else
                if (strstr(buffer, "pipe ") == buffer) {
                    (baseptr - 2)[0] = 0;
                    char *ptr = strstr(buffer, "\r");
                    if (ptr)
                        ptr[0] = 0;
                    parent_apid = AnsiString(buffer + 5).ToInt();
                }
            } else
            if (strstr(buffer, "pipe ") == buffer) {
                (baseptr - 2)[0] = 0;
                char *ptr = strstr(buffer, "\r");
                if (ptr)
                    ptr[0] = 0;
                parent_apid = AnsiString(buffer + 5).ToInt();
                SocketSay(_socket, "Pluming set");
            } else
            if (strstr(buffer, "set ") == buffer) { // set option value
                char *option = buffer + 4;
                if (!option)
                    break;

                char *buf_2 = strstr(buffer + 4, " ");
                if (!buf_2)
                    break;
                buf_2[0] = 0;

                char *buf_3 = buf_2 + 1;

                char *buf_temp = strstr(buf_3, "\r\n");
                if (buf_temp)
                    buf_temp[0] = 0;

                if (strstr(option, "servername")) {
                    SocketSay(_socket, AnsiString("Client requested session from: ") + buf_3);
                    if ((!is_cgi) && (ConceptServer::hcbp)) {
                        req_host = buf_3;
                        AnsiString temp = ConceptServer::hcbp(buf_3);
                        if (temp.Length()) {
                            SocketSay(_socket, AnsiString("Using root: ") + temp);
                            host_path = temp;
                        }
                    } else
                        SocketSay(_socket, "Request denied - CGI applications cannot set servername");
                    continue;
                } else {
                    SocketSay(_socket, ((AnsiString("Invalid request from client '") + option) + "' with value '") + buf_3);
                    break;
                }
            } else
            if (strstr(buffer, "close") == buffer) {
                SocketSay(_socket, "Close request from client");
                break;
            } else
#ifndef NOLIST
            if (strstr(buffer, "list ") == buffer) {
                (baseptr - 2)[0] = 0;
                char *ptr = strstr(buffer, "\r");
                if (ptr)
                    ptr[0] = 0;
                AnsiString path = NormalizePath(buffer + 5);

                List(/*AnsiString(buffer+5)*/ path, _socket);
            } else
#endif
            if (strstr(buffer, "run ") == buffer) {
                (baseptr - 2)[0] = 0;
                char *ptr = strstr(buffer, "\r");
                if (ptr)
                    ptr[0] = 0;
                char *arguments = strstr(buffer, "?");
                if (arguments) {
                    arguments[0] = 0;
                    arguments++;
                }

#ifdef SAFE_RECV_TIMEOUTS
                if (!is_cgi)
                    setsockopt(_socket, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv_infinite, sizeof(tv_infinite));
#endif

                AnsiString path = NormalizePath(buffer + 4);
                SocketSay(_socket, AnsiString("Crypted run ") + path); //(buffer+4));
                if (checksum_ptr)
                    *checksum_ptr = hash((host_path + (char *)"/") + path);
                res = RunConceptApplication(/*buffer+4*/ path.c_str(), arguments, _socket, SERVER_PUBLIC_KEY.c_str(), SERVER_PRIVATE_KEY.c_str(), CLIENT_PUBLIC_KEY.c_str(), pipein, pipeout, apid, parent_apid, host_path.c_str(), req_host.c_str());
                SocketSay(_socket, AnsiString("End of crypted run ") + /*(buffer+4)*/ path);
                if (res) {
                    if (res == -3) {
                        //long length=strlen(ERR_APPLICATION_ERR);
                        //send(_socket,(char *)&length,sizeof(long),0);
                        //send(_socket,ERR_APPLICATION_ERR,length,0);
                        SS.ClientError(ERR_APPLICATION_ERR, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR);
                    } else
                    if (res == -2) {
                        //long length=strlen(ERR_APPLICATION_ERR1);
                        //send(_socket,(char *)&length,sizeof(long),0);
                        //send(_socket,ERR_APPLICATION_ERR1,length,0);
                        SS.ClientError(ERR_APPLICATION_ERR1, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR1);
                    } else
                    if (res == -1) {
                        //long length=strlen(ERR_APPLICATION_ERR2);
                        //send(_socket,(char *)&length,sizeof(long),0);
                        //send(_socket,ERR_APPLICATION_ERR2,length,0);
                        SS.ClientError(ERR_APPLICATION_ERR2, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR2);
                    } else
                    if (res == 101) {
                        SS.ClientError(ERR_APPLICATION_ERRFAILED, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERRFAILED);
                    }
                }
                break;
            } else
            if (strstr(buffer, "runfast ") == buffer) {
                (baseptr - 2)[0] = 0;
                char *ptr = strstr(buffer, "\r");
                if (ptr)
                    ptr[0] = 0;
                char *arguments = strstr(buffer, "?");
                if (arguments) {
                    arguments[0] = 0;
                    arguments++;
                }

                AnsiString path = NormalizePath(buffer + 8);
#ifdef SAFE_RECV_TIMEOUTS
                if (!is_cgi)
                    setsockopt(_socket, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv_infinite, sizeof(tv_infinite));
#endif
                SocketSay(_socket, AnsiString("Uncrypted run ") + /*(buffer+8)*/ path);
                if (checksum_ptr)
                    *checksum_ptr = hash((host_path + (char *)"/") + path);
                res = RunConceptApplication(/*buffer+8*/ path.c_str(), arguments, _socket, 0, 0, 0, pipein, pipeout, apid, parent_apid, host_path.c_str(), req_host.c_str());
                SocketSay(_socket, AnsiString("End of uncrypted run ") + /*(buffer+8)*/ path);
                if (res) {
                    if (res == -3) {
                        /*long length=strlen(ERR_APPLICATION_ERR);
                         * send(_socket,(char *)&length,sizeof(long),0);
                         * send(_socket,ERR_APPLICATION_ERR,length,0);*/
                        SS.ClientError(ERR_APPLICATION_ERR, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR);
                    } else
                    if (res == -2) {
                        /*long length=strlen(ERR_APPLICATION_ERR1);
                         * send(_socket,(char *)&length,sizeof(long),0);
                         * send(_socket,ERR_APPLICATION_ERR1,length,0);*/
                        SS.ClientError(ERR_APPLICATION_ERR1, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR1);
                    } else
                    if (res == -1) {
                        /*long length=strlen(ERR_APPLICATION_ERR2);
                         * send(_socket,(char *)&length,sizeof(long),0);
                         * send(_socket,ERR_APPLICATION_ERR2,length,0);*/
                        SS.ClientError(ERR_APPLICATION_ERR2, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR2);
                    }
                }
                break;
            } else
            if (strstr(buffer, "rundebug ") == buffer) {
                (baseptr - 2)[0] = 0;
                char *ptr = strstr(buffer, "\r");
                if (ptr)
                    ptr[0] = 0;
                char *arguments = strstr(buffer, "?");
                if (arguments) {
                    arguments[0] = 0;
                    arguments++;
                }

                AnsiString path = NormalizePath(buffer + 9);

#ifdef SAFE_RECV_TIMEOUTS
                if (!is_cgi)
                    setsockopt(_socket, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv_infinite, sizeof(tv_infinite));
#endif

                SocketSay(_socket, AnsiString("Debug ") + /*(buffer+9)*/ path);
                if ((pipein > 0) && (pipeout > 0) && (parent_apid > 0))
                    res = RunDebugConceptApplication(/*buffer+9*/ path.c_str(), arguments, _socket, pipein, pipeout, apid, parent_apid, host_path.c_str(), req_host.c_str());
                else {
                    SocketSay(_socket, AnsiString("Unable to debug, due to invalid data > ") + /*(buffer+9)*/ path.c_str());
                    res = RunConceptApplication(/*buffer+9*/ path.c_str(), arguments, _socket, 0, 0, 0, pipein, pipeout, apid, parent_apid, host_path.c_str(), req_host.c_str());
                }
                SocketSay(_socket, AnsiString("End of debug ") + (buffer + 9));
                if (res) {
                    if (res == -3) {
                        /*long length=strlen(ERR_APPLICATION_ERR);
                         * send(_socket,(char *)&length,sizeof(long),0);
                         *      send(_socket,ERR_APPLICATION_ERR,length,0);*/
                        SS.ClientError(ERR_APPLICATION_ERR, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR);
                    } else
                    if (res == -2) {
                        /*long length=strlen(ERR_APPLICATION_ERR1);
                         * send(_socket,(char *)&length,sizeof(long),0);
                         *      send(_socket,ERR_APPLICATION_ERR1,length,0);*/
                        SS.ClientError(ERR_APPLICATION_ERR1, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR1);
                    } else
                    if (res == -1) {
                        /*long length=strlen(ERR_APPLICATION_ERR2);
                         * send(_socket,(char *)&length,sizeof(long),0);
                         *      send(_socket,ERR_APPLICATION_ERR2,length,0);*/
                        SS.ClientError(ERR_APPLICATION_ERR2, -2);
                        SocketSay(_socket, ERR_APPLICATION_ERR2);
                    }
                }
                break;
            }
        } else
            break;

        if (!is_cgi) {
#ifdef _WIN32
            Sleep(100);
#else
            usleep(100000);
#endif
        }
    }
    delete buffer;
    /*************************/
    if (is_cgi)
        SocketSay(_socket, "CGI disconnected");
    else
        SocketSay(_socket, "Disconnected");
#ifdef _WIN32
    if (!is_cgi)
        Sleep(1000);
    closesocket(_socket);
#else
    if (!is_cgi)
        usleep(1000000);
    close(_socket);
#endif
    /*************************/

    //std::cout << "Client " << (long)_socket << " disconnected\r\n";
    if (apid)
        ((ConceptServer *)this_reference)->RemoveAPID(apid);
    else {
        if (pipe_r > 0)
            close(pipe_r);
        if (pipe_w > 0)
            close(pipe_w);
    }
    //delete (SOCKET *)socket;
    if (pipew_name.Length())
        unlink(pipew_name.c_str());
    if (piper_name.Length())
        unlink(piper_name.c_str());

    return(0);
}

#ifdef _WIN32
DWORD WINAPI ConceptServer::ClientAPM(LPVOID P) {
#else
LPVOID ConceptServer::ClientAPM(LPVOID P) {
#endif
    INTERCOM_ELEMENT *IEL = (INTERCOM_ELEMENT *)P;
    AnsiString       main_buffer;
    AnsiString       rec_buffer;
    bool             has_data;
    while (true) {
        has_data = false;
        SEMP(IEL->lock);
        if (IEL->done) {
            // clean exit
            IEL->done = -1;
            SEMV(IEL->lock);
            break;
        }
        if (IEL->inter_buf.Length()) {
            // == SECURITY FEATURE AGAINST FLOODING == //
            if (ConceptServer::interapp_buffer > 0)
                if (rec_buffer.Length() + IEL->inter_buf.Length() >= ConceptServer::interapp_buffer) {
                    IEL->inter_buf.LoadBuffer(0, 0);
                    has_data = false;
                    SEMV(IEL->lock);
                    continue;
                }
             // END OF SECURITY FEATURE AGAINST FLOODING //
            rec_buffer += IEL->inter_buf;
            IEL->inter_buf.LoadBuffer(0, 0);
            has_data = true;
        }
        //SEMV(IEL->lock);
        if (has_data) {
            main_buffer = rec_buffer;
            rec_buffer.LoadBuffer(0, 0);
            //SEMP(IEL->lock);
            int total_written = write(IEL->pipe_out[1], main_buffer, main_buffer.Length());
            SEMV(IEL->lock);
            main_buffer.LoadBuffer(0, 0);
            has_data = false;
        } else {
            SEMV(IEL->lock);
#ifdef _WIN32
            Sleep(10);
#else
            usleep(10000);
#endif
        }
    }
    return(0);
    //LOG_DATA("AP thread done");
}

int ConceptServer::inter_write(INTERCOM_ELEMENT *IEL, char *buffer, int size, int lock) {
    if (size <= 0)
        return(0);
    if (IEL->done)
        return(0);
    if (lock)
        SEMP(IEL->lock);
    IEL->inter_buf.AddBuffer(buffer, size);
    if (lock)
        SEMV(IEL->lock);

    return(size);
}

#ifndef NOLIST
int ConceptServer::List(char *path, SOCKET sock) {
    AnsiString Path = "./";

    Path += path; //=(path[0]=='/') || (path[0]=='\\') ? path+1 : path;
    AnsiString xml_index;
    Path += PUBLIC_PATH;
    // listare director
    xml_index.LoadFile(AnsiString(Path + "/") + PUBLIC_XML);
    SocketSay(sock, AnsiString("Listing path ") + Path);

    if (!xml_index.Length())
        SocketSay(sock, "Warning : index file is empty or missing.");
    int length = htonl(xml_index.Length());
    send(sock, (char *)&length, sizeof(int), 0);
    send(sock, xml_index.c_str(), xml_index.Length(), 0);

    if (!xml_index.Length())
        return(-1);

    DIR           *dip;
    struct dirent *dit;

    if ((dip = opendir(Path.c_str())) == NULL) {
        SocketSay(sock, AnsiString("Error opening path :") + Path);
        return(-1);
    }

    // il retin ca sa evit sa-l trimit de 2 ori
    AnsiString xml_name = PUBLIC_XML;
    while ((dit = readdir(dip)) != NULL) {
        struct stat buf;
        stat(dit->d_name, &buf);

        if (((buf.st_mode) & _S_IFMT) && (xml_name != dit->d_name)) {
            AnsiString filename = Path;
            filename += "/";
            filename += dit->d_name;
            AnsiString d_name = dit->d_name;

            AnsiString file_content;
            file_content.LoadFile(filename);
            if (file_content.Length()) {
                // trimit mai intai numele fisierului
                length = htonl(d_name.Length());
                send(sock, (char *)&length, sizeof(int), 0);
                send(sock, d_name.c_str(), d_name.Length(), 0);

                // trimit contzinutul
                length = htonl(file_content.Length());
                send(sock, (char *)&length, sizeof(int), 0);
                send(sock, file_content.c_str(), file_content.Length(), 0);
            }
        }
    }
    length = htonl(0);
    send(sock, (char *)&length, sizeof(int), 0);

    return(0);
}

#endif

int ConceptServer::Stop() {
    LOG_DATA("Stopping server and child processes ...");
    SEMP(sem_lock);
    int        processes = 0;
    AnsiString err;
    for (INTERCOM_ELEMENT *IEL = this->APPLICATIONS; IEL; IEL = (INTERCOM_ELEMENT *)IEL->NEXT) {
        if (IEL->APID) {
            if (IEL->processID) {
#ifdef _WIN32
                if (!TerminateProcess((HANDLE)IEL->processID, 0x80)) {
                    err  = (char *)"Error terminating process with pid ";
                    err += (long)IEL->processID;
                    err += (char *)", error code: ";
                    err += (long)GetLastError();
                    LOG_DATA(err.c_str());
                }
#else
                if (kill(IEL->processID, SIGKILL)) {
                    err  = (char *)"Error terminating process with pid ";
                    err += (long)IEL->processID;
                    err += (char *)", error code: ";
                    err += (long)errno;
                    LOG_DATA(err.c_str());
                }
#endif
                IEL->processID = 0;
                processes++;
            }
        }
    }
    SEMV(sem_lock);
    LOG_DATA("Done. Clean server shutdown.");
    return(0);
}

ConceptServer::~ConceptServer(void) {
    INTERCOM_ELEMENT *IEL = this->APPLICATIONS;

    while (IEL) {
        INTERCOM_ELEMENT *prec = IEL;
        IEL = (INTERCOM_ELEMENT *)IEL->NEXT;
        delete prec;
    }

    for (int i = environ_base; i < environ_top; i++) {
        char *ptr = enstr[i];
        if (ptr)
            free(ptr);
    }
    free(enstr);
    enstr        = 0;
    environ_top  = 0;
    environ_base = 0;
    std::cout << "Clean Concept Server shut down\n";
    semdel(sem_lock);
    semdel(sem_web_lock);
}
