//#define DEBUG

#include <iostream>
#include <fstream>
#include <time.h>
#include <stdlib.h>

#ifdef _WIN32
 #define _WINSOCKAPI_
 #include <windows.h>
 #include <winsvc.h>
 #include <direct.h>

 #define PATH_SLASH    "\\"
#else
 #define PATH_SLASH    "/"
 #include <unistd.h>
 #include <sys/types.h>
 #include <pwd.h>
 #include <signal.h>
#endif

#include "ConceptServer.h"

#define INI_FILE        "concept.ini"
#define SERVICE_NAME    "Concept II Server SVC"

ConceptServer *SERVER_REF = 0;

#ifndef _WIN32
long GetPrivateProfileString(
    const char *lpAppName,
    const char *lpKeyName,
    const char *lpDefault,
    char       *lpReturnedString,
    long       nSize,
    const char *lpFileName) {
    FILE *in         = 0;
    long filesize    = 0;
    char *filebuffer = 0;
    char section_name[4096];
    char key_name[4096];
    int  error = 0;

    in = fopen(lpFileName, "rb");
    if (in) {
        fseek(in, 0L, SEEK_END);
        filesize = ftell(in);
        fseek(in, 0L, SEEK_SET);
        if (filesize) {
            filebuffer = new char[filesize];
            if (filebuffer) {
                if (fread(filebuffer, 1, filesize, in) == filesize) {
                    char in_name           = 0;
                    char in_quote          = 0;
                    char in_looked_section = 0;
                    char in_looked_key     = 0;
                    char in_looked_value   = 0;
                    char in_comment        = 0;
                    char in_key            = 0;
                    int  temp_pos          = 0;
                    int  key_pos           = 0;
                    int  value_pos         = 0;

                    section_name[0] = 0;
                    key_name[0]     = 0;

                    int i;
                    for (i = 0; i < filesize; i++) {
                        if (in_comment) {
                            if ((filebuffer[i] == '\n') || (filebuffer[i] == '\r'))
                                in_comment = 0;
                            continue;
                        }

                        if (((filebuffer[i] == '\n') || (filebuffer[i] == '\r')) && (in_looked_value)) {
                            lpReturnedString[value_pos++] = 0;
                            break;
                        }

                        if (!in_quote) {
                            if (filebuffer[i] == ';') {
                                if (in_looked_value)
                                    break;
                                in_comment = 1;
                                continue;
                            }

                            if (filebuffer[i] == '[') {
                                // a inceput o sectziune noua, iar sectziunea precedenta este cea cautata ...
                                if (in_looked_section) {
                                    error = 1;
                                    break;
                                }
                                temp_pos = 0;
                                in_name  = 1;
                                continue;
                            }
                            if (in_name) {
                                if (filebuffer[i] == ']') {
                                    in_name = 0;
                                    section_name[temp_pos] = 0;
                                    if (!strcmp(section_name, lpAppName))
                                        in_looked_section = 1;
                                    continue;
                                }
                                if ((filebuffer[i] == '\n') || (filebuffer[i] == '\r')) {
                                    in_name = 0;
                                    continue;
                                }
                                if (in_name)
                                    section_name[temp_pos++] = filebuffer[i];
                            }
                        }
                        // schimb flag-ul de ghilimele, daca le-am gasit ...
                        if (!in_name) {
                            if (filebuffer[i] == '"') {
                                in_quote = !in_quote;
                                continue;
                            }
                            if (((filebuffer[i] == '\n') || (filebuffer[i] == '\r')) && (in_quote)) {
                                in_quote = 0;
                                continue;
                            }
                        }

                        if (in_looked_key) {
                            if (filebuffer[i] == '=') {
                                in_looked_value = 1;
                                continue;
                            }
                            if (in_looked_value) {
                                if (value_pos < nSize - 1) {
                                    if (value_pos)
                                        lpReturnedString[value_pos++] = filebuffer[i];
                                    else if ((filebuffer[i] != ' ') && (filebuffer[i] != '\t'))
                                        lpReturnedString[value_pos++] = filebuffer[i];
                                } else {
                                    lpReturnedString[value_pos++] = 0;
                                    break;
                                }
                                continue;
                            }
                        }

                        if (in_looked_section) {
                            if (((filebuffer[i] >= '0') && (filebuffer[i] <= '9')) ||
                                ((filebuffer[i] >= 'a') && (filebuffer[i] <= 'z')) ||
                                ((filebuffer[i] >= 'A') && (filebuffer[i] <= 'Z')) ||
                                (filebuffer[i] == '_') || (filebuffer[i] == '(') ||
                                (filebuffer[i] == ')') || (filebuffer[i] == '.') ||
                                (filebuffer[i] == '-')) {
                                if (in_key == 0) {
                                    key_pos = 0;
                                    in_key  = 1;
                                }
                                key_name[key_pos++] = filebuffer[i];
                            } else if (in_key)    {
                                key_name[key_pos] = 0;
                                if (!strcmp(key_name, lpKeyName))
                                    in_looked_key = 1;
                                in_key = 0;
                            }
                            // TO DO ... sunt in sectiunea cautata ...
                        }
                    }
                    // in caz ca e ultima linie din fisier ...
                    if ((i == filesize) && (in_looked_value))
                        lpReturnedString[value_pos++] = 0;
                } else
                    error = 1;
            } else
                error = 1;
        } else
            error = 1;
        fclose(in);
    } else
        error = 1;

    if (error)
        strncpy(lpReturnedString, lpDefault, nSize);
    return((long)in);
}

#endif


AnsiString GetKey(char *ini_name, char *section, char *key, char *def) {
    char value[4096];

    value[0] = 0;
    GetPrivateProfileString(section, key, def, value, 4096, ini_name);
    return(AnsiString(value));
}

AnsiString GetDirectory() {
#ifdef _WIN32
    char buffer[4096];
    buffer[0] = 0;
    GetModuleFileName(NULL, buffer, 4096);
    for (int i = strlen(buffer) - 1; i >= 0; i--)
        if ((buffer[i] == '/') || (buffer[i] == '\\')) {
            buffer[i + 1] = 0;
            break;
        }
    return(AnsiString(buffer));
#else
    return(LINUX_PATH);
#endif
}

#ifdef _WIN32
AnsiString GetPath() {
    char buffer[4096];

    buffer[0] = 0;
    GetModuleFileName(NULL, buffer, 4096);
    return(AnsiString(buffer));
}

#endif

//-----------------------------------//
AnsiString GetRootForHost(char *host) {
    AnsiString HOME_BASE = GetDirectory();

#ifdef _WIN32
    AnsiString INI_PATH = HOME_BASE + INI_FILE;
#endif

    AnsiString key = "HostRoot(";
    key += host;
    key += ")";

    AnsiString Path = GetKey(INI_PATH, "Paths", key.c_str(), "");

    if (Path.Length())
        Path = HOME_BASE + Path + PATH_SLASH;

    //std::cerr << "PATH for [" << key.c_str() << "] is [" << Path.c_str() << "]\n";

    return(Path);
}

//-----------------------------------//
AnsiString Services;
AnsiString LoadService(int index) {
    AnsiString serv;

    if (!Services.Length())
        return(serv);

    AnsiString sid("Service");
    sid += AnsiString((long)index);
    serv = GetKey(Services, sid, "Path", "");
    AnsiString name = GetKey(Services, sid, "Name", "");

    if (serv.Length())
        std::cout << Services.c_str() << "> " << sid.c_str() << ": Starting service '" << name.c_str() << "' from " << serv.c_str() << "\n";
    else
        std::cout << Services.c_str() << "> " << sid.c_str() << ": No such service\n";

    if (ConceptServer::LogPath.Length()) {
        std::ofstream out;
        out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
        if (serv.Length())
            out << Services.c_str() << "> " << sid.c_str() << ": Starting service '" << name.c_str() << "' from " << serv.c_str() << "\n";
        else
            out << Services.c_str() << "> " << sid.c_str() << ": No such service\n";
        out.flush();
        out.close();
    }

    return(serv);
}

//-----------------------------------//

/*void SetEnvVar(char *varname, char *value) {
 * #ifdef _WIN32
 *  AnsiString my_str=varname;
 *  my_str+=(char *)"=";
 *  my_str+=value;
 *  _putenv(my_str.c_str());
 * #else
 *  setenv(varname, value, 1);
 * #endif
 * }*/
#ifdef _WIN32

BOOL leave(DWORD fdwCtrlType) {
    ConceptServer *localref = SERVER_REF;

    if (localref) {
        switch (fdwCtrlType) {
            case CTRL_C_EVENT:
            case CTRL_CLOSE_EVENT:
            case CTRL_BREAK_EVENT:
            case CTRL_LOGOFF_EVENT:
            case CTRL_SHUTDOWN_EVENT:
                SERVER_REF = 0;
                localref->Stop();
                exit(0);
                break;
        }
    }
    return(FALSE);
}

int server_main(int argc, char **argv) {
#else

void leave(int sig) {
    ConceptServer *localref = SERVER_REF;

    if (localref) {
        SERVER_REF = 0;
        localref->Stop();
        exit(0);
    }
}

int main(int argc, char **argv) {
#endif

    time_t now;
    time(&now);
    AnsiString sztime = ctime(&now);
    char       *ptr   = sztime.c_str();
    int        len    = sztime.Length();
    if (ptr[len - 1] == '\n')
        ptr[len - 1] = 0;

#ifdef _WIN32
    // don't register for services
    if ((argc) && (argv))
        SetConsoleCtrlHandler((PHANDLER_ROUTINE)leave, TRUE);
#else
    signal(SIGTERM, leave);
#endif
    ConceptServer::InheritEnv();

    AnsiString HOME_BASE = GetDirectory();
#ifdef _WIN32
    AnsiString INI_PATH = HOME_BASE + INI_FILE;
#endif
    ConceptServer::LogPath = HOME_BASE + GetKey(INI_PATH, "Paths", "LogFile", "security.log");
    ConceptServer::SetEnvVar("CONCEPT_LOGFILE", ConceptServer::LogPath.c_str());
#ifdef _WIN32
    int    retval = 0;
    HANDLE hToken = 0;

    AnsiString user = GetKey(INI_PATH, "Server", "User", "www");
    AnsiString pwd  = GetKey(INI_PATH, "Server", "UserPassword", "");

    retval = LogonUser(
        (char *)user.c_str(),
        (char *)".",
        (char *)pwd.c_str(),
        LOGON32_LOGON_SERVICE,
        LOGON32_PROVIDER_DEFAULT,
        &hToken
        );
    if (pwd.Length())
        memset(pwd.c_str(), 0, pwd.Length());

    if (!retval) {
        if (ConceptServer::LogPath.Length()) {
            std::ofstream out;
            out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
            out << ptr << ": Error switching user to '" << user.c_str() << "', error code: 0x" << std::hex << GetLastError() << "\n";
            out.flush();
            out.close();
        }
    } else {
        retval = ImpersonateLoggedOnUser(hToken);
        if (!retval) {
            if (ConceptServer::LogPath.Length()) {
                std::ofstream out;
                out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
                out << ptr << ": Error impersonating user '" << user.c_str() << "', error code: 0x" << std::hex << GetLastError() << "\n";
                out.flush();
                out.close();
            }
        } else
            std::cout << "Running as user '" << user.c_str() << "'\n";
    }

#else
    // set user id !
    AnsiString    user = GetKey(INI_PATH, "Server", "User", "www");
    struct passwd *pwd = getpwnam((const char *)user.c_str());
    if (pwd) {
        std::cout << "Using user '" << user.c_str() << "' with UID: " << pwd->pw_uid << "\n";
        setuid(pwd->pw_uid);
        setgid(pwd->pw_gid);
    } else if (ConceptServer::LogPath.Length())    {
        std::ofstream out;
        out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
        out << ptr << ": Error switching user to '" << user.c_str() << "'\n";
        out.flush();
        out.close();
    }
#endif


    //ConceptServer::SetEnvVar("TEMP",getenv("TEMP"));
    //ConceptServer::SetEnvVar("TMP",getenv("TMP"));
    //ConceptServer::SetEnvVar("PATH",getenv("PATH"));

    AnsiString Library = HOME_BASE + GetKey(INI_PATH, "Paths", "Library", "Library") + PATH_SLASH;
    ConceptServer::SetEnvVar("CONCEPT_LIBRARY", Library.c_str());
    AnsiString Include = HOME_BASE + GetKey(INI_PATH, "Paths", "Include", "Include") + PATH_SLASH;
    ConceptServer::SetEnvVar("CONCEPT_INCLUDE", Include.c_str());
    AnsiString Certificate = HOME_BASE + GetKey(INI_PATH, "Paths", "Certificate", "") + PATH_SLASH;
    ConceptServer::SetEnvVar("CONCEPT_CERTIFICATE", Certificate.c_str());
    AnsiString Path = HOME_BASE + GetKey(INI_PATH, "Paths", "ServerRoot", ".") + PATH_SLASH;
    ConceptServer::SetEnvVar("CONCEPT_SERVERROOT", Path.c_str());
    ConceptServer::SetEnvVar("CONCEPT_INIPATH", INI_PATH);
    AnsiString SafeMode = HOME_BASE + GetKey(INI_PATH, "Paths", "SafeMode", "runsafe");
    ConceptServer::Service     = HOME_BASE + GetKey(INI_PATH, "Paths", "Service", "conceptservice");
    ConceptServer::ServicePath = HOME_BASE + GetKey(INI_PATH, "Paths", "ServiceRoot", "") + PATH_SLASH;
    Services = HOME_BASE + GetKey(INI_PATH, "Paths", "ServiceList", "");
    AnsiString SHMemPool = GetKey(INI_PATH, "Server", "UseSharedMemoryPool", "0");
    ConceptServer::SetEnvVar("CONCEPT_UseSharedMemoryPool", SHMemPool.c_str());
    int port       = GetKey(INI_PATH, "Server", "Port", AnsiString(DEFAULT_PORT)).ToInt();
    int cgiport    = GetKey(INI_PATH, "Server", "CGIPort", AnsiString(DEFAULT_CGI_PORT)).ToInt();
    int usecgi     = GetKey(INI_PATH, "Server", "HandleCGIRequests", "0").ToInt();
    int max_buffer = GetKey(INI_PATH, "Server", "MaxInterAppMessageBuffer", "1024").ToInt();
    ConceptServer::CGITrustedIP = GetKey(INI_PATH, "Server", "CGITrustedIP", "127.0.0.1");
    int max_connections = GetKey(INI_PATH, "Server", "MaxConnections", AnsiString((long)MAX_CONNECTIONS).c_str()).ToInt();
    if (max_connections > 0)
        ConceptServer::max_connections = max_connections;
    int max_precontimeout = GetKey(INI_PATH, "Server", "MaxInitialIDLE", AnsiString((long)MAX_IDLE).c_str()).ToInt();
    if (max_precontimeout > 0)
        ConceptServer::max_idle = max_precontimeout;
    ConceptServer::ipv6 = GetKey(INI_PATH, "Server", "IPv6", "0").ToInt();
    if (ConceptServer::ipv6)
        ConceptServer::SetEnvVar("CONCEPT_IPv6", "1");
    else
        ConceptServer::SetEnvVar("CONCEPT_IPv6", "0");
    ConceptServer::web_timeout = GetKey(INI_PATH, "Web", "Timeout", "10").ToInt();

    if (port <= 0)
        port = DEFAULT_PORT;
    if (cgiport <= 0)
        cgiport = DEFAULT_CGI_PORT;
    ConceptServer::SetEnvVar("CONCEPT_PORT", AnsiString((long)port).c_str());
    ConceptServer::SetEnvVar("CONCEPT_CGI_PORT", AnsiString((long)cgiport).c_str());
    //AnsiString DLLPath=HOME_BASE+GetKey(INI_PATH,"Paths","SharedDLL","")+PATH_SLASH;
    //SetDllDirectory(DLLPath.c_str());

    std::cout << ptr << ": Starting Concept server with root " << Path.c_str() << "\n";

    if (ConceptServer::LogPath.Length()) {
        std::ofstream out;
        out.open(ConceptServer::LogPath.c_str(), std::ofstream::out | std::ofstream::app);
        out << ptr << ": Starting Concept server with root " << Path.c_str() << "\n";
        out.flush();
        out.close();
    }

    chdir(Path);
#ifdef _WIN32
    AnsiString ENV_PATH("CONCEPT_ROOT=");
    ENV_PATH += Path;
    _putenv(ENV_PATH.c_str());
#endif

    ConceptServer CSERVER(Certificate, Library, Include, SafeMode, port, cgiport, GetRootForHost, max_buffer);
    if (CSERVER.Init() == NO_WEB_ERROR)
        SERVER_REF = &CSERVER;
    else
        return(-1);

    if (usecgi)
        CSERVER.StartCGIServer();
    CSERVER.Start(LoadService);
    return(0);
}

#ifdef _WIN32

int install_service() {
    AnsiString HOME_BASE = GetPath();
    SC_HANDLE  NishService, scm;

    scm = OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
    if (!scm)
        return(1);
    NishService = CreateService(scm, SERVICE_NAME,
                                SERVICE_NAME,
                                SERVICE_ALL_ACCESS, SERVICE_WIN32_OWN_PROCESS, SERVICE_DEMAND_START,
                                SERVICE_ERROR_NORMAL,
                                HOME_BASE,
                                0, 0, 0, 0, 0);
    if (!NishService) {
        CloseServiceHandle(scm);
        return(1);
    }
    CloseServiceHandle(NishService);
    CloseServiceHandle(scm);
    return(0);
}

void ServiceMain(DWORD argc, LPTSTR *argv);
void ServiceCtrlHandler(DWORD nControlCode);
BOOL UpdateServiceStatus(DWORD dwCurrentState, DWORD dwWin32ExitCode,
                         DWORD dwServiceSpecificExitCode, DWORD dwCheckPoint,
                         DWORD dwWaitHint);
BOOL StartServiceThread();
DWORD ServiceExecutionThread(LPDWORD param);

HANDLE hServiceThread;
void KillService();

char *strServiceName = SERVICE_NAME;
SERVICE_STATUS_HANDLE nServiceStatusHandle;
HANDLE killServiceEvent;
BOOL   nServiceRunning;
DWORD  nServiceCurrentStatus;

int main(int argc, char *argv[]) {
    /*if (argc<2) {
     *      std::cout << "Usage : concept-server [option]\nWhere option is :\n\tinstall - installs service\n\tuninstall - uninstalls service\n\tauto - starts service automatically\n\tmanual - starts service manually\n\tstart - starts service\n\tstop - stops service\n\tnoservice - runs Concept II server without using Windows NT Services\n";
     *      return -1;
     * }*/
    if (argc == 2) {
        if (AnsiString(argv[1]) == (char *)"noservice")
            return(server_main(argc, argv));

        if (AnsiString(argv[1]) == (char *)"install")
            return(install_service());

        SC_HANDLE NishService, scm;
        scm = OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
        if (!scm)
            return(1);

        NishService = OpenService(scm, SERVICE_NAME, SERVICE_ALL_ACCESS);
        if (!NishService) {
            CloseServiceHandle(scm);
            return(1);
        }

        if (AnsiString(argv[1]) == (char *)"start") {
            std::cout << "Starting ... ";
            StartService(NishService, 0, NULL);
            std::cout << "done\n";
        } else
        if (AnsiString(argv[1]) == (char *)"stop") {
            SERVICE_STATUS m_SERVICE_STATUS;
            std::cout << "Stopping ... ";
            ControlService(NishService, SERVICE_CONTROL_STOP, &m_SERVICE_STATUS);
            std::cout << "done\n";
        } else
        if (AnsiString(argv[1]) == (char *)"auto") {
            std::cout << "Changing settings ... ";
            AnsiString HOME_BASE = GetPath();
            ChangeServiceConfig(NishService,
                                SERVICE_WIN32_OWN_PROCESS,
                                SERVICE_AUTO_START,
                                SERVICE_ERROR_NORMAL,
                                HOME_BASE,
                                0, 0, 0, 0, 0,
                                SERVICE_NAME
                                );
            std::cout << "done\n";
        } else
        if (AnsiString(argv[1]) == (char *)"manual") {
            std::cout << "Changing settings ... ";
            AnsiString HOME_BASE = GetPath();
            ChangeServiceConfig(NishService,
                                SERVICE_WIN32_OWN_PROCESS,
                                SERVICE_DEMAND_START,
                                SERVICE_ERROR_NORMAL,
                                HOME_BASE,
                                0, 0, 0, 0, 0,
                                SERVICE_NAME
                                );
            std::cout << "done\n";
        } else
        if (AnsiString(argv[1]) == (char *)"uninstall") {
            std::cout << "Removing Concept II SVC ... ";
            DeleteService(NishService);
            std::cout << "done\n";
        } else
            std::cout << "Usage : concept-server [option]\nWhere option is :\n\tinstall - installs service\n\tuninstall - uninstalls service\n\tauto - starts service automatically\n\tmanual - starts service manually\n\tstart - starts service\n\tstop - stops service\n\tnoservice - runs Concept II server without using Windows NT Services\n";

        CloseServiceHandle(NishService);
        CloseServiceHandle(scm);
        return(0);
    }

    SERVICE_TABLE_ENTRY servicetable[] =
    {
        { strServiceName, (LPSERVICE_MAIN_FUNCTION)ServiceMain },
        { NULL,           NULL                                 }
    };
    BOOL success;
    success = StartServiceCtrlDispatcher(servicetable);
    if (!success) {
        std::cout << "Usage : concept-server [option]\nWhere option is :\n\tinstall - installs service\n\tuninstall - uninstalls service\n\tauto - starts service automatically\n\tmanual - starts service manually\n\tstart - starts service\n\tstop - stops service\n\tnoservice - runs Concept II server without using Windows NT Services\n";
        return(-1);
        //return -1;
    }
    return(0);
}

void ServiceMain(DWORD argc, LPTSTR *argv) {
    BOOL success;

    nServiceStatusHandle = RegisterServiceCtrlHandler(strServiceName,
                                                      (LPHANDLER_FUNCTION)ServiceCtrlHandler);
    if (!nServiceStatusHandle)
        return;
    success = UpdateServiceStatus(SERVICE_START_PENDING, NO_ERROR, 0, 1, 3000);
    if (!success)
        return;
    killServiceEvent = CreateEvent(0, TRUE, FALSE, 0);
    if (killServiceEvent == NULL)
        return;
    success = UpdateServiceStatus(SERVICE_START_PENDING, NO_ERROR, 0, 2, 1000);
    if (!success)
        return;
    success = StartServiceThread();
    if (!success)
        return;
    nServiceCurrentStatus = SERVICE_RUNNING;
    success = UpdateServiceStatus(SERVICE_RUNNING, NO_ERROR, 0, 0, 0);
    if (!success)
        return;
    WaitForSingleObject(killServiceEvent, INFINITE);
    CloseHandle(killServiceEvent);
}

BOOL UpdateServiceStatus(DWORD dwCurrentState, DWORD dwWin32ExitCode,
                         DWORD dwServiceSpecificExitCode, DWORD dwCheckPoint,
                         DWORD dwWaitHint) {
    BOOL           success;
    SERVICE_STATUS nServiceStatus;

    nServiceStatus.dwServiceType  = SERVICE_WIN32_OWN_PROCESS;
    nServiceStatus.dwCurrentState = dwCurrentState;
    if (dwCurrentState == SERVICE_START_PENDING)
        nServiceStatus.dwControlsAccepted = 0;
    else
        nServiceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP
                                            | SERVICE_ACCEPT_SHUTDOWN;
    if (dwServiceSpecificExitCode == 0)
        nServiceStatus.dwWin32ExitCode = dwWin32ExitCode;
    else
        nServiceStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
    nServiceStatus.dwServiceSpecificExitCode = dwServiceSpecificExitCode;
    nServiceStatus.dwCheckPoint = dwCheckPoint;
    nServiceStatus.dwWaitHint   = dwWaitHint;

    success = SetServiceStatus(nServiceStatusHandle, &nServiceStatus);

    if (!success) {
        KillService();
        return(success);
    } else
        return(success);
}

BOOL StartServiceThread() {
    /*AnsiString HOME_BASE=GetDirectory();
     * AnsiString INI_PATH=HOME_BASE+INI_FILE;
     * AnsiString Path=HOME_BASE+GetKey(INI_PATH,"Paths","ServerRoot",".")+PATH_SLASH;
     * SetCurrentDirectory(Path.c_str());*/

    DWORD id;

    hServiceThread = CreateThread(0, 0,
                                  (LPTHREAD_START_ROUTINE)ServiceExecutionThread,
                                  0, 0, &id);
    if (hServiceThread == 0)
        return(false);
    else {
        nServiceRunning = true;
        return(true);
    }
}

DWORD ServiceExecutionThread(LPDWORD param) {
    while (nServiceRunning)
        if (server_main(0, 0) < 0)
            return(0);
    return(0);
}

void KillService() {
    ConceptServer *localref = SERVER_REF;

    if (localref) {
        SERVER_REF = 0;
        localref->Stop();
    }
    nServiceRunning = false;
    SetEvent(killServiceEvent);
    UpdateServiceStatus(SERVICE_STOPPED, NO_ERROR, 0, 0, 0);
}

void ServiceCtrlHandler(DWORD nControlCode) {
    BOOL success;

    switch (nControlCode) {
        case SERVICE_CONTROL_SHUTDOWN:
        case SERVICE_CONTROL_STOP:
            nServiceCurrentStatus = SERVICE_STOP_PENDING;
            success = UpdateServiceStatus(SERVICE_STOP_PENDING, NO_ERROR, 0, 1, 3000);
            KillService();
            return;

        default:
            break;
    }
    UpdateServiceStatus(nServiceCurrentStatus, NO_ERROR, 0, 0, 0);
}

#endif
