//#define DEBUG
//#define DISABLE_CHECKPOINTS

#include <stdlib.h>
#include <fcntl.h>
#ifdef _WIN32
 #include <io.h>
//#include <windows.h>
 #include <winsock2.h>
 #include <sys/utime.h>
#else
 #include <dlfcn.h>        // dlopen

 #define HANDLE     void *
 #define HMODULE    HANDLE

 #define SOCKET     int
#endif

#include "AnsiString.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef _WIN32
//#include <dir.h>
 #include <direct.h>
#else
 #define SOCKET            int
 #define INVALID_SOCKET    -1
 #define SOCKET_ERROR      -1
 #define DWORD             int
 #define LPVOID            void *

 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <netdb.h>

 #include <unistd.h>


//    #define LINUX_PATH "/opt/Concept/bin/"

#endif

#ifdef _WIN32
 #define LIBRARY    "concept-core.2.0.dll"
#else
//#define LIBRARY      AnsiString(LINUX_PATH) + "concept-core.1.0alpha.so"
 #define LIBRARY    "libconceptcore.2.0.so"
#endif

#define INI_FILE    "concept.ini"

typedef int (*GET_VARIABLE_PROC)(int operation, void *VDESC, void *CONTEXT, int Depth, char *VariableName, char *buffer, int buf_size);
typedef int (*DEBUGGER_CALLBACK)(void *VDESC, void *CONTEXT, int Depth, int line, char *filename, GET_VARIABLE_PROC GVP, void *DEBUGGER_RESERVED);

extern "C" {
typedef int (*API_INTERPRETER)(char *filename, char *inc_dir, char *lib_dir, void *fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent);
typedef int (*API_INTERPRETER2)(char *filename, char *inc_dir, char *lib_dir, void *fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, void *userdata);
}

#define     DEBUG_RUN         0x01
#define     DEBUG_STEPINTO    0x02
#define     DEBUG_RUNTO       0x03
#define     DEBUG_STEPOVER    0x04

#define MAX_DEBUG_REQUEST     8192

int __pipe_out = 0;
int __pipe_in  = 0;
int __p_apid   = 0;

typedef struct {
    long       CLIENT_NOTICE; // VARIABILA SETATA DE INTERPRETOR !!!

    int        last_line;
    int        debug_type;
    int        run_to_line;
    AnsiString FileName;
    AnsiString run_to_filename;
    void       *context_id;
    SOCKET     sock;
    SOCKET     DEBUG_SOCKET;
} ClientDebugInfo;


int  status      = 0;
bool check_point = false;

void CheckPoint(int _status) {
    status      = _status;
    check_point = true;
}

int NotifyParent(int msg_id, AnsiString DATA) {
    if ((__p_apid <= 0) || (__pipe_in <= 0))
        return(0);

    int params[3];
    params[0] = __p_apid;
    params[1] = msg_id;
    params[2] = DATA.Length();

    write(__pipe_out, params, sizeof(int) * 3);

    if (params[2])
        write(__pipe_out, DATA.c_str(), params[2]);

    return(1);
}

AnsiString SetWebDirectory(char *filename) {
    char       *buffer;
    AnsiString temp  = filename;
    AnsiString temp2 = filename;

    if (temp != (char *)"") {
        buffer = temp.c_str();
        for (int i = strlen(buffer) - 1; i >= 0; i--)
            if ((buffer[i] == '/') || (buffer[i] == '\\')) {
                temp2         = (char *)buffer + i + 1;
                buffer[i + 1] = 0;
                break;
            }
        chdir(buffer);
    }
    return(temp2);
}

void Print(char *str, int length) {
    if (length > 0)
        std::cout.write(str, length);
    else
        std::cout << str;
}

int GetParent(int *parent, int *msg_id, AnsiString *DATA) {
    if ((__p_apid <= 0) || (__pipe_out <= 0))
        return(0);

    int params[3];

    if (read(__pipe_in, params, sizeof(int) * 3) != sizeof(int) * 3)
        return(0);

    *parent = params[0];
    *msg_id = params[1];

    if (params[2]) {
        char *data = new char [params[2] + 1];
        data[params[2]] = 0;
        int rd = read(__pipe_in, data, params[2]);
        *DATA = data;
        delete[] data;
        if (rd != params[2])
            return(0);
    }

    return(1);
}

AnsiString Strip(char *filename) {
    AnsiString fname = filename;

    if (fname[0] == '"') {
        int len = fname.Length();
        if (len > 2) {
            fname = filename + 1;
            fname.c_str()[len - 2] = 0;
        }
    }
    return(fname);
}

int main(int argc, char **argv) {
    if (argc != 8) {
        std::cerr << "This program is used internaly by Concept Server and should not be instantiated from shell\n";
        return(-10);
    }
#ifdef _WIN32
    chdir(getenv("CONCEPT_ROOT"));
#endif

    char *filename   = argv[0];
    char *Include    = argv[1];
    char *Library    = argv[2];
    int  direct_pipe = AnsiString(argv[3]).ToInt();
    int  pipein      = AnsiString(argv[4]).ToInt();
    int  pipeout     = AnsiString(argv[5]).ToInt();
    int  apid        = AnsiString(argv[6]).ToInt();
    int  parent_apid = AnsiString(argv[7]).ToInt();


    AnsiString got_file = SetWebDirectory(filename);
    filename = got_file.c_str();


    __pipe_out = pipeout;
    __pipe_in  = pipein;
    __p_apid   = parent_apid;


    HMODULE          hDLL;
    API_INTERPRETER2 Concept_Execute;


#ifdef _WIN32
    hDLL = LoadLibrary(LIBRARY);
#else
    // linux module
    hDLL = dlopen(LIBRARY_PATH, RTLD_LAZY);
#endif

    if (!hDLL) {
        std::cout << "Error loading core library '" << LIBRARY << "'\r\n";
        return(-1);
    }


#ifdef _WIN32
    Concept_Execute = (API_INTERPRETER2)GetProcAddress(hDLL, "Concept_Execute");
#else
    Concept_Execute = (API_INTERPRETER2)dlsym(hDLL, "Concept_Execute");
#endif
    if (!Concept_Execute) {
        std::cout << "Corrupted library : '" << LIBRARY << "'\n";
#ifdef _WIN32
        FreeLibrary(hDLL);
#else
        dlclose(hDLL);
#endif
        return(-1);
    }

    //std::cout << "About to execute " << filename << "...\n";
    int res = Concept_Execute(filename, Include, Library, (void *)Print, 0, 0, 0, 0, 0, 0, 0, pipein, pipeout, apid, parent_apid, 0);
    //if (res==-3)
    //   std::cout << "File not found or insufficient rights : " << filename << "\n";
    if (res) {
        switch (res) {
            case -1:
            case -2:
            case -3:
                std::cout << "Result : " << res << "\n";
                break;
        }
    }
#ifdef _WIN32
    FreeLibrary(hDLL);
#else
    dlclose(hDLL);
#endif
    return(res);
}
